// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_entry.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 09/26/2005
//              :
//              :
// Description  : This file contains data structure definitions for address map
//                fields and entries. An entry is either a register or a 
//                contiguous section of memory. A field is a logical abstraction
//                of a subset of bits in a register or in a row of a memory entry. 
//
//                Each data structure comes with the necessary hooks for 
//                maintaining shadow values for scoreboarding purposes. Register
//                structs also come with backdoor access methods.
//                
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_addrmap;


// This is the basic unit of an address map. Each field
// is used to compose a row in an addrmap entry. Note 
// that register entries only have 1 row.
struct csco_addrmap_field_s {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // logical name of group of bits in memory. Name is unique within
  // each addrmap entry.
  !name : string;
    csco_print_attr name SHORT enable;

  // number of bits in field.
  !width : uint;
    csco_print_attr width SHORT enable;
    csco_print_attr width SHORT header=bits;
    csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "width" {
        result = append(dec(lsb_pos + width - 1), ":", dec(lsb_pos));
      };
    };

  // bit position of least significant bit.
  !lsb_pos : uint;
    csco_print_attr lsb_pos ALL disable;

  // software access mode, exact mirror of Blueprint defs.
  !sw_access : csco_addrmap_sw_access_t;
    csco_print_attr sw_access SHORT enable;
    csco_print_attr sw_access SHORT header=sw_acc;


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  // Returns reset value of the field.
  get_reset_val() : int(bits:*) is undefined;

  // Sets reset value of the field.
  set_reset_val(val : int(bits:*)) is undefined;

  // Returns shadow value of the field.
  get_shadow_val() : int(bits:*) is undefined;

  set_reset_and_shadow_val(val : int(bits:*)) is {
    set_reset_val(val);
    set_shadow_val(val);
  };

  // Sets shadow value of the field.
  set_shadow_val(val : int(bits:*)) is undefined;

  // returns RTL string name of field for backdoor HDL access
  get_hdl_name() : string is {
    result = name;
  };

  // This method returns whether or not the field should be 
  // checked against the shadow value on an 
  // 'update_shadow()' call.
  is_auto_comp() : bool is {
    result = (!has_attribute(HW_W) and !has_attribute(AUTOCOMP_OVERRIDE)) or
             ( has_attribute(HW_W) and  has_attribute(AUTOCOMP_OVERRIDE));
  };

  // Adds an attribute to this field.
  add_attribute(a : csco_addrmap_field_attribute_t) is {
    var ind : uint = a.as_a(uint);
    attributes[ind:ind] = 1;
  };

  // Removes an attribute to this field.
  remove_attribute(a : csco_addrmap_field_attribute_t) is {
    var ind : uint = a.as_a(uint);
    attributes[ind:ind] = 0;
  };

  // Returns all the attributes that are currently defined for the field.
  get_attributes() : list of csco_addrmap_field_attribute_t is {
    for each (attr) in all_values(csco_addrmap_field_attribute_t) {
      var tmp_ind : uint = attr.as_a(uint);
      if attributes[tmp_ind:tmp_ind] == 1 {
        result.add(attr);
      };
    };
  };

  // Returns whether or not the input attribute is defined for this
  // field.
  has_attribute(attr : csco_addrmap_field_attribute_t) : bool is {
    var tmp_ind : uint = attr.as_a(uint);
    result = (attributes[tmp_ind:tmp_ind] == 1);
  };

  // Adds a side-effect to this field.
  add_side_effect(val : csco_addrmap_side_effect_t) is {
    var ind : uint = val.as_a(uint);
    side_effects[ind:ind] = 1;
  };

  // Removes a side-effect from this field.
  remove_side_effect(val : csco_addrmap_side_effect_t) is {
    var ind : uint = val.as_a(uint);
    side_effects[ind:ind] = 0;
  };

  // Returns all the side-effects that are currently defined for the field.
  get_side_effects() : list of csco_addrmap_side_effect_t is {
    for each (se) in all_values(csco_addrmap_side_effect_t) {
      var tmp_ind : uint = se.as_a(uint);
      if side_effects[tmp_ind:tmp_ind] == 1 {
        result.add(se);
      };
    };
  };

  // Returns whether or not the input side-effect is defined for this
  // field.
  has_side_effect(se : csco_addrmap_side_effect_t) : bool is {
    var tmp_ind : uint = se.as_a(uint);
    result = (side_effects[tmp_ind:tmp_ind] == 1);
  };

  // For the specified operation (READ/WRITE) and data, return the new expected
  // value of the field based on the current shadow value.
  apply_side_effects(op : csco_addrmap_op_t, data : int(bits:*)) : int(bits:*) is {
    assert op in [READ, WRITE, FILL] else 
      error("CSCO_ADDRMAP_ERR_SE_OPT: op input to apply_side_effects() must be READ or WRITE\n");

    result = data ; // default unless a side effect below applies

    if op == READ {
      if has_side_effect(RCLR) {
        result = 0;
      };

    } // READ

    else {
      if has_side_effect(WOSET) {
        result = (get_shadow_val() | data);
      };

      if has_side_effect(WOCLR) {
        result = (get_shadow_val() & ~data);
      };
      
      if sw_access == R {
        result = get_shadow_val();
      };
    }; // WRITE
  }; // apply_side_effects()

  // For Reads, this method will compare the read response data with the expected
  // shadow value. For both reads and writes, this method will update the shadow
  // value taking into consideration all side-effects defined for this field.
  update_shadow(eptr : csco_addrmap_entry_s, op : csco_addrmap_op_t, data : int(bits:*)) is {
    var new_val : int(bits:*);

    assert op in [READ, WRITE, FILL] else 
      error("CSCO_ADDRMAP_ERR_SHDW_OPT: op input to <field>.update_shadow() must be READ or WRITE\n");

    // check read response against shadow value if needed.
    if op == READ {
      if sw_access in [W, NA] {
        if data != get_reset_val() {
          read_error(eptr, get_reset_val(), data);
        };
      }
      else if is_auto_comp() {
        if data != get_shadow_val() {
          read_error(eptr, get_shadow_val(), data);
        };
      }
      else {
        messagef(CSCO_ADDRMAP, LOW, "won't check read response (0x%x) from %s because its auto-comparison flags are disabled\n", data, name);
      };
    }; // READ

    // update shadow value
    new_val = apply_side_effects(op, data);
    set_shadow_val(new_val);
  }; // update_shadow()

  // Check if two fields have the same width and bit locations.
  compare(f : csco_addrmap_field_s) : bool is {
    result = ((f.width == width) and (f.lsb_pos == lsb_pos));
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  // SW access side-effects, encoded in a bit mask for memory efficiency reasions.
  package !side_effects : uint;
    csco_print_attr side_effects SHORT enable;
    csco_print_attr side_effects SHORT header=side_eff;

  // Special attributes of a field as a bit mask. attributes include CRITICAL,
  // AUTOCOMP_OVERRIDE, and IN_AUTOTEST.
  package !attributes : uint;
    csco_print_attr attributes SHORT enable;
    csco_print_attr attributes SHORT header=attr;

  csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
    if pf.name == "side_effects" {
      result = str_join(get_side_effects().as_a(list of string), ",");
    };

    if pf.name == "attributes" {
      result = str_join(get_attributes().as_a(list of string), ",");
    };
  };

  package read_error(eptr : csco_addrmap_entry_s, exp : int(bits:*), act : int(bits:*)) is {
    dut_error("CSCO_ADDRMAP_ERR_FIELD_RD_MISMATCH: read data mismatch for ", 
              eptr.long_name(""), ".", name) {
      outf("\n");
      csco_print(SHORT);
      outf("\n");
      out("  expected value = ", hex(exp));
      out("  actual   value = ", hex(act));
      outf("\n");
      out("  below is the entry containing the miscompared field:");
      eptr.csco_print(LONG);
    };
  };
}; // struct csco_addrmap_field_s


// This field type should be used if the field is <= 32-bits, or
// whatever the width of the default uint implementation happens to be,
// for optimal memory foot-print.
struct csco_addrmap_narrow_field_s like csco_addrmap_field_s {
  package reset_val : csco_addrmap_field_val_t;
    csco_print_attr reset_val SHORT enable;
    csco_print_attr reset_val ALL precede=side_effects;

  package !%shadow_val : csco_addrmap_field_val_t;
    csco_print_attr shadow_val SHORT enable;
    csco_print_attr shadow_val ALL precede=side_effects;

  get_reset_val() : int(bits:*) is {
    result = reset_val;
  };

  set_reset_val(val : int(bits:*)) is {
    reset_val = val;
  };

  get_shadow_val() : int(bits:*) is {
    result = shadow_val;
  };

  set_shadow_val(val : int(bits:*)) is {
    shadow_val = val;
  };
}; // struct csco_addrmap_narrow_field_s {


// This field type should be used if the field is wider than 32-bits,
// and there shouldn't be too many instances of this type around.
struct csco_addrmap_wide_field_s like csco_addrmap_field_s {
  package !reset_val : int(bits:*);
    csco_print_attr reset_val SHORT enable;
    csco_print_attr reset_val ALL precede=side_effects;

  package !%shadow_val : int(bits:*);
    csco_print_attr shadow_val SHORT enable;
    csco_print_attr shadow_val ALL precede=side_effects;

  get_reset_val() : int(bits:*) is {
    result = reset_val;
  };

  set_reset_val(val : int(bits:*)) is {
    reset_val = val;
  };

  get_shadow_val() : int(bits:*) is {
    result = shadow_val;
  };

  set_shadow_val(val : int(bits:*)) is {
    shadow_val = val;
  };
}; // struct csco_addrmap_wide_field_s


// This field type is used for "alias" fields
// which access another physical field with different attributes
struct csco_addrmap_alias_field_s like csco_addrmap_field_s {
  package !physical_field : csco_addrmap_field_s;
    csco_print_attr name ALL disable;
    csco_print_attr width ALL disable;
    csco_print_attr sw_access ALL disable;
    csco_print_attr side_effects ALL disable;
    csco_print_attr attributes ALL disable;
    csco_print_attr physical_field ALL enable;
    csco_print_attr physical_field ALL header=alias expand=ON;

  get_reset_val() : int(bits:*) is {
    result = physical_field.get_reset_val();
  };

  set_reset_val(val : int(bits:*)) is {
    physical_field.set_reset_val(val);
  };

  get_shadow_val() : int(bits:*) is {
    result = physical_field.get_shadow_val();
  };

  set_shadow_val(val : int(bits:*)) is {
    physical_field.set_shadow_val(val);
  };

  get_hdl_name() : string is only {
    result = physical_field.get_hdl_name();
  };
};


// Models either a register or a memory entry. Note that a memory
// entry may consist of multiple rows of fields, each row with the
// same field layout.
struct csco_addrmap_entry_s {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Name of the entry. Name must be unique within parent name scope.
  !name : string;
    csco_print_attr name SHORT enable;

  // pointer to parent physical region.
  !parent_region : csco_addrmap_physical_region_s;
    attribute parent_region deep_copy = reference;

  // pointer to parent name scope.
  !parent_scope : csco_addrmap_name_scope_s;
    csco_print_attr parent_scope SHORT enable;
    csco_print_attr parent_scope SHORT header=scope;
    csco_print_get_struct_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "parent_scope" and pf.format == SHORT {
        var intlist : list of byte = result.as_a(list of byte);
        if intlist.size() > 30 {
          intlist = intlist[intlist.size()-30..];
          result = intlist.as_a(string);
        };
      };
    };

  !width : csco_addrmap_addr_t;
    csco_print_attr width SHORT enable;

  // offset in bytes of the io_entry w.r.t. parent region's base adddress.
  !offset : csco_addrmap_addr_t;
    csco_print_attr offset SHORT enable;

  // List of physical fields.
  // <br>
  // For REGISTER: Stores shadow and reset value for each field
  //               in the register.
  // <br>
  // For MEMORY:   Stores the reset value for each field in each
  //               row of the memory. It also serves as the 
  //               interface for packing and unpacking the bytes
  //               in a single row of the memory.
  // <br>
  // Note that for memory entries that don't have any logical
  // structure overlaid on top of the bytes, 'fields' is not used.
  !%fields : list (key:name) of csco_addrmap_field_s;


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  // Returns width of entry in bytes.
  get_entry_width() : csco_addrmap_addr_t is {
    result = width;
  };

  // Returns number of bytes in entire entry.
  get_entry_size() : csco_addrmap_addr_t is undefined;

  // Returns number of rows in entire entry.
  get_entry_depth() : csco_addrmap_addr_t is undefined;

  // Links alias registers together.
  // The physical values of the current entry
  // will be tied to the specified alias entry
  link_alias_entry(alias : csco_addrmap_entry_s) is {
    alias.aliased_from = me;

    assert that alias.fields.size() == fields.size() else
      error("ERR_CSCO_ADDRMAP_ALIAS_MISMATCH\n",
            "aliasing ", alias.long_name(""), " to ", long_name(""),
            ", but they have different # of fields\n");

    for each (f) in alias.fields {
      if(f is a csco_addrmap_alias_field_s (af)) {

        assert(f.compare(fields[index])) else
          error(dec("ERR_CSCO_ADDRMAP_ALIAS_MISMATCH\n",
                    "aliasing ", alias.long_name(""), " to ", long_name(""),
                    ", but their field #", index, " is different\n"));

        assert fields[index] is not a csco_addrmap_alias_field_s else
          error(dec("ERR_CSCO_ADDRMAP_ALIAS_MISMATCH\n",
                    "aliasing ", alias.long_name(""), " to ", long_name(""),
                    ", but ", long_name(""), " contains alias fields"));

        af.physical_field = fields[index];
        af.attributes = fields[index].attributes;
      }
      else {
        assert(FALSE) else
          error("ERR_CSCO_ADDRMAP_ALIAS_MISMATCH\n",
                "aliasing ", alias.long_name(""), " to ", long_name(""),
                ", fields in alias entry must be csco_addrmap_alias_field_s type");
      };
    };
  };

  // Adds a field to the list of fields for this entry.
  add_field(f : csco_addrmap_field_s) is {
    for each (fld) in fields {
      assert that (fld.lsb_pos >= f.lsb_pos + f.width) or (fld.lsb_pos + fld.width <= f.lsb_pos) else
        error(appendf("CSCO_ADDRMAP_ERR_DUP_FLD_IND: %s field and %s field in %s have conflicting bit positions", fld.name, f.name, name));

      // check for duplicate names and bit positions.
      assert that fld.name != f.name else
        error(appendf("CSCO_ADDRMAP_ERR_DUP_FLD_NM: multiple fields named %s in %s", f.name, name));
    };

    fields.add(f);
  };

  // Looks up by name and returns index of a field in the entry's list of fields.
  get_field_ind(nm : string) : int is {
    result = fields.key_index(nm);
    assert that result != UNDEF else
      error(appendf("CSCO_ADDRMAP_ERR_BAD_FLD_NM: %s is not a legal field name for %s", nm, long_name("")));
  };

  // Looks up by name and returns pointer to field in the entry's list of fields.
  get_field(nm : string) : csco_addrmap_field_s is {
    result = fields[get_field_ind(nm)];
  };

  // Sets the reset value of the designated field.
  set_field_reset_val(fname : string, val : int(bits:*)) is {
    get_field(fname).set_reset_val(val);
  };

  // Returns the reset value of the designated field.
  get_field_reset_val(fname : string) : int(bits:*) is {
    result = get_field(fname).get_reset_val();
  };

  // Sets the shadow value of the designated field.
  set_field_shadow_val(fname : string, val : int(bits:*)) is {
    get_field(fname).set_shadow_val(val);
  };

  // Returns the shadow value of the designated field.
  get_field_shadow_val(fname : string) : int(bits:*) is {
    result = get_field(fname).get_shadow_val();
  };

  parse_field_val(fname : string, reg_val : int(bits:*)) : int(bits:*) is {
    var f := get_field(fname);
    result = reg_val[f.lsb_pos + f.width - 1 : f.lsb_pos];
  };

  // Packs the shadow or reset values of all the fields in this entry
  // into a single integer value and returns it.
  pack_fields(sel : csco_addrmap_field_packing_sel_t) : int(bits:*) is {
    for each (f) in fields {
      if sel == SHADOW {
        result[f.lsb_pos + f.width - 1 : f.lsb_pos] = f.get_shadow_val();
      }
      else if sel == RESET {
        result[f.lsb_pos + f.width - 1 : f.lsb_pos] = f.get_reset_val();
      };
    };
  };

  // Takes a single integer value and unpacks it into all the fields of 
  // this entry.
  unpack_fields(sel : csco_addrmap_field_packing_sel_t, entry_val : int(bits:*)) is {
    for each (f) in fields {
      if sel == SHADOW {
        f.set_shadow_val(entry_val[f.lsb_pos + f.width - 1 : f.lsb_pos]);
      }
      else {
        f.set_reset_val(entry_val[f.lsb_pos + f.width - 1 : f.lsb_pos]);
      };
    };
  };

  // The update_shadow method will compare read transaction reponses with the
  // expected shadow values and flag any differences as errors. The method 
  // will also update the shadow values affected by the transaction, taking 
  // into consideration all side-effects. 
  // <br>
  // The input to the method is a csco_addrmap transaction. This can be
  // interpreted in two ways:
  // <br>
  // 1 - The transaction for the specific interface on which it is observed
  //     is inherited from the base csco_addrmap_trans_s. In this case,
  //     the monitor simply passes in the monitored transaction as the
  //     argument, with all the base fields set appropriately.
  // <br>
  // 2 - The transaction for the specific interface on which it is observed
  //     is defined independently of the csco_addrmap_trans_s. In this case,
  //     the input argument can be considered as an aggregation of all the 
  //     information that are required by update_shadow() to do its job. It 
  //     just happens that all that information is already encapsulated in the
  //     base csco_addrmap_trans_s. The monitor in this case should translate
  //     the relevant fields from the monitored transaction into a base
  //     csco_addrmap_trans_s instance and pass it to this method.
  // <br>
  // The addrmap transaction passed to the method must have the following fields
  // correctly set:
  // <br>
  // opcode, num_bytes, data, data_mask
  // <br>
  // Note that data_mask is only required for write operations.
  update_shadow(trans : csco_addrmap_trans_s) is {
    check_trans(trans);
    if trans.read_or_write() == WRITE {
      messagef(HIGH, "updating shadow value of %s due to %s", long_name(""), trans.opcode);
    }
    else {
      messagef(LOW, "checking against and updating shadow value of %s due to %s", long_name(""), trans.opcode);
    };
  };

  // Returns the shadow value requested by transaction as a list of
  // bytes. 
  get_shadow_bytes(trans : csco_addrmap_trans_s) : list of byte is {
    check_trans(trans);
  };

  check_trans(trans : csco_addrmap_trans_s) is {
    trans.find_leaf();

    assert that trans.get_entry_offset() + trans.num_bytes <= get_entry_size() else
      error("CSCO_ADDRMAP_ERR_UPDATE_SHADOW: shadow request for ", long_name(""), 
              "exceeds entry boundary:\n", 
              "start_addr   = ", hex(trans.get_entry_offset()), "\n",
              "num_bytes    = ", trans.num_bytes, "\n",
              "entry size   = ", hex(get_entry_size()));

    assert that trans.num_bytes == trans.data.size() and trans.num_bytes == trans.data_mask.size() else
      error("CSCO_ADDRMAP_ERR_UPDATE_SHADOW: shadow request for ", long_name(""),
            " has mismatch between num_bytes and actual data/mask size:\n",
            " num_bytes         = ", trans.num_bytes, "\n", 
            " data.size()       = ", trans.data.size(), "\n", 
            " data_mask.size()  = ", trans.data_mask.size());
  };

  csco_short_name() : string is {
    if me is a csco_addrmap_reg_entry_s {
      result = append("REG: ", name);
    }
    else {
      result = append("MEM: ", name);
    };
  };

  // Returns full name scope path of entry.  The user may provide a 
  // starting node from which to begin the path. If "" is specified, 
  // then the method will start the path from the root of the name 
  // scope tree. This method can be used as the unique path-name 
  // 'entry' string to an addrmap_do action.
  // 
  long_name(start_node : string) : string is {
    result = appendf("%s.%s", parent_scope.long_name(start_node), name);
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  // When non-NULL, this entry is an alias entry, i.e.: its fields should be of type
  // csco_addrmap_alias_field_s.
  package !aliased_from : csco_addrmap_entry_s;
    csco_print_attr aliased_from SHORT enable;
    csco_print_attr aliased_from ALL precede=fields;
    csco_print_get_struct_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "aliased_from" and pf.format == SHORT {
        result = (aliased_from == NULL ? "No" : "Yes");
      };
    };

  // used for coverage.
  package !cur_access : csco_addrmap_trans_s;
    csco_print_attr cur_access ALL disable;

  package event collect_cover;
}; // struct csco_addrmap_entry_s


// A register entry in the traditional sense (i.e. implemented via flops).
struct csco_addrmap_reg_entry_s like csco_addrmap_entry_s {
  long_name(start_node : string) : string is also {
    result = append("REG: ", result);
  };

  // For register entries, the entry size is the same as the entry width.
  get_entry_size() : csco_addrmap_addr_t is {
    result = get_entry_width();
  };

  get_entry_depth() : csco_addrmap_addr_t is {
    result = 1;
  };

  // This method returns the hdl signal name for a field within the register,
  // based on BluePrint naming conventions.
  get_field_hdl_path(fn : string) : string is {
    var reg_name : string;
    reg_name = (aliased_from == NULL) ? name : aliased_from.name;

    result = parent_region.get_full_hdl_path();
    var field_hdl_name := get_field(fn).get_hdl_name();
    if(str_sub(result, str_len(result) - 1,  1) != "_") {
      result = appendf("%s.%s_%s", result, reg_name, field_hdl_name);
    } else {
      result = appendf("%s%s_%s", result, reg_name, field_hdl_name);
    };
  };

  // Set reset value of the entire register as a single integer value.
  set_reg_reset_val(val : int(bits:*)) is {
    unpack_fields(RESET, val);
  };

  // Returns reset value of the entire register as a single integer value.
  get_reg_reset_val() : int(bits:*) is {
    result = pack_fields(RESET);
  };

  // Set shadow value of the entire register as a single integer value.
  set_reg_shadow_val(val : int(bits:*)) is {
    unpack_fields(SHADOW, val);
  };

  // Returns shadow value of the entire register as a single integer value.
  get_reg_shadow_val() : int(bits:*) is {
    result = pack_fields(SHADOW);
  };

  // Note that the '@n' notation is used for backdoor access
  // to achieve immediate effect.

  // Writes the desginated field via back-door.
  write_field_bkdoor(fname : string, val : int(bits:*)) is {
    var hpath : string = get_field_hdl_path(fname);
    if hpath !~ "/^~/" {
      hpath = append("~/", hpath);
    };
    messagef(CSCO_ADDRMAP, MEDIUM, "bkdoor writing %s in %s\n", fname, long_name(""));
    messagef(CSCO_ADDRMAP, MEDIUM, "-hdl_path=%s\n", hpath);
    messagef(CSCO_ADDRMAP, MEDIUM, "-value=0x%x\n", val);
    '(hpath)@n' = val;
  };

  // Writes the entire register via back-door.
  write_reg_bkdoor(val : int(bits:*)) is {
    for each (f) in fields {
      if (f.has_attribute(IS_WIRE)) {
        messagef(CSCO_ADDRMAP, MEDIUM, "skipped write to a verilog wire\n");
      } else {
        write_field_bkdoor(f.name, (val >> f.lsb_pos)[f.width-1:0]);
      };
    };
  };

  // Reads the desginated field via back-door.
  read_field_bkdoor(fname : string) : int(bits:*) is {
    var hpath : string = get_field_hdl_path(fname);
    if hpath !~ "/^~/" {
      hpath = append("~/", hpath);
    };
    result = '(hpath)';
    messagef(CSCO_ADDRMAP, MEDIUM, "bkdoor reading %s in %s\n", fname, long_name(""));
    messagef(CSCO_ADDRMAP, MEDIUM, "-hdl_path=%s\n", hpath);
    messagef(CSCO_ADDRMAP, MEDIUM, "-value=0x%x\n", result);
  };

  // Returns the entire register via back-door.
  read_reg_bkdoor() : int(bits:*) is {
    for each (f) in fields {
      result[f.lsb_pos+f.width-1:f.lsb_pos] = read_field_bkdoor(f.name);
    };
  };

  // See method description in csco_addrmap_entry_s base type.
  update_shadow(trans : csco_addrmap_trans_s) is also {
    var trans_val : int(bits:*);

    if trans.endian == BIG {
      trans_val = (trans.data.reverse())[:];
    }
    else {
      trans_val = trans.data[:];
    };
    for each (f) in fields {
      f.update_shadow(me, trans.read_or_write(), (trans_val >> f.lsb_pos)[f.width-1:0]);
    };

    messagef(CSCO_ADDRMAP, LOW, "updated shadow value of %s due to %s: 0x%x\n", long_name(""), trans.opcode, get_reg_shadow_val());
    messagef(CSCO_ADDRMAP, HIGH, "") {
      csco_print(LONG);
    };
  }; // update_shadow()

  get_shadow_bytes(trans : csco_addrmap_trans_s) : list of byte is {
    var tmp_int := get_reg_shadow_val();
    if trans.endian == BIG {
      unpack(packing.high, tmp_int[..get_entry_width()-1], result);
    }
    else {
      unpack(packing.low, tmp_int[..get_entry_width()-1], result);
    };
  };
}; // struct csco_addrmap_reg_entry_s


// The definition of a memory in our addrmap eVC includes any set of storage bytes
// that share a common row format. Each row could be a single byte or consist of 
// several logical fields spanning several bytes.  It encompasses both true memories 
// and flop arrays. 
struct csco_addrmap_mem_entry_s like csco_addrmap_entry_s {
  long_name(start_node : string) : string is also {
    result = append("MEM: ", result);
  };

  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Number of rows in this memory entry. For byte stream 
  // memories, each row has 1 byte.
  !depth : csco_addrmap_addr_t;
    csco_print_attr depth ALL precede=fields;

  // If FALSE, then will randomize reset value of memory.
  !initialized : bool;
    csco_print_attr initialized ALL precede=fields;

  // User should extend this method to parse name of mem. entry and execute bkdoor 
  // write accordingly.
  write_mem_bkdoor(trans : csco_addrmap_trans_s) is undefined;

  // User should extend this method to parse name of mem. entry and execute bkdoor 
  // read accordingly.
  read_mem_bkdoor(trans : csco_addrmap_trans_s) : list of byte is undefined;


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  link_alias_entry(alias : csco_addrmap_entry_s) is also {
    if alias is a csco_addrmap_mem_entry_s (mem_entry) {
      assert that mem_entry.depth == depth else
        error(appendf("CSCO_ADDRMAP_ERR_ALIAS_MISM: linking %s and %s as aliases, but they have different memory depths\n",
                        long_name(""), mem_entry.long_name("")));

      mem_entry.shadow_mem = shadow_mem;
    }
    else {
      error(appendf("CSCO_ADDRMAP_ERR_ALIAS_MISM: linking %s and %s as aliases, but they have different entry types\n",
                      long_name(""), mem_entry.long_name("")));
    };
  };

  // Returns total # of bytes in memory entry.
  get_entry_size() : csco_addrmap_addr_t is {
    result = depth * get_entry_width();
  };

  get_entry_depth() : csco_addrmap_addr_t is {
    result = depth;
  };

  // Pushes data in 'fields' into shadow memory. Value of each row 
  // between 'first_row' and 'last_row' are replaced with the 
  // contents of 'fields'.
  pack_mem_fields(first_row : csco_addrmap_addr_t, last_row : csco_addrmap_addr_t) is {
    var tmp_int : int(bits:*);
    var bit_list : list of bit;
    var byte_list : list of byte;
    var write_data : list of byte;

    tmp_int = pack_fields(SHADOW);
    bit_list = tmp_int[..get_entry_width() * 8 - 1];
    unpack(packing.low, bit_list, byte_list);

    for i from first_row to last_row {
      write_data.add(byte_list);
    };
    shadow_mem.write(first_row * get_entry_width(), write_data);
  }; // pack_mem_fields()

  // Reads a specified row out of shadow memory and packs data 
  // into 'fields'. Note that if the sparse shadow memory row
  // has not been previously written, then the method will return 
  // the reset value of the row.
  unpack_mem_fields(row : csco_addrmap_addr_t) is {
    var byte_list : list of byte;

    // read values from shadow memory if entry has previously been
    // written, otherwise just return reset value for entry.
    if shadow_mem.check_alloc((row * get_entry_width())) {
      byte_list = shadow_mem.read(row * get_entry_width(), get_entry_width());
      unpack_fields(SHADOW, byte_list[:]);
    }
    else {
      unpack_fields(SHADOW, pack_fields(RESET));
    };
  }; // unpack_mem_fields()

  // See method description in csco_addrmap_entry_s base type.
  update_shadow(trans : csco_addrmap_trans_s) is also {
    var first_row : uint = (trans.get_entry_offset()) / get_entry_width();
    var last_row : uint = (trans.get_entry_offset() + trans.num_bytes - 1) / get_entry_width();
    var cur_ind : int;
    var row_val : int(bits:*);
    var exp_row_data : int(bits:*);

    if trans.get_entry_offset() % get_entry_width() != 0 {
      dut_errorf("CSCO_ADDRMAP_ERR_SHDW_ADDR_ALIGN: starting address of transaction is not aligned with memory width\n");
    };

    if trans.num_bytes % get_entry_width() != 0 {
      dut_errorf("CSCO_ADDRMAP_ERR_SHDW_SIZE_ALIGN: transaction size is not aligned with memory width\n");
    };

    for r from first_row to last_row {
      cur_ind = ((r - first_row) * get_entry_width()) % trans.data.size();
      if trans.endian == BIG {
        row_val = (trans.data[cur_ind..cur_ind+get_entry_width()-1].reverse())[:];
      }
      else {
        row_val = trans.data[cur_ind..cur_ind+get_entry_width()-1][:];
      };
      unpack_mem_fields(r);
      messagef(HIGH, "checking against and updating shadow value of row 0x%x of %s due to %s", r, long_name(""), trans.opcode);
      if trans.read_or_write() != READ {
        if trans.data_mask[cur_ind] == 1 {
          if trans.data_mask[cur_ind..cur_ind+get_entry_width()-1].has(it == 0) {
            dut_errorf("CSCO_ADDRMAP_ERR_SHDW_INVLD_MASK: non-uniform write mask detected for row %d of %s memory entry\n", r, name);
          };
        }
        else {
          if trans.data_mask[cur_ind..cur_ind+get_entry_width()-1].has(it == 1) {
            dut_errorf("CSCO_ADDRMAP_ERR_SHDW_INVLD_MASK: non-uniform write mask detected for row %d of %s memory entry\n", r, name);
          };

          // If mask for this row is disabled, then revert to shadow value for write.
          row_val = pack_fields(SHADOW);
        };
      }; // WRITE

      for each (f) in fields {
        f.update_shadow(me, trans.read_or_write(), (row_val >> f.lsb_pos)[f.width-1:0]);
      };
      pack_mem_fields(r, r);
      messagef(CSCO_ADDRMAP, LOW, "updated shadow value of row 0x%x of %s due to %s: 0x%x\n", r, long_name(""), trans.opcode, row_val);
    }; // for r from first_row to last_row
  }; // update_shadow()

  // Returns current shadow value of the memory starting at the given
  // offset.
  // Note bytes are fetched from the shadow memory as a series
  // of row reads. If a row is not allocated, then the reset
  // value for the fields are returned. This removes the need to 
  // unnecessarily populate the shadow memory if an entire memory 
  // is initialized upon reset.
  get_shadow_bytes(trans : csco_addrmap_trans_s) : list of byte is {
    var first_row : uint = trans.get_entry_offset() / get_entry_width();
    var last_row : uint = (trans.get_entry_offset() + trans.num_bytes - 1) / get_entry_width();
    var find : uint = trans.get_entry_offset() % get_entry_width();
    var lind : uint = find + trans.num_bytes - 1;
    var byte_list : list of byte;

    for i from first_row to last_row {
      unpack_mem_fields(i);             // will use reset value if a row is not yet allocated.
      if trans.endian == BIG {
        unpack(packing.high, pack_fields(SHADOW)[..(get_entry_width()*8)-1], byte_list);
      }
      else {
        unpack(packing.low, pack_fields(SHADOW)[..(get_entry_width()*8)-1], byte_list);
      };
      result.add(byte_list);
    };
    result = result[find..lind];
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  // Efficient shadow memory implementation. Each logical row is stored
  // in LITTLE-ENDIAN format.
  package !shadow_mem : csco_blk_memory;

  package !init_val : list of bit;

  run() is also {
    // Allocate shadow memory struct if not already done.
    if shadow_mem == NULL {
      shadow_mem = new;
    };

    if not initialized {
      // randomize power-on value of memory block.
      for each (f) in fields {
        gen init_val keeping {
          .size() == read_only(f.width);
        };

        f.set_reset_val(init_val[:]);
        f.set_shadow_val(init_val[:]);
      };
    };
  }; // run()
}; // struct csco_addrmap_mem_entry_s


// Byte memory stores a byte stream which has no embedded logical structure.
// Note that the 'fields' field is unused for byte memories and remains 
// empty.
struct csco_addrmap_byte_mem_entry_s like csco_addrmap_mem_entry_s {
  ////////////////////////
  // Private Fields 
  ////////////////////////
  init() is also {
    width = 1;
  };

  run() is first {
    if fields.is_empty() {
      var byte_field : csco_addrmap_field_s;
      byte_field = new csco_addrmap_narrow_field_s with {
        .name = "BYTE_FIELD";
        .width = 8;
        .lsb_pos = 0;
        .sw_access = RW;
      };
      fields.add(byte_field);
    }
    else {
      assert that fields.size() == 1 else 
        error("there can only be 1 field in a byte memory, found ", fields.size());

      assert that fields[0].width == 8 else 
        error("the field width for a byte memory must be 8, currently set as ", fields[0].width);

      assert that fields[0].lsb_pos == 0 else 
        error("the LSB position for a byte memory field must be 0, currently set as ", fields[0].lsb_pos);
    };
  }; // run() 
}; // struct csco_addrmap_byte_mem_entry_s like csco_addrmap_mem_entry_s


'>
