// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_env.e
//              :
// Title        : csco_addrmap package Env.
// Project      : csco_addrmap
//              :
// Created On   : 09/26/2005
//              :
//              :
// Purpose      : contains the root name scope and an addrmap sequence driver.
//              :
// Description  : 
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

<'
package csco_addrmap;


extend csco_addrmap_driver_u {
  // Back pointer to parent address map env.
  addrmap : csco_addrmap_env_u;
    csco_print_attr addrmap ALL disable;
};

extend csco_addrmap_space_u {
  addrmap : csco_addrmap_env_u;
    csco_print_attr addrmap ALL disable;
};

unit csco_addrmap_env_u like any_env {
  CREATE_TESTFLOW;

  package !all_space_names : list of csco_addrmap_space_t;
  pre_generate() is also {
    all_space_names = all_values(csco_addrmap_space_t).all(it != UNDEFINED);
  };

  spaces : list of csco_addrmap_space_u is instance;
    keep spaces.size() == all_space_names.size();
    keep for each (s) in spaces {
      s.name == read_only(all_space_names[index]);
      s.addrmap == me;
    };

  // The is the root of the name scope tree for the testbench under
  // which the addrmap_env is instantiated.
  scope : csco_addrmap_name_scope_s;
    keep soft scope.name == "ROOT";

  package !phy_regions : list of csco_addrmap_physical_region_s;

  event addrmaps_build_done;
  !addrmaps_built : bool;

  csco_regenerate_unit_data() is also {
    addrmaps_built = FALSE;
    for each (s) in spaces {
      s.clear_space();
    };
  };

  get_space(n : csco_addrmap_space_t) : csco_addrmap_space_u is {
    result = spaces.first(.name == n);
  };

  // Instantiates addrmap driver.
  driver : csco_addrmap_driver_u is instance;
    keep driver.default_scope == read_only(scope);
    keep driver.addrmap == me;

  short_name(): string is {
    result = "ADDRMAP";
  }; 

  // This method MUST be called for a region to be recognized by the addrmap.
  add_region(r : csco_addrmap_region_s) is {
    assert that !addrmaps_built else 
      error("CSCO_ADDRMAP_ERR_REGION_CREATED_AFTER_BUILD_ADDRMAPS:\n",
              "csco_addrmap regions must be created BEFORE build_addrmaps() is called in the csco_addrmap env unit.");

    if r is a csco_addrmap_logical_region_s (log_r) {
      assert that log_r.addr_space != UNDEFINED else
        error("CSCO_ADDRMAP_ERR_INVALID_REGION_SPACE:\n",
              log_r.name, " logical region created with UNDEFINED region space, please constrain a valid region space.");

      var space_ptr : csco_addrmap_space_u = get_space(log_r.addr_space);
      space_ptr.add_region(r);
    }
    // physical regions may belong to multiple address spaces and are 
    // populated in the address spaces after build_addrmaps().
    else if r is a csco_addrmap_physical_region_s (phy_r) {
      assert that phy_r not in phy_regions else 
        error("CSCO_ADDRMAP_ERR_DUP_PHY_REGN: \n",
              r.csco_short_name(), " already exists as a physical region in ", short_name_path());

      phy_regions.add(phy_r);
    };
  };

  delete_region(r : csco_addrmap_region_s) is {
    var del_ind : int;
    if r is a csco_addrmap_logical_region_s (log_r) {
      var space_ptr : csco_addrmap_space_u = get_space(log_r.addr_space);
      del_ind = space_ptr.log_regions.first_index(it == log_r);
      assert that del_ind != UNDEF else
        error("CSCO_ADDRMAP_ERR_INVALID_REGION_DELETION:\n",
                "trying to delete a non-existent logical region, ", r.csco_short_name(),
                " from ", space_ptr.short_name_path());

      space_ptr.delete_region(r);
    }
    else if r is a csco_addrmap_physical_region_s (phy_r) {
      del_ind = phy_regions.first_index(it == phy_r);
        error("CSCO_ADDRMAP_ERR_INVALID_REGION_DELETION:\n",
                "trying to delete a non-existent physical region, ", r.csco_short_name(),
                " from the address map.");

      phy_regions.fast_delete(del_ind);
      for each (s) in spaces {
        s.delete_region(r);
      };
    };
  };

  // Returns all entries in the address map.
  get_all_entries() : list of csco_addrmap_entry_s is {
    result = scope.get_all_child_entries();
  };

  // This method should be called once the user has hooked up all the 
  // entries and regions, and is ready to generate the system map.
  build_addrmaps() is {
    message(LOW, "printing scope tree for ", scope.long_name("")) {
      scope.csco_print_sub_tree(SHORT, "child_scopes");
    };

    messagef(CSCO_ADDRMAP, HIGH, "build_addrmaps() called\n");

    // call build_addrmap() for all the address spaces.
    for each (s) in spaces {
      message(CSCO_ADDRMAP, HIGH, "calling build_addrmap() of ", s.short_name_path());
      s.pre_build_addrmap();
      s.build_addrmap();
      s.post_build_addrmap();
      s.check_addrmap();
    };

    // Ensure every entry has both a parent scope and a parent physical region. This 
    // is checked in the env because physical regions and entries can span multiple
    // spaces.
    for each (e) in get_all_entries() {
      assert that e.parent_scope != NULL else
        error(appendf("CSCO_ADDRMAP_ERR_MISSING_PARENT_SCOPE: %s is missing a parent scope", e.csco_short_name()));

      assert that e.parent_region != NULL else
        error(appendf("CSCO_ADDRMAP_ERR_MISSING_PARENT_REGION: %s is missing a parent region", e.csco_short_name()));
    };

    for each (r) in phy_regions {
      assert that r.parent_contexts.size() > 0 else
        error("CSCO_ADDRMAP_ERR_DANGLING_PHY_REGION: ", r.csco_short_name(), "is not associated with ANY logical regions");

      for each (p) in r.parent_contexts {
        get_space(p.region.get_addr_space()).add_region(r);
      };
    };

    for each (s) in spaces {
      outf("\n");
      message(LOW, "Below is the address map for ", s.short_name_path(), " following a call to build_addrmap()") {
        s.csco_print(LONG);
      };
    };

    addrmaps_built = TRUE;
    emit addrmaps_build_done;
  };

  dut_error_dump() is {
    out("printing scope tree for ", scope.long_name(""));
    scope.csco_print_sub_tree(LONG, "child_scopes");
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  init() is also {
    if all_values(csco_addrmap_side_effect_t).size() > 32 {
      error(appendf("CSCO_ADDRMAP_ERR_SE_NUM: more than 32 (%d) csco_addrmap side effects have been defined\n", all_values(csco_addrmap_side_effect_t).size()));
    };
  };
}; // unit csco_addrmap_env_u


'>
