// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_id_routing.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 09/26/2005
//              :
//              :
// Description  : Extends csco_addrmap eVC to support address spaces with ID 
//              : based routing.
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

<'
package csco_addrmap;


// The 'ID' region route type is used to model ID routed
// address spaces.
extend csco_addrmap_route_kind_t : [ID];

struct csco_addrmap_region_id_s {
  addr_space : csco_addrmap_space_t;

  compare(targ : csco_addrmap_region_id_s) : bool is empty;
};

extend csco_addrmap_addr_s {
  get_id() : csco_addrmap_region_id_s is empty;
  set_id(csco_addrmap_region_id_s) is empty;
};

extend ID'route_kind csco_addrmap_addr_s {
  id : csco_addrmap_region_id_s;
    keep id.addr_space == read_only(addr_space);

  run() is also {
    if id == NULL {
      id = new csco_addrmap_region_id_s with {
        .addr_space = addr_space;
      };
    };
  };

  get_id() : csco_addrmap_region_id_s is { result = id; };

  set_id(i : csco_addrmap_region_id_s) is { id = i; };

  csco_short_name() : string is also {
    result = append(id.csco_short_name(), "/", result);
  };
};


extend csco_addrmap_region_s {
  get_id() : csco_addrmap_region_id_s is empty;
};

// Each ID routed region models a node in the ID routed address space. 
extend ID'route_kind csco_addrmap_logical_region_s {
  id : csco_addrmap_region_id_s;
    keep id.addr_space == read_only(addr_space);

  get_id() : csco_addrmap_region_id_s is { result = id; };

  get_connected_regions() : list of csco_addrmap_region_s is {
    var roots := get_all_parent_regions().all(.max_depth == 0);
    if me.max_depth == 0 { roots.add(me); };
    result.add(roots);
    for each (r) in roots {
      result.add(r.get_all_child_regions());
    };
  };

  get_connected_log_regions() : list of csco_addrmap_logical_region_s is {
    var conn_regions := get_connected_regions().all(it is a csco_addrmap_logical_region_s);
    result = conn_regions.as_a(list of csco_addrmap_logical_region_s);
  };

  get_connected_phy_regions() : list of csco_addrmap_physical_region_s is {
    var conn_regions := get_connected_regions().all(it is a csco_addrmap_physical_region_s);
    result = conn_regions.as_a(list of csco_addrmap_physical_region_s);
  };

  // The size of ID routed regions is simply the max size of the physical regions
  // in the region space. This is because every region is reachable via ID from 
  // every other region.
  get_region_size() : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1) is {
    for each (r) in get_connected_phy_regions() {
      result = max(result, r.get_parent_context(r).addr.offset + r.get_region_size());
    };
  };

  package get_child_addr(child : csco_addrmap_region_addr_s) : csco_addrmap_region_addr_s is also {
    if child.region is a csco_addrmap_physical_region_s (phy_r) {
      result.addr.route_kind = ID;
      result.addr.as_a(ID csco_addrmap_addr_s).id = me.id;
    };
  };

  get_region_path(child : csco_addrmap_region_s) : list of csco_addrmap_region_s is only {
    if child == me {
      result = {me};
    }
    else {
      result = {me; child};
    };
  };

  get_driving_regions(addr : csco_addrmap_addr_s) : list of csco_addrmap_region_addr_s is {
    for each (r) in get_connected_log_regions() {
      if r.get_eligible_driver() != NULL {
        var tmp_region_addr := new csco_addrmap_region_addr_s with {
          .region = r;
        };
        if addr.route_kind != ID {
          tmp_region_addr.addr.set_id(id);
          tmp_region_addr.addr.offset = addr.offset;
        }
        else {
          tmp_region_addr.addr = addr;
        };
        result.add(tmp_region_addr);
      };
    };
  };

  get_leaf(addr : csco_addrmap_addr_s) : csco_addrmap_leaf_entry_s is also {
    var tmp_ra : csco_addrmap_region_addr_s;
    var log_r := find_id_region(addr);
    assert log_r != NULL else
      error("CSCO_ADDRMAP_ERR_BAD_REGION_ID:\n",
            "The region ID ", addr.get_id().csco_short_name(), " doesn't exist in the ", 
            addr_space, " address space.");

    result = new;
    result.logical = new csco_addrmap_region_addr_s with {
      .region = log_r;
      .addr = addr;
    };

    if log_r.id.compare(addr.get_id()) {
      tmp_ra = log_r.parent_offset_to_child(addr);
      if tmp_ra != NULL and tmp_ra.region is a csco_addrmap_physical_region_s (phy_r) {
        result.physical = tmp_ra;
      };

      if result.physical != NULL {
        result.entry = result.physical.region.as_a(csco_addrmap_physical_region_s).offset_to_entry(result.physical.addr);
      };
    };
  };

  // Protocol specific ID routed region implementations can/should extend
  // or override this method to use a more specific/efficent  algorithm.
  find_id_region(addr : csco_addrmap_addr_s) : ID'route_kind csco_addrmap_logical_region_s is {
    result = get_connected_log_regions().first(.get_id().compare(addr.get_id())).as_a(ID'route_kind csco_addrmap_logical_region_s);
  };

  // The build_addrmap() methods should be extended for each relevant
  // addr_space subtype. 
}; // extend ID'route_kind csco_addrmap_logical_region_s


'>
