<'
package csco_addrmap;

import csco_arithm/e/csco_arithm_top;


type csco_addrmap_segmentation_alg_t : [ALIGNED_RANDOM, GREEDY];

type csco_addrmap_intf_alignment_op_kind_t : [ALL, READ, WRITE];


// This struct describes the set of alignment and size parameters that
// constitutes a legal transaction on any addrmap interface:
struct csco_addrmap_intf_alignment_s {
  interface : csco_addrmap_interface_t;
    keep soft interface == NONE;

  op_kind : csco_addrmap_intf_alignment_op_kind_t;
    keep soft op_kind == ALL;

  // The natural byte-alignment of the interface. <br>
  alignment : csco_addrmap_addr_t;
    keep soft alignment == 0;

  // The maximum transaction size in bytes. <br>
  max_size : csco_addrmap_addr_t;
    keep soft max_size == 0;

  // The precision of byte-enables within the interface's natural alignment. <br>
  mask_precision : csco_addrmap_addr_t;
    keep mask_precision <= alignment;
    keep soft mask_precision == 1;

  // A hard page boundary that a single transaction cannot cross. 0 indicates no boundary. <br>
  page_boundary : csco_addrmap_addr_t;
    keep soft page_boundary == 0;

  csco_short_name() : string is {
    result = append(interface, " ", op_kind, " ", dec(alignment), "-byte alignment (max=", dec(max_size), 
                    ", mask_precision=", dec(mask_precision), 
                    ", page_boundary=", dec(page_boundary), ")");
  };

  compare(a : csco_addrmap_intf_alignment_s) : bool is {
    result = (alignment == a.alignment and 
              max_size == a.max_size and
              mask_precision == a.mask_precision and
              page_boundary == a.page_boundary);
  };

  post_generate() is also {
    if interface != NONE {
      assert that alignment != 0 else 
        error("CSCO_ADDRMAP_ERR_NO_IF_MIN_ALIGNMENT: the alignment requirement for ", interface, " interface has not been constrained.");

      assert that max_size != 0 else 
        error("CSCO_ADDRMAP_ERR_NO_IF_MAX_TRANS_SIZE: the max_size requirement for ", interface, " interface has not been constrained.");

      assert that max_size % alignment == 0 else
        error("CSCO_ADDRMAP_ERR_IF_ALIGNMENT_AND_SIZE: the max_size, ", max_size, ", for the ", 
                interface, " interface is not a multiple of its alignment, ", alignment);

      assert that page_boundary % alignment == 0 else
        error("CSCO_ADDRMAP_ERR_IF_ALIGNMENT_AND_SIZE: the page_boundary, ", page_boundary, ", for the ", 
                interface, " interface is not a multiple of its alignment, ", alignment);

      assert that alignment % mask_precision == 0 else
        error("CSCO_ADDRMAP_ERR_IF_ALIGNMENT_AND_SIZE: the alignment, ", alignment, ", for the ", 
                interface, " interface is not a multiple of its mask_precision, ", mask_precision);

      if page_boundary != 0 {
        assert that page_boundary >= max_size else
          error("CSCO_ADDRMAP_ERR_IF_ALIGNMENT_AND_SIZE: the page_boundary, ", page_boundary, ", for the ", 
                  interface, " is < its maxsize, ", max_size);
      };
    };
  }; 

  merge(a : csco_addrmap_intf_alignment_s) : csco_addrmap_intf_alignment_s is {
    if a.op_kind != ALL and op_kind != ALL and a.op_kind != me.op_kind {
      // Alignments are for different operation types.
      return NULL;
    };

    // common alignment is least common multiple of the 2 alignments.
    var common_align : csco_addrmap_addr_t;
    common_align = sys.csco_arithm.lcm(a.alignment, alignment);

    // common mask precision is least common multiple of the 2 mask precisions, 
    // and common align % common mask precision must be 0.
    var common_mask_p : csco_addrmap_addr_t;
    common_mask_p = sys.csco_arithm.lcm(a.mask_precision, mask_precision);
    if common_align % common_mask_p != 0 {
      return NULL;
    };

    // common max_size must % common_align, and it should be as close to 
    // the minimum max_size as possible.
    var common_msize : csco_addrmap_addr_t;
    common_msize = (min(a.max_size, max_size) / common_align) * common_align;
    if common_msize == 0 {
      return NULL;
    };

    // common boundary has to % common_align, and should be as close to 
    // the GCD of both page boundaries as possible.
    var common_page_b : csco_addrmap_addr_t;
    if a.page_boundary != 0 and page_boundary != 0 {
      common_page_b = sys.csco_arithm.gcd(a.page_boundary, page_boundary);
    }
    else {
      common_page_b = max(a.page_boundary, page_boundary);
    };
    common_page_b = (common_page_b / common_align) * common_align;
    if common_page_b == 0 and (a.page_boundary != 0 or page_boundary != 0) {
      return NULL;
    };

    gen result keeping {
      .alignment == common_align;
      .mask_precision == common_mask_p;
      .max_size == common_msize;
      .page_boundary == common_page_b;
    };
  };

  reduce(a : csco_addrmap_intf_alignment_s) : csco_addrmap_intf_alignment_s is {
    if a.op_kind != ALL and op_kind != ALL and a.op_kind != me.op_kind {
      // Alignments are for different operation types.
      return NULL;
    };

    if alignment % a.alignment == 0 and 
       (a.max_size >= max_size) and 
       (a.mask_precision <= mask_precision) and 
       (mask_precision % a.mask_precision == 0) and
       (a.page_boundary == 0 or (page_boundary != 0 and 
                                 a.page_boundary >= page_boundary and 
                                 a.page_boundary % page_boundary == 0)) {
      result = a;
    };

    if a.alignment % alignment == 0 and 
       (max_size >= a.max_size) and 
       (mask_precision <= a.mask_precision) and 
       (a.mask_precision % mask_precision == 0) and
       (page_boundary == 0 or (a.page_boundary != 0 and 
                               page_boundary >= a.page_boundary and 
                               page_boundary % a.page_boundary == 0)) {
      result = me;
    };
  };
}; // struct csco_addrmap_intf_alignment_s

extend csco_addrmap_region_s {
  get_legal_alignments(k : csco_addrmap_intf_alignment_op_kind_t) : list of csco_addrmap_intf_alignment_s is empty;
};

// Extend logical region struct to hold a list of legal
// transaction alignments.
extend csco_addrmap_logical_region_s {
  // List of legal transaction alignments for the interface.
  legal_alignments : list of csco_addrmap_intf_alignment_s;
    keep interface == NONE => legal_alignments.size() == 0;
    keep soft legal_alignments.size() == 1;
    keep for each (l) in legal_alignments {
      l.interface == read_only(interface);
    };

  get_legal_alignments(k : csco_addrmap_intf_alignment_op_kind_t) : list of csco_addrmap_intf_alignment_s is also {
    case (k) {
      READ : { result = legal_alignments.all(.op_kind in [READ, ALL]); };
      WRITE : { result = legal_alignments.all(.op_kind in [WRITE, ALL]); };
      ALL : { result = legal_alignments.all(.op_kind == ALL); };
    };
  };

  run() is also {
    if legal_alignments.is_empty() and interface != NONE {
      gen legal_alignments;
    };
  };
};

// This struct model a segment of an addrmap transaction that
// has been formatted to use one of the legal transaction alignments
// for the source interface. The addrmap transaction's iterator
// will do the segmentation.
struct csco_addrmap_intf_segment_s {
  package !intf_alignment : csco_addrmap_intf_alignment_s;

  // This is the offset into the original addrmap transaction
  // address range of the current segment..
  seg_offset : csco_addrmap_addr_t;
    csco_print_attr seg_offset SHORT enable;
    csco_print_attr seg_offset SHORT header=offset;

  // This is the number of bytes actually transferred for the
  // original addrmap transaction, not including any adjustment
  // overhead, for the current segment.
  !seg_size : csco_addrmap_addr_t;
    csco_print_attr seg_size SHORT enable;
    csco_print_attr seg_size SHORT header=size;

  package !initial_adjustment : csco_addrmap_addr_t;

  package !final_adjustment : csco_addrmap_addr_t;

  // The actual segmented transaction is modeled using an OFFSET
  // addrmap transaction.
  !offset_trans : csco_addrmap_trans_s;
    csco_print_attr offset_trans LONG expand=ON;
};

extend csco_addrmap_trans_s {
  // The intf_iterator will take a generic addrmap transaction, and based on 
  // its source interface and its legal alignments, segment the addrmap
  // transaction into appropriate sizes and alignments that the interface
  // driver can then send onto the bus.
  intf_iterator : csco_addrmap_intf_segment_iterator_s;
    attribute intf_iterator deep_copy = reference;

  when BACK_DOOR'access csco_addrmap_trans_s {
    keep intf_iterator == NULL;
  };

  when FRONT_DOOR'access csco_addrmap_trans_s {
    keep intf_iterator.addrmap_trans == read_only(me);
    keep intf_iterator.interface == read_only(interface);

    post_generate() is also {
      var offset_tr := deep_copy(me);
      intf_iterator.cur_segment.offset_trans = offset_tr;
      intf_iterator.cur_segment.offset_trans.addr_scheme = OFFSET;
      intf_iterator.cur_segment.offset_trans.origin = TMP;
      intf_iterator.cur_segment.offset_trans.opcode = (read_or_write() == READ) ? READ : WRITE;
      intf_iterator.src_tgt_diff = source.addr.offset - target.addr.offset;
      intf_iterator.entry_alignment = ((get_entry_ptr() == NULL) ? 1 : get_entry_ptr().get_entry_width());
      intf_iterator.cur_segment.seg_offset = source.addr.offset;
    };
  };
}; 

struct csco_addrmap_intf_segment_resp_s {
  !num_bytes : csco_addrmap_addr_t;

  !start_byte : csco_addrmap_addr_t;
  !end_byte : csco_addrmap_addr_t;

  !received : bool;

  !data : list of byte;
};

struct csco_addrmap_intf_resp_collector_s {
  !resp_kind : csco_addrmap_resp_kind_t;

  !segment : list of csco_addrmap_intf_segment_resp_s;
};

struct csco_addrmap_intf_segment_iterator_s {
  // Back pointer to original addrmap transaction.
  addrmap_trans : csco_addrmap_trans_s;
    keep soft addrmap_trans == NULL;
    csco_print_attr addrmap_trans SHORT enable;

  // Source interface.
  interface : csco_addrmap_interface_t;
    keep soft interface == NONE;

  // The natural alignment of the entry, which is simply its width.
  entry_alignment : csco_addrmap_addr_t;
    keep soft entry_alignment == 0;

  // A addrmap transaction can potentially be legally segmented in
  // multiple ways for a given interface. By default, we define a 
  // GREEDY algorithm via which we fit as much data as we can into
  // each segment. But the user may also define new algorithms via
  // new sub-types.
  seg_alg : csco_addrmap_segmentation_alg_t;
    csco_print_attr seg_alg SHORT enable;

  // The current segment returned by the iterator.
  cur_segment : csco_addrmap_intf_segment_s;
    csco_print_attr cur_segment SHORT enable;
    csco_print_attr cur_segment ALL expand=ON;
    csco_print_attr cur_segment SHORT header=cur;

  // # of segments returned so far.
  package !segment_cnt : uint;
    csco_print_attr segment_cnt SHORT enable;

  // # of bytes transferred so far.
  package !bytes_transferred : csco_addrmap_addr_t;

  // Used to store read responses of various segments from a read operation.
  package resp_collector : csco_addrmap_intf_resp_collector_s;
    csco_print_attr resp_collector LONG expand=ON;

  package !src_tgt_diff : csco_addrmap_addr_t;

  // This method will return the next segment to be driven by the
  // i/f driver. If all segments have already been returned, then 
  // NULL will be returned.
  get_next_segment() : csco_addrmap_trans_s is {
    var alignment_choices : list of csco_addrmap_intf_segment_s;
    var alignment_chosen : csco_addrmap_intf_segment_s;

    if bytes_transferred >= addrmap_trans.num_bytes  {
      return NULL;
    };
    cur_segment.seg_offset += cur_segment.seg_size;

    alignment_choices = select_alignment_and_size();
    assert that alignment_choices.size() > 0 else
      error("CSCO_ADDRMAP_ERR_NO_ALIGNMENT_CHOICES: ", csco_short_name(), 
            " cannot be mapped to a legal alignment on interface ", interface);
    gen alignment_chosen keeping {
      it in alignment_choices;
    };
    cur_segment.intf_alignment = alignment_chosen.intf_alignment;
    cur_segment.seg_size = alignment_chosen.seg_size;
    cur_segment.initial_adjustment = alignment_chosen.initial_adjustment;
    cur_segment.final_adjustment = alignment_chosen.final_adjustment;
    message(CSCO_ADDRMAP, HIGH, "selected ", cur_segment.intf_alignment.csco_short_name(), 
            " for segment ", dec(segment_cnt), " of ", addrmap_trans.csco_short_name());

    update_offsets();

    update_data_and_mask();
    result = cur_segment.offset_trans;
    message(CSCO_ADDRMAP, MEDIUM, "advancing ", addrmap_trans.csco_short_name(), "'s iterator to segment ", segment_cnt) {
      csco_print(LONG);
    };
    outf("\n");

    bytes_transferred += cur_segment.seg_size;
    segment_cnt += 1;

    if addrmap_trans.post_kind == NON_POSTED {
      // Add entry in response collector for transaction just issued.
      var resp_segment := new csco_addrmap_intf_segment_resp_s with {
        .num_bytes = cur_segment.offset_trans.num_bytes;
        .start_byte = cur_segment.initial_adjustment;
        .end_byte = cur_segment.initial_adjustment + cur_segment.seg_size - 1;
      };
      resp_collector.segment.add(resp_segment);
    };
  }; // get_next_segment()

  // This method returns all the next segment choices base on the chosen interface's 
  // legal alignments. This method can be extended by various segmentation algorithms 
  // to reduce its result based on the algorithm.
  package select_alignment_and_size() : list of csco_addrmap_intf_segment_s is {
    var tmp_segment : csco_addrmap_intf_segment_s;
    var region_path : list of csco_addrmap_region_s;
    var possible_alignments : list of csco_addrmap_intf_alignment_s;
    var op_kind : csco_addrmap_intf_alignment_op_kind_t;


    // extract legal alignments by considering all legal alignments from source
    // to target.
    op_kind = (addrmap_trans.read_or_write() == READ) ? READ : WRITE;
    assert that !addrmap_trans.source.region.get_legal_alignments(op_kind).is_empty() else
      error("CSCO_ADDRMAP_ERR_NO_SOURCE_REGION_LEGAL_ALIGNMENTS_SPECIFIED:\n",
            "source region ", addrmap_trans.source.region.csco_short_name(), 
            " for ", addrmap_trans.csco_short_name(), " does not contain any legal alignment specifications.");
    region_path = addrmap_trans.source.region.get_region_path(addrmap_trans.target.region);
    possible_alignments = region_path[0].get_legal_alignments(op_kind);
    if region_path.size() > 1 {
      var tmp_a : list of csco_addrmap_intf_alignment_s;
      for each (p) in region_path[1..] {
        tmp_a = p.get_legal_alignments(op_kind);
        if tmp_a.size() > 0 {
          possible_alignments = merge_alignments(possible_alignments, tmp_a);
        };
      };
    };

    assert that possible_alignments.size() > 0 else
      error("CSCO_ADDRMAP_ERR_NO_LEGAL_ALIGNMENT:\n",
            "There's no legal alignment for ", addrmap_trans.csco_short_name(), " targetting ", 
            addrmap_trans.target.region.csco_short_name(), " from ", addrmap_trans.source.region.csco_short_name());


    //possible_alignments = reduce_alignments(possible_alignments);

    message(CSCO_ADDRMAP, HIGH, "selecting legal alignment for segment ", segment_cnt, 
            " of ", addrmap_trans.csco_short_name(), " amongst the following choices:") {
      out(possible_alignments.apply(.csco_short_name()));
    };
    for each (la) in possible_alignments {
      var dist_to_page_boundary : csco_addrmap_addr_t;
      tmp_segment = new csco_addrmap_intf_segment_s with {
        .intf_alignment = la;
        .seg_offset = cur_segment.seg_offset;
      };

      // Determine maximum transaction size.
      tmp_segment.seg_size = select_segment_size(la, addrmap_trans.num_bytes - bytes_transferred);
      assert that tmp_segment.seg_size <= addrmap_trans.num_bytes - bytes_transferred else
        error("CSCO_ADDRMAP_ERR_SEGMENT_SIZE_EXCEEDS_REM_BYTES:\n",
              "segment size (", tmp_segment.seg_size, ") returned for alignment (",
              la.csco_short_name(), ") exceeds # of remaining bytes - ", addrmap_trans.num_bytes - bytes_transferred);
      if tmp_segment.seg_size == 0 {
        continue;
      };

      if la.page_boundary > 0 {
        dist_to_page_boundary = la.page_boundary - (cur_segment.seg_offset % la.page_boundary);
        tmp_segment.seg_size = min(tmp_segment.seg_size, dist_to_page_boundary);
      };

      // Determine initial adjustment required to satisfy natural alignment.
      tmp_segment.initial_adjustment = cur_segment.seg_offset % la.alignment;
      if tmp_segment.seg_size + tmp_segment.initial_adjustment > la.max_size {
        tmp_segment.seg_size -= (la.alignment - tmp_segment.initial_adjustment);
      };

      // Determine end adjustment required to satisfy natural alignment.
      tmp_segment.final_adjustment = (la.alignment - ((cur_segment.seg_offset + tmp_segment.seg_size) % la.alignment)) % la.alignment;
      if tmp_segment.final_adjustment > 0 {
        if tmp_segment.seg_size + tmp_segment.initial_adjustment + tmp_segment.final_adjustment > la.max_size {
          // Make sure we don't exceed the max size of the alignment.
          tmp_segment.seg_size -= (la.alignment - tmp_segment.final_adjustment);
          tmp_segment.final_adjustment = 0;
        }
        else if (bytes_transferred + tmp_segment.seg_size + tmp_segment.final_adjustment) > addrmap_trans.num_bytes {
          // If we selected a segment size that was less than actual remaining bytes,
          // but the final_adjustment takes us beyond remaining bytes, then we need
          // to adjust both segment size and final adjustment.
          tmp_segment.final_adjustment = (bytes_transferred + tmp_segment.seg_size + tmp_segment.final_adjustment) - addrmap_trans.num_bytes;
          tmp_segment.seg_size = addrmap_trans.num_bytes - bytes_transferred;
        }
        else {
          tmp_segment.seg_size += tmp_segment.final_adjustment;
          tmp_segment.final_adjustment = 0;
        };
      };

      // Inspect every mask word to filter out any illegal mask patterns.
      var total_words : uint;
      var all_mask_bits : list of bit;
      var cur_mask : int(bits:*);

      total_words = (tmp_segment.seg_size + tmp_segment.initial_adjustment + tmp_segment.final_adjustment) / la.alignment;
      all_mask_bits = get_segment_mask_bits(tmp_segment);

      cur_mask[la.alignment-1:0] = (all_mask_bits[0..la.alignment-1])[:];
      if !check_first_wd_mask(cur_mask, tmp_segment) {
        continue;
      };

      if total_words > 1 {
        cur_mask[la.alignment-1:0] = (all_mask_bits[la.alignment*(total_words-1)..])[:];
        if !check_last_wd_mask(cur_mask, tmp_segment) {
          continue;
        };
      };

      if total_words > 2 {
        var bad_mask : bool;
        for i from 1 to total_words - 2 {
          cur_mask[la.alignment-1:0] = all_mask_bits[la.alignment*i..((la.alignment)*(i+1))-1][:];
          if !check_mid_wd_mask(cur_mask, tmp_segment) {
            bad_mask = TRUE;
            break;
          };
        };

        if bad_mask { continue; };
      };

      result.add(tmp_segment);
    }; // for each (la) in possible_alignments

    result = filter_alignment_and_size(result);
  }; // select_alignment_and_size()

  // This method can be extended by different seg_alg to implement
  // different selection criteria.
  select_segment_size(legal_a : csco_addrmap_intf_alignment_s, rem_bytes : csco_addrmap_addr_t) : csco_addrmap_addr_t is {
    var max_s := min(rem_bytes, legal_a.max_size);
    gen result keeping {
      it > 0;
      it <= max_s;
    };
  };

  filter_alignment_and_size(cur_result : list of csco_addrmap_intf_segment_s) : list of csco_addrmap_intf_segment_s is empty;

  package merge_alignments(list1 : list of csco_addrmap_intf_alignment_s, 
                           list2 : list of csco_addrmap_intf_alignment_s): list of csco_addrmap_intf_alignment_s is {
    var tmp_a : csco_addrmap_intf_alignment_s;
    for each (a1) in list1 {
      for each (a2) in list2 {
        tmp_a = a1.merge(a2);
        if tmp_a != NULL and !result.has(.compare(tmp_a)) {
          result.add(tmp_a);
        };
      };
    };
  };

  package reduce_alignments(a : list of csco_addrmap_intf_alignment_s) : list of csco_addrmap_intf_alignment_s is {
    var tmp_ind : uint;
    var tmp_a : csco_addrmap_intf_alignment_s;
    result = a.copy();
    while tmp_ind < result.size() - 1 {
      for i from tmp_ind + 1 to result.size() - 1 {
        tmp_a = result[tmp_ind].reduce(result[i]);
        if tmp_a != NULL {
          result[tmp_ind] = tmp_a;
          result.delete(i);
        };
      };
      tmp_ind += 1;
    };
  };

  package get_segment_mask_bits(ts : csco_addrmap_intf_segment_s) : list of bit is {
    var specified_dsize := addrmap_trans.data.size();
    var iteration_offset : csco_addrmap_addr_t;
    var specified_offset : csco_addrmap_addr_t;
    var end_offset : csco_addrmap_addr_t;
    var iteration_size : csco_addrmap_addr_t;
    var tmp_blist : list of bit;

    // Initial padding bytes are always masked out.
    result.resize(ts.initial_adjustment);

    iteration_offset = ts.seg_offset - addrmap_trans.source.addr.offset;
    specified_offset = iteration_offset % specified_dsize;
    end_offset = iteration_offset + ts.seg_size;
    while iteration_offset < end_offset {
      iteration_size = specified_dsize - specified_offset;

      if iteration_offset < specified_dsize or addrmap_trans.fill_kind == REPLICATE or addrmap_trans.read_or_write() == READ {
        result.add(addrmap_trans.data_mask[specified_offset..]);
      }
      else {
        tmp_blist.resize(iteration_size, TRUE, 1, FALSE);
        result.add(tmp_blist);
      };

      iteration_offset += iteration_size;
      specified_offset = 0;
    }; // while

    // Trucate extra trailing bits from last iteration through 'for' loop.
    result.resize(ts.seg_size + ts.initial_adjustment, FALSE, 0, TRUE);

    // Final padding bytes are always masked out.
    tmp_blist.resize(ts.final_adjustment);
    result.add(tmp_blist);
  };

  package check_first_wd_mask(cm : int(bits:*), seg : csco_addrmap_intf_segment_s) : bool is {
    result = check_mid_wd_mask(cm, seg);
  };

  package check_last_wd_mask(cm : int(bits:*), seg : csco_addrmap_intf_segment_s) : bool is {
    result = check_mid_wd_mask(cm, seg);
  };

  package check_mid_wd_mask(cm : int(bits:*), seg : csco_addrmap_intf_segment_s) : bool is {
    if seg.intf_alignment.mask_precision > 1 {
      for i from 0 to (seg.intf_alignment.alignment / seg.intf_alignment.mask_precision) -1 {
        if cm[0:0] == 0 {
          if cm[seg.intf_alignment.mask_precision-1:0] != 0 { 
            return FALSE;
          };
        }
        else {
          if (~cm)[seg.intf_alignment.mask_precision-1:0] != 0 { 
            return FALSE;
          };
        };
        cm = cm >> seg.intf_alignment.mask_precision;
      };
    };
    result = TRUE;
  };

  package update_offsets() is {
    cur_segment.offset_trans.source.addr.offset = cur_segment.seg_offset - cur_segment.initial_adjustment;
    cur_segment.offset_trans.target.addr.offset = cur_segment.offset_trans.source.addr.offset - src_tgt_diff;
    cur_segment.offset_trans.num_bytes = cur_segment.seg_size + cur_segment.initial_adjustment + cur_segment.final_adjustment;

    assert cur_segment.offset_trans.num_bytes % cur_segment.intf_alignment.alignment == 0 else
      error("CSCO_ADDRMAP_ERR_SEGMENT_ALIGNMENT: the current segment size for ", addrmap_trans.csco_short_name(), " is not aligned to ", 
              interface, "'s interface alignment. This is a csco_addrmap internal programming bug!");
  };

  package update_data_and_mask() is {
    var specified_dsize := addrmap_trans.data.size();
    var iteration_offset : csco_addrmap_addr_t;
    var specified_offset : csco_addrmap_addr_t;
    var end_offset : csco_addrmap_addr_t;
    var iteration_size : csco_addrmap_addr_t;
    var tmp_blist : list of byte;

    cur_segment.offset_trans.data.clear();
    iteration_offset = cur_segment.seg_offset - addrmap_trans.source.addr.offset;
    specified_offset = iteration_offset  % specified_dsize;
    end_offset = iteration_offset + cur_segment.seg_size;
    while iteration_offset < end_offset {
      iteration_size = specified_dsize - specified_offset;

      if iteration_offset < specified_dsize or addrmap_trans.fill_kind == REPLICATE or addrmap_trans.read_or_write() == READ {
        tmp_blist = addrmap_trans.data[specified_offset..];
      }
      else {
        case (addrmap_trans.fill_kind) {
          RANDOM : {
            gen tmp_blist keeping {
              it.size() == iteration_size;
            };
          };

          SHADOW : {
            if addrmap_trans.get_entry_ptr() is a csco_addrmap_mem_entry_s (mem_entry) {
              var tmp_trans := deep_copy(addrmap_trans);
              tmp_trans.target.addr.offset += iteration_offset;
              tmp_trans.num_bytes = iteration_size;
              tmp_blist = mem_entry.get_shadow_bytes(tmp_trans);
            }
            else {
              // probably targetting a logical region only.
              gen tmp_blist keeping {
                it.size() == iteration_size;
              };
            };
          };

          CONSTANT : {
            tmp_blist.resize(iteration_size, TRUE, addrmap_trans.fill_value, FALSE);
          };
        };
      };

      iteration_offset += iteration_size;
      specified_offset = 0;
      cur_segment.offset_trans.data.add(tmp_blist);
    }; // while 

    // truncate any trailing extra bytes.
    cur_segment.offset_trans.data.resize(cur_segment.seg_size, FALSE, 0, TRUE);

    // Initial pad bytes are set to 0 since they're masked
    // out anways.
    tmp_blist.resize(cur_segment.initial_adjustment);
    cur_segment.offset_trans.data = {tmp_blist; cur_segment.offset_trans.data};

    // Final pad bytes are set to 0 since they're masked
    // out anways.
    tmp_blist.resize(cur_segment.final_adjustment);
    cur_segment.offset_trans.data.add(tmp_blist);

    cur_segment.offset_trans.data_mask = get_segment_mask_bits(cur_segment);
  }; // package update_data_and_mask() 

  // This method is used to pass response back to addrmap segments which are non-posted.
  update_response(resp : list of byte, resp_kind : csco_addrmap_resp_kind_t, segment_ind : uint) is {
    message(CSCO_ADDRMAP, MEDIUM, "updating response data of segment ", segment_ind,
                                  " of ", addrmap_trans.csco_short_name());

    assert that addrmap_trans.post_kind == NON_POSTED else
      error("CSCO_ADDRMAP_ERR_UNEXPECTED_SEGMENT_RESP:\n",
            "update_response() called for ", csco_short_name(), ", which is a POSTED transaction.");

    assert that segment_ind < resp_collector.segment.size() else 
      error("CSCO_ADDRMAP_ERR_SEGMENT_RESP_INDEX_OUT_OF_RANGE\n",
            "segment response index ", segment_ind, " is out of range. Only ", 
            resp_collector.segment.size(), " segments are outstanding.");

    var update_segment : csco_addrmap_intf_segment_resp_s;
    update_segment = resp_collector.segment[segment_ind];

    assert that !update_segment.received else
      error("CSCO_ADDRMAP_ERR_DUPLICATE_SEGMENT_RESPONSE_UPDATE\n",
            "response data for segment ", segment_ind, " of ", addrmap_trans.csco_short_name(),
            " has already been updated.");

    update_segment.received = TRUE;
    if resp_kind.as_a(uint) > resp_collector.resp_kind.as_a(uint) {
      resp_collector.resp_kind = resp_kind;
    };

    if addrmap_trans.read_or_write() == READ {
      assert that resp_collector.segment.sum(.data.size()) < addrmap_trans.num_bytes else
        error("CSCO_ADDRMAP_ERR_EXTRA_SEGMENT_RESP:\n", 
              "received an extra read response update for ", addrmap_trans.csco_short_name());

      assert that ((resp.size() == update_segment.num_bytes) or (resp_kind != NORMAL)) else
        error("CSCO_ADDRMAP_ERR_SEGMENT_RESP_DATA_SIZE_MISMATCH\n",
              "response data size ", resp.size(), " for segment ", segment_ind, 
              " of ", addrmap_trans.csco_short_name(), " is not the expected size (", update_segment.num_bytes, ").");

      if resp_kind == NORMAL {
        update_segment.data = resp[update_segment.start_byte..update_segment.end_byte];
      };
    }
    else {
      assert that resp.is_empty() else
        error("CSCO_ADDRMAP_ERR_UPDATE_RESPONSE_SIZE:\n",
              "update_response() called with ", resp.size(), 
              " bytes of response data, but the write request doesn't expect any data in the response.");
    };

    // If all request segment for the addrmap transaction has been issued, 
    // then check whether we've received all segments.
    if bytes_transferred >= addrmap_trans.num_bytes {
      if !resp_collector.segment.has(!.received) {
        var all_data : list of byte;
        for each (s) in resp_collector.segment {
          all_data.add(s.data);
        };
        message(CSCO_ADDRMAP, MEDIUM, "all segments for ", addrmap_trans.csco_short_name(),
                                      " has received their response updates.");
        addrmap_trans.set_response(all_data, resp_collector.resp_kind);
      };
    };
  }; // update_response()
}; // struct csco_addrmap_intf_segment_iterator_s


extend GREEDY'seg_alg csco_addrmap_intf_segment_iterator_s {
  select_segment_size(legal_a : csco_addrmap_intf_alignment_s, rem_bytes : csco_addrmap_addr_t) : csco_addrmap_addr_t is also {
    result = min(rem_bytes, legal_a.max_size);
  };

  filter_alignment_and_size(cur_result : list of csco_addrmap_intf_segment_s) : list of csco_addrmap_intf_segment_s is also {
    if cur_result.size() > 0 {
      result = {cur_result.max(.seg_size)};
    };
  };
};  // extend GREEDY'seg_alg csco_addrmap_intf_segment_iterator_s


extend ALIGNED_RANDOM'seg_alg csco_addrmap_intf_segment_iterator_s {
  filter_alignment_and_size(cur_result : list of csco_addrmap_intf_segment_s) : list of csco_addrmap_intf_segment_s is also {
    if cur_result.size() > 0 {
      result = cur_result.all(.seg_offset % entry_alignment == 0 and .seg_size % entry_alignment == 0);
    };
  };
};


'>
