// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_name_scope.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 09/26/2005
//              :
//              :
// Description  : Defines the name scope struct, which is used to hold address
//                map entries. The names of the addrmap entries within a
//                name scope MUST be unique. Name scopes are hierarchical.
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003-2007 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

<'
package csco_addrmap;


// The name scope struct is a name space container for addrmap entries. 
// The scopes are hierarchical to provide a unique name space to ensure 
// that the names of all the addrmap entries that reside within a name 
// scope are unique.
struct csco_addrmap_name_scope_s {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Name of the name scope. Unique within a parent name scope.
  name : string;
    keep soft name == "";
    csco_print_attr name SHORT enable;
    csco_print_get_string_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "name" and pf.format == SHORT {
        var bytelist : list of byte = long_name("").as_a(list of byte);
        if bytelist.size() > 50 {
          bytelist = bytelist[bytelist.size()-50..];
        };
        result = bytelist.as_a(string);
      };
    };


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  // Adds a child scope to this scope and re-arranges all the
  // pointers.
  add_child_scope(c : csco_addrmap_name_scope_s) is {
    assert that c.name not in child_scopes.name else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_SCOPE: %s is a duplicate scope under %s", c.name, csco_short_name()));

    assert that c.parent_scope == NULL else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_PARENT_SCOPE: %s already has a parent scope (%s), cannot add to %s", c.csco_short_name(), 
                              c.parent_scope.csco_short_name(), csco_short_name()));

    child_scopes.add(c);
    c.parent_scope = me;
    for each (n) in {c; c.get_all_child_scopes()} {
      n.depth += depth + 1;
    };
  };

  // Remove a child scope from this scope and re-aranges all the
  // pointers.
  remove_child_scope(c : csco_addrmap_name_scope_s) is {
    var delete_ind : int = child_scopes.key_index(c.name);

    assert that delete_ind != UNDEF else
      error(appendf("CSCO_ADDRMAP_ERR_SCOPE_DEL: %s is not a child scope of %s, cannot remove it", c.csco_short_name(), csco_short_name()));

    child_scopes.delete(delete_ind);
    c.parent_scope = NULL;
    for each (n) in {c; c.get_all_child_scopes()} {
      n.depth -= (depth + 1);
    };
  };

  // integration setup utility to trim levels from a larger tree
  // keeping only those that match the specified pattern
  trim_all_except(name_re: string) is {
    for each (c) in reverse child_scopes {
      if(c.name !~ name_re) {
        remove_child_scope(c);
      };
    };
    for each (ent) in reverse child_entries {
      if(ent.name !~ name_re) {
        remove_child_entry(ent);
      };
    };
    if child_scopes.is_empty() and child_entries.is_empty() {
      warning("trim_all_except(",name_re,"), run from scope ",name,", resulted in empty list.  \n   Suggestion: use get_scope_first_match(\"<scope name>\").trim_all_except(\"",name_re,"\"); \n       where <scope name> is the scope that has children with the match string ",name_re);
    };
  };

  // Adds a child entry to this scope and re-arranges all the
  // pointers.
  add_child_entry(c : csco_addrmap_entry_s) is {
    assert that child_entries.key(c.name) == NULL else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_PHY_REGN: %s is a duplicate entry under %s", c.name, csco_short_name()));

    assert that c.parent_scope == NULL else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_PARENT_SCOPE: %s already has a parent scope (%s), cannot add to %s", c.csco_short_name(), 
                              c.parent_scope.csco_short_name(), csco_short_name()));

    child_entries.add(c);
    c.parent_scope = me;
  };

  // Remove a child entry from this scope and re-aranges all the
  // pointers.
  remove_child_entry(c : csco_addrmap_entry_s) is {
    var delete_ind : int = child_entries.key_index(c.name);

    assert that delete_ind != UNDEF else
      error(appendf("CSCO_ADDRMAP_ERR_SCOPE_ENTRY_DEL: %s is not a child entry of %s, cannot remove it", c.csco_short_name(), csco_short_name()));

    child_entries.delete(delete_ind);
    c.parent_scope = NULL;
  };

  // Adds a list of child entries to this scope and re-arranges all the
  // pointers.
  add_child_entries(children : list of csco_addrmap_entry_s) is {
    for each (c) in children {
      add_child_entry(c);
    };
  };

  // Remove a list of child entries from this scope and re-aranges all the
  // pointers.
  remove_child_entries(children : list of csco_addrmap_entry_s) is {
    for each (c) in children {
      remove_child_entry(c);
    };
  };

  // Looks up a scope by name in the current scope's sub-tree and returns
  // all scopes that match. The input search string can be a '.' delimited
  // path with the final token being the scope(s) that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used. Note that each token doesn't need to be an
  // immediate child of the previous token.
  get_scope_all_match(path : string) : list of csco_addrmap_name_scope_s is {
    var path_split : list of string = str_split(path, ".");
    if me.name == path_split[0] {
      if path_split.size() == 1 {
        result.add(me);
      }
      else {
        for each (c) in child_scopes {
          result.add(c.get_scope_all_match(str_join(path_split[1..], ".")));
        };
      };
    }
    else {
      for each (c) in child_scopes {
        result.add(c.get_scope_all_match(path));
      };
    };
  };

  // Looks up a scope by name in the current scope's sub-tree and returns
  // 1st scope that match. The input search string can be a '.' delimited
  // path with the final token being the scope that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used. Note that each token doesn't need to be an
  // immediate child of the previous token.
  get_scope_first_match(path : string) : csco_addrmap_name_scope_s is {
    var all_match : list of csco_addrmap_name_scope_s;
    all_match = get_scope_all_match(path);
    if !all_match.is_empty() {
      result = all_match[0];
    };
  }; 

  // Looks up a scope by name in the current scope's sub-tree and returns
  // a single random match. The input search string can be a '.' delimited
  // path with the final token being the scope that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used.  Note the randomization is only applied 
  // to the leaf lookup. Note that each token doesn't need to be an
  // immediate child of the previous token.
  get_scope_random_match(path : string) : csco_addrmap_name_scope_s is {
    var all_match : list of csco_addrmap_name_scope_s;
    all_match = get_scope_all_match(path);
    if all_match.size() > 0 {
      gen result keeping { 
        it in all_match;
      };
    };
  }; 

  // Looks up an entry by name in the current scope's sub-tree and returns
  // all entries that match. The input search string can be a '.' delimited
  // path with the final token being the entry(s) that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used. Note that each token doesn't need to be an
  // immediate child of the previous token.
  get_entry_all_match(path : string) : list of csco_addrmap_entry_s is {
    var path_split : list of string;
    var entry_match : csco_addrmap_entry_s;
    var scope_matches : list of csco_addrmap_name_scope_s;
    var target_entry : string;

    path_split = str_split(path,".");
    if path_split.size() == 1 {
      entry_match = child_entries.key(path_split[0]);
      if entry_match != NULL {
        result.add(entry_match);
      };

      for each (c) in child_scopes {
        result.add(c.get_entry_all_match(path));
      };
    }
    else {
      target_entry = path_split.pop();
      scope_matches = get_scope_all_match(str_join(path_split, "."));
      for each (s) in scope_matches {
        result.add(s.get_entry_all_match(target_entry));
      };
    };
  };

  // Looks up an entry by name in the current scope's sub-tree and returns
  // 1st entry that matches. The input search string can be a '.' delimited
  // path with the final token being the entry that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used. Note that each token doesn't need to be an
  // immediate child of the previous token.
  get_entry_first_match(path : string) : csco_addrmap_entry_s is {
    var all_matches : list of csco_addrmap_entry_s;
    all_matches = get_entry_all_match(path);
    if !all_matches.is_empty() {
      result = all_matches[0];
    };
  };

  // Looks up an entry by name in the current scope's sub-tree and returns
  // a single random match. The input search string can be a '.' delimited
  // path with the final token being the entry that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used.  Note the randomization is only applied 
  // to the leaf lookup. Note that each token doesn't need to be an
  // immediate child of the previous token.
  get_entry_random_match(path : string) : csco_addrmap_entry_s is {
    var all_matches : list of csco_addrmap_entry_s;
    all_matches = get_entry_all_match(path);
    if !all_matches.is_empty() {
      gen result keeping {
        it in all_matches;
      };
    };
  };

  // Returns the immediate child scopes of the current scope.
  get_child_scopes() : list of csco_addrmap_name_scope_s is {
    result = child_scopes.as_a(list of csco_addrmap_name_scope_s);
  };

  // Returns all child scopes in the sub-tree of the current scope.
  get_all_child_scopes() : list of csco_addrmap_name_scope_s is {
    var cur_ind : uint;
    var breadth_first_fifo := child_scopes.as_a(list of csco_addrmap_name_scope_s).copy();

    while cur_ind < breadth_first_fifo.size() {
      breadth_first_fifo.add(breadth_first_fifo[cur_ind].child_scopes);
      cur_ind += 1;
    };
    result = breadth_first_fifo.copy();
  };

  // Returns the immediate child entries of the current scope.
  get_child_entries() : list of csco_addrmap_entry_s is {
    result = child_entries.as_a(list of csco_addrmap_entry_s);
  };

  // Returns all child entries in the sub-tree of the current scope.
  get_all_child_entries() : list of csco_addrmap_entry_s is {
    result = {me; get_all_child_scopes()}.child_entries.as_a(list of csco_addrmap_entry_s);
  };

  // Returns the immediate child physical regions of the current scope.
  get_child_phy_regions() : list of csco_addrmap_physical_region_s is {
    result = get_child_entries().parent_region.unique(it);
  };

  // Returns all child physical regions in the sub-tree of the current scope.
  get_all_child_phy_regions() : list of csco_addrmap_physical_region_s is {
    result = get_all_child_entries().parent_region.unique(it);
  };

  csco_short_name() : string is only {
    result = long_name("");
  };

  // Returns a string of the name scope tree path leading to the current
  // scope, deliminted by '.'. The user may provide a starting node from
  // which to begin the path. If "" is specified, then the method will
  // start the path from the root of the name scope tree.
  long_name(start_node : string) : string is {
    var parent : csco_addrmap_name_scope_s = me;
    var str_list : list of string;

    str_list.add(me.name);
    while parent != NULL and parent.name != start_node {
      parent = parent.parent_scope;
      if parent != NULL {
        str_list.add(parent.name);
      };
    };
    result = str_join(str_list.reverse(), ".");

    assert that start_node == "" or str_list.top() == start_node else
      error("CSCO_ADDRMAP_ERR_GET_MY_PATH: start_node (", start_node, ") passed to ", csco_short_name(), ".long_name() is invalid");
  };

  // populate this name scope from XML metadata in <filename>
  // as generated by blueprint/reg -xmldirect custom output generator.
  // <top_scope> can be used to filter a subset of a single monolithic file
  // returns a new parent logical region containing the addressing tree
  // represented in the files.  the caller should set the .name
  // of the returned scope if desired
  populate_from_xml_file(filename: string, top_scope: string) : ADDRESS'route_kind csco_addrmap_logical_region_s is undefined;


  ////////////////////////
  // Private Fields 
  ////////////////////////
  package !depth : uint;

  // Back pointer to parent name scope.
  package !parent_scope : csco_addrmap_name_scope_s;
    csco_print_attr parent_scope ALL disable;

  // List of addrmap entries that reside in the name scope.
  package !child_entries : list (key:name) of csco_addrmap_entry_s;
    csco_print_attr child_entries SHORT enable;
    csco_print_attr child_entries SHORT expand=ON;

  // List of child name scope.
  package !child_scopes : list (key:name) of csco_addrmap_name_scope_s;
    csco_print_attr child_scopes SHORT enable;
    csco_print_attr parent_scope LONG disable;
}; // struct csco_addrmap_name_scope_s


'>
