/ ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_region.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 09/26/2005
//              :
//              :
// Description  : Defines data structures for address regions as well as address
//                translation mechansims for translating addresses across parent
//                and child regions.
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_addrmap;


// Generic utility struct to define address offset ranges.
struct csco_addrmap_offset_range_s {
  min : csco_addrmap_addr_t;
    csco_print_attr min SHORT enable;

  max : csco_addrmap_addr_t;
    csco_print_attr max SHORT enable;
    keep max >= min;

  csco_short_name() : string is also {
    result = append("min=", min, " max=", max);
  };
};

// The is struct models a generic address inside of any region.
// The base-type simply holds an offset, but it may also be
// extended to hold other routable things such as bus IDs. 
struct csco_addrmap_addr_s {
  route_kind : csco_addrmap_route_kind_t;
    csco_print_attr route_kind ALL enable;

  addr_space : csco_addrmap_space_t;
    csco_print_attr addr_space ALL enable;

  offset : csco_addrmap_addr_t;
    csco_print_attr offset ALL enable;

  csco_short_name() : string is {
    result = hex(offset);
  };
};

// Used to describe an address within a region.
struct csco_addrmap_region_addr_s {
  // Pointer to region.
  region : csco_addrmap_region_s;
    keep soft region == NULL;
    attribute region deep_copy = reference;
    csco_print_attr region SHORT enable;

  // Address within region.
  addr : csco_addrmap_addr_s;
    keep addr.route_kind == read_only(region.get_route_kind());
    keep addr.addr_space == read_only(region.get_addr_space());
    keep read_only(region != NULL and region.get_region_size() > 0) => addr.offset < region.get_region_size();
    csco_print_attr addr SHORT enable;
    csco_print_attr addr LONG expand=ON;

  csco_short_name() : string is {
    result = append(region.name, ": ", addr.csco_short_name());
  };

  run() is also {
    if addr == NULL and region != NULL {
      addr = new csco_addrmap_addr_s with {
        .route_kind = region.get_route_kind();
        .addr_space = region.get_addr_space();
      };
    };
  };

  post_generate() is also {
    var leaf : csco_addrmap_leaf_entry_s = region.get_leaf(addr);
    // Make sure generated address is aligned to the physical entry
    // if the targetted leaf region is a physical region.
    if leaf.entry != NULL {
      addr.offset = addr.offset & (-leaf.entry.get_entry_width()); 
    };
  };
};

// The struct encapsulates the results of a get_leaf() lookup
// call. The results consist of a logical region, a phyiscal region
// and an addrmap entry. Note that the physical region and the addrmap
// entry could be NULL, if there's no corresponding physical entry.
struct csco_addrmap_leaf_entry_s {
  // Lowest logical region/addr pair. Parent region of leaf
  // physical region.
  !logical : csco_addrmap_region_addr_s;
    attribute logical deep_copy = reference;
    csco_print_attr logical SHORT enable;

  // Lowest physical region/addr pair. Parent region of 
  // leaf entry.
  !physical : csco_addrmap_region_addr_s;
    attribute physical deep_copy = reference;
    csco_print_attr physical SHORT enable;

  // Leaf entry.
  !entry : csco_addrmap_entry_s;
    attribute entry deep_copy = reference;
    csco_print_attr entry SHORT enable;

   // Convenience method for memory entries to get the offset into the
   // memory.
  get_entry_offset() : csco_addrmap_addr_t is {
     result = physical.addr.offset - entry.offset;
   };
};


// Regions are used to build up the addressing topology by
// defining parent/child relationships between various I/O
// interfaces.
// <br>
// A region should be created for each input or output I/O
// interface on the block. If an interface is bi-directional,
// then two regions must be created, one for the ingress
// direction, the other for the egress direction. 
//
// Regions for input interfaces will have as child regions
// physical regions within the device, and other output 
// regions to which this particular input interface provides
// access to (typical in a bridging device).
// <br>
// Regions for output interfaces can be connected as parent
// regions to input regions in neighboring blocks.
// Note that if no DUT block is connected to the output inteface,
// then the region does not have any child regions.
// <br>
// Depending on how the testbench is put together, addrmap regions
// can be generated or new'ed. The implementation here supports
// both. 
struct csco_addrmap_region_s {
  // Name of the region. It's recommended, but not required that names of
  // each region in the address map be unique.
  name : string;
    keep soft name == "";
    csco_print_attr name SHORT enable;
    csco_print_get_string_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "name" {
        var bytelist : list of byte = result.as_a(list of byte);
        if bytelist.size() > 30 {
          bytelist = bytelist[bytelist.size()-30..];
        };
        result = bytelist.as_a(string);
      };
    };

  // Will cause this region to have an offset in its parent region which is naturally
  // aligned to this region's size, taken to the nearest 2's exponent.
  naturally_aligned : bool;

  // Alternate alignment, if not naturally aligned. '1' indicates no alignment
  // requirement.
  align : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1);
    keep soft align == 1;
    init() is also {
      align = 1;
    };

  // Calculatese the naturally aligned size of this region.
  get_natural_align() : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1) is {
    var rsize := get_region_size();
    var bit_ind : uint;
    while rsize != 0 {
      bit_ind += 1;
      rsize = rsize >> 1;
    };

    if (1 << (bit_ind - 1)) != get_region_size() {
      result = (1 << bit_ind);
    }
    else {
      result = get_region_size();
    };
  };

  get_max_align() : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1) is {
    var cregions := get_child_regions();
    result = (naturally_aligned) ? get_natural_align() : align;
    if !cregions.is_empty() {
      result = max(result, cregions.max_value(.get_max_align()));
    };
  };

  package align_offset(raw_offset : csco_addrmap_addr_t) : csco_addrmap_addr_t is {
    var max_align := get_max_align();
    if raw_offset % max_align != 0 {
      result = raw_offset + (max_align - raw_offset % max_align);
    }
    else {
      result = raw_offset;
    };
  };

  // Update backdoor path for physical unit
  // or all physical descendants under a logical unit
  set_hdl_path(u: any_unit, path: string) is undefined;

  package !max_depth : uint;

  // Used by get_all_child_regions() and get_all_parent_regions() to indicate relative 
  // depth of returned nodes from the calling node.
  package !cur_rel_depth : uint;

  // Returns the maximum # of hops from root region to this region.
  get_max_depth() : uint is {
    result = max_depth;
  };

  // Address space of the region in bytes.
  get_region_size() : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1) is undefined;

  // contexts of this region in its parent regions.
  package !parent_contexts : list of csco_addrmap_region_addr_s;
    attribute parent_contexts deep_copy = reference;
    csco_print_attr parent_contexts SHORT enable;
    csco_print_attr parent_contexts SHORT header=parent expand=ON;

  // Takes a parent region pointer and finds the corresponding entry in the 
  // current region's list of parent contexts. 
  get_parent_context(p : csco_addrmap_region_s) : csco_addrmap_region_addr_s is {
    var ind : int = parent_contexts.first_index(it.region == p);

    assert that ind != UNDEF else
      error(appendf("CSCO_ADDRMAP_ERR_NO_PARNT_RGN: cannot find parent region %s in %s's list of parents", p.name, csco_short_name()));

    result = parent_contexts[ind];
  };

  // Returns the immediate parent regions of this region.
  get_parent_regions() : list of csco_addrmap_logical_region_s is {
    result = parent_contexts.region.as_a(list of csco_addrmap_logical_region_s);
  };

  // Returns all parent regions of the current region, not just the immediately
  // adjacent ones. The result list is sorted according to each parent's 
  // proximity to the current region.
  get_all_parent_regions() : list of csco_addrmap_logical_region_s is {
    var cur_ind : uint;
    var breadth_first_fifo := get_parent_regions().copy();

    cur_rel_depth = 0;
    for each (c) in breadth_first_fifo {
      c.cur_rel_depth = 1;
    };

    while cur_ind < breadth_first_fifo.size() {
      for each (p) in breadth_first_fifo[cur_ind].get_parent_regions() {
        if p not in breadth_first_fifo {
          breadth_first_fifo.add(p);
          p.cur_rel_depth = breadth_first_fifo[cur_ind].cur_rel_depth + 1;
        };
      };
      cur_ind += 1;
    };
    result = breadth_first_fifo.copy();
  }; // get_all_parent_regions() 

  // Returns root parent regions of the current region.
  get_roots() : list of csco_addrmap_logical_region_s is {
    result = get_all_parent_regions().all(.max_depth == 0);
  };

  // Returns the immediate child regions of this region.
  get_child_regions() : list of csco_addrmap_region_s is empty;

  // Returns all child regions of the current region, not just the immediately
  // adjacent ones. The result list is sorted according to each child's 
  // proximity to the current region.
  get_all_child_regions() : list of csco_addrmap_region_s is {
    var cur_ind : uint;
    var breadth_first_fifo := get_child_regions().copy();

    cur_rel_depth = 0;
    for each (c) in breadth_first_fifo {
      c.cur_rel_depth = 1;
    };

    while cur_ind < breadth_first_fifo.size() {
      for each (c) in breadth_first_fifo[cur_ind].get_child_regions() {
        if c not in breadth_first_fifo {
          breadth_first_fifo.add(c);
          c.cur_rel_depth = breadth_first_fifo[cur_ind].cur_rel_depth + 1;
        };
      };
      cur_ind += 1;
    };
    result = breadth_first_fifo.copy();
  }; // get_all_child_regions() 

  // Looks up a region by name in the current region's sub-tree and returns
  // 1st region that matches. The input search string can be a '.' delimited
  // path with the final token being the region that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used.
  get_region_first_match(path : string) : csco_addrmap_region_s is empty;

  get_region_path(child : csco_addrmap_region_s) : list of csco_addrmap_region_s is {
    var tmp_result : list of csco_addrmap_region_s;

    if child == me {
      result = {me};
    }
    else {
      for each (c) in get_child_regions() {
        tmp_result = c.get_region_path(child);
        if tmp_result.size() > 0 {
          result = {me; tmp_result};
          return result;
        };
      };
    };
  };

  // Returns a list of connected regions that can access the given address
  // in the current region. Defined in each 'route_kind' subtype.
  get_driving_regions(addr : csco_addrmap_addr_s) : list of csco_addrmap_region_addr_s is undefined;

  // Returns the leaf region that is referenced by the given address in the
  // current region. Defined in each 'route_kind' subtype.  
  get_leaf(addr : csco_addrmap_addr_s) : csco_addrmap_leaf_entry_s is undefined;

  // This method takes an addrmap transaction that has been observed on an interface 
  // corresponding to this region, and call the update_shadow() method of the 
  // appropriate addrmap entries in the physical leaf region if the physical region
  // is local (has the same interface).
  // <br>
  // The input to the method is a csco_addrmap transaction. This can be
  // interpreted in two ways:
  // <br>
  // 1 - The transaction for the specific interface on which it is observed
  //     is inherited from the base csco_addrmap_trans_s. In this case,
  //     the monitor simply passes in the monitored transaction as the
  //     argument, with all the base fields set appropriately.
  // <br>
  // 2 - The transaction for the specific interface on which it is observed
  //     is defined independently of the csco_addrmap_trans_s. In this case,
  //     the input argument can be considered as an aggregation of all the 
  //     information that are required by update_shadow() to do its job. It 
  //     just happens that all that information is already encapsulated in the
  //     base csco_addrmap_trans_s. The monitor in this case should translate
  //     the relevant fields from the monitored transaction into a base
  //     csco_addrmap_trans_s instance and pass it to this method.
  // <br>
  // The addrmap transaction passed to the method must have the following fields
  // correctly set:
  // <br>
  // opcode, target, num_bytes, data, data_mask
  // <br>
  // Note that write_mask is only required for write operations.
  //
  // The 'local_only' parameter specifies whether the search for a physical
  // entry should pass 'interface' boundaries or not. Typically, if a monitor
  // is inside a downstream_active eVC, then 'local_only' should be set to FALSE.
  // Otherwise, 'local_only' should be set to TRUE.
  update_shadow(trans : csco_addrmap_trans_s, local_only : bool) is {
    message(CSCO_ADDRMAP, HIGH, "calling update_shadow() of ", csco_short_name(),
            " with local_only=", local_only, " and following transaction:\n") {
      trans.csco_print(LONG);
    };

    assert that trans.num_bytes != 0 else
      error("CSCO_ADDRMAP_ERR_ZERO_LENGTH_TRANS_PASSED_TO_UPDATE_SHADOW:\n",
            "update_shadow() cannot be called with a transaction whose 'num_bytes' is 0.");


    trans.find_leaf();
    if trans.leaf.entry != NULL {
      var rpath := get_region_path(trans.leaf.logical.region);
      if !rpath.has(.get_interface() != rpath[0].get_interface() and .get_interface() != NONE) or !local_only {
        trans.leaf.entry.update_shadow(trans);
      }
      else {
        message(CSCO_ADDRMAP, LOW, "addrmap transaction to ", csco_short_name(), 
                " targets a leaf entry on a different interface, no shadow value will be updated");
      };
    }
    else {
      messagef(CSCO_ADDRMAP, LOW, "addrmap transaction to %s doesn't target any reachable physical entry, no shadow will be updated\n", csco_short_name());
    };
  }; // update_shadow();

  get_shadow(trans : csco_addrmap_trans_s) : list of byte is {
    assert that trans.num_bytes != 0 else
      error("CSCO_ADDRMAP_ERR_ZERO_LENGTH_TRANS_PASSED_TO_GET_SHADOW:\n",
            "get_shadow() cannot be called with a transaction whose 'num_bytes' is 0.");

    trans.find_leaf();
    if trans.leaf.entry != NULL {
      if trans.read_or_write() == READ {
        result = trans.leaf.entry.get_shadow_bytes(trans);
      }
      else {
        message(CSCO_ADDRMAP, LOW, "addrmap transaction to ", csco_short_name(), 
                "is a write, no shadow value will be fetched");
      };
    }
    else {
      message(CSCO_ADDRMAP, LOW, "addrmap transaction to ", csco_short_name(), 
              "doesn't target a leaf entry, no shadow value will be fetched");
    };
  };

  // These two methods are defined for the region base type so that we don't have
  // to do casting to access the corresponding fields. For example, the 
  // csco_addrmap_region_addr_s struct only references the region base type.
  get_interface() : csco_addrmap_interface_t is empty;
  set_interface(intf : csco_addrmap_interface_t) is undefined;
  get_route_kind() : csco_addrmap_route_kind_t is empty;
  get_addr_space() : csco_addrmap_space_t is empty;
  get_eligible_driver() : any_sequence_driver is empty;

  csco_short_name() : string is {
    if me is a csco_addrmap_logical_region_s (log_region) {
      result = appendf("%s (%s)", name, log_region.addr_space);
    }
    else {
      result = appendf("%s (PHY)", name);
    };
  };

  print_child_tree(fmt : csco_print_format_t) is {
    csco_print_sub_tree(fmt, "child_regions")
  };

  print_parent_tree(fmt : csco_print_format_t) is {
    csco_print_sub_tree(fmt, "parent_contexts.region")
  };

  run() is also {
    assert that name != "" else
      error(appendf("CSCO_ADDRMAP_ERR_REGION_NO_NAME: each addrmap region must have a name"));
  };
}; // struct csco_addrmap_region_s


// A physical region contains address map entries (register or memory).
// The entries may have holes between them, specified via the offset 
// field of each entry.
//
// A physical region does not have any child regions.
struct csco_addrmap_physical_region_s like csco_addrmap_region_s {
  // list of memory/register entries
  package !child_entries : list (key : offset) of csco_addrmap_entry_s;
    csco_print_attr child_entries SHORT enable;
    csco_print_attr child_entries SHORT precede=parent_contexts header=size;
    csco_print_get_list_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "child_entries" and pf.format == SHORT {
        result = hex(get_region_size());
      };
    };

  // Concatenated with parent unit's hdl path to form backdoor 
  // access path.
  // Note that the parent unit of a physical region is normally
  // the addrmap_env_u instantiated in the TB's env.
  hdl_path : string;
    keep soft hdl_path == "";
    csco_print_attr hdl_path LONG header=full_hdl_path;
    csco_print_get_string_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "hdl_path" and pf.format == LONG {
        result = get_full_hdl_path();
      };
    };

  set_hdl_path(u: any_unit, path: string) is {
    set_unit(u);
    if(hdl_path == NULL) {
      hdl_path = path;
    } else {
      hdl_path = append(path, ".", hdl_path);
    };
  };

  // Concatenates this region's 'hdl_path' with its parent unit's full hdl
  // path and returns the result.
  get_full_hdl_path() : string is {
    result = get_unit().full_hdl_path();
    if(hdl_path != NULL) {
      result = append(result, ".", hdl_path);
    };
  };

  // Takes a local offset and looks up the entry in the physical region.
  get_entry(o : csco_addrmap_addr_t) : csco_addrmap_entry_s is {
    result = child_entries.key(o);
    if result == NULL {
      assert that child_entries.has(o % .get_entry_width() == 0) else
        error("CSCO_ADDRMAP_ERR_PHY_REGION_ENTRY_OFFSET_LOOKUP:\n",
                "offset ", hex(o), "is not aligned to any entries in ", csco_short_name());
    };
  };

  get_entry_by_name(nm : string) : csco_addrmap_entry_s is {
    result = child_entries.first(it.name == nm);
  };

  // Adds a new entry to the physical region.
  add_child_entry(c : csco_addrmap_entry_s) is {
    var insert_ind : int = UNDEF;

    assert that c.offset % c.get_entry_width() == 0 else 
      error(appendf("CSCO_ADDRMAP_ERR_BAD_ENTRY_OFFSET: %s's offset (%d) is not a multiple of its width (%d)", c.name, c.offset, 
                      c.get_entry_width()));

    assert that c.parent_region == NULL else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_PARENT_REGN: %s already has a parent region (%s), cannot add to %s", c.long_name(""), 
                      c.parent_region.csco_short_name(), csco_short_name()));

    for each (e) in child_entries {
      assert that e.offset != c.offset else
        error(appendf("CSCO_ADDRMAP_ERR_DUP_ENTRY_OFFSET: %s's offset in %s is the same as %s", c.long_name(""), csco_short_name(), e.long_name("")));

      if e.offset > c.offset and insert_ind == UNDEF {
        insert_ind = index;
      };
    };

    c.parent_region = me;    
    if insert_ind == UNDEF {
      child_entries.add(c);
    }
    else {
      child_entries.insert(insert_ind, c);
    };
  }; // add_child_entry()

  // Removes a child entry from the physical region.
  remove_child_entry(c : csco_addrmap_entry_s) is {
    var delete_ind : int = child_entries.key_index(c.offset);

    assert that delete_ind != UNDEF else
      error(appendf("CSCO_ADDRMAP_ERR_REGN_ENTRY_DEL: %s is not a child entry of %s, cannot remove it", c.long_name(""), csco_short_name()));

    child_entries.delete(delete_ind);
    c.parent_region = NULL;
  };

  // Adds multiple child entries to the physical region.
  add_child_entries(new_entries : list of csco_addrmap_entry_s) is {
    for each (ne) in new_entries {
      add_child_entry(ne);
    };
  };

  // Removes multiple child entries from the physical region.
  remove_child_entries(new_entries : list of csco_addrmap_entry_s) is {
    for each (ne) in new_entries {
      remove_child_entry(ne);
    };
  };

  // Matches a region name to this region.  The input search string
  // can be a '.' delimited path but since this is a leaf region, a
  // multi-level path cannot match.
  get_region_first_match(path : string) : csco_addrmap_region_s is also  {
    var levels := str_split(path, ".");

    result = (name == levels[0] and levels.size() == 1) ? me : NULL;
  };

  // Total size of a physical region is defined by the offset and size of
  // its last entry.
  get_region_size() : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1) is {
    if child_entries.size() > 0 {
      result = child_entries.max(.offset).offset + child_entries.max(.offset).get_entry_size();
    };
  };

  get_hole_offsets() : list of csco_addrmap_addr_s is {
    var cur_offset : csco_addrmap_addr_t = 0;
    var tmp_addr : csco_addrmap_addr_s;

    for each (e) in child_entries {
      if e.offset > cur_offset {
        for i from cur_offset to e.offset-1 step e.get_entry_width() {
          tmp_addr = new csco_addrmap_addr_s with {
            .route_kind = get_route_kind();
            .addr_space = get_addr_space();
            .offset = i;
          };
          result.add(tmp_addr);
        };
      };
      cur_offset = e.offset + e.get_entry_size();
    };
  };

  // Returns the addrmap entry which encompasses the given offset. The given offset
  // is relative to the physical region itself.
  offset_to_entry(addr : csco_addrmap_addr_s) : csco_addrmap_entry_s is {
    // entries are always sorted when a new one is added.
    result = child_entries.first(.offset <= addr.offset and .offset + .get_entry_size() > addr.offset);
  };

  // Returns a list of entries that intersect with the given address range.
  offset_range_to_entries(sa : csco_addrmap_addr_s, ea : csco_addrmap_addr_s) : list of csco_addrmap_entry_s is {
    var sind : int = UNDEF;
    var eind : int = UNDEF;

    for each (ce) in child_entries {
      if ce.offset <= sa.offset and ce.offset + ce.get_entry_size() > sa.offset and sind == UNDEF {
        sind = index;
      };

      if ea.offset >= ce.offset and ea.offset < ce.offset + ce.get_entry_size() {
        eind = index;
      };
    }; 

    if sind != UNDEF and eind != UNDEF {
      result = child_entries[sind..eind].as_a(list of csco_addrmap_entry_s);
    };
  }; // offset_range_to_entries()

  get_driving_regions(addr : csco_addrmap_addr_s) : list of csco_addrmap_region_addr_s is {
    assert that addr.offset < get_region_size() else
      error("CSCO_ADDRMAP_ERR_ADDR_OUT_OF_RANGE: address offset ", addr.offset, 
            " is out of ", csco_short_name(), "'s legal range (", get_region_size(), ").");

    var my_rap := new csco_addrmap_region_addr_s with {
      .region = me;
      .addr = addr;
    };

    var parent_rap : csco_addrmap_region_addr_s;
    for each (p) in get_parent_regions() {
      parent_rap = p.child_offset_to_parent(my_rap);
      result.add(p.get_driving_regions(parent_rap.addr));
    };
  };

  get_leaf(addr : csco_addrmap_addr_s) : csco_addrmap_leaf_entry_s is {
    assert that addr.offset < get_region_size() else
      error("CSCO_ADDRMAP_ERR_ADDR_OUT_OF_RANGE: address offset ", addr.offset, 
            " is out of ", csco_short_name(), "'s legal range (", get_region_size(), ").");

    result = new;
    result.physical = new csco_addrmap_region_addr_s with {
      .region = me;
    };
    result.physical.addr.offset = addr.offset;

    result.entry = offset_to_entry(addr);
  };
}; // unit csco_addrmap_physical_region_s like csco_addrmap_region_s


// Logical regions are used to group child regions into the same address
// space or to model bridging/address translation functionality of a single
// block by defining parent/child relationships between the block's I/O 
// regions.
struct csco_addrmap_logical_region_s like csco_addrmap_region_s {
  is_generated : bool;
    keep is_generated == TRUE;

  run() is also {
    assert is_generated else 
      error("CSCO_ADDRMAP_ERR_LOGICAL_REGION_NOT_GENERATED:\n",
            "logical regions must be created through generation, but ", me, "(", csco_short_name(), ") was not.");
  };

  // Indicates the routing algorithm used to navigate from one region
  // to another. By default, only ADDRESS routing is defined.
  const route_kind : csco_addrmap_route_kind_t;
    keep soft route_kind == ADDRESS;

  get_route_kind() : csco_addrmap_route_kind_t is { result = route_kind; };

  // Indicates address space that the region is a part of. By default,
  // only a single DEFAULT address space is defined.
  const addr_space : csco_addrmap_space_t;
    keep soft addr_space == DEFAULT;

  event addr_space is change(addr_space);

  on addr_space {
    error("CSCO_ADDRMAP_ERR_ROUTE_KIND_CHANGE_AFTER_RUN\n",
          "value of addr_space in a logical region must NOT change following generation\n");
  };

  get_addr_space() : csco_addrmap_space_t is { result = addr_space; };

  // I/O interface to which this region is associated. NONE means 
  // the region is not associated with an interface.  This field
  // must not be 'NONE' if its eligible driver is set.
  interface : csco_addrmap_interface_t;
    keep soft interface == NONE;

  get_interface() : csco_addrmap_interface_t is { result = interface; };

  set_interface(intf : csco_addrmap_interface_t) is { interface = intf; };

  // If the region is generated, then this field can be used as a unique
  // instance name field to allow instance specific when subtyped constraints.
  // If the region is new'ed, then this field will hold the value 'NONE' 
  // and should NOT be used, the 'name' field should then be set to hold the 
  // name of the logical region.
  gen_name : csco_addrmap_log_region_name_t;
    keep soft gen_name == NO_NAME;
    keep soft name == read_only(gen_name.as_a(string));

  // list of pointers to child memory regions(both physical and logical).
  package !child_regions : list of csco_addrmap_region_s;
    attribute child_regions deep_copy = reference;

  // Algorithm for assigning addresses to child entries/regions during map building.
  // Soft constrained to MANUAL.
  addr_alg : csco_addrmap_addr_gen_alg_t;
    keep soft addr_alg == MANUAL;

  set_hdl_path(u: any_unit, path: string) is {
    for each (child) in child_regions {
      child.set_hdl_path(u, path);
    };
  };

  // Pointers to sequence drivers who can issue direct I/O transactions 
  // to this region.
  //
  // Eligible drivers should only be defined for regions sitting at 
  // an output I/O interface. i.e., their 'interface' field must NOT
  // be 'NONE'.
  package !eligible_driver : any_sequence_driver;
    attribute eligible_driver deep_copy = reference;
    csco_print_attr eligible_driver SHORT enable;
    csco_print_attr eligible_driver SHORT header=elig_drv;

  // Sets the eligible driver for this region. The eligible driver should be
  // the interface driver that can drive transactions in this region's address
  // space.
  //
  // To remove an existing eligible driver, just call this method with 'NULL' 
  // as input.
  set_eligible_driver(d : any_sequence_driver) is {
    assert eligible_driver == NULL or d == NULL else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_DRVR: %s addrmap region already has an eligible driver: %s", csco_short_name(), eligible_driver.short_name()));

    assert interface != NONE else
      error(appendf("CSCO_ADDRMAP_ERR_INTF_NONE: cannot associate %s to an eligible driver because its interface is NONE", csco_short_name()));

    eligible_driver = d;
  };

  get_eligible_driver() : any_sequence_driver is {
    result = eligible_driver;
  };

  // Makes a parent/child association between two regions and makes
  // the necessary pointer adjustments to the region instances.
  add_child_region(c : csco_addrmap_region_s) is {
    // make sure there are no region loops in the address map.
    assert that c not in get_all_parent_regions().as_a(list of csco_addrmap_region_s) else
      error(appendf("CSCO_ADDRMAP_ERR_DUP_CHLD_REGN: child region %s is already in %s's list of parent regions", c.name, csco_short_name()));

    // make sure the parent and the child regions are of the same kind.
    assert that (route_kind == c.get_route_kind()) or (c is a csco_addrmap_physical_region_s) else
      error("CSCO_ADDRMAP_ERR_BAD_CHILD_REGN_ROUTE_KIND: child region, ", c.name, " is a ", 
              c.get_route_kind(), " region, while ", csco_short_name(), " is a ", route_kind, " region");

    // make sure the parent and the child regions are of the same kind.
    assert that (addr_space == c.get_addr_space()) or (c is a csco_addrmap_physical_region_s) else
      error("CSCO_ADDRMAP_ERR_BAD_CHILD_REGN_ADDR_SPACE: child region, ", c.name, " is a ", 
              c.get_addr_space(), " region, while ", csco_short_name(), " is a ", addr_space, " region");

    var new_ctxt := new csco_addrmap_region_addr_s with {
      .region = me;
    };

    message(CSCO_ADDRMAP, HIGH, "adding ", c.csco_short_name(), 
            " as a child region to ", csco_short_name(), ".");
    child_regions.add(c);
    c.parent_contexts.add(new_ctxt);
    for each (n) in {c; c.get_all_child_regions()} {
      n.max_depth = max(n.max_depth, max_depth + 1 + n.cur_rel_depth);
    };
  }; // add_child_region()

  // Assigns a manual offset to a child region of the current region. Only valid if
  // the region's addr_alg field is MANUAL.
  set_child_addr(c : csco_addrmap_region_s, a : csco_addrmap_addr_s) is {
    assert that addr_alg == MANUAL else
      error(appendf("CSCO_ADDRMAP_ERR_MANUAL_OFFSET : cannot manually set off set of %s in %s to %s because %s's addr_alg=%s", 
                     c.csco_short_name(), csco_short_name(), a.csco_short_name(), csco_short_name(), addr_alg));
    assert that c in child_regions else
      error(appendf("CSCO_ADDRMAP_ERR_NO_CHILD_REGION : cannot find %s in %s", c.csco_short_name(), csco_short_name()));

    assert that (a.route_kind == route_kind) or (c is a csco_addrmap_physical_region_s) else
      error("CSCO_ADDRMAP_ERR_CHILD_ADDR_ROUTE_KIND_MISMATCH: address passed to set_child_addr() has ", a.route_kind, 
              " route_kind, but region has ", route_kind, " route_kind.");

    assert that (a.addr_space == addr_space) or (c is a csco_addrmap_physical_region_s) else
      error("CSCO_ADDRMAP_ERR_CHILD_ADDR_ROUTE_KIND_MISMATCH: address passed to set_child_addr() has ", a.addr_space, 
              " addr_space, but region has ", addr_space, " addr_space.");

    //assert that a.offset < get_region_size() else
      //error("CSCO_ADDRMAP_ERR_ADDR_OUT_OF_RANGE: address offset ", a.offset, 
            //" is out of ", csco_short_name(), "'s legal range (", get_region_size(), ").");

    c.get_parent_context(me).addr = a;
  };

  // Makes a parent/child association between two regions and makes
  // the necessary pointer adjustments to the region instances. Also
  // specifies an offset for the child region within the parent region's
  // address space. This method is useful for adding regions to parent
  // regions that have MANUAL addr_alg.
  add_child_region_with_addr(c : csco_addrmap_region_s, a : csco_addrmap_addr_s) is {
    add_child_region(c);
    set_child_addr(c, a);
  };

  // Removes a region from this region's list of child regions and makes
  // the necessary pointer adjustments.
  remove_child_region(c : csco_addrmap_region_s) is {
    var delete_ind : int = child_regions.first_index(it == c);
    var delete_ind2 : int = c.parent_contexts.first_index(.region == me);
    var max_parent_depth : uint;

    assert that delete_ind != UNDEF else
      error(appendf("CSCO_ADDRMAP_ERR_REGN_DEL: %s is not a child region of %s, cannot remove it", c.csco_short_name(), csco_short_name()));

    child_regions.delete(delete_ind);
    c.parent_contexts.delete(delete_ind2);
    if c.parent_contexts.size() > 0 {
      max_parent_depth = c.parent_contexts.max_value(.region.max_depth);
    }
    else {
      max_parent_depth = 0;
    };

    for each (n) in {c; c.get_all_child_regions()} {
      if max_parent_depth > 0 {
        n.max_depth = max_parent_depth + 1 + n.cur_rel_depth;
      }
      else {
        n.max_depth = n.cur_rel_depth;
      };
    };
  };

  get_child_regions() : list of csco_addrmap_region_s is {
    result = child_regions;
  };

  // Takes the existing list of child regions and permutate 
  // their order.
  permute_child_regions() is {
    var tmp_list : list of csco_addrmap_region_s;
    gen tmp_list keeping {
      it.is_a_permutation(child_regions);
    };
    child_regions = tmp_list;
  };

  // Looks up a region by name in the current region's sub-tree and returns
  // 1st region that matches. The input search string can be a '.' delimited
  // path with the final token being the region that will be returned, if
  // found. If there are multiple matches for intermediate tokens, then a
  // first match policy is used.
  get_region_first_match(path : string) : csco_addrmap_region_s is also {
    var levels := str_split(path, ".");
    var tmp_result : csco_addrmap_region_s;

    if name == levels[0] {
      if levels[0] == path {
        result = me;
      }
      else {
        for each (c) in child_regions {
          tmp_result = c.get_region_first_match(str_join(levels[1..], "."));
          if tmp_result != NULL {
            return tmp_result;
          };
        };
      };
    }
    else {
      for each (c) in child_regions {
        tmp_result = c.get_region_first_match(path);
        if tmp_result != NULL {
          return tmp_result;
        };
      };
    };
  };

  // Simply find the child region whose address range within this region
  // encompasses the given addr. Returns NULL if a child region can't be
  // found.
  parent_offset_to_child(addr : csco_addrmap_addr_s) : csco_addrmap_region_addr_s is {
    for each (c) in child_regions {
      var ctx := c.get_parent_context(me);
      if((addr.offset >= ctx.addr.offset) and (addr.offset < ctx.addr.offset + c.get_region_size())) {
        result = new csco_addrmap_region_addr_s with {
          .region = c;
        };
        result.addr.offset = addr.offset - ctx.addr.offset;
        break;
      };
    };
  };

  // Returns the equivalent offset of 'child' in the enclosing region. Note that
  // 'child' does not have to be a direct child of the enclosing region. They 
  // just need to be connected. 
  child_offset_to_parent(child : csco_addrmap_region_addr_s) : csco_addrmap_region_addr_s is {
    var rpath : list of csco_addrmap_region_s = get_region_path(child.region);

    assert that rpath.size() > 0 else
      error("CSCO_ADDRMAP_ERR_NO_REGION_PATH:\n",  
              "regions ", csco_short_name(), " and ", child.region.csco_short_name(), " are not connected");

    rpath = rpath.reverse();
    result = new csco_addrmap_region_addr_s with {
      .region = rpath.pop0();
    };
    result.addr.offset = child.addr.offset;

    while rpath.size() > 0 {
      result = rpath.pop0().as_a(csco_addrmap_logical_region_s).get_child_addr(result);
    };
  };

  package get_child_addr(child : csco_addrmap_region_addr_s) : csco_addrmap_region_addr_s is {
    result = new csco_addrmap_region_addr_s with {
      .region = me;
    };
    result.addr.offset = child.addr.offset + child.region.get_parent_context(me).addr.offset;
  };

  get_driving_regions(addr : csco_addrmap_addr_s) : list of csco_addrmap_region_addr_s is {
    assert that addr.offset < get_region_size() else
      error("CSCO_ADDRMAP_ERR_ADDR_OUT_OF_RANGE: address offset ", addr.offset, 
            " is out of ", csco_short_name(), "'s legal range (", get_region_size(), ").");

    assert that (addr.route_kind == route_kind) else
      error("CSCO_ADDRMAP_ERR_CHILD_ADDR_ROUTE_KIND_MISMATCH: address passed to get_driving_regions() has ", addr.route_kind, 
              " route_kind, but region has ", route_kind, " route_kind.");

    assert that (addr.addr_space == addr_space) else
      error("CSCO_ADDRMAP_ERR_CHILD_ADDR_ROUTE_KIND_MISMATCH: address passed to get_driving_regions() has ", addr.addr_space, 
              " addr_space, but region has ", addr_space, " addr_space.");
  };

  get_leaf(addr : csco_addrmap_addr_s) : csco_addrmap_leaf_entry_s is {
    assert that addr.offset < get_region_size() else
      error("CSCO_ADDRMAP_ERR_ADDR_OUT_OF_RANGE: address offset ", addr.offset, 
            " is out of ", csco_short_name(), "'s legal range (", get_region_size(), ").");

    assert that (addr.route_kind == route_kind) else
      error("CSCO_ADDRMAP_ERR_CHILD_ADDR_ROUTE_KIND_MISMATCH: address passed to get_leaf() has ", addr.route_kind, 
              " route_kind, but region has ", route_kind, " route_kind.");

    assert that addr.addr_space == addr_space else
      error("CSCO_ADDRMAP_ERR_CHILD_ADDR_ROUTE_KIND_MISMATCH: address passed to get_leaf() has ", addr.addr_space, 
              " addr_space, but region has ", addr_space, " addr_space.");
  };
}; // struct csco_addrmap_logical_region_s like csco_addrmap_region_s


extend ADDRESS'route_kind csco_addrmap_logical_region_s {
  // Number of bytes in this region's address space.
  addr_space_size : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1);
    keep soft addr_space_size == 0;
    csco_print_attr addr_space_size SHORT enable;
    csco_print_attr addr_space_size SHORT precede=parent_contexts header=size;

  get_region_size() : uint(bits:CSCO_ADDRMAP_MAX_ADDR_WIDTH+1) is {
    result = addr_space_size;
  };

  add_child_region_with_offset(c : csco_addrmap_region_s, o : csco_addrmap_addr_t) is {
    var tmp_addr := new ADDRESS'route_kind csco_addrmap_addr_s with {
      .route_kind = route_kind;
      .addr_space = addr_space;
    };
    tmp_addr.offset = o;
    add_child_region(c);
    set_child_addr(c, tmp_addr);
  };

  set_child_addr(c : csco_addrmap_region_s, a : csco_addrmap_addr_s) is also {
    var top_addr : int(bits:*);

    // Expand logical region's size if necessary to accomodate
    // child regions.
    top_addr = a.offset + c.get_region_size();
    if top_addr > get_region_size() {
      addr_space_size = top_addr;
    };
  };

  // Returns all the parent regions which contain eligible drivers and which
  // can access the given addr range in the current region.  When considering 
  // parent region drivers, the method will ensure that the parent region's 
  // address translation allows access to the given offset of the child 
  // region.
  //
  // The 'alg' input can be either SEQUENTIAL or ANY. If SEQUENTIAL is specified,
  // then only regions whose addr map bulding algorithm is sequential 
  // (non_shift_addr_alg == FALSE) in nature is considered.
  get_driving_regions(addr : csco_addrmap_addr_s) : list of csco_addrmap_region_addr_s is also {
    var tmp_rap : csco_addrmap_region_addr_s;
    var parent_rap : csco_addrmap_region_addr_s;

    // Add eligible drivers for the region itself.
    if get_eligible_driver() != NULL {
      tmp_rap = new csco_addrmap_region_addr_s with {
        .region = me;
      };
      tmp_rap.addr.offset = addr.offset;
      result.add(tmp_rap);
    };

    // Consider parent regions' eligible drivers.
    tmp_rap = new csco_addrmap_region_addr_s with {
      .region = me;
    };
    tmp_rap.addr.offset = addr.offset;
    for each (p) in parent_contexts.region.as_a(list of ADDRESS'route_kind csco_addrmap_logical_region_s) {
      parent_rap = p.child_offset_to_parent(tmp_rap);
      if parent_rap != NULL {
        result.add(p.get_driving_regions(parent_rap.addr));
      };
    };
    result = result.unique(it);
  };

  // Takes the given address and goes down the child region hiearchy until
  // it reaches a leaf region.
  get_leaf(addr : csco_addrmap_addr_s) : csco_addrmap_leaf_entry_s is {
    var child_rap : csco_addrmap_region_addr_s;

    var logical := new csco_addrmap_region_addr_s with {
      .region = me;
    };
    logical.addr.offset = addr.offset;
    repeat {
      if logical.region is a ADDRESS'route_kind csco_addrmap_logical_region_s (alog_r) {
        child_rap = alog_r.parent_offset_to_child(logical.addr);
        if (child_rap != NULL ) and (child_rap.region is a csco_addrmap_logical_region_s) {
          logical = child_rap;
        };
      };
    } until logical != child_rap;

    result = new csco_addrmap_leaf_entry_s with {
      .logical = logical;
    };

    if (child_rap != NULL) and (child_rap.region is a csco_addrmap_physical_region_s) {
      result.physical = child_rap;
      result.entry = child_rap.region.as_a(csco_addrmap_physical_region_s).offset_to_entry(child_rap.addr);
    };
  };

  // ?? allow random starting offsets.
  package assign_child_offsets() is {
    var prev_offset : csco_addrmap_addr_t;
    var parent_context : csco_addrmap_region_addr_s;
    var cur_offset : csco_addrmap_addr_t;

    if addr_alg == MANUAL {
      assign_child_offsets_manual_hook();
      return;
    };

    if addr_alg == RANDOM_SHIFT {
      permute_child_regions();
    };

    for each (c) in child_regions {
      cur_offset = c.align_offset(cur_offset);
      assert that cur_offset >= prev_offset else
        error(appendf("CSCO_ADDRMAP_ERR_BULD_MAP_OUT_OF_ADDR: ran out of address space for %s when assigning child offsets, current offset = %#x", 
                        csco_short_name(), cur_offset));

      parent_context = c.get_parent_context(me);
      parent_context.addr.offset = cur_offset;
      prev_offset = cur_offset;
      cur_offset += c.get_region_size();

      // Expand logical region's size if necessary to accomodate
      // child regions.
      if cur_offset > addr_space_size {
        addr_space_size = cur_offset;
      };
    };
  }; // assign_child_offsets() 

  assign_child_offsets_manual_hook() is empty;

  // Ensure that child regions don't have overlapping address ranges within
  // same parent region.
  package check_built_map() is {
    var max_addr : csco_addrmap_addr_t;
    var c_offset : csco_addrmap_addr_t;
    var c_max_addr : csco_addrmap_addr_t;

    assert that addr_space_size != 0 else 
      error(appendf("CSCO_ADDRMAP_ERR_LOG_REGN_EMPTY: address space size for %s cannot be 0", csco_short_name()));

    for each (c) in child_regions.sort(.get_parent_context(me).addr.offset) {
      c_offset = c.get_parent_context(me).addr.offset;
      c_max_addr = c_offset + c.get_region_size() - 1;

      if index != 0 {
        assert that c_offset > max_addr else
          error("CSCO_ADDRMAP_ERR_BUILD_MAP: address range of ", c.csco_short_name(), 
                " is not unique within parent region ", csco_short_name());
      };
      max_addr = c_max_addr;
    };
  };
}; // extend ADDRESS'route_kind csco_addrmap_logical_region_s


extend DEFAULT'addr_space csco_addrmap_logical_region_s {
  keep route_kind == ADDRESS;
};


'>
