// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_seq.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 08/15/2005
//              :
//              :
// Description  : Sequence interface for executing csco_addrmap transactions.
//                There are 2 main types of transactions, with different 
//                addressing schemes:
//
//                OFFSET'addr_scheme:
//                  starting offset is explicitly supplied by the caller 
//                  accompanied by a target region. This scheme should only
//                  be used to access logical regions and byte-width physical
//                  memory entries.
//                  
//                ENTRY'addr_scheme:
//                  starting offset is derived from an addrmap entry specified 
//                  by the caller. This scheme can be used to access any physical
//                  entry, including byte-width memory entries.
//
//                Several limitations apply when specifying an addrmap transaction:
//
//                  - Each transaction may target at most 1 addrmap entry. It's 
//                    possible for a transaction with OFFSET addressing to target
//                    a leaf logical region, in which case no addrmap entry is
//                    referenced.
//
//                  - Currently, only sequential addressing is supported when 
//                    addrmap transactions are mapped from parent regions to child
//                    regions and vice-versa.
//
//                  - Register entries are atomically accessed. i.e., they are 
//                    always read and written as a single entity via an addrmap
//                    transaction. Similarly, each row of a memory entry are atomically
//                    accessed.
//
//                  - 'do' actions in an addrmap sequence must be specified through 
//                    the 'addrmap_do' macro. The macro syntax is identical to the 
//                    native eRM sequence 'do' syntax.
//
//                  - IMPORTANT : each sequence item represents an I/O operation on a 
//                    contiguous segment of the address map. When randomly selecting
//                    parent source regions, it's assumed that the address translation
//                    from the source address to the target address is also sequential.
//                    So the constraints for source regions are written such that only
//                    regions that use sequential addressing down to child regions are
//                    considered.  The user can always segment the target address range
//                    further if he wishes to include more source regions during the
//                    selection process.
//
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_addrmap;


extend csco_trans_kind_t : [ADDRMAP];

// Used to describe data values on a specific field in a specific row of an entry.
struct csco_addrmap_field_value_pair_s {
  !row : csco_addrmap_addr_t;
    csco_print_attr row SHORT enable;

  !name : string;
    csco_print_attr name SHORT enable;

  !value_kind : csco_addrmap_fill_kind_t;
    csco_print_attr value_kind SHORT enable;

  !value : int(bits:*);
    csco_print_attr value SHORT enable;
};


// Used to describe data bytes and corresponding byte mask bits.
struct csco_addrmap_data_and_mask_s {
  !data : list of byte;
  !mask : list of bit;
};


struct csco_addrmap_target_region_addr_s like csco_addrmap_region_addr_s {
  keep addr.offset < read_only(region.get_region_size());
};

// 
// This is the addrmap transaction item. It contains fields that are 
// common to ALL bus based transactions. Fields that are unique to a 
// particular interface should be added in a when sub-type using 
// 'interface' as the when-determinant
struct csco_addrmap_trans_s like csco_trans_s {
  // Indicates how the transaction addresses the target region. 
  // <br>
  // OFFSET : The user supplies a target region pointer and 
  //          an offset within the target region. This should only
  //          be used for transactions targetted for byte memory 
  //          entries or logical regions.
  // <br>
  // ENTRY  : The user supplies an addmap entry name, as well as
  //          a list of field/value pairs. All transactions except
  //          ones targetted for byte memory entries MUST use this
  //          scheme.
  // <br>
  // Note that before an ENTRY transaction is forwarded to the
  // bus interface driver, it will always be converted into an OFFSET
  // transaction, since the interface drivers only undertand address 
  // and data transaction info.
  addr_scheme : csco_addrmap_trans_addr_scheme_t;
    keep soft addr_scheme == ENTRY;

  // Used to determine how to map multi-byte entry rows into the
  // addrmap transaction's byte list. Note that the endianness
  // described here may not be the same as the actual endianness
  // on the interface bus of the source driver.
  endian : csco_addrmap_endian_t;
    keep soft endian == LITTLE;

  // Whether or not this is a backdoor transaction.
  access : csco_addrmap_access_t;
    keep soft access == FRONT_DOOR;
    csco_print_attr access SHORT enable;

  // Indicates the type of I/O operation. 
  opcode : csco_addrmap_op_t;
    csco_print_attr opcode SHORT enable;

  // Used to qualify a FILL opcode with the types of values that should
  // be used for locations where an explicit value is not specified.
  fill_kind : csco_addrmap_fill_kind_t;
    keep fill_kind != SKIP;
    keep soft fill_kind == REPLICATE;

  // Only applicable when fill_kind==CONSTANT, used to hold a constant
  // fill value.
  fill_value : byte;

  // Whether or not this transaction requires a response (non-posted).
  // Note that reads are always non-posted.
  post_kind : csco_addrmap_post_kind_t;
    keep opcode in [READ] => post_kind == NON_POSTED;
    csco_print_attr post_kind SHORT enable;
    csco_print_attr post_kind SHORT header=post;

  // Whether the 'addrmap_do' action will block until the transaction is sent or
  // execute in 0 time.
  // <br>
  // For posted transactions: blocking will wait until TD is on bus.
  // For non-posted transactions: blocking will wait until response has been received.
  // If blocking is FALSE, then the 'do' on this transaction will return
  // immediately.
  blocking_kind : csco_addrmap_trans_return_t;
    keep soft blocking_kind == UNKNOWN;


  // For the fields below, the user can directly constrain them for transactions
  // using OFFSET addressing scheme. For transaction using ENTRY addressing scheme,
  // they are automatically constrained to values derived from the entry and field
  // value pair specifications, so the user MUST NOT constrain these fields for
  // ENTRY addressing scheme.
  //
  // target
  // num_bytes
  // data
  // data_mask

  // Contains pointer to target region and starting address within the
  // target region.
  target : csco_addrmap_target_region_addr_s;
    keep soft target.region == NULL;
    csco_print_attr target SHORT enable;

  !leaf : csco_addrmap_leaf_entry_s;
    attribute leaf deep_copy = reference;

  // Total number of bytes in the transaction. The user should only directly constrain
  // this field if the transaction has an OFFSET'addr_scheme.
  num_bytes : csco_addrmap_addr_t;
    keep num_bytes <= read_only(target.region.get_region_size() - target.addr.offset);
    keep soft num_bytes > 0;

  // Data of this transaction. Index 0 corresponds to byte offset 0, and so on.
  data : list of byte;
    keep opcode == READ => all of {
      data.size() == num_bytes;
      for each (b) in data {
        b == 0;
      };
    };
    csco_print_attr data LONG header=data_and_mask;
    csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "data" and pf.format == LONG {
        var ind := pf.header.as_a(int);
try {
        result = append(result, "(", data_mask[ind], ")");
} else {
out("LOOK HERE");
  print me ;
};
      };
    };

  // A '1' in the bit location means the corresponding byte is enabled.
  // <br>
  // The user should only directly constrain
  // this field if the transaction has an OFFSET'addr_scheme.
  data_mask : list of bit;
    keep data_mask.size() == data.size();
    keep for each (m) in data_mask {
      soft m == 1;
    };
    csco_print_attr data_mask ALL disable;

  keep gen (target) before (num_bytes);

  // Source region and address for transaction. Typically,
  // the user should not need to constrain this field.
  source : csco_addrmap_region_addr_s;
    csco_print_attr source LONG expand=ON;

  // Indicates interface of source region from which the the addrmap transaction.
  // will be launched.
  // <br>
  // Also a when-determinant that can be used to create new sequence item 
  // sub-types for each unique bus interface. The fields that may live in
  // such a sub-type include:
  // <br>
  // o segmentation algorithm to determine how a large logical addrmap transaction
  //   should be frgmented into discrete transactions on the physical interface.
  // <br>
  // o interface specific bus attributes.
  // <br>
  // o etc...
  //
  // The user shouldn't constrain this field. To constrain the transaction to use 
  // only a given interface, the user should constrain the 'source' field
  // instead. e.g.: 
  //
  //   keep source in avail_sources.all(.region.get_interface() == XXX);
  //
  interface : csco_addrmap_interface_t;

  // Indicates 'route_kind' of the source region that will drive the transaction.
  //
  //
  // The user shouldn't constrain this field. To constrain the transaction to use 
  // only a given route_kind, the user should constrain the 'source' field
  // instead. e.g.: 
  //
  //   keep source in avail_sources.all(.region.get_route_kind() == YYY);
  //
  // This field is only intended as a when determinant for defining address space
  // specific fields and methods.
  route_kind : csco_addrmap_route_kind_t;

  // Indicates 'addr_space' of the source region that will drive the transaction.
  //
  //
  // The user shouldn't constrain this field. To constrain the transaction to use 
  // only a given addr_space, the user should constrain the 'source' field
  // instead. e.g.: 
  //
  //   keep source in avail_sources.all(.region.get_addr_space() == YYY);
  //
  // This field is only intended as a when determinant for defining address space
  // specific fields and methods.
  addr_space : csco_addrmap_space_t;

  keep gen (interface, route_kind, addr_space) before (tracking, rate_stats);
  keep gen (access) before (avail_sources);

  when FRONT_DOOR'access csco_addrmap_trans_s {
    keep avail_sources == get_avail_sources(addrmap_parent_seq, target);
    keep source in read_only(avail_sources);
    keep interface == read_only(source.region.get_interface());
    keep route_kind == read_only(source.region.get_route_kind());
    keep addr_space == read_only(source.region.get_addr_space());
  };

  // Specman really needs to have a 'gen_first' constraint.
  keep gen (interface, route_kind, addr_space) before (opcode, post_kind, num_bytes, data, 
                                                       data_mask, fill_kind, fill_value, 
                                                       blocking_kind, endian);

  when BACK_DOOR'access csco_addrmap_trans_s {
    keep avail_sources.size() == 0;
    keep source == NULL;
    keep interface == NONE;
    keep route_kind == ADDRESS;
    keep addr_space == DEFAULT;
  };

  // Indicates response valid.
  !response_done_flag : bool;
    csco_print_attr response_done_flag ALL header=response_done;

  !response_kind : csco_addrmap_resp_kind_t;

  // Emitted when the set_response() method is called, indicating that the response
  // data is now safe to sample.
  event response_done;

  // User hook to filter out source driving region from the list of all eligible sources.
  filter_avail_sources(pseq : csco_addrmap_seq, all_elig : list of csco_addrmap_region_addr_s) : list of csco_addrmap_region_addr_s is {
    result = all_elig;
  };


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  addrmap_parent_seq : csco_addrmap_seq;
    keep soft addrmap_parent_seq == NULL;
    attribute addrmap_parent_seq deep_copy=ignore;

  // *********** IMPORTANT ************
  // The methods should not be called by the user as part of constraining
  // the addrmap transaction. The methods are intended to be used by logic to 
  // query the addrmap transaction AFTER it has been generated.

  // Categorize opcode into either READ or WRITE, regardless of 
  // direction. 
  read_or_write() : csco_addrmap_op_t is {
    if opcode in [READ] {
      result = READ;
    }
    else if opcode in [WRITE, FILL] {
      result = WRITE;
    }
    else {
      assert FALSE else
        error(appendf("CSCO_ADDRMAP_ERR_UNRECOG_OP: unable to categorize %s as either read or write opcode", opcode));
    };
  };

  // This method should ALWAYS be used to return response data to a non-posted
  // addrmap request transaction. It will unblock a BLOCKING request transaction.
  set_response(resp : list of byte, resp_kind : csco_addrmap_resp_kind_t) is {
    message(CSCO_ADDRMAP, MEDIUM, "set_response() called for ", csco_short_name());
    assert that post_kind == NON_POSTED else
      error("CSCO_ADDRMAP_ERR_UNEXPECTED_RESP: set_response() called for ", csco_short_name(), ", which is a POSTED transaction.");

    response_kind = resp_kind;
    if read_or_write() == READ {
      assert that resp.size() == num_bytes or response_kind != NORMAL else
        error("CSCO_ADDRMAP_ERR_SET_RESPONSE_SIZE: set_response() called with ", resp.size(), 
                " bytes of response data, but the read request asked for ", num_bytes, " bytes.");

      data = resp.copy();
    }
    else {
      assert that resp.size() == 0 else
        error("CSCO_ADDRMAP_ERR_SET_RESPONSE_SIZE: set_response() called with ", resp.size(), 
                " bytes of response data, but the write request doesn't expect any data in the response.");
    };
  };

  // Returns source region, from which this transaction will be launched.
  get_source_region() : csco_addrmap_logical_region_s is {
    if source != NULL and source.region is a csco_addrmap_logical_region_s (log_r) {
      result = log_r;
    };
  };

  // Returns address of source region. This is the address that the driving
  // sequence driver should place onto the bus.
  get_source_address() : csco_addrmap_addr_s is {
    if source != NULL {
      result = source.addr;
    };
  };

  // Returns the interface driver that will push this transaction onto the
  // bus.
  get_source_driver() : any_sequence_driver is {
    if source != NULL {
      result = source.region.get_eligible_driver();
    };
  };

  // Returns a pointer to the addrmap entry to which this transaction is
  // targeting. Note that this method could return NULL, if the transaction
  // is targetted toward a leaf logical region as an OFFSET transaction.
  get_entry_ptr() : csco_addrmap_entry_s is {
    if leaf != NULL {
      result = leaf.entry;
    };
  };

  // Returns offset of the addrmap entry to which this transaction is 
  // targeting. Only valid if get_entry_ptr() returns non-NULL.
  get_entry_offset() : csco_addrmap_addr_t is {
    if leaf.entry != NULL {
      result = leaf.get_entry_offset();
    };
  };

  csco_short_name() : string is only {
    result = appendf("%s_%s_%s_TR(0x%x)", tracking.origin, tracking.trans_kind, opcode, tracking.get_id());
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  macro_call : bool;
    keep soft macro_call == FALSE;
    csco_print_attr macro_call ALL disable;

  package is_write_req : bool;
    keep is_write_req == (opcode in [WRITE, FILL]);
    csco_print_attr is_write_req ALL disable;

  package is_read_req : bool;
    keep is_read_req == (opcode in [READ]);
    csco_print_attr is_read_req ALL disable;

  // addrmap transactions are always generated.
  keep tracking.trans_kind == ADDRMAP;

  // holds list of eligible driving regions.
  avail_sources : list of csco_addrmap_region_addr_s;
    csco_print_attr avail_sources ALL disable;


  package get_avail_sources(pseq : csco_addrmap_seq, target_i : csco_addrmap_region_addr_s) : list of csco_addrmap_region_addr_s is {
    result = target_i.region.get_driving_regions(target_i.addr);
    if result.is_empty() {
      try {
        csco_print(LONG);
      }; -- use of "try" avoids potential failure in print

      if addr_scheme == OFFSET {
        assert FALSE else
          error("CSCO_ADDRMAP_ERR_NO_SOURCE_REGION_FOR_OFFSET_TRANS:\n",
                "Cannot execut addrmap ", opcode,  " to target region ", target_i.region.csco_short_name(), 
                ". Could not find a source region that could drive target region. ", 
                " Make sure that addrmap regions are properly linked ", 
                "and that the 'addr_alg' and 'eligible_driver' fields of all logical regions have been properly set.");
      }
      else if me is a ENTRY csco_addrmap_trans_s (entry_t) {
        assert FALSE else
          error("CSCO_ADDRMAP_ERR_NO_SOURCE_REGION_FOR_ENTRY_TRANS:\n",
                "Cannot execut addrmap ", opcode, " to addrmap entry ", entry_t.entry_ptr_priv.long_name(""),
                ". Could not find a source region that could drive target region ", target_i.region.csco_short_name(),
                " Make sure that addrmap regions are properly linked ",
                "and that the 'addr_alg' and 'eligible_driver' field of all logical regions have been properly set.");
      };
    }

    else if not result.region.has(.get_eligible_driver() != NULL) {
      try {
        csco_print(LONG);
      }; -- use of "try" avoids potential failure in print

      if addr_scheme == OFFSET {
        assert FALSE else
          error("CSCO_ADDRMAP_ERR_NO_ELGB_DRVR: cannot execut addrmap ", opcode, " to target region ", target_i.region.csco_short_name(), 
                    ". Could not find a source region that contains eligible driver.",
                    "Make sure that each interface region has set its eligible driver via the set_eligible_driver() method",
                    "Make sure that all register interfaces have corresponding interface regions");
      }
      else {
        assert FALSE else
          error("CSCO_ADDRMAP_ERR_NO_ELGB_DRVR: cannot execut addrmap ", opcode, " to addrmap entry ", get_entry_ptr().long_name(""),
                    ". Could not find a source region that contains an eligible driver.",
                    "Make sure that each interface region has set its eligible driver via the set_eligible_driver() method",
                    "Make sure that the addrmap name scope tree only contains nodes which are accessible from the current testbench");
      };
    };

    result = filter_avail_sources(pseq, result);
  }; // get_avail_sources()

  post_generate() is also {
    assert that blocking_kind != UNKNOWN else 
      error("The project lead should soft constrain the 'blocking_kind' field of ", 
              "'csco_addrmap_trans_s' to a reasonable default value (BLOCKING/NON_BLOCKING");

    if opcode == FILL and fill_kind == REPLICATE {
      assert that data.size() > 0 else 
        error("CSCO_ADDRMAP_ERR_INVLD_REPLICATE_FILL_OP: cannot execute FILL-REPLICATE operation when no data has been specified");
    };

    find_leaf();
  };

  find_leaf() is {
    if leaf == NULL {
      var end_addr : csco_addrmap_addr_s = deep_copy(target.addr);
      end_addr.offset += num_bytes - 1;

      leaf = target.region.get_leaf(target.addr);
      var end_leaf := target.region.get_leaf(end_addr);

      assert that leaf.entry == end_leaf.entry else
        error("CSCO_ADDRMAP_ERR_MULTIPLE_TARGET_ENTRY:\n",
              "each addrmap transaction must not target more than 1 entry.");

      if leaf.logical == NULL and access == FRONT_DOOR {
        var all_pregions := leaf.physical.region.get_parent_regions();
        if all_pregions.size() > 0 {
          var lregion := all_pregions.first(.addr_space == source.region.get_addr_space());
          if lregion != NULL {
            leaf.logical = lregion.get_child_addr(leaf.physical);
          };
        };
      };
    };
  };
}; // struct csco_addrmap_trans_s like csco_trans_s


extend OFFSET'addr_scheme csco_addrmap_trans_s {
  when is_write_req OFFSET'addr_scheme csco_addrmap_trans_s {
    keep data.size() <= num_bytes;
    keep opcode == WRITE => data.size() == num_bytes;
    keep opcode == FILL and fill_kind == REPLICATE => data.size() > 0;
  };

  when is_read_req OFFSET'addr_scheme csco_addrmap_trans_s {
    //keep data.size() == 0;
    post_generate() is also {
      //data.resize(num_bytes, TRUE, 0, FALSE);
      //data_mask.resize(num_bytes, TRUE, 1, FALSE);
    };
  };


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////


  ////////////////////////
  // Private Fields 
  ////////////////////////
  set_response(resp : list of byte, resp_kind : csco_addrmap_resp_kind_t) is also {
    response_done_flag = TRUE;
    emit response_done;
  };

  post_generate() is also {
    //assert that target.region is not a csco_addrmap_physical_region_s else
    //  error("CSCO_ADDRMAP_ERR_OFFSET_TR_TO_PHYSICAL_REGION:\n",
    //        "OFFSET transactions to physical regions are not allowed, use ENTRY transactions instead.");

    //if route_kind == ADDRESS {
    //  assert that target.region.get_child_regions().is_empty() else 
    //    error("CSCO_ADDRMAP_ERR_OFFSET_TR_TO_NON_LEAF_REGION:\n",
    //            "OFFSET based addrmap transactions can only be issued to leaf target regions.\n",
    //            "Otherwise, it might access an unintended physical entry.");
    //};
  };

  run() is also {
    if get_entry_ptr() != NULL {
      assert that get_entry_offset() + num_bytes <= get_entry_ptr().get_entry_size() else
        error("CSCO_ADDRMAP_ERR_INVLD_TR_SZ:\n",
              "addrmap transaction size exceeds target memory entry boundary");

      assert that get_entry_offset() % get_entry_ptr().get_entry_width() == 0 else
        error("CSCO_ADDRMAP_ERR_OFFSET_TRANS_START_ALIGNMENT:\n", 
              csco_short_name(), " to offset ,", hex(get_entry_offset()),
              ", is not aligned with target entry (", get_entry_ptr().long_name(""), ")'s width (", 
              get_entry_ptr().get_entry_width(), ").");
    };

    if get_entry_ptr() == NULL and opcode == FILL and num_bytes > data.size() {
      assert fill_kind != SHADOW else
        error("CSCO_ADDRMAP_ERR_FILL_KIND : ", csco_short_name(), 
              " doesn't target a physical addrmap entry, cannot use SHADOW fill_kind.");
    };
  }; // run() 
}; // extend OFFSET'addr_scheme csco_addrmap_trans_s


// Transactions that use ENTRY addressing scheme should NEVER explicitly
// constrain the target, num_bytes, data, and data_mask fields in 
// the base transaction type. Those base fields will be automatically 
// set based on the field value pairs provided in the ENTRY transaction.
extend ENTRY'addr_scheme csco_addrmap_trans_s {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Scope under which to search for 'entry'.  By default, the scope is 
  // constrained to use the parent sequence's scope.
  scope : csco_addrmap_name_scope_s;
    attribute scope deep_copy = reference;
    // JHZ_HACK: This constraint doesn't work because in our addrmap_do macro, we gen
    // the item first, and Specman doesn't correctly set 'parent_sequence'
    // during a gen (only during a do).  We've opened an RFE with Cadence.
    keep read_only(addrmap_parent_seq != NULL) => soft scope == addrmap_parent_seq.default_scope;


  // Name/Path of addrmap entry to access. The string can be a '.' delimited path
  // which includes scope tree node names. The sequence interface will use the 
  // supplied path, starting at 'scope', to search for the entry name. The last token in 
  // the 'entry' string MUST be the name of an entry.
  entry : string;
    keep soft entry == "";
    csco_print_attr entry ALL disable;

  // This is an alternative way to specify the target entry for this transaction. Instead
  // of using the 'entry' name scope path string field above, the user can just directly
  // pass in a pointer to the entry he wishes to access.
  entry_ptr : csco_addrmap_entry_s;
    keep soft entry_ptr == NULL;
    attribute entry_ptr deep_copy = reference;
    csco_print_attr entry_ptr ALL disable;

  // Always 0 for register entry transactions. For memory entry, indicates starting
  // row in memory covered by this transaction. For a WRITE/FILL transaction, this field
  // is automatically constrained to the minimum specified row value in 'fields'.
  start_row : csco_addrmap_addr_t;
    keep start_row < read_only(entry_ptr_priv.get_entry_depth());
    keep entry_ptr_priv is a csco_addrmap_reg_entry_s => start_row == 0;

  // Always 0 for register entry transactions. For memory entry, indicates ending
  // row in memory covered by this transaction. For a WRITE transaction, this field is
  // automatically constrained to the maximum specified row value in 'fields'. For a FILL
  // transaction, 'end_row' can be any value >= the maximum specified row value in 'fields'.
  end_row : csco_addrmap_addr_t;
    keep end_row < read_only(entry_ptr_priv.get_entry_depth());
    keep end_row >= start_row;
    keep entry_ptr_priv is a csco_addrmap_reg_entry_s => end_row == 0;

  // For unspecified fields between 'start_row' and 'end_row', this indicates the type of value
  // that is used for those unspecified fields. 'SKIP' is only applicable for a particular row 
  // if NONE of fields from that row has been specified. Otherwise, 'SKIP' takes on the same 
  // behavior as 'SHADOW'.
  unspecified_field_kind : csco_addrmap_fill_kind_t;
    keep unspecified_field_kind in [SHADOW, RANDOM, SKIP];
    keep soft unspecified_field_kind == SKIP;

  // WRITES:
  // <br>
  // =======
  // <br>
  // For a single row in an entry, the user should use the set_field*() 
  // and set_reg*() methods to set the desired values for select fields.  
  // The 'row' field of each field_value_pair in this case are set to 
  // 'start_row'.
  // <br>
  // For multi-row memory entry writes, the user should use the 
  // set_field_and_row*() and set_row*() methods to set the desired values 
  // for select fields in select rows. The driver will fill in any unspecified 
  // fields between 'start_row' and 'end_row' according to 'unspecified_field_kind'.
  // <br>
  // For FILL writes, the driver will take all rows from 'start_row' to
  // the last row with at least 1 specified field, fill in any unspecified 
  // fields according to 'unspecified_field_kind', then fill all the rows over 
  // the entire 'start_row' 'end_row' range according to 'fill_kind'. If 
  // 'fill_kind' is REPLICATE, then the pattern from 'start_row' to the last
  // row with at least 1 specified field will be repeated to fill up to 'end_row'.
  //
  // READS:
  // <br>
  // ======
  // <br>
  // This list will be automatically set based on start_row and end_row,
  // the user should NEVER explicity constrain this list for a read trans.
  // <br>
  // The interface driver MUST fill in all the field value pairs from
  // start_row to end_row once the read completes.
  fields : list of csco_addrmap_field_value_pair_s;
    keep gen (entry_ptr_priv) before (fields);
    keep soft fields.size() == 0;


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  // The set_field(), set_reg(), set_field_and_row(), and set_row() are methods
  // used to define constraints on the 'fields' list within a csco_addrmap 
  // write transaction.  The constraint block should look something like:
  // 
  // keep fields == {
  //                  .set_field("<field_name>", <val>);
  //                  .set_reg(<val>);
  //                  .set_field_and_row("<row_num>", <field_name>, <val>);
  //                  .set_row("<row_num>", <row_val>);
  //                  ...
  //                };
  // <br>
  // <br>
  // set_field_and_row() returns a field/value tuple for the specified row with the specified value.
  set_field_and_row(r:csco_addrmap_addr_t, fn:string, val:int(bits:*)) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(r, fn, CONSTANT, val);
  }; 

  // Returns a field/value tuple for all fields in the specified row with the specified value.
  set_row(r : csco_addrmap_addr_t, val : int(bits:*)) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(r, "__ALL_FIELDS__", CONSTANT, val);
  };

  // Returns a field/value tuple for the specified field in row 0 with the specified value.
  set_field(fn : string, val : int(bits:*)) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(0, fn, CONSTANT, val);
  }; 

  // Returns a field/value tuple for all fields in row 0 with the specified value.
  set_reg(val : int(bits:*)) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(0, "__ALL_FIELDS__", CONSTANT, val);
  };

  // set_field_and_row() returns a random field/value tuple for the specified row.
  set_field_and_row_rand(r:csco_addrmap_addr_t, fn:string) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(r, fn, RANDOM, 0);
  }; 

  // Returns a random field/value tuple for all fields in the specified row.
  set_row_rand(r : csco_addrmap_addr_t) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(r, "__ALL_FIELDS__", RANDOM, 0);
  };

  // Returns a random field/value tuple for the specified field in row 0.
  set_field_rand(fn : string) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(0, fn, RANDOM, 0);
  }; 

  // Returns a field/value tuple for all fields in row 0.
  set_reg_rand() : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(0, "__ALL_FIELDS__", RANDOM, 0);
  };

  // set_field_and_row() returns a field/value tuple for the specified row using the current shadow value.
  set_field_and_row_shadow(r:csco_addrmap_addr_t, fn:string) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(r, fn, SHADOW, 0);
  }; 

  // Returns a field/value tuple for all fields in the specified row using the current shadow value.
  set_row_shadow(r : csco_addrmap_addr_t) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(r, "__ALL_FIELDS__", SHADOW, 0);
  };

  // Returns a field/value tuple for the specified field in row 0 using the current shadow value.
  set_field_shadow(fn : string) : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(0, fn, SHADOW, 0);
  }; 

  // Returns a field/value tuple for all fields in row 0 using the current shadow value.
  set_reg_shadow() : csco_addrmap_field_value_pair_s is {
    result = set_field_priv(0, "__ALL_FIELDS__", SHADOW, 0);
  };

  // Randomly returns a list of field/value tuples for the specified entry, without exceeding
  // the 'max_bytes' constraint.
  set_fields_rand(entry_ptr : csco_addrmap_entry_s, max_bytes : csco_addrmap_addr_t) : list of csco_addrmap_field_value_pair_s is {
    var srow : csco_addrmap_addr_t;
    var erow : csco_addrmap_addr_t;
    var max_rows : csco_addrmap_addr_t;

    max_rows = max_bytes / entry_ptr.get_entry_width();

    gen srow keeping {
      it < read_only(entry_ptr.get_entry_size() / entry_ptr.get_entry_width());
    };

    gen erow keeping {
      it < read_only(entry_ptr.get_entry_size() / entry_ptr.get_entry_width());
      it >= read_only(srow);
      it <= read_only(srow + max_rows - 1);
    };

    result.add(set_row_rand(srow));
    if erow > srow {
      result.add(set_row_rand(erow));
    };
  };

  package set_field_priv(r:csco_addrmap_addr_t, fn:string, vkind : csco_addrmap_fill_kind_t, val:int(bits:*)) : csco_addrmap_field_value_pair_s is {
    assert that vkind in [CONSTANT, SHADOW, RANDOM] else 
      error("CSCO_ADDRMAP_ERR_SET_FIELD: value kind of ", vkind, " is not supported by the set_field() methods.");

    result = new csco_addrmap_field_value_pair_s with {
      .row = r;
      .name = fn;
      .value_kind = vkind;
      .value = val;
    };
  };

  // Returns the value of a specific field in a specific row. For register entry,
  // 0 should be used as the 'r' input parameter.  Can be used after a read trans.
  // to fetch read responses.
  get_field_and_row(r : uint, fname : string) : int(bits:*) is {
    var fvpair : csco_addrmap_field_value_pair_s;

    fvpair = fields.first(.row == r and .name == fname);
    assert that fvpair != NULL else 
      error(appendf("CSCO_ADDRMAP_ERR_INVLD_FLD_ROW: invalid field/row specified to get_field_and_row(): %s/%d", fname, r));

    result = fvpair.value;
  }; // get_field_value()

  // Returns a row of fields as an integer. Can be used after a read transaction
  // to fetch read responses.
  get_row(r : uint) : int(bits:*) is {
    var fptr : csco_addrmap_field_s;
    for each (f) in fields.all(.row == r) {
      fptr = entry_ptr_priv.get_field(f.name);
      result[fptr.lsb_pos+fptr.width-1:fptr.lsb_pos] = f.value;
    };
  };

  // Returns all fields in 'start_row' as an integer. Can be used to fetch read response
  // after a register read.
  get_reg() : int(bits:*) is {
    result = get_row(start_row);
  };

  // Returns a specific field in 'start_row' as an integer. Can be used to fetch read response
  // after a register read.
  get_field(fname : string) : int(bits:*) is {
    result = get_field_and_row(start_row, fname);
  };

  // This method is used to process the user specified 'fields' and translate them
  // into actual values based on the 'value_kind' attribute of each tuple. 
  package process_fields(fvp : list of csco_addrmap_field_value_pair_s, ent : csco_addrmap_entry_s) : list of csco_addrmap_field_value_pair_s is {
    var cur_row : csco_addrmap_addr_t;
    var fptr : csco_addrmap_field_s;

    // Do some error checking. Also, for each field specification whose .row==UNDEF, change .row to start_row.
    // For each field specification whose .name == "__ALL_FIELDS__", create a new list of fields spanning all
    // fields of the entry and assign each the correct bit-slice value from the original value specification.
    for each (f) in fvp {
      var val : int(bits:*);
      var tmp_f : csco_addrmap_field_value_pair_s;

      case (f.value_kind) {
        SHADOW : {
          if fptr == NULL or cur_row != f.row {
            if ent is a csco_addrmap_reg_entry_s (reg_entry) {
              val = reg_entry.get_reg_shadow_val();
            }
            else if ent is a csco_addrmap_mem_entry_s (mem_entry) {
              mem_entry.unpack_mem_fields(f.row);
              val = mem_entry.pack_fields(SHADOW);
            };
          };
        };

        RANDOM : {
          var bit_list : list of bit;
          gen bit_list keeping {
            .size() == ent.get_entry_width() << 3;
          };
          val = bit_list[:];
        };

        CONSTANT : {
          val = f.value;
        };
      };

      if f.name == "__ALL_FIELDS__" {
        for i from 0 to ent.fields.size() - 1 {
          fptr = ent.fields[i];
          tmp_f = new csco_addrmap_field_value_pair_s with {
            .row = f.row;
            .name = fptr.name;
            .value_kind = f.value_kind;
            .value = (val >> fptr.lsb_pos)[fptr.width-1:0];
          };
          result.add(tmp_f);
        };
      }
      else {
        fptr = ent.get_field(f.name);
        tmp_f = new csco_addrmap_field_value_pair_s with {
          .row = f.row;
          .name = f.name;
          .value_kind = f.value_kind;
          .value = (.value_kind == CONSTANT) ? f.value : (val >> fptr.lsb_pos)[fptr.width-1:0];
        };
        result.add(tmp_f);
      };
    }; // for each (f) in fvp
  }; // processe_fields()

  // Converts field specifications into data/data_mask fields.
  package set_data(flds : list of csco_addrmap_field_value_pair_s, entry_ptr : csco_addrmap_entry_s) : csco_addrmap_data_and_mask_s is {
    var min_specified_row : csco_addrmap_addr_t;
    var max_specified_row : csco_addrmap_addr_t;
    var tmp_int : int(bits:*);
    var bit_list : list of bit;
    var byte_list : list of byte;
    var fptr : csco_addrmap_field_s;

    result = new;
    min_specified_row = flds.is_empty() ? 0 : flds.min_value(.row);
    max_specified_row = flds.is_empty() ? 0 : flds.max_value(.row);

    assert that min_specified_row <= entry_ptr.get_entry_depth() else 
      error("CSCO_ADDRMAP_ERR_ILLEGAL_TRANS_FIELDS_SPEC\n:", 
              "minimum specified row in fields ( ", min_specified_row, ") must be <= depth (", 
               entry_ptr.get_entry_depth(), ") of target entry ", entry_ptr.long_name(""));


    if entry_ptr is a csco_addrmap_reg_entry_s (reg_entry) {
      result.mask.resize(reg_entry.get_entry_width(), TRUE, 1, FALSE);

      // Initialize result according to unspecified_field_kind.
      case (unspecified_field_kind) {
        RANDOM : {
          gen bit_list keeping {
            it.size() == reg_entry.get_entry_width() * 8;
          };
          tmp_int = bit_list[:];
        };

        SHADOW : {
          tmp_int = reg_entry.get_reg_shadow_val();
        };
      };

      for each (f) in flds {
        fptr = entry_ptr.get_field(f.name);
        tmp_int[fptr.lsb_pos+fptr.width-1:fptr.lsb_pos] = f.value;
      };
      if endian == BIG {
        unpack(packing.high, tmp_int[..entry_ptr.get_entry_width()*8-1], result.data);
      }
      else {
        unpack(packing.low, tmp_int[..entry_ptr.get_entry_width()*8-1], result.data);
      };
    }

    else if entry_ptr is a csco_addrmap_mem_entry_s (mem_entry) {
      for r from min_specified_row to max_specified_row {
        case (unspecified_field_kind) {
          RANDOM : {
            gen bit_list keeping {
              it.size() == mem_entry.get_entry_width() * 8;
            };
            tmp_int = bit_list[:];
          };

          [SHADOW, SKIP] : {
            mem_entry.unpack_mem_fields(r);
            tmp_int = mem_entry.pack_fields(SHADOW);
          };
        };

        if flds.has(.row == r) {
          bit_list.resize(mem_entry.get_entry_width(), TRUE, 1, FALSE);
          for each (f) in flds.all(.row == r) {
            fptr = entry_ptr.get_field(f.name);
            tmp_int[fptr.lsb_pos+fptr.width-1:fptr.lsb_pos] = f.value;
          };
        }
        else if unspecified_field_kind == SKIP {
          bit_list.resize(mem_entry.get_entry_width(), TRUE, 0, FALSE);
          tmp_int = 0;
        }
        else {
          bit_list.resize(mem_entry.get_entry_width(), TRUE, 1, FALSE);
        };

        if endian == BIG {
          unpack(packing.high, tmp_int[..entry_ptr.get_entry_width()*8-1], byte_list);
        }
        else {
          unpack(packing.low, tmp_int[..entry_ptr.get_entry_width()*8-1], byte_list);
        };
        result.data.add(byte_list);
        result.mask.add(bit_list);
      };
    };
  }; // set_data()

  // This extension for ENTRY transactions will also use the reponse list of bytes
  // to fill the 'fields' of the transaction.
  set_response(resp : list of byte, resp_kind : csco_addrmap_resp_kind_t) is also {
    var tmp_int : int(bits:*);
    var fptr : csco_addrmap_field_s;
    var ind1 : csco_addrmap_addr_t;
    var ind2 : csco_addrmap_addr_t;

    if read_or_write() == READ {
      assert that fields.size() == get_entry_ptr().fields.size() * (end_row - start_row + 1) else
        error("CSCO_ADDRMAP_ERR_SET_RESPONSE_FIELDS: the 'fields' of ", csco_short_name(), 
                " has not been properly initialized to accept read response data.",
                " Expect ", get_entry_ptr().fields.size() * (end_row - start_row + 1), " entries, but only ", 
                fields.size(), " are present.");

      if resp.size() == num_bytes {
        if get_entry_ptr() is a csco_addrmap_reg_entry_s (reg_entry) {
          if endian == BIG {
            tmp_int = (resp.reverse())[:];
          }
          else {
            tmp_int = resp[:];
          };
          for each (f) in fields {
            fptr = reg_entry.get_field(f.name);
            f.value = tmp_int[fptr.lsb_pos+fptr.width-1:fptr.lsb_pos];
          };
        }
        else if get_entry_ptr() is a csco_addrmap_mem_entry_s (mem_entry) {
          for r from start_row to end_row {
            ind1 = (r - start_row) * mem_entry.get_entry_width();
            ind2 = ((r - start_row + 1) * mem_entry.get_entry_width()) - 1;

            if endian == BIG {
              tmp_int = (resp[ind1..ind2].reverse())[:];
            }
            else {
              tmp_int = resp[ind1..ind2][:];
            };
            for each (f) in fields.all(.row == r) {
              fptr = mem_entry.get_field(f.name);
              f.value = tmp_int[fptr.lsb_pos+fptr.width-1:fptr.lsb_pos];
            };
          };
        };
      };
    };
    response_done_flag = TRUE;
    emit response_done;
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  // If the entry has multiple matches within the specified
  // name scope, and the user does not provide further constraints,
  // then the entry is randomly chosen amongst the matches.
  package entry_ptr_priv : csco_addrmap_entry_s;
    keep entry_ptr_priv == lookup_entry(scope, entry, entry_ptr);
    keep read_only(entry_ptr_priv) is a csco_addrmap_reg_entry_s => all of {
      unspecified_field_kind != SKIP;
      soft unspecified_field_kind == SHADOW;
    };
    attribute entry_ptr_priv deep_copy = reference;
    csco_print_attr entry_ptr_priv SHORT enable;
    csco_print_attr entry_ptr_priv SHORT header=entry_name;

  package post_processed_fields : list of csco_addrmap_field_value_pair_s;

  keep target.region == read_only(entry_ptr_priv.parent_region);
  keep target.addr.offset == read_only(entry_ptr_priv.offset + start_row * entry_ptr_priv.get_entry_width());
  keep num_bytes == read_only((end_row - start_row + 1) * entry_ptr_priv.get_entry_width());
  keep gen (entry_ptr_priv, start_row, end_row) before (target, num_bytes) ; -- fixes bug where read_only() does not force this order

  when TRUE'is_write_req ENTRY'addr_scheme csco_addrmap_trans_s {
    package min_fields_row : csco_addrmap_addr_t;
      keep min_fields_row == ((fields.is_empty()) ? 0 : fields.min_value(.row));
      keep start_row == min_fields_row;

    package max_fields_row : csco_addrmap_addr_t;
      keep max_fields_row == ((fields.is_empty()) ? 0 : fields.max_value(.row));
      keep end_row >= max_fields_row;
      keep opcode == WRITE => end_row == max_fields_row;

    keep post_processed_fields == process_fields(fields, entry_ptr_priv);
    post_generate() is also {
      fields = post_processed_fields;
      var data_mask_pair : csco_addrmap_data_and_mask_s = set_data(fields, get_entry_ptr());
      data = data_mask_pair.data;
      data_mask = data_mask_pair.mask;
    };
  };

  when TRUE'is_read_req ENTRY'addr_scheme csco_addrmap_trans_s {
    post_generate() is also {
      fields = set_read_fields(entry_ptr_priv, start_row, end_row);
      post_processed_fields = fields;
      var data_and_mask := set_data(fields, get_entry_ptr());
      data = data_and_mask.data;
      data_mask = data_and_mask.mask;
    };

    // Initialize list of field/value pairs to hold eventual read response data.
    package set_read_fields(ent : csco_addrmap_entry_s, sr : csco_addrmap_addr_t, er : csco_addrmap_addr_t) : list of csco_addrmap_field_value_pair_s is {
      var fvp : csco_addrmap_field_value_pair_s;
      for r from sr to er {
        for each (f) in ent.fields {
          fvp = new csco_addrmap_field_value_pair_s with {
            .row = r;
            .name = f.name;
            .value_kind = CONSTANT;
            .value = 0;
          };
          result.add(fvp);
        };
      };
    }; // set_read_fields()
  }; 

  package lookup_entry(s : csco_addrmap_name_scope_s, n : string, p : csco_addrmap_entry_s) : csco_addrmap_entry_s is {
    var str_entry : csco_addrmap_entry_s;

    assert that s != NULL else
      error("CSCO_ADDRMAP_ERR_NULL_ADDRMAP_ENTRY_ITEM_SCOPE:\n",
            "The current transaction's (", csco_short_name(), ") scope has not been correctly constrained, ",
            "the scope is NULL", csco_short_name(), ".\nPossible reasons are:\n",
            " - addrmap_do or SIMPLE_ENTRY sequence wasn't used to generate item.\n",
            " - a parent sequence leading up to item was explicitly generated (without using 'do')\n",
            "   without constraining its driver or scope field");

    assert that n != "" or p != NULL else 
      error("CSCO_ADDRMAP_ERR_NULL_ENTRY:\n",
            "the current transaction's (", csco_short_name(), 
            ") entry has not been correctly constrained");

    if n != "" {
      str_entry = s.get_entry_first_match(n);
    };

    if p != NULL {
      if n != "" {
        assert str_entry == p else
          error("CSCO_ADDRMAP_ERR_ENTRY_CONFLICT: the entry string (", n, ") and pointer (", p.long_name(""), ") conflict with each other");
      };
      result = p;
    }
    else {
      result = str_entry;
    };

    assert that result != NULL else 
      error(appendf("CSCO_ADDRMAP_ERR_SEQ_ENTRY_LKUP: cannot find the addrmap entry %s in the transaction's scope: %s", n, s.csco_short_name()));
  };
}; // extend ENTRY'addr_scheme csco_addrmap_trans_s


// Define transaction forwarding interface between addrmap driver and 
// any (interface) sequence driver.
extend any_sequence_driver {
  // This method should be extended by the interface driver to define
  // its behavior when the addrmap driver forwards a transaction to it.
  send_addrmap_item(addrmap_item : csco_addrmap_trans_s) is undefined;

  // This method should be extended by the interface driver to signal
  // backpressure to the addrmap driver whenever necessary.
  addrmap_ready() : bool is empty;
};

sequence csco_addrmap_seq using
  item = csco_addrmap_trans_s,
  created_driver = csco_addrmap_driver_u,
  created_kind = csco_addrmap_seq_kind_t;

// All 'do' actions in an addrmap sequence should use
// the macro below.
define <addrmap_do'action> "addrmap_do[ <subtype'exp>] <item'name>[ keeping <constraints'block>]" as {
  gen <subtype'exp> <item'name> keeping {
    .addrmap_parent_seq == read_only(me);
    .macro_call == TRUE;
    .driver == driver;
    read_only(prev_source != NULL and !allow_per_trans_rand_driver and prev_source.region in .avail_sources.region) => all of {
      .source == .avail_sources.first(.region == prev_source.region);
    };
    all of <constraints'block|{.macro_call == TRUE}>;
  };

  // Tell is_relevant() what the source driver for this transaction is.
  prev_source = <item'name>.source;

  do macro_item keeping { 
    it == <item'name>; 
  };
}; // define <addrmap_do'action> 

extend csco_addrmap_seq {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Soft constrained down to each item in the sequence. This is by default
  // the scope of the addrmap driver. 
  default_scope : csco_addrmap_name_scope_s;
    keep soft default_scope == read_only(parent_sequence is a csco_addrmap_seq (aseq) ? aseq.default_scope : driver.default_scope);

  // Allows each sequence to use a different source driver for its 1st transaction 
  // than parent sequence.  By default, the 1st transaction of each sequence will 
  // use the same source driver as the last executed transaction of the parent 
  // sequence, if the target address allows it.
  allow_seq_rand_driver : bool;
    keep soft allow_seq_rand_driver == FALSE;

  // Allows each transaction in a sequence to use a different source driver.
  // By default, each transaction within a sequence will use the same source 
  // driver as the previous transaction of the same sequence, if the target 
  // address allows it.
  allow_per_trans_rand_driver : bool;
    keep soft allow_per_trans_rand_driver == FALSE;

  // Provided for use in "addrmap_do XXX entry_trans" from within a 
  // addrmap sequence.
  !entry_trans : ENTRY'addr_scheme csco_addrmap_trans_s ;

  // Provided for use in "addrmap_do XXX offset_trans" from within a
  // addrmap sequence.
  !offset_trans : OFFSET'addr_scheme csco_addrmap_trans_s ;

  // Provided for use in "addrmap_do addrmap_seq" from within a addrmap
  // sequence.
  !addrmap_seq : csco_addrmap_seq ;

  // Source region/address pair of current item being issued by a 
  // 'addrmap_do' action. Also used to ensure (by default) that 
  // addrmap_do actions in the same sequence attempt to use the
  // same source driver.
  prev_source : csco_addrmap_region_addr_s;
    keep soft prev_source == (read_only(parent_sequence is a csco_addrmap_seq (aseq) and 
                                        !allow_seq_rand_driver) ? read_only(aseq.prev_source) : NULL);


  /////////////////
  // Private Fields 
  /////////////////
  // Used by addrmap_do macro.
  !macro_item : csco_addrmap_trans_s;

  // Only schedule the sequence if its current item's source driver 
  // is ready to accept new addrmap requrests.
  is_relevant() : bool is also {
    if prev_source != NULL {
      assert that prev_source.region.get_eligible_driver() != NULL else
        error("CSCO_ADDRMAP_ERR_NO_ELIGIBLE_DRIVER_IN_SOURCE\n",
              "could not find an eligible driver for ", prev_source.region.csco_short_name(), " source region.");

      result = prev_source.region.get_eligible_driver().addrmap_ready();
    }
    else {
      // No need to wait for interface driver availablity if transaction
      // is backdoor.
      result = TRUE;
    };
    message(CSCO_ADDRMAP, FULL, "is_relevant() of ", kind, " addrmap sequence called and returned ", result);
  }; // is_relevant()

  post_do(s : any_sequence_item) is {
    if s is a csco_addrmap_seq (addrmap_seq) {
      prev_source = addrmap_seq.prev_source;
    };
  };

  post_do_tcm(s : any_sequence_item) @sys.any is {
    if s is a csco_addrmap_trans_s (addrmap_trans) {
      message(CSCO_ADDRMAP, HIGH, "executing post_do logic for ", addrmap_trans.csco_short_name());
      if addrmap_trans.blocking_kind == BLOCKING {
        if addrmap_trans.post_kind == POSTED {
          message(CSCO_ADDRMAP, MEDIUM, "blocking on xmt_done event for ", addrmap_trans.csco_short_name());
          sync cycle@addrmap_trans.tracking.xmt_done;
          message(CSCO_ADDRMAP, MEDIUM, "finished blocking on xmt_done event for ", addrmap_trans.csco_short_name());
        }
        else {
          all of {
            { 
              message(CSCO_ADDRMAP, MEDIUM, "blocking on all_done event for ", addrmap_trans.csco_short_name());
              sync cycle@addrmap_trans.tracking.all_done; 
              message(CSCO_ADDRMAP, MEDIUM, "finished blocking on all_done event for ", addrmap_trans.csco_short_name());
            };

            { 
              message(CSCO_ADDRMAP, MEDIUM, "blocking on response_done event for ", addrmap_trans.csco_short_name());
              sync cycle@addrmap_trans.response_done; 
              message(CSCO_ADDRMAP, MEDIUM, "finished blocking on response_done event for ", addrmap_trans.csco_short_name());
            };
          };
        };
      };
      message(CSCO_ADDRMAP, HIGH, "finished executing post_do logic for ", addrmap_trans.csco_short_name());
    };
  }; // post_do_tcm()
}; // extend csco_addrmap_seq

extend csco_addrmap_seq_kind_t : [CSCO_RESET, CSCO_INIT_TEST, CSCO_PRE_TEST, CSCO_MAIN_TEST, CSCO_POST_TEST] ;
extend csco_addrmap_seq {
   keep soft kind not in [CSCO_RESET, CSCO_INIT_TEST, CSCO_PRE_TEST, CSCO_MAIN_TEST, CSCO_POST_TEST];
};

extend csco_addrmap_driver_u {
  event clock is only @sys.any;

  CREATE_TESTFLOW clock;

  short_name() : string is {
    result = "DRIVER";
  };

  keep bfm_interaction_mode == PUSH_MODE;

  // Scope of the addrmap driver. By default, this will be constrained from
  // the addrmap_env to be the root of the name scope tree.
  default_scope : csco_addrmap_name_scope_s;
    keep soft default_scope == NULL;

  // The sequence that will execute automatically in the reset phase
  !reset_seq : CSCO_RESET'kind csco_addrmap_seq ;
    keep reset_seq.driver == me ;

  csco_reset() @clock is also {
    gen reset_seq;
    reset_seq.start_sequence() ;
  };

  // The sequence that will execute automatically in the init_test phase
  !init_test_seq : CSCO_INIT_TEST'kind csco_addrmap_seq ;
    keep init_test_seq.driver == me ;

  csco_init_test() @clock is also {
    gen init_test_seq;
    init_test_seq.start_sequence() ;
  };

  // The sequence that will execute automatically in the pre_test phase
  !pre_test_seq : CSCO_PRE_TEST'kind csco_addrmap_seq ;
    keep pre_test_seq.driver == me ;

  csco_pre_test() @clock is also {
    gen pre_test_seq;
    pre_test_seq.start_sequence() ;
  };

  // The sequence that will execute automatically in the main_test phase
  !main_test_seq : CSCO_MAIN_TEST'kind csco_addrmap_seq ;
    keep main_test_seq.driver == me ;

  csco_main_test() @clock is also {
    gen main_test_seq;
    main_test_seq.start_sequence() ;
  };

  // The sequence that will execute automatically in the post_test phase
  !post_test_seq : CSCO_POST_TEST'kind csco_addrmap_seq ;
    keep post_test_seq.driver == me ;

  csco_post_test() @clock is also {
    gen post_test_seq;
     post_test_seq.start_sequence() ;
  };

  send_to_bfm(seq_item : csco_addrmap_trans_s) @clock is {
    // error checking against use of regular 'do'.
    assert seq_item.macro_call == TRUE else
      error(appendf("CSCO_ADDRMAP_ERR_DO_ERR: %s item was not launched via an addrmap_do", short_name()));

    assert seq_item.opcode in [WRITE, READ, FILL] else 
      error(appendf("CSCO_ADDRMAP_ERR_INVLD_BFM_OPCODE: %s item is not a request", short_name()));

    if seq_item is a ENTRY csco_addrmap_trans_s (entry_trans) {
      messagef(LOW, "issuing a %s addrmap transaction to %s (%s)\n", seq_item.opcode, entry_trans.get_entry_ptr().long_name(""), seq_item.csco_short_name());
    }
    else {
      messagef(LOW, "issuing a %s addrmap transaction to offset %s of %s (%s)\n", seq_item.opcode, seq_item.target.addr.csco_short_name(), 
                      seq_item.target.region.csco_short_name(), seq_item.csco_short_name());
    };

    if seq_item.access == BACK_DOOR {
      var byte_list : list of byte;

      messagef(MEDIUM, "executing %s via backdoor\n", seq_item.csco_short_name());
      assert that seq_item.get_entry_ptr() != NULL else
        error("CSCO_ADDRMAP_ERR_NO_BKDOOR_ENTRY: cannot back-door execute ", seq_item.csco_short_name(), 
                " because it is not targetted toward any physical addrmap entry.");

      // Backdoor transactions will consider all side-effects of the transaction
      // on the affected fields, and the value applied will be as if a front-door
      // transaction has just taken place on the fields.
      if seq_item.get_entry_ptr() is a csco_addrmap_reg_entry_s (reg_entry) {
        if seq_item.is_read_req {
          unpack(packing.low, reg_entry.read_reg_bkdoor()[..reg_entry.get_entry_size()*8-1], byte_list); 
          seq_item.set_response(byte_list, NORMAL);
        };

        // update_shadow() will return expected shadow value after all side-effects
        // are considered. The value is then bkdoor written into the register entry.
        reg_entry.update_shadow(seq_item);
        reg_entry.write_reg_bkdoor(reg_entry.get_reg_shadow_val());
      } // (reg_entry)

      else if seq_item.get_entry_ptr() is a csco_addrmap_mem_entry_s (mem_entry) {
        if seq_item.is_read_req {
          seq_item.set_response(mem_entry.read_mem_bkdoor(seq_item), NORMAL);
        };

        mem_entry.update_shadow(seq_item);
        var tmp_trans := deep_copy(seq_item);
        tmp_trans.opcode = WRITE;
        tmp_trans.data = mem_entry.get_shadow_bytes(seq_item);
        mem_entry.write_mem_bkdoor(tmp_trans);
      }; // (mem_entry)

// ?? how to handle bkdoor side-effects for indirect transactions.

      if seq_item.read_or_write() == WRITE and seq_item.post_kind == NON_POSTED {
        seq_item.set_response({}, NORMAL);
      };
      seq_item.tracking.bkdoor_done();
      messagef(HIGH, "finished backdoor access to %s below\n", seq_item.csco_short_name()) {
        seq_item.csco_print(LONG);
      };
    } // if seq_item.bkdoor

    else {
      messagef(MEDIUM, "forwarding %s to %s\n", seq_item.csco_short_name(), seq_item.get_source_driver().short_name_path());
      messagef(CSCO_ADDRMAP, MEDIUM, "") {
        seq_item.csco_print(LONG);
      };
      outf("\n");
      seq_item.get_source_driver().send_addrmap_item(seq_item);
    };
  }; // send_to_bfm()
}; // extend csco_addrmap_driver_u


'>
