// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_space.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 09/26/2005
//              :
//              :
// Description  : Defines data structures for modeling an address space, holding
//                a mesh of addrmap regions.
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2006 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

<'
package csco_addrmap;


// The addrmap_space unit is a container for all regions within a single address space.
// There could be multiple orthongonal addr space instances in a given system.
unit csco_addrmap_space_u {
  // Name of the address space.
  name : csco_addrmap_space_t;
    csco_print_attr name SHORT enable;

  // Determines how transactions are routed between regions in the address space.
  route_kind : csco_addrmap_route_kind_t;
    csco_print_attr route_kind SHORT enable;

  // auto-maintained flat list of all regions that have been created in the environment.
  // Note that the phy_regions list will not be populated until AFTER the build_addrmap()
  // phase and hence any methods which operate on 'phy_regions' should not be called 
  // until AFTER the addrmap has been generated.
  package !phy_regions : list of csco_addrmap_physical_region_s;
    csco_print_attr phy_regions SHORT enable;
    csco_print_attr phy_regions LONG disable;

  package !log_regions : list of csco_addrmap_logical_region_s;
    csco_print_attr log_regions SHORT enable;
    csco_print_attr log_regions LONG disable;

  short_name(): string is {
    result = append(name);
  }; 

  // This method MUST be called for a region to be recognized by the address map.
  package add_region(r : csco_addrmap_region_s) is {
    if r is a csco_addrmap_logical_region_s (log_r) {
      assert that log_r not in log_regions else 
        error(appendf("CSCO_ADDRMAP_ERR_DUP_LOG_REGN: %s already exists as a logical region in %s", r.csco_short_name(), short_name_path()));

      log_regions.add(log_r);
    }
    else if r is a csco_addrmap_physical_region_s (phy_r) {
      assert that phy_r not in phy_regions else 
        error(appendf("CSCO_ADDRMAP_ERR_DUP_LOG_REGN: %s already exists as a physical region in %s", r.csco_short_name(), short_name_path()));

      phy_regions.add(phy_r);
    };
  };

  package delete_region(r : csco_addrmap_region_s) is {
    var del_ind : int;
    if r is a csco_addrmap_logical_region_s (log_r) {
      del_ind = log_regions.first_index(it == log_r);
      log_regions.fast_delete(del_ind);
    }
    else if r is a csco_addrmap_physical_region_s (phy_r) {
      del_ind = phy_regions.first_index(it == phy_r);
      if del_ind != UNDEF {
        phy_regions.fast_delete(del_ind);
      };
    };
  };

  package clear_space() is {
    log_regions.clear();
    phy_regions.clear();
  };

  // Note that since region names are not required to be unique, this method
  // will return the first key match in case of multiple matches.
  get_region(nm : string) : csco_addrmap_region_s is {
    result = get_physical_region(nm);
    if result == NULL {
      result = get_logical_region(nm);
    };
  };

  get_physical_region(nm : string) : csco_addrmap_physical_region_s is {
    result = phy_regions.first(.name == nm);
  };

  get_logical_region(nm : string) : csco_addrmap_logical_region_s is {
    result = log_regions.first(.name == nm);
  };

  // Returns the root regions in the address map.
  get_top_regions() : list of csco_addrmap_logical_region_s is {
    result.add(log_regions.all(.max_depth == 0));
  };

  // Returns all regions in the address map.
  get_all_regions() : list of csco_addrmap_region_s is {
    result = {log_regions; phy_regions};
    result = result.sort_by_field(max_depth).as_a(list of csco_addrmap_region_s);
  };

  get_all_entries() : list of csco_addrmap_entry_s is {
    for each (p) in phy_regions {
      result.add(p.child_entries);
    };
  };

  package !all_regions : list of csco_addrmap_region_s;
    csco_print_attr all_regions LONG header=region;
    csco_print_get_fields_user_hook(format : csco_print_format_t) is also {
      all_regions = get_all_regions();
    };

  dut_error_dump() is {
    csco_print(LONG);
  };

  // Allocates address space of this region amongst all the child regions under it. The
  // package provides 3 default algorithms:
  // <br>
  //  SEQUENTIAL: place each child region in a contiguous section of the parent address space in the order in which it's defined.
  // <br>
  //  MANUAL : offset is generated/assigned by user.
  // <br>
  //  RANDOM_SHIFT: randomly place each child region in a contiguous section of the parent address space.
  // <br>
  //
  // If new algorithms are defined by the user, then this method needs to be extended under the
  // <NEW_ALG>'addr_alg sub-type.
  // <br>
  // Note that the offset of io entries within a physical region is NEVER randomized. 
  pre_build_addrmap() is empty;
  build_addrmap() is undefined;
  post_build_addrmap() is empty;
  check_addrmap() is empty;
}; // unit csco_addrmap_space_u


extend DEFAULT'name csco_addrmap_space_u {
  keep route_kind == ADDRESS;
};


extend ADDRESS'route_kind csco_addrmap_space_u {
  // This method should be called once the user has hooked up all the 
  // entries and regions, and is ready to generate the system map.
  build_addrmap() is {
    if log_regions.size() > 0 {
      var rlist := log_regions.copy().sort(.max_depth);
      var cur_r : ADDRESS'route_kind csco_addrmap_logical_region_s;
      while !rlist.is_empty() {
        cur_r = rlist.pop().as_a(ADDRESS'route_kind csco_addrmap_logical_region_s);
        cur_r.assign_child_offsets();
      };
    };
  };

  check_addrmap() is {
    for each (r) in log_regions.as_a(list of ADDRESS'route_kind csco_addrmap_logical_region_s) {
      r.check_built_map();
    };
  };
}; // extend ADDRESS'route_kind csco_addrmap_space_u


'>
