// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_seq_intermix.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 08/15/2005
//              :
//              :
// Description  : The sequence will run the intermix algorithm over addrmap entries.
//                The user can configure whether or not to include reg/mem entries.
//
//                If no scoreboards are instantiated in the system, then
//                the user should set local_scoreboard == TRUE, which
//                will allow checking of the read response values.
//
//
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_addrmap;

import csco_addrmap/e/csco_addrmap_top;


extend csco_addrmap_seq_kind_t : [INTERMIX];

extend INTERMIX'kind csco_addrmap_seq {
  // If infinite_len is TRUE, then the sequence will continue until POST_TEST phase is reached.
  // If infinite_len is FALSE, then the sequence will only go through each entry in all_entries 
  // exactly once before terminating.
  keep soft csco_infinite_len == TRUE;
  keep soft csco_abort_phase == CSCO_POST_TEST;

  // Whether chip is doing normal operations in addition to running the intermix test. If
  // TRUE, then the intermix sequence is basically running as a background sequence.
  mission_mode : bool;
    keep soft mission_mode == read_only(csco_infinite_len);

  // By default, includes all the registers under the sequence scope. But
  // user may also supply an explcity list of registers.
  // <br>
  // In mission mode, we assume that the user has constrained the 'all_entries' field to
  // contain only registers that are safe to write/read during normal device operation.
  // <br>
  // In non-mission mode, the user can just use the soft constraint to pick up all 
  // the entries under the default_scope. We'll write/read to any register field that is 
  // not marked CRITICAL.
  all_entries : list of csco_addrmap_entry_s;
    keep soft all_entries.size() == 0;

  // Whether register entries should be inlcuded in the intermix.
  include_regs : bool;
    keep soft include_regs == TRUE;

  // Whether memory entries should be inlcuded in the intermix.
  include_mems : bool;
    keep soft include_mems == TRUE;

  include_wo_regs : bool;
    keep soft include_wo_regs == TRUE;

  include_ro_regs : bool;
    keep soft include_ro_regs == TRUE;

  max_trans_size : csco_addrmap_addr_t;
    keep soft max_trans_size == 1K;

  // User can use this field to define a delay distribution between each transaction.
  rand_delay : uint;
    keep soft rand_delay <= 500;

  // Target size of outstanding_writes queue.
  target_writes_out : uint;
    keep target_writes_out > 0;
    keep soft target_writes_out in [1..50];     // arbitrary default

  // User can use this field to constrain a percentage of the total noise traffic to be write traffic.
  !do_write : bool;
    keep outstanding_writes.is_empty() => do_write;
    keep writable_entries.is_empty() => !do_write;
    keep soft do_write == select {              // randomly tend to have outstanding_count transactions
      (112/(4*outstanding_writes.size()/target_writes_out + 1) - 12)         : TRUE;
      (112/(max(1, 9 - 4*outstanding_writes.size()/target_writes_out)) - 12) : FALSE;
    };

  // Whether or not the sequence should do local scoreboarding. Should
  // be TRUE if the rest of the TB doesn't have scoreboard.
  local_scoreboard : bool;
    keep soft local_scoreboard == TRUE;


  // List of write transactions that have not been followed up with a read.
  package !outstanding_writes : list of ENTRY csco_addrmap_trans_s;

  keep soft entry_trans.access == FRONT_DOOR;

  // Filtered version of all_entries, excluding registers, or memories,
  // or entries with side-effects, if so configured.
  package !writable_entries : list of csco_addrmap_entry_s;

  post_generate() is also {
    // The default is to target all entries under the default_scope.
    if all_entries.is_empty() {
      all_entries = default_scope.get_all_child_entries();
    };

    writable_entries = filter_entries(all_entries);
  };

  filter_entries(orig : list of csco_addrmap_entry_s) : list of csco_addrmap_entry_s is {
    for each (e) in orig {
      if e is a csco_addrmap_reg_entry_s and !include_regs {
        continue;
      };

      if e is a csco_addrmap_mem_entry_s and !include_mems {
        continue;
      };

      if e.fields.has(.has_attribute(CRITICAL)) {
        continue;
      };

      if e.fields.has(.sw_access == R) and !include_ro_regs {
        continue;
      };

      if e.fields.has(.sw_access == W) and !include_wo_regs {
        continue;
      };

      for each (f) in e.fields {
        if !mission_mode {
          f.add_attribute(IN_AUTOTEST);
        };
      };

      if not e.fields.has(.is_auto_comp()) {
        warning("ADDRMAP_INTERMIX: ", e.long_name(default_scope.name), " fields don't have any auto-comparable fields");
      };

      result.add(e);
    };
  };

  body() @driver.clock is also {
    var idx : uint;

    messagef(MEDIUM, "ADDRMAP_INTERMIX: launching INTERMIX test with scope as : %s\n", default_scope.long_name(""));
    messagef(CSCO_ADDRMAP, MEDIUM, "ADDRMAP_INTERMIX: Following addrmap entries will be accessed during INTERMIX test:\n") {
      for each (e) in writable_entries {
        out(e.long_name(""));
      };
    };

    while writable_entries.size() > 0 or outstanding_writes.size() > 0 {
      gen do_write;

      if do_write {
        gen idx keeping { it < writable_entries.size(); };

        addrmap_do entry_trans keeping {
          .opcode == WRITE;
          .entry_ptr == read_only(writable_entries[idx]);
          .fields == .set_fields_rand(.entry_ptr, max_trans_size);
          .fill_kind == RANDOM;
          .unspecified_field_kind == RANDOM;
        };
        outstanding_writes.add(entry_trans);

        if !csco_infinite_len {
          writable_entries.fast_delete(idx);
        };
      }
      else {
        gen idx keeping { it < outstanding_writes.size(); };

        addrmap_do READ entry_trans keeping {
          .entry_ptr == read_only(outstanding_writes[idx].get_entry_ptr());
          .start_row == read_only(outstanding_writes[idx].start_row);
          .end_row == read_only(outstanding_writes[idx].end_row);
          soft .blocking_kind == BLOCKING;
        };

        if !mission_mode and !csco_infinite_len {
          // remove autotest attribute from affected register fields.
          for each (f) in entry_trans.get_entry_ptr().fields {
            f.remove_attribute(IN_AUTOTEST);
          };
        };

        outstanding_writes.fast_delete(idx);
      };

      if local_scoreboard {
        messagef(CSCO_ADDRMAP, HIGH, "ADDRMAP_INTERMIX: calling update_shadow() from %s for %s because local_scoreboard = TRUE\n", 
                  kind, entry_trans.csco_short_name());
        entry_trans.get_entry_ptr().update_shadow(entry_trans);
      };

      gen rand_delay;
      messagef(CSCO_ADDRMAP, FULL, "ADDRMAP_INTERMIX: wait %d cycles before issuing next random access\n", rand_delay);
      wait [rand_delay];
    }; // while 
  }; // body() 
}; // extend INTERMIX'kind csco_addrmap_seq


'>
