// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_seq_random_rw.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 08/15/2005
//              :
//              :
// Description  : The sequence will issue random reads and writes to a user 
//                specified list of entries.
//
//                This sequence will only target registers without ANY critical 
//                fields.
//
//                If no scoreboards are instantiated in the system, then
//                the user should set local_scoreboard == TRUE, which
//                will allow checking of the read response values.
//
//
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_addrmap;

import csco_addrmap/e/csco_addrmap_top;


extend csco_addrmap_seq_kind_t : [RANDOM_RW];

extend RANDOM_RW'kind csco_addrmap_seq {
  // If infinite_len is TRUE, then the sequence will continue until POST_TEST phase is reached.
  // If infinite_len is FALSE, then the sequence will only go through each entry in the readable 
  // writable entries list exactly once before terminating.
  keep soft csco_infinite_len == TRUE;
  keep soft csco_abort_phase == CSCO_POST_TEST;

  // Whether chip is doing normal operations in addition to running the random test. If
  // TRUE, then the random sequence is basically running as a background sequence.
  mission_mode : bool;
    keep soft mission_mode == read_only(csco_infinite_len);

  // By default, includes all the registers under the sequence scope. But
  // user may also supply an explcity list of registers.
  // <br>
  // In mission mode, we assume that the user has constrained the entries to
  // contain only registers that are safe to write/read during normal device 
  // operation.
  // <br>
  // In non-mission mode, the user can just use the soft constraint to pick up all 
  // the entries under the default_scope. We'll write/read to any register field 
  // that is not marked critical.
  write_entries : list of csco_addrmap_entry_s;
    keep soft write_entries.size() == 0;

  read_entries : list of csco_addrmap_entry_s;
    keep soft read_entries.size() == 0;

  // Whether register entries should be inlcuded in the random test.
  include_regs : bool;
    keep soft include_regs == TRUE;

  // Whether memory entries should be inlcuded in the random test.
  include_mems : bool;
    keep soft include_mems == TRUE;

  allow_ro_writes : bool;
    keep soft allow_ro_writes == TRUE;

  allow_wo_reads : bool;
    keep soft allow_wo_reads == TRUE;

  // User can use this field to define a delay distribution between each transaction.
  rand_delay : uint;
    keep soft rand_delay <= 500;

  max_trans_size : csco_addrmap_addr_t;
    keep soft max_trans_size == 1K;

  // User can use this field to constrain a percentage of the total noise traffic to be write traffic.
  !do_write : bool;
    keep filtered_write_entries.is_empty() => do_write == FALSE;
    keep filtered_read_entries.is_empty() => do_write == TRUE;

  // Whether or not the sequence should do local scoreboarding. Should
  // be TRUE if the rest of the TB doesn't have scoreboard.
  local_scoreboard : bool;
    keep soft local_scoreboard == TRUE;

  keep soft entry_trans.access == FRONT_DOOR;

  package !filtered_write_entries : list of csco_addrmap_entry_s;
  package !filtered_read_entries : list of csco_addrmap_entry_s;

  post_generate() is also {
    if write_entries.is_empty() {
      gen write_entries keeping {
        it.is_a_permutation(default_scope.get_all_child_entries());
      };
    };

    if read_entries.is_empty() {
      gen read_entries keeping {
        it.is_a_permutation(default_scope.get_all_child_entries());
      };
    };

    filtered_write_entries = filter_write_entries(write_entries);
    filtered_read_entries = filter_read_entries(read_entries);
  };

  filter_write_entries(orig : list of csco_addrmap_entry_s) : list of csco_addrmap_entry_s is {
    for each (e) in orig {
      if e is a csco_addrmap_reg_entry_s and !include_regs {
        continue;
      };

      if e is a csco_addrmap_mem_entry_s and !include_mems {
        continue;
      };

      if e.fields.has(.has_attribute(CRITICAL)) {
        continue;
      };

      if e.fields.has(.sw_access == R) and !allow_ro_writes {
        continue;
      };

      result.add(e);
    };

  };

  filter_read_entries(orig : list of csco_addrmap_entry_s) : list of csco_addrmap_entry_s is {
    for each (e) in orig {
      if e is a csco_addrmap_reg_entry_s and !include_regs {
        continue;
      };

      if e is a csco_addrmap_mem_entry_s and !include_mems {
        continue;
      };

      if e.fields.has(.sw_access == W) and !allow_wo_reads {
        continue;
      };

      if e.fields.has(.has_attribute(CRITICAL)) {
        continue;
      };

      for each (f) in e.fields {
        if !mission_mode {
          f.add_attribute(IN_AUTOTEST);
        };
      };

      if not e.fields.has(.is_auto_comp()) {
        warning("ADDRMAP_RANDOM_RW: ", e.long_name(default_scope.name), " fields don't have any auto_comparable fields");
      };

      result.add(e);
    };
  };

  body() @driver.clock is also {
    messagef(MEDIUM, "ADDRMAP_RANDOM_RW: launching RANDOM_RW test with scope as : %s\n", default_scope.long_name(""));
    messagef(CSCO_ADDRMAP, MEDIUM, "ADDRMAP_RANDOM_RW: Following addrmap entries will be written during RANDOM_RW test:\n") {
      for each (e) in filtered_write_entries {
        out(e.long_name(default_scope.name));
      };
    };
    messagef(CSCO_ADDRMAP, MEDIUM, "ADDRMAP_RANDOM_RW: Following addrmap entries will be read during RANDOM_RW test:\n") {
      for each (e) in filtered_read_entries {
        out(e.long_name(default_scope.name));
      };
    };

    while filtered_write_entries.size() > 0 or filtered_read_entries.size() > 0 {
      var idx : uint;

      gen do_write;
      if do_write {
        gen idx keeping { it < filtered_write_entries.size(); };

        addrmap_do entry_trans keeping {
          .opcode == WRITE;
          .entry_ptr == read_only(filtered_write_entries[idx]);
          .fields == .set_fields_rand(.entry_ptr, max_trans_size);
          .fill_kind == RANDOM;
          .unspecified_field_kind == RANDOM;
        };

        if !csco_infinite_len {
          filtered_write_entries.fast_delete(idx);
        };
      }
      else {
        gen idx keeping { it < filtered_read_entries.size(); };

        addrmap_do READ entry_trans keeping {
          .entry_ptr == read_only(filtered_read_entries[idx]);
          .end_row - .start_row + 1 <= read_only(max_trans_size / .entry_ptr.get_entry_width());
        };

        if !mission_mode and !csco_infinite_len {
          // remove autotest attribute from affected register fields.
          for each (f) in entry_trans.get_entry_ptr().fields {
            f.remove_attribute(IN_AUTOTEST);
          };
        };

        if !csco_infinite_len {
          filtered_read_entries.fast_delete(idx);
        };
      };

      if local_scoreboard {
        messagef(CSCO_ADDRMAP, HIGH, "ADDRMAP_RANDOM_RW: calling update_shadow() from %s for %s because local_scoreboard = TRUE\n", 
                  kind, entry_trans.csco_short_name());
        entry_trans.get_entry_ptr().update_shadow(entry_trans);
      };

      gen rand_delay;
      messagef(CSCO_ADDRMAP, FULL, "ADDRMAP_RANDOM_RW: wait %d cycles before issuing next random access\n", rand_delay);
      wait [rand_delay];
    }; // while 
  }; // body() 
}; // extend RANDOM_RW'kind csco_addrmap_seq


'>
