// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_addrmap_seq_walk.e
//              :
// Project      : csco_addrmap
//              :
// Created On   : 08/15/2005
//              :
//              :
// Description  : This is a directed and automatic addrmap entry walk one/zero test.
//                The sequence will walk either a zero or 1 pattern across
//                all bits of all entries in the sequence name scope.
//                Note that critical field bits are skipped and maintain
//                their power-on value during the walk.
//
//                If no scoreboards are instantiated in the system, then
//                the user should set local_scoreboard == TRUE, which
//                will allow checking of the read response values.

//
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_addrmap;

import csco_addrmap/e/csco_addrmap_top;


type csco_addrmap_walk_value_t : [ZERO, ONE];

struct csco_addrmap_walk_group_s {
  entries : list of csco_addrmap_entry_s;
};


extend csco_addrmap_seq_kind_t : [REG_WALK];

// This sequence walks both register and memory entries with configurable
// 0 or 1 patterns.
extend REG_WALK'kind csco_addrmap_seq {
  walk_value : csco_addrmap_walk_value_t;

  // Whether or not the sequence should do local scoreboarding. Should
  // be TRUE if the rest of the TB doesn't have scoreboard.
  local_scoreboard : bool;
    keep soft local_scoreboard == TRUE;

  // By default, includes all the addrmap entries under the sequence scope. But
  // user may also supply an explcity list of entries.
  all_entries : list of csco_addrmap_entry_s;
    keep soft all_entries.size() == 0;

  walk_reg : bool;
    keep soft walk_reg == TRUE;

  walk_mem : bool;
    keep soft walk_mem == FALSE;

  include_wo_regs : bool;
    keep soft include_wo_regs == TRUE;

  include_ro_regs : bool;
    keep soft include_ro_regs == TRUE;

  // If TRUE, will write to critical fields with their power-on values.
  include_critical : bool;
    keep soft !include_critical;

  // Will launch each walk group in parallel.
  do_parallel : bool;
    keep soft do_parallel == TRUE;

  package !filtered_entries : list of csco_addrmap_entry_s;
  package !walk_groups : list of csco_addrmap_walk_group_s;

  post_generate() is also {
    if all_entries.is_empty() {
      all_entries = default_scope.get_all_child_entries();
    };

    filtered_entries = filter_entries(all_entries).sort(.aliased_from != NULL);
    var phy_list : list of csco_addrmap_entry_s;
    var alias_list : list of csco_addrmap_entry_s;
    var tmp_ind := filtered_entries.first_index(.aliased_from != NULL);
    if tmp_ind != UNDEF {
      phy_list = filtered_entries[..tmp_ind-1];
      alias_list = filtered_entries[tmp_ind..];
    }
    else {
      phy_list = filtered_entries;
    };

    for each (e) in alias_list {
      tmp_ind = phy_list.first_index(it == e.aliased_from);
      phy_list.insert(tmp_ind+1, e);
    };

    var tmp_grp : csco_addrmap_walk_group_s;
    for each (e) in phy_list {
      if e.aliased_from == NULL {
        tmp_grp = new csco_addrmap_walk_group_s with {
          .entries.add(e);
        };
        walk_groups.add(tmp_grp);
      }
      else {
        tmp_grp.entries.add(e);
      };
    };

    gen walk_groups keeping {
      it.is_a_permutation(walk_groups);
    };
  };

  filter_entries(orig : list of csco_addrmap_entry_s) : list of csco_addrmap_entry_s is {
    for each (e) in orig {
      if e is a csco_addrmap_reg_entry_s and !walk_reg { continue; };
      if e is a csco_addrmap_mem_entry_s and !walk_mem { continue; };
      if e.fields.has(.sw_access == R) and !include_ro_regs { continue; };
      if e.fields.has(.sw_access == W) and !include_wo_regs { continue; };
      if e.fields.has(.has_attribute(CRITICAL)) and !include_critical { continue; };
      result.add(e);
    };
  };

  body() @driver.clock is also {
    messagef(MEDIUM, "ADDRMAP_WALK: launching WALK%d test with scope as : %s\n", walk_value, default_scope.long_name(""));
    var cur_group : csco_addrmap_walk_group_s = new;
    for each (g) in walk_groups {
      cur_group.entries.add(g.entries);
      gen do_parallel;
      if index == walk_groups.size() - 1 {
        do_parallel = TRUE;
      };

      if do_parallel {
        gen WALK_REG_GROUP addrmap_seq keeping {
          .driver == driver;
          .group == cur_group;
          .walk_value == read_only(walk_value);
          .local_scoreboard == read_only(local_scoreboard);
        };
        addrmap_seq.start_sequence();
        cur_group = new;
      };
    }; 
  }; // body() 
}; // extend REG_WALK'kind csco_addrmap_seq


extend csco_addrmap_seq_kind_t : [WALK_REG_GROUP];
extend WALK_REG_GROUP csco_addrmap_seq {
  group : csco_addrmap_walk_group_s;

  walk_value : csco_addrmap_walk_value_t;

  local_scoreboard : bool;
    keep soft local_scoreboard == TRUE;

  // User can use this field to define a delay distribution between each transaction.
  rand_delay : uint;
    keep soft rand_delay <= 100;

  body() @driver.clock is {
    var permuted_entries : list of csco_addrmap_entry_s;
    gen permuted_entries keeping {
      it.is_a_permutation(group.entries);
    };

    messagef(CSCO_ADDRMAP, MEDIUM, "ADDRMAP_WALK: now walking following register group: \n") {
      for each (e) in group.entries {
        out(e.long_name(default_scope.name));
      };
    };

    for each (e) in permuted_entries {
      do WALK_SINGLE_REG addrmap_seq keeping {
        .entry_ptr == e;
        .walk_value == walk_value;
        .local_scoreboard == local_scoreboard;
      };

      gen rand_delay;
      messagef(CSCO_ADDRMAP, FULL, "ADDRMAP_WALK: wait %d cycles before issuing next random access\n", rand_delay);
      wait [rand_delay];
    };
  };
};


extend csco_addrmap_seq_kind_t : [WALK_SINGLE_REG];
extend WALK_SINGLE_REG csco_addrmap_seq {
  entry_ptr : csco_addrmap_entry_s;
    keep soft entry_ptr == NULL;

  walk_value : csco_addrmap_walk_value_t;

  local_scoreboard : bool;
    keep soft local_scoreboard == TRUE;

  !walk_trans : csco_addrmap_trans_s;
    keep soft walk_trans.access == FRONT_DOOR;


  body() @driver.clock is {
    var reset_val : int(bits:*);
    var cur_val : int(bits:*);
    var cur_masked_val : int(bits:*);
    var write_val : int(bits:*);
    var critical_mask : int(bits:*);
    var max_row_num : csco_addrmap_addr_t;

    critical_mask = 0;
    for each (f) in entry_ptr.fields {
      if f.has_attribute(CRITICAL) {
        // critical bit locations are set to 1.
        critical_mask[f.lsb_pos + f.width - 1 : f.lsb_pos] = UNDEF;
      };

      f.add_attribute(IN_AUTOTEST);
    }; // for each (f) in entry_ptr.fields

    if entry_ptr is a csco_addrmap_mem_entry_s (mem_entry) {
      max_row_num = mem_entry.depth - 1;
    }
    else {
      max_row_num = 0;
    };

    messagef(HIGH, "ADDRMAP_WALK: now walking %s\n", entry_ptr.long_name(default_scope.name));
    if not entry_ptr.fields.has(.is_auto_comp()) {
      warning("ADDRMAP_WALK: ", entry_ptr.long_name(default_scope.name), " doesn't have any auto_comparable fields");
    };

    for cur_row from 0 to max_row_num {
      cur_val = 1;
      repeat {
        reset_val = entry_ptr.pack_fields(SHADOW);
        cur_masked_val = cur_val & (~critical_mask);     // The current bit is only walked if it's not critical.
        if walk_value == ONE {
          write_val = cur_masked_val | (reset_val & critical_mask);
        }
        else {
          write_val = ~cur_masked_val & (reset_val | ~critical_mask);
        };
        message(CSCO_ADDRMAP, HIGH, "ADDRMAP_WALK: now applying pattern, ", hex(write_val), " to ",  entry_ptr.long_name(default_scope.name));

        if entry_ptr is a csco_addrmap_byte_mem_entry_s (byte_mem) {
          var write_byte_list : list of byte;
          unpack(packing.low, write_val[..entry_ptr.get_entry_width()*8-1], write_byte_list);

          addrmap_do OFFSET'addr_scheme walk_trans keeping {
            .opcode == WRITE;
            .target.region == read_only(byte_mem.parent_region);
            .target.addr.offset == read_only(byte_mem.offset + cur_row * byte_mem.get_entry_width());
            .num_bytes == read_only(byte_mem.get_entry_width());
            .data == read_only(write_byte_list);
          };
        }
        else {
          // All the writes and reads for the same register use the same source driver.
          addrmap_do ENTRY'addr_scheme walk_trans keeping {
            .opcode == WRITE;
            .entry_ptr == read_only(entry_ptr);
            .fields == { .set_row(cur_row, write_val[entry_ptr.get_entry_width()*8-1:0]) };
          };
        };

        if local_scoreboard {
          messagef(CSCO_ADDRMAP, HIGH, "ADDRMAP_WALK: calling update_shadow() from %s for %s because local_scoreboard = TRUE\n", 
                    kind, walk_trans.csco_short_name());
          walk_trans.get_entry_ptr().update_shadow(walk_trans);
        };

        message(CSCO_ADDRMAP, HIGH, "ADDRMAP_WALK: now reading back pattern, ", hex(write_val), " from ",  entry_ptr.long_name(default_scope.name));
        if entry_ptr is a csco_addrmap_mem_entry_s (byte_mem) {
          addrmap_do OFFSET walk_trans keeping {
            soft .blocking_kind == BLOCKING;
            .opcode == READ;
            .target.region == read_only(byte_mem.parent_region);
            .target.addr.offset == read_only(byte_mem.offset + cur_row * byte_mem.get_entry_width());
            .num_bytes == read_only(byte_mem.get_entry_width());
          };
        }
        else {
          addrmap_do ENTRY walk_trans keeping {
            soft .blocking_kind == BLOCKING;
            .opcode == READ;
            .entry_ptr == read_only(entry_ptr);
            .start_row == cur_row;
            .end_row == cur_row;
          };
        };

        if local_scoreboard {
          messagef(CSCO_ADDRMAP, HIGH, "ADDRMAP_WALK: calling update_shadow() from %s for %s because local_scoreboard = TRUE\n", 
                    kind, walk_trans.csco_short_name());
          walk_trans.get_entry_ptr().update_shadow(walk_trans);
        };

        cur_val = cur_val << 1;
      } until cur_val[entry_ptr.get_entry_width()*8-1:0] == 0;
    }; // for cur_row from 0 to max_row_num

    // remove autotest attribute from affected register fields.
    for each (f) in entry_ptr.fields {
      f.remove_attribute(IN_AUTOTEST);
    };
  }; // body() 
};


'>
