#!/usr/bin/perl -w
#-----------------------------------------------------------------------------
# filename: test.pl
# authors: spadix, lihuang
#
# description:
#   script to run a test from the standalone environment. try running
#     perldoc `sn_which csco_blk_memory/bin/test.pl`
#   or
#     `sn_which csco_blk_memory/bin/test.pl` -man
#
#----------------------------------------------------------------------------
#>>>>>>>>>>>>>>>>>>>>>>>>>>> COPYRIGHT NOTICE <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
#----------------------------------------------------------------------------
# Copyright 2002-2005 (c) Cisco Systems Inc.
#----------------------------------------------------------------------------
# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License
# at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is "Cisco generic block-based memory model".
# The Initial Developer of the Original Code is "Lisa Huang and Jeff Brown".
# Portions created by the Initial Developer are Copyright (C) 2002-2005
# Cisco Systems Inc. All Rights Reserved.
# Contributor(s): Lisa Huang and Jeff Brown
#----------------------------------------------------------------------------
use File::Find;
use Getopt::Long;
use Pod::Usage;
use strict;

my $package = 'csco_blk_memory';

my $pkgdir = `sn_which.sh $package`;
chomp($pkgdir);
-d $pkgdir ||
  die("could not locate package: $package\n");

my @optspec = qw(help man verbose debug number:i seed=s test=s@
                 list|file=s@ commands|cmds=s with|wrapper=s compiled <>
                 only=s max_size=s);

# defaults and special handling
my %opt = (
  seed => 'random',
  number => 1,
  commands => '',
  with => '',
  list => \&parse_list,
  test => \&parse_test,
  '<>' => \&parse_test,
);

my @runs; # parsed list of tests to run
GetOptions(\%opt, @optspec) &&
  (-t STDIN || parse_list('-', \*STDIN)) || pod2usage(2);

pod2usage(-verbose => 1) if($opt{help});
pod2usage(-verbose => 2) if($opt{man});

# pre-compile if requested
my $pkgtop = $package . '_top';
if($opt{compiled}) {
  # check all .e files under e/
  # FIXME could have alternatives...
  my @src_times;
  find({
         no_chdir => 1,
         wanted => sub { -f && /\.e$/ && push(@src_times, (stat(_))[9]) },
       }, $pkgdir . '/e');
  if(!-f $pkgtop || (sort({$b <=> $a} @src_times))[0] > (stat(_))[9]) {
    # recompile required
    local $ENV{SPECMAN_PRE_COMMANDS} = 'set log sn_compile';
    my $topfile = $pkgdir . '/e/' . $pkgtop;
    system('sn_compile.sh', $topfile) &&
      die("compilation failed for: $topfile\n");
  }
  exit(0) if(!@runs);
}

pod2usage(-message => "no tests specified to run?") if(!@runs);

# specman command seperator fixups
$opt{commands} .= '; '
  if($opt{commands} !~ /(^|;)\s*$/);
$ENV{SPECMAN_PRE_COMMANDS} ||= "";
$ENV{SPECMAN_PRE_COMMANDS} .= '; '
  if($ENV{SPECMAN_PRE_COMMANDS} !~ /(^|;)\s*$/);

my @cmd = qw(time sn_edit);
if($opt{compiled}) {
  push(@cmd, 'specsim', "./$pkgtop");
} else {
  push(@cmd, 'specman');
}
unshift(@cmd, split(' ', $opt{with}))
  if(defined($opt{with}));

if($opt{debug} || $opt{verbose}) {
  print STDERR ("using base command:\n  ", join(" ", @cmd) . "\n");
}

my $status = 0;
foreach my $run (@runs) {
  # FIXME filter duplicate runs...
  $run->{test} = "$package/tests/$run->{test}"
    if($run->{test} !~ /\//);

  my @cfg;
  push(@cfg, "define CFG_$run->{only}_ONLY")
    if(defined($run->{only}));
  push(@cfg, "define CFG_MAX_SIZE $run->{max_size}")
    if(defined($run->{max_size}));
  $run->{seed} = 'random'
    if(!defined($run->{seed}) ||
       $run->{seed} eq substr('random', 0, length($run->{seed})) );
  $run->{number} = 1 # override
    if($run->{seed} ne 'random');

  for (1..$run->{number}) {
    my $seed = $run->{seed};
    $seed = int(rand(2**31))
      if($seed eq 'random');

    local $ENV{SPECMAN_PRE_COMMANDS} = $ENV{SPECMAN_PRE_COMMANDS};
    $ENV{SPECMAN_PRE_COMMANDS} .= join('; ', (
        "set log " . run2log($run, $seed),
        @cfg,
        "load $run->{test}",
        $opt{commands} .
       "test -seed=$run->{seed}",
        "show mem",
        "quit",
      )) . ';';

    print('starting: ' . join(" ", @cmd) . "\n");
    print('SPECMAN_PRE_COMMANDS=' . $ENV{SPECMAN_PRE_COMMANDS} . "\n");
    $status ||= system(@cmd) if(!$opt{debug});
  }
  # FIXME write results to history
}

exit($status);


# parse single test argument
# add to list of tests to run
sub parse_test {
  shift if(@_ > 1);
  push(@runs, log2run(@_));
}

# parse test list file argument
# iterate each line and feed to GetOptions::Long
sub parse_list {
  my $file = $_[1];
  local (@ARGV, $Getopt::Long::error);
  my $list;
  if(!ref($file)) {
    my @files = grep({-r} ($file, map("$pkgdir/$_/$file", qw(etc bin tests .))));
    die("could not find list file: $file\n")
      if(!@files || $?);
    $file = $files[0];

    open($list, $file) ||
      die("could not open list file: $file: $!\n");
  } else {
    $list = $file;
  }

  my %saved = %opt;
  my %defaults = %opt;
  while(<$list>) {
    s[\s*(//|#).*$][];
    next if(/^\s*$/);
    my $lastrun = @runs;
    @ARGV = split;  # FIXME could be smarter...
    GetOptions(\%opt, @optspec) || pos2usage(2);
    if($lastrun == @runs) {
      %defaults = %opt; # new defaults
    } else {
      %opt = %defaults; # reset to defaults
    }
  }
  close($list);
  %opt = %saved;
}

sub run2log {
  my ($run, $seed) = @_;
  $run->{test} =~ m'([^/]+)(?:.e)?$' ||
    die("unrecognized testcase name: $run->{test}\n");
  my $test = $1;
  my $only = ($run->{only}) ? ('_' . $run->{only}) : '';
  my $max_size = ($run->{max_size}) ? ('_' . $run->{max_size}) : '';
  return($test . $only . $max_size . '-' . $seed);
}

sub log2run {
  $_[0] =~ /(.*?)(?:(?:_(FLAT|HASH|BLK))?(?:_([A-Z][_A-Z]*))?(?:-(\d+))(?:\.elog)?)?$/;
  my $run = {
    %opt,
    'test' => $1,
  };
  if($2) { $run->{only} = $2 }
  if($3) { $run->{max_size} = $3 }
  if($4) { $run->{seed} = $4 }
  return($run);
}


__END__

#----------------------------------------------------------------------------

=head1 NAME

test.pl - run test(s) from eRM package

=head1 SYNOPSIS

test.pl [options] [testcase ...]

=head1 DESCRIPTION

runs a testcase or complete regression from the standalone test
environment. if multiple testcase names or the B<-list> or B<-file>
options are specified, a regression is started. regressions default to
sequential execution in the forground, but the B<-with> option can be
used to queue the tests in the background via LSF or similar
commands. see the B<-list> option for details about the testlist file
format.

testcases can be specified as logfile names (w/options and seed
appended). this is useful for rerunning failures.

options are parsed in the order that they appear on the command line,
when a test name/list is encountered, options accumulated to that
point will be applied. all options may be abbreviated down to the
shortest unique prefix.

=head1 OPTIONS

=over 4

=item B<-help>

print a brief help message and exit

=item B<-man>

produce the full man page

=item B<-number>=I<N>

run the following test(s) each I<N> times. ignored if B<-seed> is not
I<random>

=item B<-seed>=I<SEED>

run the following test(s) with I<SEED>. if I<SEED> is I<random> (or
some abbreviation) run tests with a random seed. implies B<-number>=1

=item B<-list>=I<FILE>, B<-file>=I<FILE>

read options and testcases from I<FILE>. blank lines and lines
starting with "#", "//" or "--" are ignored. defaults from the command
line are applied first. lines containing only options (starting with
"-") set new defaults for subsequent lines. the remaining lines
contain combinations of options and testcase names which are executed
serially or in parallel depending on the behavior of the B<-with> program

=item B<-commands>=I<specman commands>

commands appended to those sent to specman, separated by semicolons as
usual. useful for debugging, etc.

=item B<-with>=I<command>, B<-wrapper>=I<command>

run the test(s) with the specified I<command>. options to I<command>
can be specified (appropriately quoted)

=back

=head2 Package Specific Options

=over 4

=item B<-only>=I<MODE>

run the test(s) with only the memory implementation specified by
I<MODE>, which should be one of: FLAT, HASH, BLK. the default is to
run with all implementations for comparison

=item B<-max_size>=I<SIZE>

limit the maximum transaction size (and memory size for the I<flat>
implementation) for the test to I<SIZE>. I<SIZE> should be selected
from the range names defined in F<tests/testbench.e>: BYTE, SMALL,
MEDIUM, LARGE, KILO_SMALL, KILO_MEDIUM, KILO_LARGE, MEG_SMALL

=back

=head1 EXAMPLES

run the default random testcase with a random seed:

  test.pl test_random

submit sanity regression to LSF with various fixed and random seeds:

  test.pl -seed=1 -list=testlist.sanity \
          -n=10 -seed=random -list=testlist.sanity \
          -with='csub -o/dev/null -qlinux'

regress all failing tests in the current directory

  grep -l Error: *.elog | test.pl --with='csub -o/dev/null -qlinux'

what were those options again?

  test.pl -h

=head1 AUTHOR

Lisa Huang and Jeff Brown <spadix+evc@cisco.com>

=head1 COPYRIGHT

Copyright 2002-2005 (c) Cisco Systems Inc.

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
the License for the specific language governing rights and limitations
under the License.

The Original Code is "Cisco generic block-based memory model".  The
Initial Developer of the Original Code is "Lisa Huang and Jeff Brown".
Portions created by the Initial Developer are Copyright (C) 2002-2005
Cisco Systems Inc. All Rights Reserved.  Contributor(s): Lisa Huang
and Jeff Brown

=cut
