//-----------------------------------------------------------------------------
// filename: csco_blk_memory_top.e
//
// description:
//   generic 'e' memory model.
//   attempts to reach a compromise between a flat implementation
//   (used for fully/mostly implemented memories) and a keyed list
//   implementation (used for very sparsely implemented memories) by
//   using dynamically sized blocks to represent implemented regions
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// Copyright 2002-2005 (c) Cisco Systems Inc.
//----------------------------------------------------------------------------
// The contents of this file are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in
// compliance with the License. You may obtain a copy of the License
// at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "Cisco generic block-based memory model".
// Portions created by the Initial Developer are Copyright (C) 2002-2005
// Cisco Systems Inc. All Rights Reserved.
//----------------------------------------------------------------------------
<'
package csco_blk_memory;

import evc_util/e/evc_util_top;

define CSCO_BLK_MEMORY_VERSION_1_0;

define CSCO_BLK_MEMORY_VERSION_1_0_OR_LATER;
define CSCO_BLK_MEMORY_VERSION_0_9_OR_LATER;
define CSCO_BLK_MEMORY_VERSION_0_1_OR_LATER;

//========================================================================
// user interface definition
//========================================================================

// new tag for messages from this package
extend message_tag : [ CSCO_BLK_MEMORY ];

// these assertions are not necessary and
// are only enabled for debugging this package
//#define CSCO_BLK_MEMORY_DEBUG_ASSERTIONS

// instance name aspect
type csco_blk_memory_name_t : [
  MEM = 0 // generic default (backward compatibility)
];

// basic address type
type csco_blk_memory_addr_t : uint(bits: 64);

//------------------------------------------------------------
// base memory instance
// new code should use this struct

struct csco_blk_memory {
  name : csco_blk_memory_name_t;
    keep soft name == MEM;

  // default value of a memory location
  // used to fill memory if an address is read before it is written
  default_value : byte;
    keep soft default_value == 0;

  // general access interface
  //   writes allocate memory if necessary
  //   reads return default_value for unallocated memory
  write(addr: csco_blk_memory_addr_t, data: list of byte) is undefined;
  read(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : list of byte is undefined;

  // single byte access interface
  write_byte(addr: csco_blk_memory_addr_t, data: byte) is undefined;
  read_byte(addr: csco_blk_memory_addr_t) : byte is undefined;

  // deallocate memory byte(s) (if previously allocated)
  // any stored data is lost
  free(addr: csco_blk_memory_addr_t) is undefined;
  free_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) is undefined;

  // deallocate all memory
  clear() is undefined;

  // check if memory has been previously allocated
  check_alloc(addr: csco_blk_memory_addr_t) : bool is undefined;
  check_alloc_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : bool is undefined;

  // get allocated ranges in specified region
  // returns a list of <start, end> pairs
  get_alloc_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : list of csco_blk_memory_addr_t is undefined;

  // output methods
  print() is undefined;
  get_string() : string is undefined;

  // customization controls for HIGH verbosity partial data dumps
  // (FULL verbosity always dumps everything)

  // dump everything for transactions up to this threshold
  partial_dump_threshold : uint;
    keep partial_dump_threshold[3:0] == 0;
    keep soft partial_dump_threshold == 0x100;

  // for larger transactions, only dump this many bytes
  // from the beginning of the transaction
  partial_dump_head : uint;
    keep partial_dump_head[3:0] == 0;
    keep soft partial_dump_head == 0x80;

  // and this many bytes from the end
  partial_dump_tail : uint;
    keep partial_dump_tail[3:0] == 0;
    keep soft partial_dump_tail == 0x40;
};


//========================================================================
// implementation details
//========================================================================

//------------------------------------------------------------
// internal representation for a block of allocated memory

package struct csco_blk_memory_blk {
  // block location
  !start_addr : csco_blk_memory_addr_t;
  // last address in block
  !end_addr : csco_blk_memory_addr_t;
  // block contents
  !data : list of byte;

  // binary tree node children
  package !left : csco_blk_memory_blk;
  package !right : csco_blk_memory_blk;

  // tree balancing property
  !mark : bool;

  // test if specific address is owned by this block
  has_addr(addr: csco_blk_memory_addr_t) : bool is {
    result = ((addr >= start_addr) && (addr <= end_addr));
  };

  // test if address interval is completely contained by this block
  contains_interval(istart: csco_blk_memory_addr_t, iend: csco_blk_memory_addr_t) : bool is {
    result = ((istart >= start_addr) && (iend <= end_addr));
  };

  // dump contents for debug
  debug_dump_node() is {
    outf("[%x %x]%s %x (%s)\n", start_addr, end_addr,
         ((mark) ? "+" : "-"), data.size(),
         str_join(data.apply(appendf("%02x", it)), ""));
  };

  debug_dump_subtree(indent: string) is {
    outf("%s", indent);
    debug_dump_node();

    if(left != NULL) {
      left.debug_dump_subtree(append(indent, "<   "))
    };
    if(right != NULL) {
      right.debug_dump_subtree(append(indent, ">   "))
    };
  };

  verify_subtree(h: *int, n: *int) : bool is {
    result = TRUE;
    var ok := result;
    var hl := 0;
    var hr := 0;
    var nl := 0;
    var nr := 0;

    ok = (data.size() == (end_addr - start_addr + 1));
    ok &&= (end_addr >= start_addr);
    result &&= ok;
    if(!ok) {
      outf("failed data size and end_addr >= start_addr check\n");
      debug_dump_node();
    };

    if(left != NULL) {
      ok = (start_addr > (left.end_addr + 1));
      ok &&= (!mark || !left.mark);
      result &&= ok;
      if(!ok) {
        outf("failed left blk test\n");
        debug_dump_node();
      };
      result &&= left.verify_subtree(hl, nl);
    };

    if(right != NULL) {
      ok = ((end_addr + 1) < right.start_addr);
      ok &&= (!mark || !right.mark);
      result &&= ok;
      if(!ok) {
        outf("failed right blk test\n");
        debug_dump_node();
      };
      result &&= right.verify_subtree(hr, nr);
    };

    // r/b tree invariant
    h = max(hl, hr) + (!mark).as_a(bit);  // black height
    n = 1 + nl + nr;                      // total subtree nodes
    ok = ((n + 1) >= (1 << (h / 2)));
    ok &&= (hl == hr);
    result &&= ok;
    if(!ok) {
      outf("failed invariant test (hl=%02x, hr=%02x, nl=%02x, nr=%02x)\n",
           hl, hr, nl, nr);
      debug_dump_node();
    };
  };
};

//------------------------------------------------------------
// memory implementation

extend csco_blk_memory {
  // root of tree of memory blocks
  package !root : csco_blk_memory_blk;
  // last accessed search path
  package !path : list of csco_blk_memory_blk;

  //----------------------------------------
  // interface method definitions

  // single byte access interface
  write_byte(addr: csco_blk_memory_addr_t, data: byte) is {
    write(addr, { data });
  };

  read_byte(addr: csco_blk_memory_addr_t) : byte is {
    result = read(addr, 1)[0];
  };

  write(addr: csco_blk_memory_addr_t, data: list of byte) is {
    var end_addr : csco_blk_memory_addr_t = addr + data.size() - 1;

    messagef(CSCO_BLK_MEMORY, MEDIUM, "write addr=%x size=%x\n",
             addr, data.size()) {
      if(at_message_verbosity(HIGH)) {
        out(get_data_string(addr, data));
      };
    };

    // first check for cache hit
    if((!path.is_empty()) && (path.top().contains_interval(addr, end_addr))) {
      var iblk := path.top();           // cache hit (TRACE)
      iblk.data[(addr - iblk.start_addr)..(end_addr - iblk.start_addr)] = data;
    } else {
      // cache miss, search for start block
      var iblk := find_next_blk((addr == 0) ? 0 : (addr - 1));

      if((iblk == NULL) || (iblk.start_addr > (end_addr + 1))) {
        if(iblk != NULL) {
          compute find_blk(addr);       // refresh path (TRACE)
        };
        iblk = new csco_blk_memory_blk with {
          .start_addr = addr;
          .end_addr = end_addr;
          .data = data.copy();
        };
        insert(iblk);                   // create new block (TRACE)
      } else if(iblk.contains_interval(addr, end_addr)) {
        iblk.data[(addr - iblk.start_addr)..(end_addr - iblk.start_addr)] = data;  // no structural change (TRACE)
      } else {
        // merge required
        // FIXME conserve rotations w/alternate traversal...
        var prepend_data : list of byte;
        var append_data : list of byte;
        
        if(addr > iblk.start_addr) {
          // copy prefix data from start block
          prepend_data = iblk.data[0..(addr - iblk.start_addr - 1)];
          addr = iblk.start_addr;       // start within block (TRACE)
        } else if((end_addr + 1) == iblk.start_addr) {
          append_data = iblk.data;
          end_addr = iblk.end_addr;     // start adjacency (TRACE)
        };
        var nblk := iblk;
        while((nblk != NULL) && (nblk.start_addr <= (end_addr + 1))) {
          if(nblk != iblk) {
            delete(nblk);               // clear extra blocks (TRACE)
          };
          if(end_addr < nblk.end_addr) {
            append_data = nblk.data[(end_addr - nblk.start_addr + 1)..(nblk.data.size() - 1)];
            end_addr = nblk.end_addr;   // end within block (TRACE)
            break;
          } else if(end_addr == nblk.end_addr) {
            break;                      // existing block end (TRACE)
          };
          nblk = find_next_blk(nblk.end_addr + 1);
        };
        // stretch start block to fill entire cleared range
        iblk.start_addr = addr;
        iblk.end_addr = end_addr;
        iblk.data = { prepend_data; data; append_data };
      };
    };
    #ifdef CSCO_BLK_MEMORY_DEBUG_ASSERTIONS {
      consistency_check();
    };
  };

  read(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : list of byte is {
    var end_addr : csco_blk_memory_addr_t = addr + size - 1;
    var iblk : csco_blk_memory_blk;
    if(!path.is_empty() && path.top().has_addr(addr)) {
      iblk = path.top();                // cache hit (TRACE)
    } else {
      iblk = find_next_blk(addr);       // cache miss (TRACE)
    };
    result.resize(size, TRUE, default_value, FALSE); // FIXME perf??
    while((iblk != NULL) && (iblk.start_addr <= end_addr)) {
      var blk_start : csco_blk_memory_addr_t = max(addr, iblk.start_addr);
      var blk_end : csco_blk_memory_addr_t = min(end_addr, iblk.end_addr);
      result[(blk_start - addr)..(blk_end - addr)] =
        iblk.data[(blk_start - iblk.start_addr)..(blk_end - iblk.start_addr)];
      if(iblk.end_addr < end_addr) {
        iblk = successor(iblk);         // check for more data (TRACE)
      } else {
        break;                          // filled end of block (TRACE)
      };
    };
    messagef(CSCO_BLK_MEMORY, MEDIUM, "read addr=%x size=%x\n", addr, size) {
      if(at_message_verbosity(HIGH)) {
        out(get_data_string(addr, result));
      };
    };
  };

  // deallocate memory byte(s) (if previously allocated)
  // any stored data is lost
  free(addr: csco_blk_memory_addr_t) is {
    free_region(addr, 1);
  };

  free_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) is {
    messagef(CSCO_BLK_MEMORY, MEDIUM, "free addr=%x size=%x", addr, size);
    var end_addr : csco_blk_memory_addr_t = addr + size - 1;
    var iblk : csco_blk_memory_blk;
    if(!path.is_empty() && path.top().has_addr(addr)) {
      iblk = path.top();
    } else {
      iblk = find_next_blk(addr);
    };
    while((iblk != NULL) && (iblk.start_addr < end_addr)) {
      var free_blk_start := (iblk.start_addr >= addr);
      var free_blk_end := (iblk.end_addr <= end_addr);
      case {
        (free_blk_start && free_blk_end): {
          delete(iblk);                    // free entire block (TRACE)
        };
        (free_blk_start && !free_blk_end): {
          iblk.data = iblk.data[(end_addr + 1 - iblk.start_addr)..];
          iblk.start_addr = end_addr + 1;  // free block start (TRACE)
          break;
        };
        (!free_blk_start && free_blk_end): {
          iblk.data = iblk.data[..(addr - 1 - iblk.start_addr)];
          iblk.end_addr = addr - 1;        // free block end (TRACE)
        };
        default: {
          var nblk := new csco_blk_memory_blk with {
            .start_addr = end_addr + 1;
            .end_addr = iblk.end_addr;
            .data = iblk.data[(end_addr + 1 - iblk.start_addr)..];
          };
          iblk.end_addr = addr - 1;
          iblk.data = iblk.data[..(addr - 1 - iblk.start_addr)];
          compute find_blk(addr);          // FIXME should be able to opt out?
          insert(nblk);                    // split block (TRACE)
          break;
        };
      };
      iblk = successor(iblk);
    };
    #ifdef CSCO_BLK_MEMORY_DEBUG_ASSERTIONS {
      consistency_check();
    };
  };

  // deallocate all memory
  clear() is {
    root = NULL;
    path.clear();
  };

  // check if memory has been previously allocated
  check_alloc(addr: csco_blk_memory_addr_t) : bool is {
    result = ((!path.is_empty() && path.top().has_addr(addr)) ||
              (find_blk(addr) != NULL));
  };

  check_alloc_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : bool is {
    var end_addr : csco_blk_memory_addr_t = addr + size - 1;
    result = (!path.is_empty() && path.top().contains_interval(addr, end_addr));
    if(!result) {
      var iblk := find_blk(addr);
      result = ((iblk != NULL) && iblk.contains_interval(addr, end_addr));
    };
  };

  // get allocated ranges in specified region
  // returns a list of <start, end> pairs
  get_alloc_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : list of csco_blk_memory_addr_t is {
    var end_addr : csco_blk_memory_addr_t = addr + size - 1;
    var iblk : csco_blk_memory_blk;
    if(!path.is_empty() && path.top().has_addr(addr)) {
      iblk = path.top();                // cache hit (TRACE)
    } else {
      iblk = find_next_blk(addr);       // cache miss (TRACE)
    };
    while((iblk != NULL) && (iblk.start_addr <= end_addr)) {
      result.add(max(addr, iblk.start_addr));
      result.add(min(end_addr, iblk.end_addr));
      if(iblk.end_addr < end_addr) {
        iblk = successor(iblk);         // check for more data (TRACE)
      } else {
        break;                          // filled end of block (TRACE)
      };
    };
  };

  // output methods
  print() is {
    out(get_string());
  };

  // FIXME these string routines are likely terribly inefficient...
  get_string() : string is {
    var lines : list of string;
    var unallocated := "  ";
    var aline : list of string;
    aline.resize(0x20);
    for gap from 0x2 to 0xe step 2 {
      aline[gap] = " ";
      aline[gap + 0x10] = " ";
    };
    aline[0x10] = "-";

    // start at min blk
    var blk := find_next_blk(0);
    if(blk == NULL) {
      return("    <empty>");
    };
    var addr := blk.start_addr;
    addr &= ~0xf;
    var byte : uint;

    while(blk != NULL) {
      byte = (addr & 0xf);
      if(byte == 0) {
        aline[0] = appendf("    [%010x] ", addr);
      };
      aline[(byte << 1) + 1] =
        ((addr >= blk.start_addr)
         ? appendf("%02x", blk.data[addr - blk.start_addr])
         : unallocated);

      addr += 1;
      if(byte == 0xf) {
        lines.add(str_join(aline, ""));
      };
      if(addr > blk.end_addr) {
        // iterate by successor
        blk = successor(blk);
        if((blk != NULL) && ((addr >> 4) != (blk.start_addr >> 4))) {
          lines.add(str_join(aline[0..(byte << 1) + 1], ""));
          addr = blk.start_addr;
          addr &= ~0xf;
        };
      };
    };

    if(byte != 0xf) {
      lines.add(str_join(aline[0..(byte << 1) + 1], ""));
    };
    result = str_join(lines, "\n");
  };

  get_data_string(start_addr: csco_blk_memory_addr_t, data: list of byte) : string is {
    var lines : list of string;
    var aline : list of string;
    for each (b) in data {
      if(at_message_verbosity(FULL) ||
         (data.size() <= partial_dump_threshold) ||
         (index < partial_dump_head) ||
         ((index + partial_dump_tail) >= (((data.size() + 0xf) & ~0xf)))) {
        if((index & 0xf) == 0) {
          aline = { appendf("    [%010x]", start_addr + index) };
        };
        aline.add(appendf(((index & 0xf) != 0x7) ? " %02x" : "-%02x", b));
        if(((index & 0xf) == 0xf) ||
           ((index + 1) == data.size())) {
          lines.add(str_join(aline, ""));
          if(!at_message_verbosity(FULL) &&
             (data.size() > partial_dump_threshold) &&
             ((index + 1) == partial_dump_head)) {
            lines.add("    ...");
          };
        };
      };
    };
    result = str_join(lines, "\n");
  };

  //----------------------------------------
  // internal methods

  package consistency_check() is {
    var n : int;
    var h : int;
    if((root != NULL) && !root.verify_subtree(h, n)) {
      debug_dump_tree();
      error("consistency check failed!!!");
    };
  };

  // find block allocated at address
  // maintains path even if not found (for subsequent insert)
  package find_blk(addr: csco_blk_memory_addr_t) : csco_blk_memory_blk is {
    path.clear();
    result = root;
    while(result != NULL) {
      if(addr < result.start_addr) {
        path.push(result);
        result = result.left;           // look left for match (TRACE)
      } else if(addr > result.end_addr) {
        path.push(result);
        result = result.right;          // look right for match (TRACE)
      } else {
        path.push(result);
        break;                          // match current blk (TRACE)
      };
    };
  };

  // find block allocated at or after address
  package find_next_blk(addr: csco_blk_memory_addr_t) : csco_blk_memory_blk is {
    path.clear();
    var blk := root;
    while(blk != NULL) {
      if(addr < blk.start_addr) {
        path.push(blk);                 // record path
        result = blk;                   // record last match
        blk = blk.left;                 // look left for closer match (TRACE)
      } else if(addr > blk.end_addr) {
        path.push(blk);                 // record path
        blk = blk.right;                // look left for match (TRACE)
      } else {
        path.push(blk);
        result = blk;                   // hit current blk (TRACE)
        break;
      };
    };

    if(result != NULL) {
      while(result != path.top()) {
        compute path.pop();             // remove false path (TRACE)
      };
    };
  };

  // find adjacent block (similar to find_next_blk)
  // assumes path is to blk and updates it to new blk.
  // if there is no successor, the path is unmodified
  package successor(blk: csco_blk_memory_blk) : csco_blk_memory_blk is {
    if(blk.right != NULL) {
      result = blk.right;
      path.push(result);
      while(result.left != NULL) {
        result = result.left;
        path.push(result);
      };
    } else {
      var depth := path.size() - 1;
      while((depth > 0) && (path[depth] != path[depth - 1].left)) {
        depth -= 1;
      };
      depth -= 1;
      if(depth >= 0) {
        path = path[0..depth];
        result = path[depth];
      };
    };
  };

  // rotate subtree left at given path index
  package rotate_left(depth: int) is {
    var x := path[depth];
    var y := x.right;
    x.right = y.left;
    y.left = x;
    if(depth == 0) {
      root = y;                         // blk replaces root (TRACE)
    } else if(x == path[depth - 1].left) {
      path[depth - 1].left = y;         // blk replaces parent left (TRACE)
    } else {
      path[depth - 1].right = y;        // blk replaces parent right (TRACE)
    };
    if((path.size() > (depth + 1)) && (y == path[depth + 1])) {
      // y is in path
      if((path.size() <= (depth + 2)) || (path[depth + 2] == y.right)) {
        path.delete(depth);             // remove x from path (TRACE)
      } else {
        path[depth] = y;                // swap x and y in path (TRACE)
        path[depth + 1] = x;
      };
    } else {
      path.insert(depth, y);            // add y to path (TRACE)
    };
  };

  // rotate subtree right at given path index
  package rotate_right(depth: int) is {
    var x := path[depth];
    var y := x.left;
    x.left = y.right;
    y.right = x;
    if(depth == 0) {
      root = y;                         // blk replaces root (TRACE)
    } else if(x == path[depth - 1].left) {
      path[depth - 1].left = y;         // blk replaces parent left (TRACE)
    } else {
      path[depth - 1].right = y;        // blk replaces parent right (TRACE)
    };
    if((path.size() > (depth + 1)) && (y == path[depth + 1])) {
      // y is in path
      if((path.size() <= (depth + 2)) || (path[depth + 2] == y.left)) {
        path.delete(depth);             // remove x from path (TRACE)
      } else {
        path[depth] = y;                // swap x and y in path (TRACE)
        path[depth + 1] = x;
      };
    } else {
      path.insert(depth, y);            // add y to path (TRACE)
    };
  };

  // insert new block into tree and rebalance
  package insert(new_blk: csco_blk_memory_blk) is {
    new_blk.left = NULL;
    new_blk.right = NULL;
    if(root == NULL) {
      root = new_blk;                   // empty tree case (TRACE)
      path.push(root);
    } else {
      // start w/normal insert
      var parent := path.top();
      if((new_blk.end_addr < parent.start_addr) &&
         (parent.left == NULL)) {
        parent.left = new_blk;          // insert to left (TRACE)
      } else if((new_blk.start_addr > parent.end_addr) &&
                (parent.right == NULL)) {
        parent.right = new_blk;         // insert to right (TRACE)
      } else {
        new_blk.debug_dump_node();
        parent.debug_dump_subtree("");
        error("consistency check failure during insertion");
      };
      path.push(new_blk);
      var depth := (path.size() - 1);

      // rebalance back up to root
      new_blk.mark = TRUE;
      while((depth > 0) && (path[depth - 1].mark)) {
        if(path[depth - 1] == path[depth - 2].left) {
          if((path[depth - 2].right != NULL) && path[depth - 2].right.mark) {
            path[depth - 1].mark = FALSE;
            path[depth - 2].mark = TRUE;
            path[depth - 2].right.mark = FALSE;
            depth -= 2;                 // left case 1 (TRACE)
          } else {
            if(path[depth] == path[depth - 1].right) {
              rotate_left(depth - 1);   // left case 2 (TRACE)
            };
            depth -= 1;
            path[depth].mark = FALSE;
            path[depth - 1].mark = TRUE;
            rotate_right(depth - 1);    // left case 3 (TRACE)
          };
        } else {
          if((path[depth - 2].left != NULL) && path[depth - 2].left.mark) {
            path[depth - 1].mark = FALSE;
            path[depth - 2].mark = TRUE;
            path[depth - 2].left.mark = FALSE;
            depth -= 2;                 // right case 1 (TRACE)
          } else {
            if(path[depth] == path[depth - 1].left) {
              rotate_right(depth - 1);  // right case 2 (TRACE)
            };
            depth -= 1;
            path[depth].mark = FALSE;
            path[depth - 1].mark = TRUE;
            rotate_left(depth - 1);     // right case 3 (TRACE)
          };
        };
      };
    };
    root.mark = FALSE;
    #ifdef CSCO_BLK_MEMORY_DEBUG_ASSERTIONS {
      consistency_check();
    };
  };

  // delete block from tree and rebalance
  // path is to blk, to blk's parent on return
  package delete(blk: csco_blk_memory_blk) is {
    var x : csco_blk_memory_blk;  // replacement node
    if((blk.left != NULL) && (blk.right != NULL)) {
      // remove successor from tree and swap w/blk
      // x will _always_ have at least one free child
      x = successor(blk);
      #ifdef CSCO_BLK_MEMORY_DEBUG_ASSERTIONS {
        assert((x.left == NULL) || (x.right == NULL));
      };
      delete(x);
      // refresh path to blk (this is kindof cheezy...)
      compute find_blk(blk.start_addr);
      // replace blk w/successor
      x.left = blk.left;
      x.right = blk.right;
      x.mark = blk.mark;
      blk.mark = TRUE;  // skip rebalance
    } else {
      // otherwise replace blk w/blk's only child
      x = (blk.left != NULL) ? blk.left : blk.right;
    };
    var depth := (path.size() - 1);
    path[depth] = x;                    // replace deleted node
    if(depth == 0) {
      root = x;                         // deleted root (TRACE)
    } else if(path[depth - 1].left == blk) {
      path[depth - 1].left = x;         // deleted left child (TRACE)
    } else {
      #ifdef CSCO_BLK_MEMORY_DEBUG_ASSERTIONS {
        assert(path[depth - 1].right == blk);
      };
      path[depth - 1].right = x;        // deleted right child (TRACE)
    };
    if(!blk.mark) {
      // rebalance tree
      while((depth > 0) && ((path[depth] == NULL) || !path[depth].mark)) {
        if(path[depth] == path[depth - 1].left) {
          var w := path[depth - 1].right;
          if(w.mark) {
            w.mark = FALSE;
            path[depth - 1].mark = TRUE;
            rotate_left(depth - 1);     // left case 1 (TRACE)
            depth += 1;
            w = path[depth - 1].right;
          };
          if(((w.left == NULL) || (!w.left.mark)) &&
             ((w.right == NULL) || (!w.right.mark))) {
            w.mark = TRUE;              // left case 2 (TRACE)
            depth -= 1;
          } else {
            if((w.right == NULL) || (!w.right.mark)) {
              w.left.mark = FALSE;
              w.mark = TRUE;
              // FIXME...
              var path_save := path.copy();
              path = path[0..depth];
              path[depth] = w;
              rotate_right(depth);      // left case 3 (TRACE)
              path = path_save;
              w = path[depth - 1].right;
            };
            w.mark = path[depth - 1].mark;
            path[depth - 1].mark = FALSE;
            w.right.mark = FALSE;
            rotate_left(depth - 1);     // left case 4 (TRACE)
            break;
          };
        } else {
          var w := path[depth - 1].left;
          if(w.mark) {
            w.mark = FALSE;
            path[depth - 1].mark = TRUE;
            rotate_right(depth - 1);     // right case 1 (TRACE)
            depth += 1;
            w = path[depth - 1].left;
          };
          if(((w.left == NULL) || (!w.left.mark)) &&
             ((w.right == NULL) || (!w.right.mark))) {
            w.mark = TRUE;              // right case 2 (TRACE)
            depth -= 1;
          } else {
            if((w.left == NULL) || (!w.left.mark)) {
              w.right.mark = FALSE;
              w.mark = TRUE;
              // FIXME...
              var path_save := path.copy();
              path = path[0..depth];
              path[depth] = w;
              rotate_left(depth);       // left case 3 (TRACE)
              path = path_save;
              w = path[depth - 1].left;
            };
            w.mark = path[depth - 1].mark;
            path[depth - 1].mark = FALSE;
            w.left.mark = FALSE;
            rotate_right(depth - 1);    // left case 4 (TRACE)
            break;
          };
        };
      };
      if((depth >= 0) && (path[depth] != NULL)) {
        path[depth].mark = FALSE;
      };
    };
    compute path.pop();  // leave path at blk's parent
    #ifdef CSCO_BLK_MEMORY_DEBUG_ASSERTIONS {
      consistency_check();
    };
  };

  debug_dump_tree() is {
    if(root != NULL) {
      root.debug_dump_subtree("");
    };
  };
};


//------------------------------------------------------------
// unit memory instance
// backward compatibility interface wrapper
// not recommended for new designs

unit csco_blk_memory_u {
  name : csco_blk_memory_name_t;
    keep soft name == MEM;

  short_name() : string is {
    result = name.as_a(string);
  };

#ifdef SPECMAN_VERSION_4_2_OR_LATER {
  logger : message_logger is instance;
} #else {
  logger : message_logger;
};

  // default value of a memory location
  // used to fill memory if an address is read before it is written
  default_value : byte;
    keep soft default_value == 0;

  package mem : csco_blk_memory;
    keep mem.default_value == value(default_value);
    keep mem.name == value(name);

  //----------------------------------------
  // general access interface
  //   writes allocate memory if necessary
  //   reads allocate memory w/the default_value
  write(addr: csco_blk_memory_addr_t, data: list of byte) is {
    mem.write(addr, data);
  };

  read(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : list of byte is {
    result = mem.read(addr, size);
  };

  // single byte access interface
  write_byte(addr: csco_blk_memory_addr_t, data: byte) is {
    mem.write(addr, { data });
  };

  read_byte(addr: csco_blk_memory_addr_t) : byte is {
    result = mem.read(addr, 1)[0];
  };

  // deallocate memory byte(s) (if previously allocated)
  // any stored data is lost
  free(addr: csco_blk_memory_addr_t) is {
    mem.free(addr);
  };

  free_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) is {
    mem.free_region(addr, size);
  };

  // deallocate all memory
  clear() is {
    mem.clear();
  };

  // check if memory has been previously allocated
  check_alloc(addr: csco_blk_memory_addr_t) : bool is {
    result = mem.check_alloc(addr);
  };

  check_alloc_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : bool is {
    result = mem.check_alloc_region(addr, size);
  };

  // get allocated ranges in specified region
  // returns a list of <start, end> pairs
  get_alloc_region(addr: csco_blk_memory_addr_t, size: csco_blk_memory_addr_t) : list of csco_blk_memory_addr_t is {
    result = mem.get_alloc_region(addr, size);
  };

  // output methods
  print() is {
    mem.print();
  };

  get_string() : string is {
    result = mem.get_string();
  };
};

'>
