//-----------------------------------------------------------------------------
// filename: testbench.e
//
// description:
//   simple standalone environment for sanity testing
//   and comparisons with other memory models
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// Copyright 2002-2005 (c) Cisco Systems Inc.
//----------------------------------------------------------------------------
// The contents of this file are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in
// compliance with the License. You may obtain a copy of the License
// at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "Cisco generic block-based memory model".
// Portions created by the Initial Developer are Copyright (C) 2002-2005
// Cisco Systems Inc. All Rights Reserved.
//----------------------------------------------------------------------------
<'
package csco_blk_memory;
import csco_blk_memory/e/csco_blk_memory_top;

import flat_mem;    // simple list based comparison model
import hash_mem;    // hash based comparison model (64 bit version of shr_ram)

// range of data sizes used by a test/transaction
type tb_size_range_t : [
  BYTE        =    1,
  SMALL       =   16,
  MEDIUM      =  128,
  LARGE       =    1k,
  KILO_SMALL  =   16k,
  KILO_MEDIUM =  128k,
  KILO_LARGE  =    1m,
  MEG_SMALL   =   16m,
  MEG_MEDIUM  =  128m,
  MEG_LARGE   = 1024m
];

type tb_operation_t : [ WRITE, READ ];

//------------------------------------------------------------
// testbench instances

extend csco_blk_memory_name_t : [ BLK_MEM ];
extend flat_mem_name_t : [ FLAT_MEM ];
extend shr_ram_name_t : [ HASH_MEM ];

extend sys {
  short_name() : string is only { result = "TEST" };
  keep soft logger.verbosity == MEDIUM;
  keep soft logger.tags == { NORMAL; CSCO_BLK_MEMORY };

  // test model
  blk_mem : BLK_MEM'name csco_blk_memory;
  enable_blk_mem : bool;
    keep soft enable_blk_mem;

  // global test size control
  max_test_size : tb_size_range_t;
    // largest sizes only by explicit request
    keep soft max_test_size not in [MEG_MEDIUM..MEG_LARGE];

  // reference models
  hash_mem : HASH_MEM'name shr_ram;
  enable_hash_mem : bool;
    keep soft enable_hash_mem;

  flat_mem : FLAT_MEM'name flat_mem;
  enable_flat_mem : bool;
    keep soft enable_flat_mem;
    keep flat_mem.size == value((enable_flat_mem)
                                ? max_test_size.as_a(uint)
                                : 0);

  // some memory must be enabled
  keep enable_blk_mem or enable_hash_mem or enable_flat_mem;

  // some quick cmdline configs
  #ifdef CFG_BLK_ONLY {
    keep enable_blk_mem && !enable_hash_mem && !enable_flat_mem;
  };
  #ifdef CFG_HASH_ONLY {
    keep !enable_blk_mem && enable_hash_mem && !enable_flat_mem;
  };
  #ifdef CFG_FLAT_ONLY {
    keep !enable_blk_mem && !enable_hash_mem && enable_flat_mem;
  };
  #ifdef CFG_MAX_SIZE {
    keep max_test_size == CFG_MAX_SIZE;
  };

  post_generate() is also {
    message(LOW, "test configuration:");
    messagef(LOW, "  enabled models:") {
      if(enable_blk_mem) { outf(" BLK") };
      if(enable_hash_mem) { outf(" HASH") };
      if(enable_flat_mem) { outf(" FLAT") };
    };
    message(LOW, "  max_test_size: ", max_test_size);
  };
};


//------------------------------------------------------------
// testbench memory access transaction

struct tb_mem_transaction {
  operation : tb_operation_t;
  size_range : tb_size_range_t;
    keep soft size_range in [ BYTE..KILO_SMALL ];
  size : csco_blk_memory_addr_t;

  data : list of byte;
  keep gen (data) before (addr);
  addr : csco_blk_memory_addr_t;

  when WRITE'operation tb_mem_transaction {
    keep data.size() == size;
  };
  when READ'operation tb_mem_transaction {
    keep data.size() == 0;
  };

  keep gen (operation, size_range) before (data);
  keep gen (data) before (addr);

  keep soft size_range == select {
     8k: BYTE;
     4k: SMALL;
     1k: MEDIUM;
    256: LARGE;
     64: KILO_SMALL;
     16: KILO_MEDIUM;
      4: KILO_LARGE;
      1: MEG_SMALL;
  };

  when BYTE tb_mem_transaction        { keep size == 1 };
  when SMALL tb_mem_transaction       { keep size in [1+1..16] };
  when MEDIUM tb_mem_transaction      { keep size in [16+1..128] };
  when LARGE tb_mem_transaction       { keep size in [128+1..1k] };
  when KILO_SMALL tb_mem_transaction  { keep size in [1k+1..16k] };
  when KILO_MEDIUM tb_mem_transaction { keep size in [16k+1..128k] };
  when KILO_LARGE tb_mem_transaction  { keep size in [128k+1..1m] };
  when MEG_SMALL tb_mem_transaction   { keep size in [1m+1..16m] };
  when MEG_MEDIUM tb_mem_transaction  { keep size in [16m+1..128m] };
  when MEG_LARGE tb_mem_transaction   { keep size in [128m+1..1024m] };

  keep soft size == select {
    2k: 1;            2k: [2..4];        1k: [4..8];        1k: [8..16];
   512: [16..32];    512: [32..64];     256: [64..128];    256: [128..256];
   128: [256..512];  128: [512..1k];     64: [1k..2k];      64: [2k..4k];
    32: [4k..8k];     32: [8k..16k];     16: [16k..32k];    16: [32k..64k];
     8: [64k..128k];   8: [128k..256k];   4: [256k..512k];   4: [512k..1m];
     2: [1m..2m];      2: [2m..4m];       1: [4m..8m];       1: [8m..16m];
  };

  event transferred;

  cover transferred using radix=HEX is {
    item operation;
    item size_range;
    item addr_lo : uint(bits: 5) = addr[4:0];
    item addr_k : uint(bits: 5) = addr[14:10];
    item addr_m : uint(bits: 5) = addr[24:20];

    cross operation, size_range;
    cross operation, size_range, addr_m;
  };
};

extend sys {
  !transaction : tb_mem_transaction;
    keep soft transaction.size_range.as_a(uint) < value(max_test_size).as_a(uint);
    keep transaction.addr <= (max_test_size.as_a(csco_blk_memory_addr_t) - transaction.size);
  when BYTE sys        { keep transaction.size_range == BYTE };
  when SMALL sys       { keep transaction.size_range in [BYTE..SMALL] };
  when MEDIUM sys      { keep transaction.size_range in [BYTE..MEDIUM] };
  when LARGE sys       { keep transaction.size_range in [BYTE..LARGE] };
  when KILO_SMALL sys  { keep transaction.size_range in [BYTE..KILO_SMALL] };
  when KILO_MEDIUM sys { keep transaction.size_range in [BYTE..KILO_MEDIUM] };
  when KILO_LARGE sys  { keep transaction.size_range in [BYTE..KILO_LARGE] };
  when MEG_SMALL sys   { keep transaction.size_range in [BYTE..MEG_SMALL] };
  when MEG_MEDIUM sys  { keep transaction.size_range in [BYTE..MEG_SMALL] };
  when MEG_LARGE sys   { keep transaction.size_range in [BYTE..MEG_SMALL] };

  count : uint;
    keep soft count == 10k;

  post_generate() is also {
    messagef(LOW, "  transaction count: %x\n", count);
  };

  // test runs automatically w/o simulator
  run() is also {
    start send_loop();
    emit clk;
  };

  event clk is { @clk; [1] };

  // normal test executes transactions
  send_loop() @clk is {
    for i from 1 to count {
      wait [1];
      gen (transaction);
      execute(transaction);
    };
    stop_run();
  };

  execute(t: tb_mem_transaction) is {
    if(t.operation == WRITE) {
      if(enable_blk_mem) {
        blk_mem.write(t.addr, t.data);
      };
      if(enable_hash_mem) {
        hash_mem.write(t.addr, t.data);
      };
      if(me is a enable_flat_mem sys (efm)) {
        efm.flat_mem.write(t.addr, t.data);
      };
      message(HIGH, "current memory contents:\n") { blk_mem.print() };
    } else {
      if(enable_blk_mem) {
        t.data = blk_mem.read(t.addr, t.size);
      };
      if(enable_hash_mem) {
        var hash_data := hash_mem.read(t.addr, t.size);
        if(enable_blk_mem) {
          assert(t.data == hash_data);
        } else {
          t.data = hash_data;
        };
      };
      if(enable_flat_mem) {
        var flat_data := flat_mem.read(t.addr, t.size);
        if(enable_blk_mem || enable_hash_mem) {
          assert(t.data == flat_data);
        } else {
          t.data = flat_data;
        };
      };
    };
    emit t.transferred;
  };
};


//------------------------------------------------------------
// extra debug/test code

extend csco_blk_memory {
  total_blks() : int is {
    result = (root == NULL) ? 0 : root.total_blks();
  };
};

extend csco_blk_memory_blk {
  total_blks() : int is {
    result = 1;
    if(left != NULL) {
      result += left.total_blks();
    };
    if(right != NULL) {
      result += right.total_blks();
    };
  };
};


//------------------------------------------------------------
// basic specman config

extend sys {
  setup() is also {
    set_config(print, line_size, 120, radix, HEX);
    set_config(print, list_index_radix, HEX);
    set_config(print, list_starts_on_right, FALSE);
    set_config(gen, max_structs, 4m);
    set_config(gen, absolute_max_list_size, 1024m + 1m);
    #ifdef SPECMAN_VERSION_4_2_1_OR_LATER {
      set_config(gen, static_analysis_opt, TRUE);
      set_config(gen, unit_reference_rule, TRUE);
    };
    set_config(memory, otf_threshold, 1024m);
    set_config(memory, gc_threshold, 512m);
    set_config(memory, gc_increment, 256m);
    set_config(memory, max_size, 3100m);
    set_config(memory, absolute_max_size, 3102m);
    set_config(memory, retain_printed_structs, FALSE);
    set_config(run, tick_max, 1000000);
    //set_config(run, use_manual_tick, TRUE);
  };

  finalize() is also {
    // take out the trash
    transaction = NULL;
    do_otf_gc();
  };
};

extend global {
  finalize_test() is also {
    // simple cpu performance metric
    out("--------------------------------------------------");
    outf("executed %d transactions in %d seconds user time\n",
         sys.count, session.user_time);
    var Mtps : time = sys.count * 1000;
    if(session.user_time > 0) {
      Mtps /= session.user_time;
      outf("  = %d.%03d transactions per second\n",
           Mtps / 1000, Mtps % 1000);
    };
    out("--------------------------------------------------");
  };
};

'>
