//------------------------------------------------------------------------
// File name    : csco_clkgen_top.e
// Project      : csco_clkgen
//              :
// Created On   : 05/24/2001
//              :
//              :
// Description  : clock generator unit, maps to csco_clkgen verilog module.
//              : to use this just create an instance and point hdl_path()
//              : to the csco_clkgen isntance you want to control.
//              : at a minimum the period should be constrained
//              : the other parameters should pick sane defaults
//              : you might want to do a keep .reset_soft() if you want
//              : the other parameters to vary
//              :
//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Copyright 2001-2006 (c) Cisco Systems Inc.
//------------------------------------------------------------------------
<'
package csco_clkgen;

define CSCO_CLKGEN_2_0;
define CSCO_CLKGEN_2_0_OR_LATER;
define CSCO_CLKGEN_1_0_OR_LATER;

// eRM utilities
import evc_util/e/evc_util_top;

// config tree support
import csco_config/e/csco_config_top;

// verbosity package
import csco_verbosity/e/csco_verbosity_top;

//========================================================================
// API definition
//========================================================================

// clock generator debug messages
extend message_tag : [ CSCO_CLKGEN ];

// clock type name for clkgen unit
type csco_clkgen_name_t : [ DEFAULT_CLK ];

type csco_clkgen_edge_mode_t : [ DISABLED, POSEDGE, NEGEDGE, ANYEDGE ];

// type of model to auto-instance
type csco_clkgen_model_t : [
  NO_SIMULATOR,  // for running without a simulator
                 //   emits clock every period/1000 occurrances of sys.any
  SIMPLE,        // original model with simple skew and jitter
  SHAGGY         // more complex model w/divider/multiplier and cyclic skewing
];

// HDL signal type to drive
type csco_clkgen_hdl_type_t : [
  NONE,    // do not connect clock to DUT (specman only clock)
  REG,     // drive verilog "reg" type
  WIRE     // drive (otherwise undriven) verilog "wire" type
];


//------------------------------------------------------------
// clock generator unit
unit csco_clkgen_u like any_env {
  // clock type name for clkgen instance
  name : csco_clkgen_name_t;
    keep name == config.name;
                                                           
  short_name() : string is also { 
    // assuming name follows style of <PROJ>_<NAME>, this code will return just <NAME>
    // if there is no "_" in the name, it will return the whole name
    var substrings := str_split_all(name.as_a(string),"_");
    if substrings.size() > 2 {
      result = str_join(substrings[2..],"") ;
    } else {
      result = name.as_a(string);
    };
  };

  short_name_style() : vt_style is also { result = PURPLE };

  // display general status information
  show_status() is also { }; // pre-defined for any_env

  // method interface to update enabled state
  // and reprogram the verilog HDL driver.
  // ignored for PASSIVE clock monitors.
  // Useful for OIR.
  set_enable(new_state: bool) @sys.any is undefined;

  //----------------------------------------
  // clock edge event

  // configurable for positive, negative or both edges
  event clock;

  // clock signal to sample or drive
  sig_clock : in simple_port of bit is instance;
    keep read_only(sig_clock.hdl_path() != NULL) =>
      bind(sig_clock, external);
};


//------------------------------------------------------------
// clock configuration struct
DEFINE_CFG_STRUCT begin
  config_prefix = csco_clkgen,
  parent_unit_type = csco_clkgen_u 
end;

extend csco_clkgen_config_s {
  // Name of the clock type, usually refers to a clock domain
  name : csco_clkgen_name_t;
    keep soft (name == DEFAULT_CLK);

  // ACTIVE driver or PASSIVE monitor:
  //   ACTIVE mode generates clocks using HDL model
  //   PASSIVE mode expects some HDL source (DUT) to provide a clock
  //     which this model samples and distributes to its clients
  active_passive : erm_active_passive_t;
    keep soft active_passive == PASSIVE;

  // model can be enabled/disabled and re-programmed to start/stop clock
  enable : bool;
    keep soft enable;

  // control whether HDL clock driver module
  // is automatically instanced in specman stubs
  auto_instance : bool;
    keep soft auto_instance;

  // model to instance
  auto_model : csco_clkgen_model_t;
    keep soft auto_model == SIMPLE;

  // target verilog signal type to drive
  auto_drive_type : csco_clkgen_hdl_type_t;
    keep soft auto_drive_type == REG;

  // each event also has individual enable

  // enable event at rising edge of clock
  edge_mode : csco_clkgen_edge_mode_t;
    keep soft edge_mode == POSEDGE;

  // all clk params in units of 1/1000 of timescale (ps?)
  period : uint;
    keep soft period == MAX_UINT;

  // percentage of period that clock will be high
  duty_cycle : uint;
    keep (duty_cycle in [1..99]);
    keep soft (duty_cycle == 50);

  // delay from programming time to first edge transition
  skew : uint;
    keep soft (skew == 0);
    keep skew < period;

  // clock "phase"
  //   FALSE => start clk at 1'b0, first edge is rising after "skew" delay
  //   TRUE  => start clk at 1'b1, first edge is falling after "skew" delay
  inv_phase : bool;
    keep soft !inv_phase;

  // here we define "jitter" to be the size of the window
  // in which the clock is allowed to transition
  // this window will _start_ from "skew" so if you need
  // it centered you better do a "skew" -= "jitter" / 2
  jitter : uint;
    keep soft (jitter == 0);

  // configuration coverage
  event generated;

  cover generated is {
    item name using per_instance,
      ignore = (name == DEFAULT_CLK);
    item config_mode;
    item period;
    item duty_cycle;
    item skew;
    item jitter;
  };
};


//========================================================================
// implementation details
//========================================================================

extend csco_clkgen_u {
  // copy config subtypes
  package !active_passive : erm_active_passive_t;
  package !auto_instance : bool;
  package !auto_model : csco_clkgen_model_t;
  package !auto_drive_type : csco_clkgen_hdl_type_t;
  package !hdl_gen : bool;
  package !edge_mode : csco_clkgen_edge_mode_t;

  csco_config_update_subtypes() is {
    active_passive = config.active_passive;
    auto_instance = config.auto_instance;
    auto_model = config.auto_model;
    auto_drive_type = config.auto_drive_type;
    hdl_gen = config.auto_model != NO_SIMULATOR;
    edge_mode = config.edge_mode;
  };

  when hdl_gen POSEDGE'edge_mode csco_clkgen_u {
    event clock is only rise(sig_clock$) @sim;
  };
  when hdl_gen NEGEDGE'edge_mode csco_clkgen_u {
    event clock is only fall(sig_clock$) @sim;
  };
  when hdl_gen ANYEDGE'edge_mode csco_clkgen_u {
    event clock is only change(sig_clock$) @sim;
  };

  // auto-instance differentiator
  package inst_id : string;
    keep inst_id == appendf("%s_%08x", name, e_path().as_a(list of byte).crc_32(0, str_len(e_path())));

  // extract verilog path from specman path-like version
  package get_full_hdl_path() : string is {
    result = str_replace(sig_clock.full_hdl_path(), "/^~[/]/", "");
    var pc := str_split_all(result, "/\\s+|[\\]/");
    var esc := FALSE;
    for each (c) in pc {
      if(c ~ "/\\s+|[\\]/") {
        esc = (c == "\\");
      } else if(!esc) {
        pc[index] = str_replace(c, "/[/]/", ".");
      };
    };
    result = str_join(pc, "");
  };

  short_name() : string is also {
    if(result !~ "/cl(oc)?k/") {
      result = append(result, "_CLK");
    };
  };

  show_banner() is also {
    out("Copyright 2001-2006 (c) Cisco Systems Inc.");
    out("    instance : ", name, " (", config.active_passive, ")");
  };

  show_status() is also {
    config.csco_print(LONG);
    out("(id=", inst_id, ")");
  };

  set_enable(new_state: bool) @sys.any is {
    config.enable = new_state;
    program();
  };

  package program() @sys.any is empty;

  when ACTIVE hdl_gen csco_clkgen_u {
    // automatically start active clock at runtime
    run() is also {
      start program();
    };

    // program parameters into verilog model and
    // create an edge on the 'run' reg causing clock to start.
    // runs automatically in run() or can be activated by
    // set_enable to re-program the clock
    package program() @sys.any is also {
      assert(config.enable => (config.period < MAX_UINT)) else
        error("ERR_CSCO_CLKGEN_UNCONSTRAINED_PERIOD\n",
              "ACTIVE, enabled clkgen must have period constrained\n",
              "  clkgen: ", get_unit().e_path());

      var path : string;
      if(config.auto_instance) {
        path = append("~/specman/i", inst_id, "/");
        '~/specman/(inst_id)_en' = !config.enable;
      };
      '(path)PERIOD'    = (config.enable) ? config.period : 0;
      '(path)DUTYCYCLE' = config.duty_cycle;
      '(path)SKEW'      = config.skew;
      '(path)JITTER'    = config.jitter;
      '(path)INV_PHASE' = config.inv_phase;

      // create an edge on the run signal
      '(path)run' = 1'b0;
      wait delay(1);
      '(path)run' = 1'b1;

      messagef(CSCO_CLKGEN, MEDIUM, "csco_clkgen\n") { config.csco_print(SHORT) };
    };
  };

  when PASSIVE csco_clkgen_u {
    run() is also {
      messagef(CSCO_CLKGEN, HIGH, "csco_clkgen\n") { config.csco_print(SHORT) };
    };
  };

  when ACTIVE FALSE'hdl_gen csco_clkgen_u {
    keep bind(sig_clock, empty);
    package !period : uint;
    package !period_frac : int;

    run() is also {
      start driver();
    };

    package program() @sys.any is {
      if(edge_mode == ANYEDGE) {
        period = config.period / 2000;
        period_frac = (config.period % 2000);
      }
      else {
        period = config.period / 1000;
        period_frac = (config.period % 1000) * 2;
      };
    };

    package driver() @sys.any is {
      messagef(CSCO_CLKGEN, MEDIUM, "csco_clkgen\n") { config.csco_print(SHORT) };
      program();
      var delay : uint = config.skew / 1000;     // integral ns period
      var frac : int = (config.skew % 1000) * 2; // half-ps accumulator
      while(TRUE) {
        if(frac >= 1000) {
          frac -= 2000;
          assert(frac >= -1000);
          delay += 1;
        };
        if(delay > 0) {
          wait [delay];
        } else {
          sync [0];
        };
        emit clock;
        delay = period;
        frac += period_frac;
        assert(period > 0 || period_frac > 0);
      };
    };
  };

  when FALSE'auto_instance hdl_gen csco_clkgen_u {
    // default clock port name
    keep soft sig_clock.hdl_path() == "clk";
  };

  when ACTIVE auto_instance hdl_gen csco_clkgen_u {
    // hack verilog into stubs file
    verilog code {
      dec("wire ", inst_id, ";");            // clock output wire
      dec("reg ", inst_id, "_en;");          // output enable control
      ((config.auto_drive_type == REG)
       ? dec("always @(", inst_id, ") ",
             get_full_hdl_path(), " <= ", inst_id, ";")
       : (config.auto_drive_type == WIRE)
       ? dec("assign ", get_full_hdl_path(), " = ", inst_id, ";")
       : "");
    };
  };

  when NONE'auto_drive_type ACTIVE auto_instance hdl_gen csco_clkgen_u {
    keep soft sig_clock.hdl_path() == read_only(append("~/specman/i", inst_id, "/clk"));
  };

  // choice of clock generator modules
  when SIMPLE ACTIVE auto_instance hdl_gen csco_clkgen_u {
    verilog code {
      // instance csco_clkgen verilog module
      dec("csco_clkgen_simple i", inst_id, " (", // instance name
          ".clk(", inst_id, "), ",           // clock output
          "._en(", inst_id, "_en));");       // output enable
    };
  };

  when SHAGGY ACTIVE auto_instance hdl_gen csco_clkgen_u {
    verilog code {
      // instance csco_clkgen verilog module
      dec("wire [3:0]", inst_id, "_div;");   // divider input (N/C)
      dec("wire [3:0]", inst_id, "_mul;");   // multiplier input (N/C)
      dec("csco_clkgen_shaggy i", inst_id, " (", // instance name
          ".clk(", inst_id, "), ",           // clock output
          "._en(", inst_id, "_en), ",        // output enable
          ".div(", inst_id, "_div), ",       // divider
          ".mul(", inst_id, "_mul));");      // multiplier
    };
  };
};


//------------------------------------------------------------
extend csco_clkgen_config_s {
  post_generate() is also {
    emit generated;
  };

  csco_print_attr name SHORT enable;
  csco_print_attr enable SHORT enable;
  csco_print_attr active_passive SHORT enable;
  csco_print_attr active_passive SHORT header=locale;
  when ACTIVE csco_clkgen_config_s {
    csco_print_attr period SHORT enable;
    csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
      if pf.name == "period" {
         result = appendf("%2d.%3.3d", (period / 1000), (period % 1000))
      };
      if pf.name == "skew" {
         result = appendf("%2d.%3.3d", (skew / 1000), (skew % 1000))
      };
      if pf.name == "jitter" {
         result = appendf("%2d.%3.3d", (jitter / 1000), (jitter % 1000))
      };
    };
  };

  csco_print_attr auto_model LONG disable;
  when TRUE'auto_instance csco_clkgen_config_s {
    csco_print_attr auto_model LONG enable;
  };
  csco_print_get_scalar_field_value(pf : csco_print_field_s) : string is also {
    if pf.name == "auto_drive_type" and pf.format == LONG {
--print me;
--if parent_unit == NULL {
--   out("parent unit is NULL");
--} else {
      result = append(result, " (hdl_path=", parent_unit.sig_clock.full_hdl_path(), ")");
--};
    };
  };

  csco_print_attr period LONG disable;
  csco_print_attr duty_cycle LONG disable;
  csco_print_attr skew LONG disable;
  csco_print_attr inv_phase LONG disable;
  csco_print_attr jitter LONG disable;
  when ACTIVE csco_clkgen_config_s {
    csco_print_attr period LONG enable;
    csco_print_attr duty_cycle LONG enable;
    csco_print_attr duty_cycle ALL radix=DEC;
    csco_print_attr skew LONG enable;
    csco_print_attr inv_phase LONG enable;
    csco_print_attr jitter LONG enable;
  };
};


'>
