//-------------------------------------------------------------------------------------------------------------
//
//   Filename : csco_clkgen_fullchip_dev_testbench_env.e
//   Language : 'e'
//   Author   : Sean Smith (Cisco)
//
//   General Description: Example Unit level Testbench to illustrate the usage of the csco_clkgen model
//                        in several scenarios.  This model illustrates how to instantiate the model,
//                        configure its output and connect it to your system.
//
//   History:
//   01/07/02  SWS  Initial Coding
//
//--------------------------------------------------------------------------------------------------------------

<'

//========================================================================
// UNIT FOR TOP LEVEL TB
//========================================================================

extend tb_clock_region_t: [fullchip_tb_clkgen];            // Define 1 new isntance names for the clkgen models
                                                           // we want to use.

// This example shows a random clock setup
//
extend DEFAULT'config_mode fullchip_tb_clkgen csco_clkgen_config_s {
  keep (period in [9_900..10_100]);                        // Clk period is now random in ps
  keep (duty_cycle.reset_soft());                          // Reset the soft constraint of 50% to randomize duty cycle
  keep (duty_cycle in [41..59]);                           // Duty cycle specified in percent (Randomize Now)
  keep (skew.reset_soft());                                // Clk skew specified in pico seconds (Randome Skew)
  keep (inv_phase.reset_soft());                           // Invert Phase of Clock 1=TRUE 0=FALSE (Random Phase)
  keep (jitter.reset_soft());                              // Remove soft constraints on Jitter so we can randomize
  keep (jitter     in [100..500]);                         // Clock Jitter in pico seconds
  keep (posedge_enable == TRUE);                           // Turn on the Specman Event for this instance
};                                                         // End Extend


// We have two clocks in the first UNIT TB example and 1 in the
// 2nd unit tb example.  For full chip sim we still need one of those
// clocks to simulate an external input
// but the other clock is now provided by some signal in the chip
// so we need to put that clock driver in passive mode.  In addition
// rather than refer to the HDL signal the new top level clock driver
// provides we can be more efficient by reffering to the specific
// specman event in the new clock driver we need. In order to do this
// we add a pointer to the new top level clock gen so we can
// use its clock event in the passive clock driver.
//

extend cpu_clk_gen csco_clkgen_u {
  fullchip_tb_clkgen : csco_clkgen_u;                      // Add pointer to new clock gen instance which provides clock event

  event posedge is only @fullchip_tb_clkgen.posedge;       // Overload event definition to get clock from clkgen model
};                                                         // End Extend

extend cpu_clk_gen csco_clkgen_config_s {
  keep locale == PASSIVE;                                  // put unit level TB#1  clock gen in passive mode
  keep use_hdl_clk == FALSE;                               // tell passive clock driver to not try and find
                                                           // and HDL clock signal for this passive clkgen
                                                           // we will overload clock event natively in specman
                                                           // preffered approach if possible
};

extend pci_clk_gen csco_clkgen_config_s {                  // put unit level TB#2  clock gen in passive mode
  keep (locale == PASSIVE);                                // tell passive clock driver to not try and find
  keep (use_hdl_clk == TRUE);                              // In this case we are using the hdl clock signal
                                                           // with the passive clock gen. This is NOT the preffered approach
  keep (clk_sig == "~/specman.fullchip_tb_clkgen_INST_0");
};                                                         // End Extend


extend csco_clkgen_unit_tb_config_inst_t : [INST_1];       // Add a new instance because we have have two instances of the unit TB1
extend csco_clkgen_config_inst_t : [INST_1];               // Add a new instance because we have have two instances of the unit TB1


extend csco_clkgen_u {
  short_name() : string is only {                          // Change the message format to accomodate the instance names
    result = appendf("%s_%s", name,config.inst_name);
  };
};                                                         // End Extend

extend csco_clkgen_unit_tb_env {
  short_name() : string is only {                          // Change the message format to accomodate the instance names
    result = appendf("CLKGEN_UNIT_TB_%s", config.inst_name);
  };
};                                                         // End Extend


unit csco_clkgen_fullchip_tb_env like any_env {};          // See ERM
                                                           // Create a new config kind (instance name) for the unit TB's config block

// Extend the unit tb config block and add the two clock gen config blocks
// the combination of the below instances and two constraints below actually
// relocate the config block here from the clkgen modules themselves.  When
// this unit testbench is instantiated in the full chip level testbench its
// config block will be placed in the full chips config block and it will
// actually move up.  In this system the acutal config block belongs at the
// root of the config tree and all child references are actually pointers.
// We will use the value() constraint to control the generation order to
// force the REAL config block to the root of the tree and all other
// references will be pointers...  See config block doc for more details.
//

DEFINE_CFG_STRUCT csco_clkgen_fullchip_tb csco_clkgen_fullchip_tb_env;     // Use the Macro to define the config struct

ADD_CHILD_CFG (csco_clkgen_fullchip_tb_config_s) unit_tb_0 (INST_0 csco_clkgen_unit_tb_config_s);
ADD_CHILD_CFG (csco_clkgen_fullchip_tb_config_s) unit_tb_1 (INST_1 csco_clkgen_unit_tb_config_s);
ADD_CHILD_CFG (csco_clkgen_fullchip_tb_config_s) unit_tb_2 (csco_clkgen_unit_tb2_config_s);

extend INST_0 csco_clkgen_unit_tb_config_s {
  keep core_clk_config.inst_name == INST_0;
};

extend INST_1 csco_clkgen_unit_tb_config_s {
  keep core_clk_config.inst_name == INST_1;
};                                                         // End Extend

extend csco_clkgen_fullchip_tb_config_s  {

    keep unit_tb_0.inst_name == INST_0;
    keep unit_tb_0.cpu_clk_config.inst_name == INST_0;
    keep unit_tb_0.core_clk_config.inst_name == INST_0;

    keep unit_tb_1.inst_name == INST_1;
    keep unit_tb_1.cpu_clk_config.inst_name == INST_1;
    keep unit_tb_1.core_clk_config.inst_name == INST_1;



  fullchip_tb_clk_config : fullchip_tb_clkgen csco_clkgen_config_s;// Add our new Top level Clock Driver

};                                                         // End Extend


// There are two methods to hook a specman eVC to a clock generator.  First Method is used if the
// specman model expects the user to define an HDL path string to the HDL clock signal.  This method
// is demonstrated with sample_evc1.  The 2nd method is used if the specman model provides a hook/empty
// event that the user of the model is to overload.  In this case if we wish to make the code portable
// we need to setup a pointer to the clkgen model instance that is providing the clock to the eVC.
// if we were to use a hard hierarchial instance name here when we re-use the unit level TB in
// the fullchip TB or any other it will break.  The followwing technique demonstrates the portable way to
// do this. Method 2 may be considered the preferred method because it is more efficient.  Using method1
// each occurence and instance of the clockevent will cause a Verilog PLI Call.  With method2 we do 1
// Verilog PLI call per clock event regardless of the number of instances.  PLI calls are expensive so
// eliminating redundant calls are always a good idea.
//
//


//
// Here's the Actual testbench env unit declaration
//

extend csco_clkgen_fullchip_tb_env {                       // See ERM

  //short_name() : string is only {                        // This is implied so I eliminate it for brevity
  //  result = appendf("FULLCHIP_TB");                     // aka there is only 1 top so why print it
  //};


  //
  // Instance Models and BFM's here.....
  //

  fullchip_tb_clkgen : fullchip_tb_clkgen csco_clkgen_u is instance;       // Instance the Core Clock Generator
    keep fullchip_tb_clkgen.config == value(config.fullchip_tb_clk_config);// Here we force the constraint to be unidirectional
                                                           // This causes the instance in the TB not the CLK gen
                                                           // to be generated first and contain the actual
                                                           // config struct. the config instance in the clock gen
                                                           // model is a pointer to the real instance in the TB.

  unit_tb_0 : csco_clkgen_unit_tb_env is instance;         // First instance of the example #1 unit TB
    keep unit_tb_0.config == value(config.unit_tb_0);      // Here we force the constraint to be unidirectional
   keep unit_tb_0.hdl_path() == "fubar";

  unit_tb_1 : csco_clkgen_unit_tb_env is instance;         // Another instance of the example #1  unit TB
    keep unit_tb_1.config == value(config.unit_tb_1);      // Here we force the constraint to be unidirectional
   keep unit_tb_1.hdl_path() == "fubar2";

  unit_tb_2 : csco_clkgen_unit_tb2_env is instance;        // First instance of the example #2  unit TB
    keep unit_tb_2.config == value(config.unit_tb_2);      // Here we force the constraint to be unidirectional
   keep unit_tb_2.hdl_path() == "fubar3";


  // Constrain the pointers in the unit TB we added to point to the new top level clock driver unit
  keep unit_tb_0.cpu_clk.fullchip_tb_clkgen == value(fullchip_tb_clkgen);  // Make sure we generate the top first
  keep unit_tb_1.cpu_clk.fullchip_tb_clkgen == value(fullchip_tb_clkgen);  // Make sure we generate the top first

};

extend pci_clk_gen csco_clkgen_u {
  keep hdl_path() == "ipci_clk_gen";                       // set some bogus HDL path. Shows we can override hdl path with clk_sig settings
};                                                         // End Extend

'>
