//-------------------------------------------------------------------------------------------------------------
//
//   Filename : csco_clkgen_unit_dev2_testbench_env.e
//   Language : 'e'
//   Author   : Sean Smith (Cisco)
//
//   General Description: Example Unit level Testbench to illustrate the usage of the csco_clkgen model
//                        in several scenarios.  This model illustrates how to create a unit level TB env
//                        to instantiate eVCs/models, and configure its output and connect it to your system.
//
//   History:
//   01/07/02  SWS  Initial Coding
//
//--------------------------------------------------------------------------------------------------------------

<'

//========================================================================
// UNIT FOR TOP LEVEL TB
//========================================================================

unit csco_clkgen_unit_tb2_env like any_env {               // See ERM

  short_name() : string is also {                          // Change the message format
      result = appendf("CLKGEN_UNIT_TB2");
  };

  //
  // Instance Models and BFM's here.....
  //

  pci_clk : pci_clk_gen csco_clkgen_u is instance;         // Instance the Core Clock Generator
    keep pci_clk.config == value(config.pci_clk_config);   // Here we force the constraint to be unidirectional
                                                           // This causes the instance in the TB not the CLK gen
                                                           // to be generated first and contain the actual
                                                           // config struct. the config instance in the clock gen
                                                           // model is a pointer to the real instance in the TB.



  sample_evc3 : csco_clkgen_sample_evc_3 is instance;      // Instance Sample1 eVC with Clocks Hooked up through HDL
    keep sample_evc3.hdl_path() == "foo";
    keep sample_evc3.clk_sig == value(pci_clk.config.full_clk_sig);// Method 1: Constrain the models HDL signal to the clkgen
                                                                   // models clock output signal. Use full_clk_sig to avoid
                                                                   // hdl_path() problems.

  sample_evc4 : csco_clkgen_sample_evc_4 is instance;      // Instance Sample2 eVC with Clocks hooked up through events
    keep sample_evc4.hdl_path() == "bar";
    keep sample_evc4.clock_gen == value(pci_clk);          // Method2 hook backpointer we added to the sample eVC
                                                           // to the clock gen model it needs the clock from.  The overloading
                                                           // of the event is in the extension where we added the
                                                           // pointer to clkgen directly below this code.

};

extend tb_clock_region_t: [pci_clk_gen];                   // Define 2 new isntance names for the clkgen models
                                                           // we want to use.

// First Example shows a directed clock setup
// All time units to clock gen model are specified in ps
//

extend DEFAULT'config_mode pci_clk_gen csco_clkgen_config_s {
  keep (period == 15_000);                                 // Clk period specified in pico seconds
  keep (duty_cycle == 50);                                 // Duty cycle specified in percent
  keep (skew       == 0);                                  // Clk skew specified in pico seconds
  keep (inv_phase  == 0);                                  // Invert Phase of Clock 1=TRUE 0=FALSE
  keep (jitter     == 0);                                  // Clock Jitter in pico seconds
  keep (posedge_enable == TRUE);                           // Turn on the Specman Event for this instance
  keep (auto_hdl_path == FALSE);
};                                                         // End Extend


DEFINE_CFG_STRUCT csco_clkgen_unit_tb2 csco_clkgen_unit_tb2_env;       // Use the Macro to define the config struct

ADD_CHILD_CFG (csco_clkgen_unit_tb2_config_s) pci_clk_config (pci_clk_gen csco_clkgen_config_s);


// There are two methods to hook a specman eVC to a clock generator.  First Method is used if the
// specman model expects the user to define an HDL path string to the HDL clock signal.  This method
// is demonstrated with sample_evc1.  The 2nd method is used if the specman model provides a hook/empty
// event that the user of the model is to overload.  In this case if we wish to make the code portable
// we need to setup a pointer to the clkgen model instance that is providing the clock to the eVC.
// if we were to use a hard hierarchial instance name here when we re-use the unit level TB in
// the fullchip TB or any other it will break.  The followwing technique demonstrates the portable way to
// do this. Method 2 may be considered the preferred method because it is more efficient.  Using method1
// each occurence and instance of the clockevent will cause a Verilog PLI Call.  With method2 we do 1
// Verilog PLI call per clock event regardless of the number of instances.  PLI calls are expensive so
// eliminating redundant calls are always a good idea.
//
//

extend csco_clkgen_sample_evc_4 {
  clock_gen : csco_clkgen_u;                               // Extend Sample_evc2 and add a pointer to the clock gen
                                                           // instance that we need to hook it to.

  event clk_rise is only @clock_gen.posedge;               // Overload event definition to get clock from clkgen model
                                                           // In this example we use event posedge. Also supported are
                                                           // negedge and anyedge.

};                                                         // End Extend



//
// Here's the Actual testbench env unit declaration
//


'>
