//=============================================================================
//
//   Program  : csco_clkgen_simple.v
//
//   Purpose  : Generate a single clock with programmable freq, duty cycle,
//              phase, skew offset, and jitter
//
//   Author   : sesmith, spadix
//   Created  : 02/28/97
//
//   Revision History
//
//   Date     Author  Comments
//   ======== ======= =========================================================
//   02/28/97 SWS     Initial version of basic clock driver model for ISL
//   12/12/97 KMK     Stole from Sean and altered for Toaster
//   05/24/01 spadix  updated to work better w/blackwater and specman
//                    also added jitter parameter
//   01/31/03 SWS     Started upgrades for more integration with specman
//                    and changed names and structure to comply with ERM
//                    packaging and naming requireements.
//
//=============================================================================

`timescale 1ns / 1ps

module csco_clkgen_simple(clk, _en);
  parameter STOPPED = 0,
            STARTING = 1,
            RUNNING = 2;

  output     clk;    // clock output
  input      _en;    // clock output enable

  // all timing in units of ps!!
  // (actually 1/1000 of timescale)
  // these can be set directly from specman
  reg [31:0] PERIOD;

  // [1..99] percentage of period that clock will be high
  integer    DUTYCYCLE;

  // direction of first edge transition
  reg        INV_PHASE;

  // time from (posedge run) to first edge transition
  integer    SKEW;

  // here we define JITTER to be the size of the window
  // in which the clock is allowed to transition
  // this window will _start_ from SKEW, so if you want
  // it centered you better do a SKEW -= JITTER / 2
  reg [31:0] JITTER;

  // set this reg after the clock is programmed
  // the first edge will be SKEWed and started from there
  reg        run;
  integer    state;

  reg        clk;
  real       period;
  real       posclk;
  real       negclk;
  real       skew;
  real       jitter;

  always @(posedge run)
    state = STARTING;

  initial begin
    clk = 1'bz;
    state = STOPPED;

    forever begin
      while(state != STARTING)
        @(state);

      clk <= INV_PHASE;

      if(PERIOD == 0)
        state = STOPPED;
      else begin
        state = RUNNING;
        period = PERIOD / 1000.0;
        skew   = SKEW   / 1000.0;

        posclk = (period * DUTYCYCLE / 100.0);
        negclk = period - posclk;

        $display("%t (%m) clkgen: period %3.3fns (+%3.3f, -%3.3f), duty cycle %0d%%",
                 $time, period, posclk, negclk, DUTYCYCLE);
        $display("%t (%m)         skew %3.3fns, jitter %3.3fns, inv_phase %b",
                 $time, skew, (JITTER / 1000.0), INV_PHASE);

        #(skew);  // skew start of clock

        while(state == RUNNING) begin
          jitter =  ($random % JITTER) / 1000.0;
          clk <= #(jitter) ((_en === 1'b0) ? ~INV_PHASE : 1'bz);
          #(posclk);
          jitter = ($random % JITTER) / 1000.0;
          clk <= #(jitter) ((_en === 1'b0) ? INV_PHASE : 1'bz);
          #(negclk);
        end
      end
    end
  end

endmodule
