//============================================================================
//============================================================================
//
// File name    : csco_config_main.e
//              :
// Title        : Macro Definitions to Implement Staic Config Tree
// Project      : Shaggy ASIC
//              :
// Created On   : 12/24/2002
//              :
//              :
// Purpose      :
//              :
// Description  :
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
//============================================================================
//============================================================================
//Copyright 2003 (c) Cisco Systems Inc.
//
//============================================================================
//============================================================================
//02/26/2003  JHZ  Add code to handle run-time regeneration
//11/22/2005  MDS  Added INSTANCE_CFG macro (obsoletes ADD_CHILD_CFG macro)


<'
package csco_config;


extend any_unit {
  // Utility method to set the config struct of the enclosing unit.
  csco_config_set_cfg(new_val : csco_config_base_struct_s) is undefined;

  // To be extended with assignments from config fields to copies of the 
  // fields in the parent unit, to be used as when-determinants.
  csco_config_update_subtypes() is empty;
};

struct csco_config_internals_s {
  // This list contains all the units which reference this config struct instance.
  // This list is used during run-time regeneration to re-align the config fields
  // of the associated units to point to the newly generated config struct instances.
  !associated_units : list of any_unit;

  !parent_cfg : csco_config_base_struct_s;
    attribute parent_cfg deep_copy=ignore;
};

struct csco_config_base_struct_s {
  csco_config_guts : csco_config_internals_s ;
    csco_print_attr csco_config_guts ALL disable;

  // This method takes a pointer to the old config struct instance (the one being
  // replaced by the regeneration event) and does the following:
  // 
  // 1) goes through its list of associated units and resets the config reference field 
  //    of each unit to the newly generated config instance.
  //
  // 2) call set_unit() of the newly generated instance to set its parent unit pointer 
  //    to that of the old config struct instance. 
  //
  // 3) copies the associated_units list of the old config instance to the new one.
  realign_unit_cfg_references(old_cfg : csco_config_base_struct_s) is {
    var parent_unit := old_cfg.get_enclosing_unit(any_unit);
    set_unit(parent_unit);

    for each (u) in old_cfg.csco_config_guts.associated_units {
      u.csco_config_set_cfg(me);
      u.csco_config_update_subtypes();
    };

    csco_config_guts.associated_units = old_cfg.csco_config_guts.associated_units;
  };
}; // struct csco_config_base_struct_s


-- 
define <DEFINE_CFG_STRUCT_ERROR'statement> "DEFINE_CFG_STRUCT <junk'any>" as {
    extend sys {
      init() is also {
       error("Syntax error ",source_location()," with DEFINE_CFG_STRUCT.  Correct usage is\n  DEFINE_CFG_STRUCT begin\n    config_prefix= ___,\n    parent_unit_type=___\n  end");
       };
    };
};
define <DEFINE_CFG_STRUCT'statement> 
"DEFINE_CFG_STRUCT begin [ ][config_instance_name[ ]=[ ]<inst_name'name>[ ],][ ]config_prefix[ ]=[ ]<cfg_name_prefix'name>[ ],[ ]parent_unit_type[ ]=[ ]<parent_unit'Type> [ ]end" as {
    type <cfg_name_prefix'name>_config_mode_t : [DEFAULT];
    type <cfg_name_prefix'name>_config_inst_t : [INST_0];

    struct <cfg_name_prefix'name>_config_s like csco_config_base_struct_s {
      // 'config_mode' is used to encapsulate a set of constraints that
      // places the unit/units into a specific configuration. It also allows us
      // to build a library of config modes such that testbench integrators and
      // testcases can simply choose a configuratiohn by constrain the
      // 'config_mode' field as opposed to individually constraining all the
      // config fields.  e.g.,
      //
      // extend pci_config_mode_t : [PCI_MODE_1];
      //
      // extend PCI_MODE_1 PCI csco_config_s {
      //     keep speculation == TRUE;
      //     keep host_bridge == FALSE;
      //     <etc...>
      // };
      //
      // then from a testcase or the tb instantiation file:
      //
      // extend PCI csco_config_s {
      //     keep config_mode == PCI_MODE_1;
      // };
      //
      config_mode : <cfg_name_prefix'name>_config_mode_t;
        keep soft config_mode == DEFAULT;

      // Pointer to the enclosing parent unit. This is useful for accessing
      // internal fields of the parent unit when defining method hooks and
      // events from within the config struct.
      !parent_unit : <parent_unit'Type>; 

      realign_unit_cfg_references(old_cfg : csco_config_base_struct_s) is also {
        // make sure the regenerated config struct instance has the proper 
        // back pointer to the enclosing parent unit.
        parent_unit = old_cfg.as_a(<cfg_name_prefix'name>_config_s).parent_unit;
      };

      pre_generate() is also {
        #ifdef CSCO_CONFIG_DEBUG then {
          out("generating CONFIG <cfg_name_prefix'name>");
        };
      };

      run() is also {
        if parent_unit == NULL {
          error("config <cfg_name_prefix'name>_config_s has been instanced in the config tree (probably with INSTANCE_CFG), but its parent unit <parent_unit'Type> has not been instanced in the unit tree (i.e. starting with sys) ");
        };
      };
    }; -- struct <cfg_name_prefix'name>_config_s like csco_config_base_struct_s

    // If a config struct is also shared by child units, then the child units
    // must reference the config struct through the parent unit.
    extend <parent_unit'Type> {
      !<inst_name'name|config> : <cfg_name_prefix'name>_config_s;

      csco_config_get_cfg() : <cfg_name_prefix'name>_config_s is {
        result = <inst_name'name|config> ;
      };

      csco_config_set_cfg(new_val : csco_config_base_struct_s) is {
        <inst_name'name|config> = new_val.as_a(<cfg_name_prefix'name>_config_s);
      };

      // This method is called during pre_generate() to set the config struct of the
      // enclosing unit. If the unit holds the root of the config tree, then the 
      // config struct will be generated (default implementation here). Otherwise,
      // the config struct will be set to the corresponding node on the parent unit's
      // config tree (default behavior will be overwritten in INSTANCE_CFG and 
      // INSTANCE_CFG_LIST macros).
      csco_config_generate() is {
        #ifdef CSCO_CONFIG_DEBUG then {
          out("generating top of config tree -- ", e_path(), ".<cfg_name_prefix'name> ") ;
        };
        gen <inst_name'name|config> ;
      };

      pre_generate() is also { 
        #ifdef CSCO_CONFIG_DEBUG then {
          out("generating ", e_path()) ;
        };
        csco_config_generate() ;
        csco_config_update_subtypes();

        // Set parent back pointer in config struct to this unit.
        <inst_name'name|config>.parent_unit = me;

        // Indicate the unit association so that message outputs from the config block
        // will indicate the correct unit path, even after the physical instance of the
        // config block has been propagated upwards.
        <inst_name'name|config>.set_unit(me);
        <inst_name'name|config>.csco_config_guts.associated_units.add(me);
      };
    };

    extend sys {
      init() is also {
        #ifdef CSCO_CONFIG_DEBUG then {
          out("DEFINE_CFG_STRUCT expanded into ") ;
          out("   extend <parent_unit'Type> {") ;
          out("      !<inst_name'name|config> : <cfg_name_prefix'name>_config_s;");
          out("   };\n\n");
        };
      };
    };

}; -- define <DEFINE_CFG_STRUCT'statement> 
'>

-- This macro performs the necessary operations to promote the config struct of a child unit
-- into the config struct of the parent unit and makes sure that the config field of the 
-- child unit is a reference pointer to its corresponding config instance in the parent unit's
-- config struct.  
 
The correct usage is
  INSTANCE_CFG begin
    config_instance_name= ____,     -- instance name of the child within the parent
    child_config_type= ____,   -- type of the child config, e.g. XXX evc_config_s
    parent_config_type= ____, -- type of the parent config, e.g. XXX csco_tb_config_s
    parent_unit_type= ____, -- type of the unit that holds the parent config, e.g. XXX csco_tb_u
    child_unit_type= ___  -- type of the unit that holds the child config 
  end

<'
define <INSTANCE_CFG_ERROR'statement> "INSTANCE_CFG <junk'any>" as {
    extend sys {
      init() is also {
       error("Syntax error ",source_location()," with INSTANCE_CFG.  Correct usage is\n  INSTANCE_CFG begin\n    config_instance_name= ___,\n    child_config_type=___,\n    parent_config_type=___,\n    parent_unit_type=___,\n    child_unit_type=___\n  end");
       };
    };
};

define <INSTANCE_CFG'statement> 
"INSTANCE_CFG begin [ ]config_instance_name[ ]=[ ]<child_cfg'name>[ ],[ ]child_config_type[ ]=[ ]<child_cfg'Type>[ ],[ ]parent_config_type[ ]=[ ]<parent_config'Type>[ ],[ ]parent_unit_type[ ]=[ ]<parent_unit'Type>[ ],[ ]child_unit_type[ ]=[ ]<child_unit'Type> [ ]end" as {

    extend <parent_config'Type> {
      <child_cfg'name> : <child_cfg'Type>;
        csco_print_attr <child_cfg'name> LONG expand=ON;

      post_generate() is also {
        if <child_cfg'name> != NULL {
          <child_cfg'name>.csco_config_guts.parent_cfg = me;
        };
      };

      realign_unit_cfg_references(old_cfg : csco_config_base_struct_s) is also {
        <child_cfg'name>.realign_unit_cfg_references(old_cfg.as_a(<parent_config'Type>).<child_cfg'name>);
      };
    }; -- extend <parent_config'Type>

    -- Make sure 'config' field of child unit becomes a pointer reference to 
    -- the corresponding config instance in the parent unit's config struct.
    -- Type casting is needed in the constraint because the caller of the macro may
    -- specify a sub-type of the base parent config struct.
    extend <child_unit'Type> {
      csco_config_generate() is only { 
        var parent_unit := try_enclosing_unit(<parent_unit'Type>) ;
        assert that parent_unit != NULL else
          error("CSCO_CONFIG_ERR_NO_PARENT_UNIT_TYPE:\n",
                "<parent_unit'Type> is not a parent unit of ", e_path(), 
                " make sure INSTANCE_CFG for <<child_cfg'Type> is using the correct parent unit type");

        var child_config := parent_unit.csco_config_get_cfg().as_a(<parent_config'Type>).<child_cfg'name>;
        csco_config_set_cfg(child_config) ;
        #ifdef CSCO_CONFIG_DEBUG then {
          out(e_path(), "'s config struct has been assigned to ", child_config) ;
        };
      };

      post_generate() is also {
        var parent_unit := try_enclosing_unit(<parent_unit'Type>) ;
        if parent_unit != NULL {
          var child_config := parent_unit.csco_config_get_cfg().as_a(<parent_config'Type>).<child_cfg'name>;
          assert that csco_config_get_cfg() == child_config else
            error("CSCO_CONFIG_ERR_WRONG_CONFIG_ASSIGN:\n", 
                    e_path(), "'s config struct (, ", csco_config_get_cfg(), ") != corresponding node (", 
                    child_config, ") in config tree!\n",
                    "This is most likely caused by multiple calls to the INSTANCE_CFG macro using the same child\n",
                    " unit type in each call. The solution is to replace the multiple INSTANCE_CFG macros calls\n",
                    " with a single call to INSTACE_CFG_LIST. Or, uniquify the child unit type argument of the\n",
                    " multiple INSTANCE_CFG macro calls, possibly using unique subtypes.");
        };
      };
    };
    extend sys {
      init() is also {
        #ifdef CSCO_CONFIG_DEBUG then {
          out("INSTANCE_CFG expanded into") ;
          out("   extend <parent_config'Type> {") ;
          out("      <child_cfg'name> : <child_cfg'Type>;");
          out("   };\n\n");
        };
      };
    };

}; -- define <INSTANCE_CFG'statement> 
'>

<'
define <INSTANCE_CFG_LIST_ERROR'statement> 
"INSTANCE_CFG_LIST [ ]<junk'any>" as {
    extend sys {
      init() is also {
        error("Syntax error ",source_location()," with INSTANCE_CFG_LIST.  Correct usage is\n  INSTANCE_CFG_LIST begin\n    config_instance_name= ___,\n    child_config_type=___,\n    parent_config_type=___,\n    child_unit_path=___,\n    child_unit_type=___,\n    parent_unit_type=___\n  end");
      };
    };
};

define <INSTANCE_CFG_LIST'statement> 
"INSTANCE_CFG_LIST[ begin [ ]config_instance_name[ ]=[ ]<child_cfg'name>[ ],[ ]child_config_type[ ]=[ ]<child_cfg'Type>[ ],[ ]parent_config_type[ ]=[ ]<parent_cfg'Type>[ ],[ ]parent_unit_type[ ]=[ ]<parent_unit'Type>[ ],[ ]child_unit_type[ ]=[ ]<child_unit'Type>] [ ]end" as{
    extend <parent_cfg'Type> {
      <child_cfg'name> : list of <child_cfg'Type>;
        csco_print_attr <child_cfg'name> LONG expand=ON;

      !<child_cfg'name>_list_ctr : uint ;

      post_generate() is also {
        for each (c) in <child_cfg'name> {
          if c != NULL {
            c.csco_config_guts.parent_cfg = me;
          };
        };
      };

      realign_unit_cfg_references(old_cfg : csco_config_base_struct_s) is also {
        for each (c) in <child_cfg'name> {
            c.realign_unit_cfg_references(old_cfg.as_a(<parent_cfg'Type>).<child_cfg'name>[index]);
        };
      };
    }; -- extend <parent_config'Type>

    extend <parent_unit'Type> {
      post_generate() is also {
        if config.as_a(<parent_cfg'Type>).<child_cfg'name>.size() != config.as_a(<parent_cfg'Type>).<child_cfg'name>_list_ctr {
           error("Testbench Configuration Error: A child config list, <child_cfg'name>, has been added to <parent_cfg'Type> but the size of the corresponding child unit list of <child_unit'Type> under <parent_unit'Type> is ",dec(config.as_a(<parent_cfg'Type>).<child_cfg'name>_list_ctr)," while it should be constrained to the size of this config list, ",dec(config.as_a(<parent_cfg'Type>).<child_cfg'name>.size()));
        };
      };
    };

    -- Make sure 'config' field of child unit becomes a pointer reference to 
    -- the corresponding config instance in the parent unit's config struct.
    extend <child_unit'Type> {
      csco_config_generate() is only { 
        var parent_unit := try_enclosing_unit(<parent_unit'Type>) ;
        if parent_unit != NULL {
          var parent_cfg := parent_unit.csco_config_get_cfg().as_a(<parent_cfg'Type>);
          if parent_cfg.<child_cfg'name>_list_ctr == parent_cfg.<child_cfg'name>.size() {
            error("Testbench Configuration Error: A child config list, <child_cfg'name>, has been added to <parent_cfg'Type> but the size of the corresponding child unit list of <child_unit'Type> under <parent_unit'Type> has not been constrained to equal the size of this config list, which is ",dec(parent_cfg.<child_cfg'name>.size()));
          };
          csco_config_set_cfg(parent_cfg.<child_cfg'name>[parent_cfg.<child_cfg'name>_list_ctr]) ;
          #ifdef CSCO_CONFIG_DEBUG then {
            out(e_path(), "'s config struct has been assigned to ", parent_cfg.<child_cfg'name>[parent_cfg.<child_cfg'name>_list_ctr]) ;
          };
          parent_cfg.<child_cfg'name>_list_ctr += 1 ;
        };
      };
    };

    extend sys {
      init() is also {
        #ifdef CSCO_CONFIG_DEBUG then {
          out("INSTANCE_CFG_LIST expanded into") ;
          out("   extend <parent_cfg'Type> {") ;
          out("      <child_cfg'name> : list of <child_cfg'Type>;");
          out("   };\n\n");
        };
      };
    };

}; -- define <INSTANCE_CFG_LIST'statement>


define <USE_PARENT_CFG_ERROR'statement> "USE_PARENT_CFG <junk'any>" as {
    extend sys {
      init() is also {
       error("Syntax error ",source_location()," with USE_PARENT_CFG.  Correct usage is\n  USE_PARENT_CFG begin\n    parent_config_type=___,\n    parent_unit_type=___,\n    child_unit_type=___  end");
       };
    };
};

-- This macro provides the necessary logic to create a config struct inside a child unit that references the config
-- struct of its parent unit. This macro should be used on units who don't have their own config struct type.
--
define <USE_PARENT_CFG'struct_member> 
"USE_PARENT_CFG begin [ ]parent_config_type[ ]=[ ]<parent_config'Type>[ ],[ ]parent_unit_type[ ]=[ ]<parent_unit'Type>[[ ],[ ]config_instance_name[ ]=[ ]<child_cfg'name>] [ ]end" as {
    !<child_cfg'name|config> : <parent_config'Type>;

    csco_config_set_cfg(new_val : csco_config_base_struct_s)  is {
      config = new_val.as_a(<parent_config'Type>);
    };

    pre_generate() is also {
      config = get_enclosing_unit(<parent_unit'Type>).csco_config_get_cfg().as_a(<parent_config'Type>) ;
      csco_config_update_subtypes() ;
      config.csco_config_guts.associated_units.add(me);
    };
}; -- define <USE_PARENT_CFG'statement> 




-- This macro is used to create a sub-typed version of the config struct for a unit. This is handy if there
-- are lots of internal code in the unit that need to access config fields of the particular config struct
-- sub-type. By creating a new config struct reference of the corresponding sub-type, the user won't need
-- to use 'as_a' everywhere in the internal code.
define <COPY_CFG_AS_SUBTYPE'struct_member> 
"COPY_CFG_AS_SUBTYPE begin [ ]orig_instance_name[ ]=[ ]<orig_name'name>[ ],[ ]copy_instance_name[ ]=[ ]<copy_name'name>[ ],[ ]copy_type[ ]=[ ]<copy_type'Type> [ ]end" as {
    !<copy_name'name> : <copy_type'Type>;

    csco_config_update_subtypes() is first {
      <copy_name'name> = <orig_name'name>.as_a(<copy_type'Type>);
    };
}; -- "COPY_CFG_AS_SUBTYPE \(<unit'Type>\) <cfg'name> <cfg_sub'Type>"


-- Regeneration of a node should be sourced from the root of othe config tree so that all
-- constraints on the node are picked up.  
define <REGEN_CFG_BRANCH'action>
"REGEN_CFG_BRANCH <full_path_starting_at_sys'exp>" as {
    var reset_cfg_branch_str_list := str_split("<full_path_starting_at_sys'exp>", ".");

    if reset_cfg_branch_str_list[0] != "sys" {
      error("input argument to REGEN_CFG_BRANCH must start with 'sys.', got: ", "'", reset_cfg_branch_str_list[0], "'");
    }
    else {
      messagef(LOW, "regenerating config sub-tree: path=%s, instance=%s\n", "<full_path_starting_at_sys'exp>", <full_path_starting_at_sys'exp>);

      // Remeber node about to be regenerated
      var tmp_cfg := <full_path_starting_at_sys'exp>;

      // Regenerated specified config branch in top-down fashion
      gen <full_path_starting_at_sys'exp>;

      // Realign config pointers of all the units that reference the regenerated config branch.
      <full_path_starting_at_sys'exp>.realign_unit_cfg_references(tmp_cfg);

      messagef(MEDIUM, "regenerated config node instance = %s\n", <full_path_starting_at_sys'exp>);
    };
}; -- "REGEN_CFG_BRANCH <full_path_starting_at_sys'exp>" as


'>
