//------------------------------------------------------------------------
// Title        : csco_event package Env.
// Project      : csco_event
//              :
// Created On   : 01/10/2006
//              :
//              :
// Description  : 
//              :
//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Copyright 2006 (c) Cisco Systems Inc.
//------------------------------------------------------------------------
<'
package csco_event;

// Event object corresponds to a leaf or summary event.
struct csco_event_s {
  // field id
  !name : string; 
  // event types
  !types : list of csco_event_name_t;
  // blockname fixme remove?
  !blockname : string;  
  // typename used when moving between blocks in the addrmap
  !typename : string;
  // long description (for messaging)
  !longdesc : string;  
  // is a summary bit
  !is_summary : list of bool;   
  // is enabled for propagation
  !is_enabled : list of bool;  
  // is it a part of the first event capture
  !is_first : list of bool;
  // severity of event
  !severity : string; 
  // action to be taken on occurence
  !action : string;   
  //Min number of events expected
  !min_count : list of uint;       
  //Max number of events expected
  !max_count : list of uint;    
  //Actual number of events detected 
  //during the simulation
  !actual_count : list of uint;  
  //first index in event type. 
  //Needed for event type list size > list size in instance
  !first_idx : uint;
  //wait flag for use in autotest
  !wait_on_detect: bool;
  //wait flag for use in autotest
  !wait_on_clear: bool;
  //back pointer to parent in the tree
  !parent: csco_event_s;
  //pointers to leaf nodes
  !children: list (key: name) of csco_event_s;

  //debug dump method called by dump_tree()
  debug_dump(ind: string, event_type: csco_event_name_t) is {
   
    var idx:= event_type.as_a(uint) - first_idx;
    var clrd_string: string;

    if (!is_summary[idx]) {
      clrd_string = (is_enabled[idx])?
      vt.text_style(DARK_GREEN, 
        appendf( "%s+ %-14.13s max %4.3d min %4.3d actual %4.3d",
        ind, name, max_count[idx], min_count[idx], actual_count[idx])):
      vt.text_style(RED, 
        appendf( "%s- %-14.13s max %4.3d min %4.3d actual %4.3d", 
        ind, name, max_count[idx], min_count[idx], actual_count[idx]));
    } else {
      //No counts needed if summary bit    
      clrd_string = (is_enabled[idx])?
        vt.text_style(DARK_GREEN, 
          appendf( "%s+ %-14.13s",ind, name)):
        vt.text_style(RED, 
          appendf( "%s- %-14.13s", ind, name));
    };
      
    messagef(CSCO_EVENT, LOW, append(ind, clrd_string));
    if (!children.is_empty()) {
      ind = append(ind, "  ");
      for each (child) in children {
        child.debug_dump(ind, event_type);
      };
    };
  };

  //Debug print method to dump event tree along with expect and
  //actual counts
  dump_tree() is {
    messagef(CSCO_EVENT, LOW, 
      (vt.text_style(DARK_BLUE, 
        appendf( "\nEvent Hierarchy (+ = ENABLED, - = DISABLED):"))));
    for each (t) in types {
      messagef(CSCO_EVENT, LOW, 
        (vt.text_style(DARK_BLUE, appendf( "\n%s Tree:\n", t))));
      for each (child) in children {
        child.debug_dump(" ", t);
      };
    };
  };

  //Check if event count exceeded maximum allowed
  check_max(event_type: csco_event_name_t) is {
    if (actual_count[event_type.as_a(uint)-first_idx] > max_count[event_type.as_a(uint)-first_idx]) {
      dut_error("ERR_CSCO_EVENT_MAX_MISMATCH\n",
                "Register ", parent.name,
                " Event ", event_type, " ", name, "\n") {
        outf("Maximum = %d :: Actual = %d\n",
             max_count[event_type.as_a(uint)-first_idx], actual_count[event_type.as_a(uint)-first_idx]);
        outf(" Description %s\n", longdesc);
      };
    };
  };

  //Check if event count is below minimum allowed
  check_min(event_type: csco_event_name_t) is {
    if (actual_count[event_type.as_a(uint)-first_idx] < min_count[event_type.as_a(uint)-first_idx]) {
      dut_error("ERR_CSCO_EVENT_MIN_MISMATCH\n",
                "Register ", parent.name,
                " Event ", event_type, " ", name, "\n") {
        outf("Minimum = %d :: Actual = %d\n",
             min_count[event_type.as_a(uint)-first_idx], actual_count[event_type.as_a(uint)-first_idx]);
        outf("Description %s\n", longdesc);
      };
    };
  };

  //Update actual count and check if it has exceeded max allowed
  update_count(event_type: csco_event_name_t) is {
    //update actual count
    actual_count[event_type.as_a(uint)-first_idx] += 1;
    check_max(event_type);
    //collect coverage
    ev_type = event_type;
    error_test = !auto_test;
    emit detect;
    //turn off auto_test
    auto_test = FALSE;
  };

  //End of test checks
  check() is also {
    for each (t) in types {
      if (is_summary.has(it==FALSE)) {
        check_min(t);
        check_max(t);
      };
    };
  };

  //API methods
  
  //Method to update enable bit
  //name = path.bitname, Also allow =>
  //*.* OR *.foo OR *.*.bar OR *.foo.*
  //enable = enable(true)/disable(false), 
  enable(pathname: string, enable: csco_event_enable_t, event_type: csco_event_name_t) is {

    //Dealing with *.* notation
    if (pathname ~ "/\*/") {
      
      var path_list:= str_split(pathname,".");
      var current_name := path_list[0];
      path_list.delete(0);
      pathname = str_join(path_list, ".");

      for each (c) in children {
        if ((current_name == c.name) or (current_name ~ "/\*/"))  {
          c.enable(pathname, enable, event_type);
        };
      };
        
    //if path name is not empty there and there are no wild cards
    } else if (pathname != "")  {
      var current_scope: csco_event_s = get_child_scope(pathname);
      current_scope.is_enabled[event_type.as_a(uint)-first_idx] = enable.as_a(bool);
    } else if (pathname == "") {
       is_enabled[event_type.as_a(uint)-first_idx] = enable.as_a(bool);  
    } else {
    //if path name is empty there is no identifier. This is the current scope  
      assert(FALSE) else
        error("ERR_CSCO_EVENT_ILLEGAL_PATH\n",
              "Event path incorrect. Nowhere to go from ", name);
    };
  };

  //Method to update all enable bits
  enable_all(enable: csco_event_enable_t, event_type: csco_event_name_t) is {

    for each (c) in children {
      c.enable_all(enable, event_type);
    };
    if (!is_enabled.is_empty()) {
      is_enabled[event_type.as_a(uint)-first_idx] = enable.as_a(bool);
    };
  };

  //Method to update expect count for an event. Name can be a dotted string.
  //min_inc = # to increment min_count
  //max_inc = # to increment max_count
  //event_type = type of the event
  update_exp_count(name: string, 
                   min_inc: uint, 
                   max_inc: uint,
                   event_type: csco_event_name_t) is {
    
    var current_scope :csco_event_s = get_child_scope(name);
   
    //fixme assert not summary bit
    if (will_it_propagate(event_type)) {
      current_scope.min_count[event_type.as_a(uint)-first_idx] += min_inc;
      current_scope.max_count[event_type.as_a(uint)-first_idx] += max_inc;
      wait_on_detect = TRUE;
      wait_on_clear = TRUE;
      messagef(CSCO_EVENT, HIGH, 
        "Updated expect count for %s.\n Min Exp = %d; Max Exp = %d\n",
          name,
          current_scope.min_count[event_type.as_a(uint)-first_idx],
          current_scope.max_count[event_type.as_a(uint)-first_idx]);
    } else {
      messagef(CSCO_EVENT, LOW, "WARNING %s won't  propagate. Cnt not updated",
        name);
    };
  };

  //Method to be called by user to force a wait until clear of an event after
  //update_exp_count() has been called
  wait_on_detect()@sys.any is {
    wait true(!wait_on_detect);
  };
  
  //Method to be called by user to force a wait until clear of an event after
  //update_exp_count() has been called
  wait_on_clear()@sys.any is {
    wait true(!wait_on_clear);
  };

  //Method to move down the scope
  //name = blockname1.blockname2....
  get_child_scope(name: string): csco_event_s is {

    var name_path: list of string = 
      str_split(name, ".");
      
    var current_scope: csco_event_s = me;

    if (current_scope.name == name_path[0]) {
      name_path.delete(0);
    };
    
    //Move scope to block defined in path
      for each (s) in name_path {
        var ev := current_scope.children.key(s);
        if (ev != NULL) {
          current_scope = ev;
        } else {
          outf("\n#Event tree search failed at level %s. Tree nodes available...", s);
          current_scope.dump_tree();
          assert(FALSE) else
            error("ERR_CSCO_EVENT_ILLEGAL_PATH\n",
                  "Event path incorrect ", s, " in ", name);
        };
      };
    return(current_scope);  
  };

  //Method to move up the scope
  //name = blockname
  get_scope_up(name: string): csco_event_s is {

    if (me.name != name) {
      assert (me.parent != NULL) else
        error("ERR_CSCO_EVENT_ILLEGAL_PATH\n",
              "Parent block does not exist: ", name);
    } else {
      return( get_scope_up(name)); 
    };
      
  };
  
  //Returns the addrmap_scope to be used in a register access
  addrmap_scope(): string is {
    if (typename != NULL) {
      //if moving between blocks, the typename identifies the 
      //leaf block register
      return(appendf("%s.%s", blockname, typename));
    };
  };

  //Check if the event will propagate up the tree
  will_it_propagate(event_type: csco_event_name_t): bool is {
    if (parent == NULL) {
      return(TRUE);
    } else {
      return( is_enabled[event_type.as_a(uint)-first_idx] and
        parent.will_it_propagate(event_type));
    };
  };

  //Returns list of field value pairs with all but
  //current events value reversed 
  set_only_relevant(val: bit): list of csco_addrmap_field_value_pair_s is {
  
    var field_value_pairs: list of csco_addrmap_field_value_pair_s;

    for each (e) in parent.children {
      var field_value:= new csco_addrmap_field_value_pair_s with {
        .name = e.name;
        .value = ((e.name == name)? val:~val);
        .value_kind = CONSTANT;
      };
      field_value_pairs.add(field_value);
    };
    return(field_value_pairs);
  };

  set_all(val: bit): list of csco_addrmap_field_value_pair_s is {
  
    var field_value_pairs: list of csco_addrmap_field_value_pair_s;

    for each (e) in parent.children {
      var field_value:= new csco_addrmap_field_value_pair_s with {
        .name = e.name;
        .value = val;
        .value_kind = CONSTANT;
      };
      field_value_pairs.add(field_value);
    };
    return(field_value_pairs);
  };

  //Returns a uint with all but
  //current events value reversed 
  set_only_relevant_phy(val: bit): uint(bits:32) is {
    for each (e) in parent.children {
      result[index:index] = (e.name == name)? val:~val;
    };
  };

  //Returns a uint with all but
  //current events value reversed 
  set_all_phy(val: bit): uint(bits:32) is {
    for each (e) in parent.children {
      result[index:index] = val;
    };
  };


  //Integration setup utility to trim levels from a larger tree
  //keeping only those that match the specified pattern
  trim_all_except(name_re: string) is {
    for each (e) in reverse children {
      if(e.blockname !~ name_re) {
        e.children.clear();
      };
    };
  };

  //Coverage fields

  //cover event
  event detect;
  //event type for coverage
  ev_type: csco_event_name_t;
  //is the event generated by auto-test
  auto_test: bool;
  error_test: bool;

  cover detect is {
    item hier_name: string = append(parent.name, ".", name) using no_collect;
    item ev_type using no_collect;
    item auto_test using ignore = (auto_test == FALSE), no_collect;
    item error_test using ignore = (error_test == FALSE), no_collect;
    cross hier_name, ev_type, auto_test using name = name_x_ev_type_x_auto;
    cross hier_name, ev_type, error_test using name = name_x_ev_type_x_error;
  };
    
};

'>
