//-----------------------------------------------------------------------------
//
//   filename : csco_event_phy_top.e
//   author(s): dsuryan
//
//   description:
//     simple event monitor to ease connection of events to testbench
//
//   history:
//   09/03/06 : dsuryan : copied over from csco_interrupt.
//                        this will be paired with csco_event
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//  Copyright 2002 (c) Cisco Systems Inc.
//----------------------------------------------------------------------------
<'

package csco_event_phy;

import csco_testflow/e/csco_testflow_top;
import csco_config/e/csco_config_top;

//========================================================================
// API definition
//========================================================================

// debug messages
extend message_tag : [ CSCO_EVENT_PHY ];

// logical names of instances
type csco_event_phy_evc_instance_t : [ NONE ];
type csco_event_phy_agent_instance_t: [ NONE ];

type csco_event_phy_sensitivity_t : [
  LEVEL,                                // level sensitive event
  EDGE                                  // EDGE sensitive event
];

type csco_event_phy_polarity_t : [
  NEGATIVE = 0,                         // active low or falling EDGE sensitive
  POSITIVE = 1                          // active high or rising EDGE sensitive
] (bits: 1);

type csco_event_phy_locale_t : [
  PASSIVE,                         // inactive state
  ACTIVE                           // active state
];

type csco_event_phy_drive_t : [
  OFF,                         // inactive state
  ON                           // active state
];


// method port for event processing
method_type csco_event_phy_push_t(name: csco_event_phy_trans_s) @sys.any;


//------------------------------------------------------------
// event monitor environment
// activates appropriate hooks when event occurs

unit csco_event_phy_env_u like any_env {
  evc_instance : csco_event_phy_evc_instance_t;

  short_name() : string is { result = evc_instance.as_a(string) };

};


DEFINE_CFG_STRUCT begin
  config_prefix = csco_event_phy_env,
  parent_unit_type = csco_event_phy_env_u
end;

extend csco_event_phy_env_config_s {
  evc_instance: csco_event_phy_evc_instance_t;
};

struct csco_event_phy_trans_s {
  name : csco_event_phy_evc_instance_t;
    keep soft name == NONE;
  scope: string;
    keep soft scope == "";
};

unit csco_event_phy_agent_u {
  CREATE_TESTFLOW;

  //name of evc
  evc_instance: csco_event_phy_evc_instance_t;
  
  //ACTIVE/PASSIVE agent
  locale: csco_event_phy_locale_t;

  //Name of agent
  agent_instance: csco_event_phy_agent_instance_t;

};

//event agent config

DEFINE_CFG_STRUCT begin
  config_prefix = csco_event_phy_agent,
  parent_unit_type = csco_event_phy_agent_u
end;

extend csco_event_phy_agent_config_s {
  // instance uniquifier
  evc_instance : csco_event_phy_evc_instance_t;

  agent_instance : csco_event_phy_agent_instance_t;

  // events are only processed when monitor is enabled,
  // otherwise they are ignored
  enable : bool;
    keep soft enable;

  // event sensitivity mode
  sensitivity : csco_event_phy_sensitivity_t;
    keep soft sensitivity == EDGE;

  // event signal polarity
  driver_polarity : csco_event_phy_polarity_t;
    keep soft driver_polarity == POSITIVE;
  
  // event signal polarity
  monitor_polarity : csco_event_phy_polarity_t;
    keep soft monitor_polarity == POSITIVE;

  // time to delay between event processing requests
  propagation_delay : time;
    keep soft propagation_delay == 200 ns;
};


//========================================================================
// implementation details
//========================================================================

extend csco_event_phy_env_u {

  //monitor agent to detect and call event hook
  monitor_agents: list of PASSIVE csco_event_phy_agent_u is instance;
    keep soft monitor_agents.size() == 0;
    
  //driver agents to drive signals from lower down in event hierarchy
  driver_agents: list of ACTIVE csco_event_phy_agent_u is instance;
    keep soft driver_agents.size() == 0;

};

extend csco_event_phy_agent_u {

  keep evc_instance == read_only(config.evc_instance);
  
  //target scope in event tree
  scope: string;
  
  // signal change notification
  event event_edge is change(sig_event$) @sim;

  // HDL event signal
  sig_event : in simple_port of csco_event_phy_polarity_t is instance;
    keep bind(sig_event, empty);
    keep read_only(sig_event.hdl_path() != "") => bind(sig_event, external);
  
  // reporting
  event_notify : out method_port of csco_event_phy_push_t is instance;
    keep bind(event_notify, empty);

  // basic coverage
  event detected;
  cover detected is {
    item ev_name: csco_event_phy_agent_instance_t = agent_instance using per_instance;
  };

  // driveable events
  driver : out simple_port of csco_event_phy_polarity_t is instance;  
    keep bind(driver, empty);
    keep read_only(driver.hdl_path() != "") => bind(driver, external);

};

extend ACTIVE'locale csco_event_phy_agent_u {

  active: bool;
    keep soft !active;

  init_driver() @sys.any is {
    wait true (csco_get_cur_test_phase() != CSCO_PRE_RESET);
    if (active) { drive_sig(OFF); };
  };

  drive_sig(state: csco_event_phy_drive_t) @sys.any is {
    driver$ = ((state == ON) ? config.driver_polarity :
       (~config.driver_polarity.as_a(uint)).as_a(csco_event_phy_polarity_t));
  };

  run() is also {
    start init_driver();
  };

};

extend PASSIVE'locale csco_event_phy_agent_u {

  !event_outstanding : bool;

  on event_edge {
    event_outstanding = (sig_event$ == config.monitor_polarity);
    message(HIGH, "New event detected on signal ", sig_event.full_hdl_path());
  };

  event_monitor() @event_edge is {
    while(TRUE) {
      while(event_outstanding &&
            config.enable &&
            (csco_get_cur_test_phase() != CSCO_PRE_RESET)) {
        event_outstanding = FALSE;
        process_event();
        wait delay(config.propagation_delay) @sim;
        if(config.sensitivity == LEVEL) {
          event_outstanding = (sig_event$ == config.monitor_polarity);
        };
      };
      event_outstanding = FALSE;
      wait [1];
    };
  };

  run() is also {
    start event_monitor();
  };

  process_event() @sys.any is {
    message(MEDIUM, config.evc_instance, " event detected");
    emit detected;
    var trans := new csco_event_phy_trans_s with {
      .name = config.evc_instance;
      .scope = scope;
    };
    event_notify$(trans);
  };

};

'>
