// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_rate_stats_param.e
//              :
// Project      : csco_rate_stats
//              :
// Created On   : 11/10/2005
//              :
//              :
// Description  : This file defines the base type for any rate/stat parameter that
//                can be calculated/checked for a rate/stat stream. The parameters
//                are instantiated inside of a csco_rate_stats_stream_u unit
//                to configure/calculate various statistics on a stream.
//
//                Users may also add customized parameters by extending the base
//                type. New parameter types must be defined from file modules of
//                package csco_rate_stats.
//
//                To define a new parameter type:
//
//                1) package csco_rate_stats;  // indicate file module belongs to csco_rate_stats package.
//
//                2) extend csco_rate_stats_param_t : [NEW_PARAM];
//        
//                3) extend NEW_PARAM csco_rate_stats_param_s {
//                     ...
//                     <NEW_PARAM fields>
//                     ...
//
//                     get_print_unit() : string {
//                       <return string that should be appended to any value calculated for this parameter, e.g.: "bytes per second">
//                        // should return "time" if the new parameter calculates a time value.
//                     };
//
//                     calc_stat() : int(bits:64) {
//                       <fetch all relevant transaction from parent stream's input and output interfaces>
//                       <calculate rate/stat result based on each transactions's 'tracking' fields>
//                     };
//                   };
//                
//
//                Some common rate/stats parameters have already been pre-defined:
//
//                INPUT_DATA_RATE OUTPUT_DATA_RATE DATA_LOSS 
//                INPUT_TRANS_RATE OUTPUT_TRANS_RATE TRANS_LOSS
//                AVG_LATENCY MIN_LATENCY MAX_LATENCY
//
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_rate_stats;


type csco_rate_stats_param_t : [
  INPUT_DATA_RATE, 
  OUTPUT_DATA_RATE, 
  DATA_LOSS, 
  INPUT_TRANS_RATE, 
  OUTPUT_TRANS_RATE, 
  TRANS_LOSS,
  AVG_LATENCY,
  MIN_LATENCY,
  MAX_LATENCY
];


struct csco_rate_stats_param_s {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  name : csco_rate_stats_param_t;
    csco_print_attr name SHORT enable;
    csco_print_attr name SHORT header=rate_stat;

  enable_min_check : bool;
    keep soft enable_min_check == TRUE;
    csco_print_attr enable_min_check SHORT enable;
    csco_print_attr enable_min_check SHORT header=min_en;

  min_val: int(bits:64);
    keep soft min_val == MAX_INT;
    csco_print_attr min_val SHORT enable;
    csco_print_attr min_val SHORT header=min;

  enable_max_check : bool;
    keep soft enable_max_check == FALSE;
    csco_print_attr enable_max_check SHORT enable;
    csco_print_attr enable_max_check SHORT header=max_en;

  max_val: int(bits:64);
    keep soft max_val == 0;
    csco_print_attr max_val SHORT enable;
    csco_print_attr max_val SHORT header=max;

  // scratch field to store last calculated value of this rate/stats param.
  !cur_val : int(bits:64);
    csco_print_attr cur_val SHORT enable;

  // The minimum # of transactions that must have been collected over 
  // the stream for rate/stats parameters to be considered valid. 
  min_required_trans : uint;
    keep soft min_required_trans == 5;

  // If TRUE, then if min_required_trans transactions are not collected 
  // for the stream, a dut_error will be raised. 
  inactivity_fatal : bool;
    keep soft inactivity_fatal == FALSE;

  when INPUT_DATA_RATE'name csco_rate_stats_param_s {
    // Indicates whether to drop the first or the last packet
    // when making rate calculation.
    calc_type : csco_rate_stats_calculation_type;
--      keep soft calc_type == AVG_GAP ;
      keep  calc_type == AVG_GAP ;
      csco_print_attr calc_type SHORT enable;
  }; 

  when OUTPUT_DATA_RATE'name csco_rate_stats_param_s {
    // Indicates whether to drop the first or the last packet
    // when making rate calculation.
    calc_type : csco_rate_stats_calculation_type;
--      keep soft calc_type == AVG_GAP ;
      keep  calc_type == AVG_GAP ;
      csco_print_attr calc_type SHORT enable;
  };

  when INPUT_TRANS_RATE'name csco_rate_stats_param_s {
    // Indicates whether to drop the first or the last packet
    // when making rate calculation.
    calc_type : csco_rate_stats_calculation_type;
--      keep soft calc_type == AVG_GAP ;
      keep  calc_type == AVG_GAP ;
      csco_print_attr calc_type SHORT enable;
  }; 

  when OUTPUT_TRANS_RATE'name csco_rate_stats_param_s {
    // Indicates whether to drop the first or the last packet
    // when making rate calculation.
    calc_type : csco_rate_stats_calculation_type;
--      keep soft calc_type == AVG_GAP ;
      keep  calc_type == AVG_GAP ;
      csco_print_attr calc_type SHORT enable;
  };


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  // When determinant, automatically constrained by parent stream. The
  // User can use this field as a when determinant to configuring a
  // parameter instance.
  stream_name : csco_rate_stats_stream_name_t;

  csco_short_name() : string is {
    result = appendf("%s", name);
  };

  // Define extra/custom formatting for rate/stat value fields.
  csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
    if pf.name == "min_val" {
      result = value_to_str(min_val);
    };

    if pf.name == "max_val" {
      result = value_to_str(max_val);
    };

    if pf.name == "cur_val" {
      cur_val = calc_stat();
      result = value_to_str(cur_val);
    };
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  // Pointer back to parent stream unit. This pointer allows the parameter to access
  // all the transactions in the input and output interfaces of the parent stream
  // when calculating rates/stats.
  package parent_stream : csco_rate_stats_stream_u;
    keep soft parent_stream == NULL;

  package is_rate : bool;
    keep is_rate == (name in [INPUT_DATA_RATE, OUTPUT_DATA_RATE, INPUT_TRANS_RATE, OUTPUT_TRANS_RATE]);

  package is_loss : bool;
    keep is_loss == (name in [DATA_LOSS, TRANS_LOSS]);

  package is_latency : bool;
    keep is_latency == (name in [MIN_LATENCY, MAX_LATENCY, AVG_LATENCY]);

  package get_relevant_trans() : list of csco_trans_s is undefined;

  // This method must be extended by the package developer for each new parameter
  // type that is added.
  package calc_stat() : int(bits:64) is undefined;

  // This method must be extended by the package developer for each new parameter
  // type that is added.
  //
  // For new rate/stats parameters that return a time, the package developer 
  // should extend this method to return "time".
  package get_print_unit() : string is empty;

  package value_to_str(val : int(bits:64)) : string is {
    var local_val : int(bits:64);
    var prefix : string;
    var maj_str : string;
    var val_str : string;
    var is_time : bool;
    var print_unit : string;

    if get_print_unit() != "time" {
      local_val = val;
      is_time = FALSE;
      val_str = appendf("%d", val);
      print_unit = get_print_unit();
    }
    else {
      local_val = from_specman_scale(val, fs);
      is_time = TRUE;
      val_str = appendf("%d", local_val);
      print_unit = "";
    };

    if (local_val / 1_000_000_000_000) > 0 {
      prefix = (is_time) ? "ms" : "T";
      maj_str = appendf("%d", local_val / 1_000_000_000_000);
    }
    else if (local_val / 1_000_000_000) > 0 {
      prefix = (is_time) ? "us" : "B";
      maj_str = appendf("%d", local_val / 1_000_000_000);
    }
    else if (local_val / 1_000_000) > 0 {
      prefix = (is_time) ? "ns" : "M";
      maj_str = appendf("%d", local_val / 1_000_000);
    }
    else if (local_val / 1_000) > 0 {
      prefix = (is_time) ? "ps" : "K";
      maj_str = appendf("%d", local_val / 1_000);
    }
    else if is_time and (local_val / 1_000_000_000_000_000) > 0 {
      prefix = "s";
      maj_str = appendf("%d", local_val / 1_000_000_000_000_000);
    }
    else {
      maj_str = val_str;
      if is_time {
        print_unit = "fs";
      };
    };

    if maj_str == val_str {
      result = appendf("%s %s", maj_str, print_unit);
    }
    else {
      result = appendf("%s.%s%s %s", maj_str, str_sub(val_str, str_len(maj_str), 3), prefix, print_unit);
    };
  }; // value_to_str()

  run() is also {
    assert that parent_stream != NULL else
      error(appendf("CSCO_RATE_STATS_ERR_NO_PARENT: %s parameter is not associated with a parent stream", name));

    assert that min_val <= max_val or !enable_min_check or !enable_max_check else
      error(appendf("CSCO_RATE_STATS_ERR_BAD_MIN_MAX: min value (%d) > max value (%d) for %s of %s", min_val, max_val, name, stream_name));
  };

  package min_trans_met() : bool is undefined;

  check_param() is {
    var calc_val : int(bits:64);

    if parent_stream.enable {
      if !min_trans_met() {
        if inactivity_fatal {
          dut_error("CSCO_RATE_STATS_ERR_INACTIVITY: must have ", min_required_trans, " to calculate ", name, " for ", 
                        stream_name, ", collected ", parent_stream.get_input_stats().all_trans.size(), " for the input channel, ", 
                        "and ", parent_stream.get_output_stats().all_trans.size(), " for the output channel.");
        };
        return;
      };

      calc_val = calc_stat();

      if enable_min_check and calc_val < min_val {
        dut_errorf("CSCO_RATE_STATS_ERR_%s_MIN_CMPR: %s actual value (%s) < minimum expected value (%s)", name, stream_name, value_to_str(calc_val), 
                                                                                                          value_to_str(min_val));
      }
      else if enable_min_check {
        messagef(LOW, "successfully checked minimum value (%s) for %s of %s\n", value_to_str(min_val), name, parent_stream.short_name());
      };

      if enable_max_check and calc_val > max_val {
        dut_errorf("CSCO_RATE_STATS_ERR_%s_MAX_CMPR: %s actual value (%s) > maximum expected value (%s)", name, stream_name, value_to_str(calc_val), 
                                                                                                          value_to_str(max_val));
      }
      else if enable_max_check {
        messagef(LOW, "successfully checked maximum value (%s) for %s of %s\n", value_to_str(max_val), name, parent_stream.short_name());
      };
    }
    else {
      messagef(CSCO_RATE_STATS, HIGH, "will not check minimum value for %s of %s because the parent stream is disabled\n", name, parent_stream.short_name());
    };
  };
}; // struct csco_rate_stats_param_s


extend TRUE'is_rate csco_rate_stats_param_s {
  package get_print_unit() : string is {
    if name in [INPUT_DATA_RATE, OUTPUT_DATA_RATE] {
      result = appendf("%s/s", parent_stream.get_data_unit());
    }
    else {
      result = appendf("trans/s");
    };
  };

  package min_trans_met() : bool is {
    result = (get_stats_ptr().all_trans.size() >= min_required_trans);
  };

  package get_relevant_trans() : list of csco_trans_s is also {
    result = get_stats_ptr().get_eligible_trans();
    if result.size() > 0 {
      case get_calc_type() {
        DROP_FIRST : {
          compute result.pop0();
        };
        DROP_LAST : {
          compute result.pop();
        };
      };
    };

    //assert result.size() > 0 else
      //error(appendf("CSCO_TRANS_ERR_RELVNT_TRANS: no relevant transactions were found when computing %s for %s", name, parent_stream.short_name()));
  }; // get_relevant_trans() 

  package calc_stat() : int(bits:64) is {
    if get_relevant_trans().size() > 0 {
      var total_time := get_elapsed_time() ;
      if total_time == 0 {
        result = 0 ;
      } else {
        if name in [INPUT_DATA_RATE, OUTPUT_DATA_RATE] {
          result = get_relevant_trans().sum(.rate_stats.data_size);
        } else {
          result = get_relevant_trans().size();
        };
        result *= to_specman_scale(1, sec) ; -- scale up to avoid precision loss
        result /= total_time ; -- average over time
      };
    };
  };

  package get_calc_type() : csco_rate_stats_calculation_type is {
    if me is a INPUT_DATA_RATE csco_rate_stats_param_s (idrate) {
      result = idrate.calc_type;
    }
    else if me is a OUTPUT_DATA_RATE csco_rate_stats_param_s (odrate) {
      result = odrate.calc_type;
    };
  };

  package get_stats_ptr() : csco_rate_stats_io_s is {
    if name == INPUT_DATA_RATE {
      result = parent_stream.get_input_stats();
    }
    else {
      result = parent_stream.get_output_stats();
    };
  };

  // Calculate the amount of elapsed time, based on the selected algorithm.
  package get_elapsed_time() : time is {
    var trans_list := get_stats_ptr().get_eligible_trans() ;
    case get_calc_type() {
      DROP_FIRST : {
        result = trans_list.top().tracking.get_xmt_done_time() - trans_list.top0().tracking.get_xmt_done_time();
      };
      DROP_LAST : {
        result = trans_list.top().tracking.get_start_time() - trans_list.top0().tracking.get_start_time();
      };
      AVG_GAP : {
        result = trans_list.top().tracking.get_xmt_done_time() - trans_list.top0().tracking.get_start_time() + get_avg_gap(trans_list);
      };
    }
  }; // get_elapsed_time() 

  package get_avg_gap(trans_list:list of csco_trans_s) : time is {
    if trans_list.size() <= 1 { 
      result = 0 ;
    } else {
      var last_end_time : time ;
      var gap_total : time ;
      for each (trans) in trans_list {
        if last_end_time < trans.tracking.get_xmt_done_time() {
          last_end_time = trans.tracking.get_xmt_done_time() ;
          if last_end_time < trans.tracking.get_start_time() {
            gap_total += (trans.tracking.get_start_time() - last_end_time) ;
          };
        };
      };
      //average gaps over one less than the packet count (as packets are on both ends)
      result = gap_total / (trans_list.size() - 1) ;
    };
  };
}; // extend TRUE'is_rate csco_rate_stats_param_s

extend TRUE'is_loss csco_rate_stats_param_s {
  keep soft !enable_min_check;
  keep soft enable_max_check;

  package get_print_unit() : string is {
    result = "%";
  };

  package min_trans_met() : bool is {
    result = (parent_stream.get_input_stats().all_trans.size() >= min_required_trans and
              parent_stream.get_output_stats().all_trans.size() >= min_required_trans);
  };

  package calc_stat() : int(bits:64) is {
    var itrans : list of csco_trans_s = parent_stream.get_input_stats().get_eligible_trans();
    var otrans : list of csco_trans_s = parent_stream.get_output_stats().get_eligible_trans();
    var tmp_int1 : int(bits:64);
    var tmp_int2 : int(bits:64);

    if name == DATA_LOSS {
      tmp_int1 = itrans.sum(.rate_stats.data_size);
      tmp_int2 = otrans.sum(.rate_stats.data_size);
    }
    else {
      tmp_int1 = itrans.size();
      tmp_int2 = otrans.size();
    };

    if tmp_int1 != 0 {
      result = (tmp_int1 - tmp_int2) / tmp_int1;
    };
  };
}; // extend TRUE'is_loss csco_rate_stats_param_s

extend TRUE'is_latency csco_rate_stats_param_s {
  keep soft enable_min_check == FALSE;
  keep soft enable_max_check == TRUE;

  package get_print_unit() : string is {
    result = appendf("time");
  };

  package min_trans_met() : bool is {
    result = (get_relevant_trans().size() >= min_required_trans);
  };

  package get_relevant_trans() : list of csco_trans_s is also {
    result = parent_stream.get_input_stats().get_eligible_trans().all(.tracking.get_hop_done_flag() == TRUE);
  };

  package calc_stat() : int(bits:64) is {
    var latency_list : list of time;

    for each (t) in get_relevant_trans() {
      latency_list.add(t.tracking.get_hop_done_time() - t.tracking.get_start_time());
    };

    case name {
      MIN_LATENCY : {
        result = latency_list.min_value(it);
      };

      MAX_LATENCY : {
        result = latency_list.max_value(it);
      };

      AVG_LATENCY : {
        result = latency_list.sum(it) / latency_list.size();
      };
    };
  };
}; // extend TRUE'is_latency csco_rate_stats_param_s


'>
