// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_rate_stats_stream.e
//              :
// Project      : csco_rate_stats
//              :
// Created On   : 11/10/2005
//              :
//              :
// Description  : This file defines a facility for calculating rates and statistics
//                for transactions tracked over a rate/stats stream. Each stream
//                consists of an input interface and an output interface. The model
//                is that transactions are injected over a stream's input interface
//                and leaves via the stream's output interface. The stream allows
//                us to calculate transaction rates, data rates, and latency measures
//                over all transactions observed over the stream.
//
//                The transactions flowing through a stream are assumed to inherit 
//                from the "csco_trans_s" base type. 
//
//                There are 2 kinds of streams:
//
//                  PHYICAL : includes input and output interfaces as described above.
//                  MERGED  : a collection of child (physical and merged) streams whose 
//                            statistics are merged into a single logical stream.
//
//                For MERGED streams, several different merge algorithms are defined:
//
//                  ALL_TRANS:      All transactions from all the child streams are merged,
//                                  per INPUT/OUTPUT interface.
//                
//                  ACTIVE_OVERLAP: Per INPUT/OUTPUT interface, the new transaction
//                                  list is constructed by setting the merge start
//                                  time to be the latest time stamp amongst the 1st
//                                  transaction in all the child interfaces. And setting
//                                  the merge end time to be the earliest time stamp
//                                  amongst the last transaction in all the child 
//                                  interfaces. The resulting transaction list is then
//                                  all the transactions of all the child interfaces
//                                  whose time stamp falls between the merge start and
//                                  end times.
//                
//                  MANUAL:         The user defines the merge start and end times.
//                
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

<'
package csco_rate_stats;


// Encapsulates all stat parameters for a single interface. Note that
// a rate/stats I/O interface pair should be instantiated inside of a
// stream. An interface by itself does not yield any throughput information.
package struct csco_rate_stats_io_s {
  // Ingress or Egress side of a stream.
  direction : csco_rate_stats_io_dir_t;
    csco_print_attr direction SHORT enable;
    csco_print_attr direction SHORT header=dir;

  parent_stream : csco_rate_stats_stream_u;
    csco_print_attr parent_stream ALL disable;

  !all_trans : list of csco_trans_s;
    csco_print_attr all_trans SHORT enable;

  package !first_trans : csco_trans_s;
  package !last_trans : csco_trans_s;
  package !total_elig_data : int(bits:64);

  // Update the interface with a new transaction.
  add_trans(tr : csco_trans_s) is {
    assert that (tr not in all_trans) else 
      error(appendf("CSCO_RATE_STATS_ERR_DUP_TR: %s has already been added to %s's input stats", tr.csco_short_name(), parent_stream.csco_short_name()));

    if first_trans == NULL {
      first_trans = tr;
    };
    all_trans.add(tr);
    last_trans = tr;
  }; 

  remove_trans(tr : csco_trans_s) is {
    var ind : int = all_trans.first_index(it == tr);
    assert that ind != UNDEF else
      error(appendf("CSCO_RATE_STATS_ERR_REMOVE_TR: %s is not part of %s's input stats, cannot remove.", tr.csco_short_name(), parent_stream.csco_short_name()));

    all_trans.delete(ind);
    if all_trans.size() > 0 {
      first_trans = all_trans[0];
    };
    if all_trans.size() > 1 {
      last_trans = all_trans.top();
    }
  };

  get_eligible_trans() : list of csco_trans_s is {
    result = all_trans.all(.rate_stats.dont_include == FALSE);
  };

  clear() is {
    all_trans.clear();
  }; // clear() is 

  csco_print_get_fields_user_hook(format : csco_print_format_t) is also {
    total_elig_data = get_eligible_trans().sum(.rate_stats.data_size);
  };

  csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
    if pf.name == "total_elig_data" {
      result = append(result, " ", parent_stream.get_data_unit());
    };
  };
}; // package struct csco_rate_stats_io_s


unit csco_rate_stats_stream_u like any_env {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Physical or Merged stream. Note that both stream types share the same statistical
  // attributes, the difference lies in how they're calculated.
  kind : csco_rate_stats_stream_kind_t;
    keep soft kind == PHYSICAL;
    csco_print_attr kind SHORT enable;

  // Used for per-instance sub-typing.
  name : csco_rate_stats_stream_name_t;
    csco_print_attr name SHORT enable;

  // Unique instance name for each stream type (name).
  inst_name : string;
    keep soft inst_name == name.as_a(string);
    csco_print_attr inst_name SHORT enable;
    csco_print_attr inst_name SHORT header=inst;

  // Whether the stream is enabled or not. A disabled stream will not collect 
  // transactions (if PHYSICAL stream) nor check min/max values at the end of 
  // the sim (both PHYSICAL and MERGED).
  enable : bool;
    keep soft enable == TRUE;
    csco_print_attr enable SHORT enable;

  // List of rate/stats parameters that will be calculated
  // for this stream.
  params : list of csco_rate_stats_param_s;
    keep soft params.size() == 0;
    keep params.size() <= all_values(csco_rate_stats_param_t).size();
    keep for each (p) in params {
      p.parent_stream == me;
      p.stream_name == name;
      index > 0 => p.name not in params[..index-1].name;
    };
    csco_print_attr params SHORT enable;
    csco_print_attr params LONG expand=ON;

  // A short-cut for user to instantiate all defined rate/stats 
  // parameters in the stream.
  use_all_params : bool;
    keep soft use_all_params == FALSE;
    when TRUE'use_all_params csco_rate_stats_stream_u {
      keep params.size() == all_values(csco_rate_stats_param_t).size();
      keep for each (p) in params {
        p.name == all_values(csco_rate_stats_param_t)[index];
        !enable => all of {
          p.enable_min_check == FALSE;
          p.enable_max_check == FALSE;
        };
      };
    };

  logger: message_logger is instance;
    keep soft logger.tags == {CSCO_RATE_STATS};
    keep soft logger.verbosity == LOW;


  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  // Looks up a parameter by name and returns its current value.
  get_stat(nm : csco_rate_stats_param_t) : int(bits:64) is {
    result = get_param(nm).calc_stat();
  };

  csco_short_name(): string is {
    result = appendf("%s_RSTATS", inst_name);
  };

  check_rate_stats() is {
    if enable {
      message(LOW, "checking parameters of ", csco_short_name(), ":") {
        csco_print(LONG);
      };
      outf("\n");
      for each (p) in params {
        p.check_param();
      };
      outf("\n");
    };
  };

  clear() is {
    message(CSCO_RATE_STATS, MEDIUM, "clearing rate/stats in ", csco_short_name());
    data_unit = "";
  }; 


  ////////////////////////
  // Private Fields 
  ////////////////////////
  // Unit of data for transactions on this interface (bytes, bits, etc...)
  package !data_unit : string;

  package get_param(nm : csco_rate_stats_param_t) : csco_rate_stats_param_s is {
    result = params.first(.name == nm);
    assert that result != NULL else 
      error(appendf("%s parameter has not been instantiated inside %s", nm, csco_short_name()));
  };

  package get_data_unit() : string is {
    result = data_unit;
    if result == "" {
      result = "<data_size_unit>";
    };
  };

  package check_data_unit(du : string) is {
    if data_unit != "" {
      assert that data_unit == du else 
        error(appendf("CSCO_RATE_STATS_ERR_DATA_UNIT: inconsistent data size units (%s and %s) detected for %s", data_unit, du, csco_short_name()));
    }
    else {
      data_unit = du;
    };
  };

  // Virtual methods to be extended by PHYSICAL and MERGE sub-types.
  package get_input_stats() : csco_rate_stats_io_s is undefined;
  package get_output_stats() : csco_rate_stats_io_s is undefined;
}; // unit csco_rate_stats_stream_u

extend PHYSICAL csco_rate_stats_stream_u {
  // Captures input interface stats.
  package input_stats: csco_rate_stats_io_s;
    keep input_stats.parent_stream == me;
    keep input_stats.direction == INPUT;
    csco_print_attr input_stats LONG expand=ON; 

  // Captures output interface stats.
  package output_stats: csco_rate_stats_io_s;
    keep output_stats.parent_stream == me;
    keep output_stats.direction == OUTPUT;
    csco_print_attr output_stats LONG expand=ON; 

  package get_input_stats() : csco_rate_stats_io_s is {
    result = input_stats;
  };

  package get_output_stats() : csco_rate_stats_io_s is {
    result = output_stats;
  };

  clear() is also {
    input_stats.clear();
    output_stats.clear();
  };

  // Method port to add a new transaction to the input interface. This port must
  // be bound to the appropriate output method port in the monitor/scoreboard.
  add_input_trans : in method_port of csco_trans_push_mp is instance;
    keep bind(add_input_trans, undefined);

  // Method port to add a new transaction to the output interface. This port must
  // be bound to the appropriate output method port in the monitor/scoreboard.
  add_output_trans : in method_port of csco_trans_push_mp is instance;
    keep bind(add_output_trans, undefined);

  // Update the input statistics.
  package add_input_trans(tr: csco_trans_s) is {
    add_input_tr(tr);
  };

  add_input_tr(tr : csco_trans_s) is {
    if enable {
      messagef(CSCO_RATE_STATS, MEDIUM, "adding a new %s transaction to %s's input interface\n", tr.csco_short_name(), csco_short_name());
      check_data_unit(tr.rate_stats.data_unit);
      input_stats.add_trans(tr);
    }
    else {
      messagef(CSCO_RATE_STATS, HIGH, "will not add %s transaction to %s's input interface because stream is disabled\n", tr.csco_short_name(), csco_short_name());
    };
  };

  remove_input_tr(tr : csco_trans_s) is {
    input_stats.remove_trans(tr);
  };

  // Update the output statistics.
  package add_output_trans(tr: csco_trans_s) is {
    add_output_tr(tr);
  };  

  add_output_tr(tr : csco_trans_s) is {
    if enable {
      messagef(CSCO_RATE_STATS, MEDIUM, "adding a new %s transaction to %s's output interface\n", tr.csco_short_name(), csco_short_name());
      check_data_unit(tr.rate_stats.data_unit);
      output_stats.add_trans(tr);
    }
    else {
      messagef(CSCO_RATE_STATS, HIGH, "will not add %s transaction to %s's output interface because stream is disabled\n", tr.csco_short_name(), csco_short_name());
    };
  };

  remove_output_tr(tr : csco_trans_s) is {
    output_stats.remove_trans(tr);
  };
}; // extend PHYSICAL csco_rate_stats_stream_u


extend MERGED csco_rate_stats_stream_u {
  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  merge_alg : csco_rate_stats_merge_alg_t;
    keep soft merge_alg == ALL_TRANS;

  // A merged stream is an amalgamation of child (physical and merged) streams.
  child_streams : list of csco_rate_stats_stream_u;
    keep soft child_streams.size() == 0;

  dut_error_dump() is also {
    csco_print_sub_tree(LONG, "child_streams");
  };



  ////////////////////////
  // User Read-Only Fields 
  ////////////////////////
  add_child_stream(st : csco_rate_stats_stream_u) is {
    child_streams.add(st);
  };

  // Used by user to set merge start time for MANUAL merge mode.
  set_merge_start_time(t : time) is {
    merge_start_time = t;
  };

  // Used by user to set merge end time for MANUAL merge mode.
  set_merge_end_time(t : time) is {
    merge_end_time = t;
  };

  clear() is also {
    for each (s) in child_streams {
      s.clear();
    };
  };


  ////////////////////////
  // Private Fields 
  ////////////////////////
  package !merge_start_time : time;

  package !merge_end_time : time;

  package !merged_istats : csco_rate_stats_io_s;
    csco_print_attr merged_istats LONG expand=ON;

  package !merged_ostats : csco_rate_stats_io_s;
    csco_print_attr merged_ostats LONG expand=ON;

  // Merge multiple interface's statistics.
  package merge_io_stats(stats : list of csco_rate_stats_io_s) : csco_rate_stats_io_s is {
    var valid_stats : list of csco_rate_stats_io_s;

    valid_stats = stats.all(.all_trans.size() > 0);
    result = new;
    result.parent_stream = me;

    if valid_stats.size() > 0 {
      assert that valid_stats.parent_stream.unique(.data_unit).size() == 1 else
        error(appendf("CSCO_RATE_STATS_ERR_MRGE_DATA_UNIT: merging rate/stats channels with different data size units"));

      data_unit = valid_stats[0].parent_stream.data_unit;
      case merge_alg {
        ALL_TRANS : {
          merge_start_time = valid_stats.min_value(.all_trans.top0().tracking.get_start_time());
          merge_end_time = valid_stats.max_value(.all_trans.top().tracking.get_xmt_done_time());
          for each (s) in valid_stats {
            result.all_trans.add(s.all_trans);
          };
        };

        ACTIVE_OVERLAP : {
          if valid_stats.size() > 0 {
            merge_start_time = valid_stats.max_value(.all_trans.top0().tracking.get_start_time());
            merge_end_time = valid_stats.min_value(.all_trans.top().tracking.get_xmt_done_time());
          };

          for each (s) in valid_stats {
            result.all_trans.add(s.all_trans.all(.tracking.get_start_time() >= merge_start_time and .tracking.get_xmt_done_time() <= merge_end_time));
          };
        };

        MANUAL : {
          assert that merge_start_time < merge_end_time else
            error(appendf("CSCO_RATE_STATS_ERR_MERGE_TIMES: merge start time (%s) is >= merge_end_time (%s)", merge_start_time, merge_end_time));

          for each (s) in valid_stats {
            result.all_trans.add(s.all_trans.all(.tracking.get_start_time() >= merge_start_time and .tracking.get_xmt_done_time() <= merge_end_time));
          };
        };
      };

      result.all_trans = result.all_trans.sort(.tracking.get_id());
      result.all_trans = result.all_trans.unique(it);
    };
  }; // merge_io_stats()

  package get_input_stats() : csco_rate_stats_io_s is {
    for each (s) in child_streams {
      if result == NULL {
        result = s.get_input_stats();
      }
      else {
        result = merge_io_stats({result; s.get_input_stats()});
      };
    }; // for each (s) in child_streams
  };

  package get_output_stats() : csco_rate_stats_io_s is {
    for each (s) in child_streams {
      if result == NULL {
        result = s.get_output_stats();
      }
      else {
        result = merge_io_stats({result; s.get_output_stats()});
      };
    }; // for each (s) in child_streams
  };

  csco_print_get_fields_user_hook(format : csco_print_format_t) is also {
    merged_istats = get_input_stats();
    merged_ostats = get_output_stats();
  };
}; // extend MERGED csco_rate_stats_stream_u


'>
