//------------------------------------------------------------------------
// File name    : csco_sequence_extensions_top.e
//              :
//              :
//              :
// Description  : extensions to the base any_sequence struct for:
//              :   - started sequence state tracking
//              :   - fork/merge of arbitrary lists of parallel sequences
//              :   - blocking sequence execution until specified
//              :     dependency sequences have finished
//              :
//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Copyright 2002-2003,2005-2006 (c) Cisco Systems Inc.
//------------------------------------------------------------------------
<'
package csco_seq_ext;

define CSCO_SEQ_EXT_VERSION_2_0;
define CSCO_SEQ_EXT_VERSION_2_0_OR_LATER;
define CSCO_SEQ_EXT_VERSION_1_0_OR_LATER;

//========================================================================
// API definition
//========================================================================

// current running state of a *started* sequence
type csco_seq_state_t : [
  NA = 0,       // either not yet started or was executed via "do"
  WAIT,         // "blocking" for dependencies to complete
  STARTED,      // was started w/start_sequence() and is running
  ENDED         // has finished it's body() and is no longer running
];

//------------------------------------------------------------
// transaction passing method ports

// non-time consuming PUSH interface.
// a transaction source uses this to PUSH to sink(s) and return immediately.
// sinks must use their own threading for time consuming actions.
// used, eg, for passing transactions from monitors to drivers/scoreboards
method_type csco_seq_push_mp(trans : any_sequence_item);

// time consuming PUSH interface.
// a transaction source uses this to PUSH to sink(s) and block for processing.
// used, eg, for passing transactions from drivers to BFMs in PUSH_MODE
method_type csco_seq_push_tcmp(trans : any_sequence_item) @sys.any;

// non-time consuming PULL interface.
// a transaction sink uses this to PULL from a source.
// NULL is returned if no transaction is availble.
// used, eg, for a try_next_item() style BFM/driver interaction
method_type csco_sesq_pull_mp() : any_sequence_item;

// time consuming PULL interface.
// a transaction sink uses this to PULL from a source.
// the call will block until a transaction is available.
// used, eg, for a get_next_item() style BFM/driver interaction
method_type csco_seq_pull_tcmp() : any_sequence_item @sys.any;


//------------------------------------------------------------
// utility extensions to any_sequence
extend any_sequence {

  //----------------------------------------
  // sequence fork/merge interface
  // for executing sub-sequences in parallel

  // fork off sub-sequences.
  // they must have already been generated
  csco_fork(seq: any_sequence) is undefined;
  csco_fork_list(seqs: list of any_sequence) is undefined;

  // wait for specified sequences to complete
  csco_merge(seq: any_sequence) @sys.any is undefined;
  csco_merge_list(seqs: list of any_sequence) @sys.any is undefined;

  // wait for all children started from this sequence to complete
  csco_merge_all() @sys.any is undefined;

  // fork the specified (pre-generated) sequences
  // and wait for them all (and any others!) to complete
  csco_do_all(seqs: list of any_sequence) @sys.any is undefined;

  // legacy interface
  csco_fork_children(seqs: list of any_sequence) is { csco_fork_list(seqs) };
  csco_merge_children() @sys.any is { csco_merge_all() };

  //----------------------------------------
  // sequence dependency interface.
  // a sequence will WAIT in pre_body
  // until all dependencies have ENDED.
  // relevant sequences must be *started*, not "do"ne

  // list of sequences which this sequence will WAIT on
  !csco_seq_deps : list of any_sequence;

  // interface to add a dependency after sequences have been generated.
  // NULL sequence references are silently ignored
  csco_add_dep(seq: any_sequence) is undefined;
  csco_add_dep_list(seqs: list of any_sequence) is undefined;

  // interface to remove a previously added dependency.
  // no error is generated if the dependency was never added
  csco_remove_dep(seq: any_sequence) is undefined;
  csco_remove_dep_list(seqs: list of any_sequence) is undefined;

  //----------------------------------------
  // query current sequence extension state

  // obtain current running state of started sequences
  csco_get_state() : csco_seq_state_t is undefined;

  // obtain current count of started sequences
  // which have not yet ended or been terminated
  csco_get_seq_count() : int is undefined;

  // obtain current list of unsatisfied dependencies
  csco_get_deps() : list of any_sequence is undefined;
};


//========================================================================
// implementation details
//========================================================================
    
extend any_sequence {
  // generic counter to track started parallel sequences 
  // which can be used to block this parent sequence
  // waiting for the spawned children to complete
  package csco_seq_count : int;
    keep soft csco_seq_count == 0;

  csco_get_seq_count() : int is { result = csco_seq_count };

  // state to determine if sequence was invoked using .start()
  package !csco_seq_state : csco_seq_state_t;

  csco_get_state() : csco_seq_state_t is { result = csco_seq_state };


  //------------------------------------------------------------
  // sequence fork-merge interface

  csco_fork(seq: any_sequence) is {
    seq.parent_sequence = me;  // *still* required after 4 years
    seq.start_sequence();
  };

  csco_fork_list(seqs: list of any_sequence) is {
    for each in seqs {
      .parent_sequence = me;  // *still* required after 4 years
      .start_sequence();
    };
  };

  csco_merge(seq: any_sequence) @sys.any is {
    sync true(seq.csco_seq_state == ENDED) @sys.tick_start;
  };

  csco_merge_list(seqs: list of any_sequence) @sys.any is {
    for each in seqs {
      sync true(.csco_seq_state == ENDED) @sys.tick_start;
    };
  };

  csco_merge_all() @sys.any is {
    assert(csco_seq_count >= 0);
    if(csco_seq_count > 0) {
      sync true(csco_seq_count == 0) @sys.tick_start;
    };
  };

  csco_do_all(seqs: list of any_sequence) @sys.any is {
    csco_fork_list(seqs);    // start the children
    sync [0];                // wait for children to be started
    csco_merge_list(seqs);   // wait for them to complete
  };

  csco_do_all_random(seqs: list of any_sequence) @sys.any is {
    var reorder_seqs : list of any_sequence;
    gen reorder_seqs keeping {
      it.is_a_permutation(seqs);
    };
    csco_fork_list(reorder_seqs);   // start the children
    sync [0];                       // wait for children to be started
    csco_merge_list(reorder_seqs);  // wait for them to complete
  };


  //------------------------------------------------------------
  // dependency setup & query

  csco_get_deps() : list of any_sequence is {
    result = csco_seq_deps.all(.csco_seq_state != ENDED);
  };

  csco_add_dep(seq: any_sequence) is {
    if((seq != NULL) && !csco_seq_deps.has(it == seq)) {
      csco_seq_deps.add(seq);
    };
  };

  csco_add_dep_list(seqs: list of any_sequence) is {
    for each in seqs {
      // FIXME O(n^2)
      csco_add_dep(it);
    };
  };

  csco_remove_dep(seq: any_sequence) is {
    csco_seq_deps = csco_seq_deps.all(it != seq);
  };

  csco_remove_dep_list(seqs: list of any_sequence) is {
    for each (seq) in reverse csco_seq_deps {
      if(seqs.has(it == seq)) {
        csco_seq_deps.fast_delete(index);
      };
    };
  };

  //------------------------------------------------------------
  // sequence state tracking

  pre_body() @sys.any is also {
    if(parent_sequence != NULL) {
      assert(parent_sequence.csco_seq_count >= 0);
      parent_sequence.csco_seq_count += 1;
    };
    // wait for all dependencies to be satisfied
    csco_seq_state = WAIT;
    for each in csco_seq_deps {
      // FIXME find a better sampling event
      // (maybe based on @ended? but what abt @quit?)
      sync true(.csco_seq_state == ENDED) @sys.tick_start;
    };
    csco_seq_state = STARTED;
  };

  post_body() @sys.any is also {
    // FIXME for some reason quit() sequences now call post_body()?!
    //assert(csco_seq_state == STARTED);
    assert(csco_seq_state != NA);
    if(csco_seq_state in [STARTED, WAIT]) {
      csco_seq_state = ENDED;
      if(parent_sequence != NULL) {
        assert(parent_sequence.csco_seq_count > 0);
        parent_sequence.csco_seq_count -= 1;
      };
    };
  };

  quit() is also {
    // decrement sequence counter if the sequence was started (w/ .start())
    //   but quit() explicitly or due to a rerun() (or stop_sequence()?)
    if((parent_sequence != NULL) && (csco_seq_state in [STARTED, WAIT])) {
      assert(parent_sequence.csco_seq_count > 0);
      parent_sequence.csco_seq_count -= 1;
    };
    csco_seq_state = ENDED;
  };
};

'>
