// ----------------------------------------------------------------------------
// File name    : csco_testflow_dut_reset_base.e
//              :
// Project      : csco_testflow
//              :
// Created On   : 05/11/2006
//              :
//              :
// Description  : This file contains the base logic for converting environment
//              : test phases into simulator activity that causes the DUT to
//              : enter and leave the corresponding phases
//              :
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2006 (c) Cisco Systems Inc.
// ----------------------------------------------------------------------------
<'
package csco_testflow;

// current state of DUT reset (according to testbench)
type csco_testflow_dut_reset_state_t : [
  UNKNOWN,      // reset has not yet been initialized by testbench
  ACTIVE,       // reset is in (or transitioning to) its active state
  INACTIVE      // reset is in (or transitioning to) its inactive state
];

// indication of how reset is connected/driven at the DUT.
// often different between testbench levels (unit, fullchip, system, etc).
// users may also create additional driver types for extended functionality
type csco_testflow_dut_reset_driver_t : [

  // no reset will be driven (ie, set_state$() will not be activated).
  // get_state$() will still be activated to check the DUT value if available
  NONE,

  // indicates that the current layer of testbench hierarchy
  // has no corresponding DUT reset register.
  // so no reset pin will be driven directly;
  // instead, underlying dut_reset units will be activated
  VIRTUAL,

  // reset is a single HDL primary input pin to the DUT
  // over which the testbench has exclusive control
  PIN,

  // reset is controlled by a register field.
  // the DUT activates the field for any higher-level reset,
  // but it must be programmed to the inactive state
  // at the start of the appropriate test phase
  REG_AUTO,

  // reset controlled by a register field.
  // the DUT activates the field during a higher-level PRE_RESET,
  // but all other state changes must be programmed
  REG_PRE_RESET
];

// reset application hook.
// invoked by dut_reset_base unit to activate/deactivate reset.
// should block until operation has completed at target (or very close)
method_type csco_testflow_dut_reset_set_mp(reset: csco_testflow_dut_reset_u) @sys.any;

// reset checking hook.
// used to sanity check current reset state of DUT.
method_type csco_testflow_dut_reset_get_mp(reset: csco_testflow_dut_reset_u) : csco_testflow_dut_reset_state_t;


//------------------------------------------------------------------------
unit csco_testflow_dut_reset_u {

  // type of reset.
  // specifically, the last testflow phase in which the reset should be active
  phase : csco_testflow_stage_t;

  run() is also {
    assert(phase in [CSCO_PRE_RESET, CSCO_RESET, CSCO_INIT_TEST]) else
      error("ERR_CSCO_TESTFLOW_DUT_RESET_ILLEGAL_PHASE\n",
            "reset phase ", phase, "is not in [PRE_RESET, RESET, INIT_TEST]");
  };

  // current reset driver type: PIN, REG, etc.
  driver : csco_testflow_dut_reset_driver_t;
    keep soft driver == PIN;

  short_name() : string is {
    result = append("dut_reset(", phase, " ", driver, ")");
  };

  get_string() : string is empty;

  // current testbench setting of DUT reset
  state : csco_testflow_dut_reset_state_t;
    keep soft state == UNKNOWN;

  // user hook to apply new reset state to DUT
  set_state : out method_port of csco_testflow_dut_reset_set_mp is instance;
    keep bind(set_state, empty);

  // user hook to return current reset state at DUT
  get_state : out method_port of csco_testflow_dut_reset_get_mp is instance;
    keep bind(get_state, empty);

  // propagation delay (holdoff next test phase after set())
  propagation_delay : time;
    keep soft propagation_delay == 250 ns;

  //----------------------------------------
  // PIN driver connection

  // pin connection for simple resets implemented using a single pin
  pin : inout simple_port of bit is instance;
    keep bind(pin, empty);
    keep (pin.hdl_path() != "") => bind(pin, external);

  // active value of a simple single pin reset
  pin_active_value : bit;
    keep soft pin_active_value == 1;

  get_pin_state() : csco_testflow_dut_reset_state_t is {
    if(!bind(pin, empty)) {
      result = (pin$ == pin_active_value) ? ACTIVE : INACTIVE;
    };
  };

  //----------------------------------------
  // REG_* driver register field connection

  // reset register entry name
  entry : string;

  // field within register entry which control this reset
  field : string;

  // active value of reset field
  field_active_value : bit;

  //----------------------------------------
  // VIRTUAL driver 

  virtual_parent : csco_testflow_dut_reset_u;
    keep soft virtual_parent == NULL;

  // testflow unit doesn't keep is_reset_root around,
  // so we make a persistent local copy
  !is_root : bool;

  run() is also {
    is_root = csco_get_testflow_parent().csco_get_testflow_unit().is_reset_root();
  };

  // check if this unit or it's virtual parent are the root of reset
  is_root() : bool is {
    result = is_root || ((virtual_parent != NULL) && virtual_parent.is_root());
  };

  //----------------------------------------
  // check that current sampled state of DUT
  // matches expectations set by previous test phases
  check_dut_state() is {
    if(state != UNKNOWN) {
      var pin_state := get_pin_state();
      check ((pin_state == UNKNOWN) || (pin_state == state)) else
        dut_errorf("ERR_CSCO_TESTFLOW_UNEXPECTED_DUT_RESET_PIN_STATE\nIn test_phase %s, dut_reset for %s expects pin %s to be %s, but get_pin_state() reported %s", 
                      csco_get_cur_test_phase(), phase, pin.full_hdl_path(), state, pin_state);

      var dut_state := get_state$(me);
      check ((dut_state == UNKNOWN) || (dut_state == state)) else
        dut_errorf("ERR_CSCO_TESTFLOW_UNEXPECTED_DUT_RESET_STATE\nIn test_phase %s, dut_reset for %s expects to be in state %s, but get_state$() reported %s", 
                      csco_get_cur_test_phase(), phase, state,dut_state);
    };
  };

  // default is to set() only for reset_root (or virtual root).
  // overridden by drivers
  enable() : bool is {
    result = is_root();
  };

  // driver extensible hook
  set() @sys.any is {
    message(MEDIUM, "SET ", get_string());
    set_state$(me);
  };

  event pin_change is change(pin$)@sim ;
  event going_active is true(pin$ == pin_active_value)@pin_change; -- state is transitioning to ACTIVE
  event going_inactive is true(pin$ != pin_active_value)@pin_change ; -- state is transitioning to INACTIVE

  // testflow control. sets new state if not current
  ensure(new_state: csco_testflow_dut_reset_state_t) @sys.any is {
    // never re-apply current state
    if(state != new_state) {
      state = new_state;
      if(enable()) {
        set();
        wait delay(propagation_delay) @sim;
        var pin_state := get_pin_state();
        if((pin_state != UNKNOWN) && (pin_state != state)) {
          sync true((pin$ == pin_active_value) == (state == ACTIVE)) @pin_change;
        };
      };
    };
  };

  //----------------------------------------
  // testflow hooks.
  // in each "reset propagation" phase:
  //   - check the current DUT state
  //   - activate resets for current and later phases (if not already active)
  //   - deactivate reset from immediately previous test phase
  // also re-check current DUT state at start of next phase

  CREATE_TESTFLOW;

  csco_pre_reset_prop() @sys.any is also {
    check_dut_state();
    case(phase) {
      CSCO_PRE_RESET { ensure(ACTIVE) };
      CSCO_RESET     { ensure(ACTIVE) };
      CSCO_INIT_TEST { ensure(ACTIVE) };
    };
  };

  csco_pre_reset() @sys.any is also {
    check_dut_state();
  };

  csco_reset_prop() @sys.any is also {
    check_dut_state();
    case(phase) {
      CSCO_PRE_RESET { ensure(INACTIVE) };
      CSCO_RESET     { ensure(ACTIVE)  };
      CSCO_INIT_TEST { ensure(ACTIVE) };
    };
  };

  csco_reset() @sys.any is also {
    check_dut_state();
  };

  csco_init_test_prop() @sys.any is also {
    check_dut_state();
    case(phase) {
      CSCO_PRE_RESET { };
      CSCO_RESET     { ensure(INACTIVE) };
      CSCO_INIT_TEST { ensure(ACTIVE) };
    };
  };

  csco_init_test() @sys.any is also {
    check_dut_state();
  };

  csco_pre_test_prop() @sys.any is also {
    check_dut_state();
    case(phase) {
      CSCO_PRE_RESET { };
      CSCO_RESET     { };
      CSCO_INIT_TEST { ensure(INACTIVE) };
    };
  };

  csco_pre_test() @sys.any is also {
    check_dut_state();
  };

  //----------------------------------------
  get_string() : string is also {
    if((entry != NULL) && (driver.as_a(string) ~ "/(.+_)?REG(_.+)?/")) {
      result = append(entry, ".", field, "=", field_active_value, " ");
    };
    if((result == NULL) && (pin.hdl_path() != NULL)) {
      result = append(pin.hdl_path(), "=", get_pin_value_from_state(state), " ");
    };
    result = append(result, "[", state, "]");
  };

  get_pin_value_from_state(state:csco_testflow_dut_reset_state_t) : uint is {
    if state == ACTIVE {
      result = pin_active_value ;
    } else if state == INACTIVE {
      var pin_width := pin_active_value[..].size() ;
      result = (~pin_active_value)[pin_width-1:0] ;
    };
  };
};


//------------------------------------------------------------
// NONE driver ignores state changes

extend NONE'driver csco_testflow_dut_reset_u {
  enable() : bool is only {
    result = FALSE;  // never drive when disabled
  };
};


//------------------------------------------------------------
// PIN driver pokes reset state to DUT pin value.
// testbench is exclusive driver

extend PIN'driver csco_testflow_dut_reset_u {
  enable() : bool is only {
    result = TRUE;  // always drive pin value
  };
  set() @sys.any is also {
    pin$ = (state == ACTIVE) ? pin_active_value : ~pin_active_value;
  };
};


//------------------------------------------------------------
// VIRTUAL state changes are propagated to lower layers
// FIXME look for a case where driver wants to be VIRTUAL and PIN?

extend VIRTUAL'driver csco_testflow_dut_reset_u {
  check_dut_state() is first {
    if(is_root()) {
      // skip check if reset is not activated at this level
      return;
    };
  };
};


//------------------------------------------------------------
// REG_AUTO state is automatically reset by DUT when appropriate

extend REG_AUTO'driver csco_testflow_dut_reset_u {
  // assume any higher-level reset also resets the controlling reg field.
  // Note that 'state' is the new value that we're trying to set.
  enable() : bool is also {
    var test_phase := csco_get_cur_test_phase().as_a(int);
    var reset_phase := phase.as_a(int);
    result ||= ((state == INACTIVE) ||
                (test_phase > reset_phase)); // redundant?
  };
};


//------------------------------------------------------------
// REG_PRE_RESET state is automatically reset by DUT when appropriate

extend REG_PRE_RESET'driver csco_testflow_dut_reset_u {
  // assume higher-level PRE_RESET also resets the controlling reg field
  // Note that 'state' is the new value that we're trying to set.
  enable() : bool is also {
    result ||= ((state == INACTIVE) ||
                (csco_get_cur_test_phase() != CSCO_PRE_RESET_PROP));
  };
};

'>
