// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_testflow_main.e
//              :
// Project      : csco_testflow
//              :
// Created On   : 11/08/2005
//              :
//              :
// Description  : This file contains the main logic for extending the run()
//                phase of each unit into 7 distinct time consuming phases, and
//                for synchronizing each phase amongst all the units that 
//                participate in testflow. The 7 test phases are defined as
//                follows:
//
//                CSCO_PRE_RESET : waiting for analog circuitry such as PLLs and oscillators 
//                                 to lock and settle
//                CSCO_RESET :     digital reset for the device.
//
//                CSCO_INIT_TEST : programming the device and the testbench.
//
//                CSCO_PRE_TEST :  wait for certain initial conditions before injecting main
//                                 stimulus. 1 example would be frame/clock locking of some
//                                 protocols.
//
//                CSCO_MAIN_TEST : main stimulus injection phase.
//
//                CSCO_POST_TEST : waiting for system to quiesce.
//
//                CSCO_POST_TEST_CHECK : time consuming checks performed following stimulus
//                                       injection.
//                
//
//                Each unit that instantiates testflow using the CREATE_TESTFLOW macro
//                will have 7 method hooks corresponding to the 7 phases above:
//
//                csco_pre_reset()
//                csco_reset()
//                csco_init_test()
//                csco_pre_test()
//                csco_main_test()
//                csco_post_test()
//                csco_post_test_check()
//
//                The test phases of each unit execute in synchronization with all other
//                testflow units. i.e., a phase transition does not occur until all units
//                have executed all their test phase hooks.
//
//                The user may also raise/drop objections to the current testphase by calling
//                the following methods:
//
//                csco_raise_phase_objection()
//                csco_drop_phase_objection()
//
//                Care MUST be taken to ensure that the calls in occur matching pairs!!!
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003-2007 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

<'
package csco_testflow;

import csco_testflow/e/csco_testflow_types;


// Although not mandatory, it is HIGHLY recommended that the user use the
// CREATE_TESTFLOW macro to instantiate testflow inside their units.  
//
// This macro can be called from within any user-defined unit.  It adds an instance
// of the testflow unit and definitions of all of the testflow TCM's, which will
// be invoked via method port calls from the testflow unit instance.  The macro 
// can be used with or without a clock event specified; if a clock is specified,
// it will be used as the clocking event for the TCMs, otherwise sys.any is the 
// clocking event.
//
define <CREATE_TESTFLOW'struct_member> "CREATE_TESTFLOW[ <clock'exp>]" as {
  // Instantiate testflow unit.
  csco_testflow : csco_testflow_u is instance;
    keep csco_testflow.macro_flag == TRUE;

  csco_get_testflow_unit() : csco_testflow_u is {
    result = csco_testflow;
  };

  // NB port binding is automated using specman reflection facility
  keep bind(csco_pre_reset_prop, empty);
  keep bind(csco_pre_reset, empty);
  keep bind(csco_pre_reset_or_later, empty);
  keep bind(csco_reset_prop, empty);
  keep bind(csco_reset, empty);
  keep bind(csco_reset_or_later, empty);
  keep bind(csco_init_test_prop, empty);
  keep bind(csco_init_test, empty);
  keep bind(csco_pre_test_prop, empty);
  keep bind(csco_pre_test, empty);
  keep bind(csco_main_test, empty);
  keep bind(csco_post_test, empty);
  keep bind(csco_post_test_check, empty);

  csco_pre_reset_prop : in method_port of csco_testflow_hook_t is instance;
  csco_pre_reset : in method_port of csco_testflow_hook_t is instance;
  csco_pre_reset_or_later : in method_port of csco_testflow_hook_t is instance;
  csco_reset_prop : in method_port of csco_testflow_hook_t is instance;
  csco_reset : in method_port of csco_testflow_hook_t is instance;
  csco_reset_or_later : in method_port of csco_testflow_hook_t is instance;
  csco_init_test_prop : in method_port of csco_testflow_hook_t is instance;
  csco_init_test : in method_port of csco_testflow_hook_t is instance;
  csco_pre_test_prop : in method_port of csco_testflow_hook_t is instance;
  csco_pre_test : in method_port of csco_testflow_hook_t is instance;
  csco_main_test : in method_port of csco_testflow_hook_t is instance;
  csco_post_test : in method_port of csco_testflow_hook_t is instance;
  csco_post_test_check : in method_port of csco_testflow_hook_t is instance;

  csco_pre_reset_prop() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the PRE_RESET phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_pre_reset() @<clock'exp|sys.any> is empty;

  csco_pre_reset_or_later() @<clock'exp|sys.any> is empty;

  csco_reset_prop() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the RESET phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_reset() @<clock'exp|sys.any> is empty;

  csco_reset_or_later() @<clock'exp|sys.any> is empty;

  csco_init_test_prop() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the INIT phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_init_test() @<clock'exp|sys.any> is empty;

  csco_pre_test_prop() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the PRE_TEST phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_pre_test() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the MAIN_TEST phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_main_test() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the POST_TEST phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_post_test() @<clock'exp|sys.any> is empty;

  // The user should extend this method to add logic for the POST_TEST_CHECK phase.
  // Only valid if testflow has been instantiated inside the unit.
  csco_post_test_check() @<clock'exp|sys.any> is empty;
};


// extend utility unit (instanced under sys)
extend csco_testflow_utils_u {
  // holds a list of all the testflow root units under sys.
  package !testflow_roots : list of any_unit;
};

extend sys {
  run() is also {
    #ifndef SPECMAN_VERSION_4_2_1_OR_LATER {
      assert FALSE else 
        error("testflow package does not support Specman versions earlier than 4.2.1\n");
    };
  };
}; // extend sys


// This unit should be instantiated inside of any unit that wishes
// to participate in testflow.
unit csco_testflow_u {
  /////////////////////////////////
  // User Writable Fields
  /////////////////////////////////
  // Short descriptive name used in message actions.
  short_name() : string is only {
    result = "TESTFLOW";
  };

  // By default, all units belong to same GLOBAL test flow thread. If the user
  // constrains a value other than 'GLOABL' for this field, then each sub-unit
  // under the enclosing unit will belong to the new testflow domain, unless
  // a different explicit value is specified for the sub-unit.
  domain : csco_testflow_domain_t;
    keep soft domain == CSCO_GLOBAL;

  // NB individual timer fields are deprecated,
  // use get/set_phase_timer() instead

  // Expiration timer for PRE_RESET phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  pre_reset_timer : time;
    keep pre_reset_timer > 0;
    keep soft pre_reset_timer == 10 us;

  // Expiration timer for RESET phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  reset_timer : time;
    keep reset_timer > 0;
    keep soft reset_timer == 50 us;

  // Expiration timer for INIT_TEST phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  init_test_timer : time;
    keep init_test_timer > 0;
    keep soft init_test_timer ==  200 us;

  // Expiration timer for PRE_TEST phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  pre_test_timer : time;
    keep pre_test_timer > 0;
    keep soft pre_test_timer ==  200 us;

  // Expiration timer for MAIN_TEST phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  main_test_timer : time;
    keep main_test_timer > 0;
    keep soft main_test_timer ==  1 ms;

  // Expiration timer for POST_TEST phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  post_test_timer : time;
    keep post_test_timer > 0;
    keep soft post_test_timer == 50 us;

  // Expiration timer for POST_TEST_CHECK phase.
  // Note that this counter is only used for units who are either top-level
  // units or units who may be reset. i.e., any unit who might become the root
  // of their testflow hierarchy.
  post_test_check_timer : time;
    keep post_test_check_timer > 0;
    keep soft post_test_check_timer == 10 us;

  post_generate() is also {
    set_phase_timer(CSCO_PRE_RESET, pre_reset_timer);
    set_phase_timer(CSCO_RESET, reset_timer);
    set_phase_timer(CSCO_INIT_TEST, init_test_timer);
    set_phase_timer(CSCO_PRE_TEST, pre_test_timer);
    set_phase_timer(CSCO_MAIN_TEST, main_test_timer);
    set_phase_timer(CSCO_POST_TEST, post_test_timer);
    set_phase_timer(CSCO_POST_TEST_CHECK, post_test_check_timer);
  };

  // list of expiration timeout values for each phase, indexed by phase enum.
  // Note that these timers are only used for units who are either top-level
  // units or units who may be reset. i.e., any unit that becomes the root
  // of its testflow hierarchy
  phase_timers : list of time;
    keep phase_timers.size() == read_only(all_values(csco_testflow_stage_t).size());
    keep for each in phase_timers {
      soft it == 10 us;
    };

  // set new timeout value for a phase
  set_phase_timer(phase: csco_testflow_stage_t, timer: time) is {
    phase_timers[phase.as_a(int)] = timer;
  };

  // retrieve current timeout value for a phase
  get_phase_timer(phase: csco_testflow_stage_t) : time is {
    result = phase_timers[phase.as_a(int)];
  };

  // The user may override this method with 'is only' to impose any special 
  // ordering/conditions while launching the test phase hooks of the child units.
  launch_child_run_phases(stage : csco_testflow_stage_t, obj : csco_testflow_objection_kind_t) is {
    for each (child) in get_parent_unit().csco_get_testflow_children() do {
      child.csco_check_testflow_unit().run_phase(stage, obj);
    };
  };

  check_current_objections() is {
    print_objecting_units(get_current_objecting_units()) ;
  };


  get_current_objecting_units() : list of any_unit is {
    var testflow_root : any_unit ;
    testflow_root = sys.csco_testflow_utils.testflow_roots.first(.csco_get_testflow_domain() == domain);
    for each (u) in {testflow_root; testflow_root.csco_get_all_testflow_children()} {
      if u.get_objection_counter(cur_test_objection) + u.csco_get_testflow_unit().get_objection_counter(cur_test_objection) > 0 {
        result.add(u);
      };
    };
  };


  // method to print which units are preventing testflow from advancing from the current phase
  print_objecting_units(objecting_units: list of any_unit) is {
    if objecting_units.is_empty() {
      out("No units are objecting to end of current phase ",cur_test_phase," in domain ",domain) ;
    } else {
      out("Following units are currently objecting to end of phase ", cur_test_phase," in domain ",domain);
      out("\n");
      out("Obj. Cnt Unit");
      out("-------- ----");
      var obj_cnt : uint ;
      for each (u) in objecting_units {
        obj_cnt = u.get_objection_counter(cur_test_objection) + u.csco_get_testflow_unit().get_objection_counter(cur_test_objection) ;
        outf("%7d  %s\n", obj_cnt, u.e_path());
      };
    };
    out("\n");
  };
    

  /////////////////////////////////
  // User Read-Only Fields
  /////////////////////////////////
  // Event hook for test phase change. One use would be for coverage collection.
  event test_phase_change;

  is_past_phase(min_phase:csco_testflow_stage_t) : bool is {
    result = (cur_test_phase.as_a(uint) > min_phase.as_a(uint)) ;
  };

  // Returns the # of times this unit has been reset, not including the initial 
  // power-on reset. It can be used to code different behavior for the same do
  // hooks depending on the # of times the unit has been reset.
  get_rerun_counter() : uint is {
    result = rerun_counter;
  };

  // Returns the current reset type of the unit. POWER_ON vs PRE_RESET vs RESET 
  // vs INIT_TEST.
  get_reset_kind() : csco_testflow_reset_kind_t is {
    result = reset_kind;
  };

  run() is also {
    if get_reset_kind() == CSCO_POWER_ON and testflow_parent == sys { is_reset_root = TRUE };
  };

  is_reset_root() : bool is {
    result = is_reset_root;
  };

  // method port to be connected to method describing behavior of each phase.
  // If the user uses the CREATE_TESTFLOW macro to instantiate testflow
  // for the enclosing unit, then the port will be automatically
  // connected to a method hook of the same name in the enclosing unit, 
  // created by the macro.
  do_phases : list of out method_port of csco_testflow_hook_t is instance;
    keep do_phases.size() == read_only(all_values(csco_testflow_stage_t).size());
    keep for each in do_phases { bind(it, empty); };

  // special case in addition to normal "do_phases"
  do_pre_reset_or_later : out method_port of csco_testflow_hook_t is instance;
    keep bind(do_pre_reset_or_later, empty);

  // special case in addition to normal "do_phases"
  do_reset_or_later : out method_port of csco_testflow_hook_t is instance;
    keep bind(do_reset_or_later, empty);

  // method to return whether this unit and all its testflow parents are enabled
  is_enabled() : bool is {
    result = (enable == TRUE) and (branch_is_disabled == FALSE) ;
  };

  // method to change enable value; has side effect of propagating value of branch_is_disabled
  set_enable(e:bool) is {
    enable = e ;
    for each (child_unit) in get_parent_unit().csco_get_testflow_children() {
      child_unit.csco_check_testflow_unit().set_branch_is_disabled(not e);
    };
  };


  /////////////////////////////////
  // Private Fields
  /////////////////////////////////
  macro_flag : bool;
    keep soft macro_flag == FALSE;

  // Testflow may be instantiated, but not enabled.
  package enable : bool;
    keep soft enable == TRUE;

  // testflow can be disabled for this whole branch of the testflow hierarchy, based on enable
  // value of parent
  package !branch_is_disabled : bool ;

  // Current test phase the enclosing unit is in.
  package cur_test_phase : csco_testflow_stage_t;
    keep cur_test_phase == UNDEFINED;

  // Objection type for current test phase.
  package cur_test_objection : csco_testflow_objection_kind_t;
    keep cur_test_objection == CSCO_PRE_RESET_DONE;

  // Indicates whether the unit is the top testflow unit for the testbench environment.
  package !is_top_flow : bool;

  // Closest parent unit to this unit in the same testflow domain.
  package !testflow_parent : any_unit;

  // Pointers to child units whose phase hooks will be launched by this unit
  package !testflow_children : list of any_unit;

  // This field is a means for the reset_unit() method to tell run() which reset has been
  // requested. This is so that the RESET sequences will be able to launch the appropriate
  // logic based on the type of reset that's occuring. POWER_ON is the intial value.
  package !reset_kind : csco_testflow_reset_kind_t;

  // This is a flag used by reset_unit() to indicate to run() that a reset_unit() had 
  // been called for this unit. When the run() method sees that this flag is set, it 
  // will launch the do_testflow() method of the unit to kick off the test flow again.
  // Note that during a dynamic reset, only the top-unit being reset will have this 
  // flag set.
  package !is_reset_root : bool;

  // Indicates that rerun()/quit() has been called for the current tick and that a 
  // 'quit' event will be emitted at the end of the tick to kill all the TCMs.
  package !is_being_rerun : bool;

  // This is used by do_testflow() after a local reset to store the phase at which
  // the test flow will re-merge with the top level test flow.
  package !join_phase : csco_testflow_stage_t;

  // utility field allowing such behavior as limiting the number of times a unit is rerun
  // during one simulation
  package !rerun_counter : uint;

  // This event is used by do_testflow() to determine when to transition to the next
  // test stage. Keep in mind that do_testflow is only invoked by run() if this is the
  // top level unit or if reset_unit() has been called to reset this unit.
  package event phase_objections_dropped;

  // auto-detect (using reflection) whether a hook has been
  // implemented/extended by the parent struct and, if so,
  // bind to specified internal hook port
  package connect_hook(parent: any_unit, parent_type: rf_struct, hook: string,
                       out_port: out method_port of csco_testflow_hook_t) is
  {
    var hook_m := parent_type.get_method(hook);
    var hook_f := parent_type.get_field(hook);
    if((hook_m != NULL) && (hook_f != NULL) &&
       (hook_f.get_type().get_name() == "in method_port of csco_testflow_hook_t") &&
       (hook_m.get_relevant_layers(parent_type)
          .count(.get_extension_mode() not in [empty, undefined, none]) > 0))
    {
      var hook_v := hook_f.get_value(parent);
      var in_port : in method_port of csco_testflow_hook_t;
      in_port = hook_v.get_value().unsafe();
      if(in_port != NULL) {
        do_bind(in_port, out_port);
      };
    };
  };

  connect_ports() is also {
    var parent := get_parent_unit();
    var parent_type := rf_manager.get_exact_subtype_of_instance(parent);
    for each (phase) in all_values(csco_testflow_stage_t) {
      connect_hook(parent, parent_type, str_lower(phase.as_a(string)),
                   do_phases[phase.as_a(int)]);
    };
    connect_hook(parent, parent_type, "csco_pre_reset_or_later",
                 do_pre_reset_or_later);
    connect_hook(parent, parent_type, "csco_reset_or_later",
                 do_reset_or_later);
  };

  // This method kicks off the testflow for the unit sub-tree with the enclosing unit as
  // the root. The method handles the transitions from 1 phase to the next by
  // monitoring the phase_objections_dropped event, which is emitted by
  // all_objections_dropped().
  //
  // When transitioning from 1 phase to the next, a check is made to ensure that if the
  // unit is not the top testflow unit, the testflow will not be advanced beyond the
  // current phase of the top-level testflow. This allows us to merge the testflow of a
  // unit that has been locally reset with that of the overall environment.
  package do_testflow() @sys.any is {
    outf("\t\n");
    messagef(CSCO_TESTFLOW, NONE, "kicking off %s testflow\n", domain);

    if is_top_flow {
      messagef(NONE, "==================================================================\n");
      messagef(NONE, "Testflow Setup for %s Domain with %s as Top-level Unit\n", domain, get_parent_unit().short_name());
      messagef(NONE, "==================================================================\n");
      messagef(NONE, "simulator              = %s\n", agent());
      messagef(NONE, "current time           = %s\n", sys.time);
      messagef(NONE, "rerun_counter          = %d\n", rerun_counter);
      for each (phase) in all_values(csco_testflow_stage_t) {
        if(phase != UNDEFINED) {
          messagef(NONE, "phase_timers[%s]\t= %s\n",
                   phase, get_phase_timer(phase));
        };
      };

      if rerun_counter == 0 {
        // objections to TEST_DONE are raised and dropped in the 'sys' unit because the local
        // unit may be rerun() and therefore its objections automatically dropped, which is
        // not what we want.
        sys.raise_objection(TEST_DONE);
      };
    };
    if !is_top_flow {
      join_phase = testflow_parent.csco_get_cur_test_phase();
      messagef(CSCO_TESTFLOW, NONE, "will re-merge with top-level test flow at %s phase\n", join_phase);
    };

    var start_phase : csco_testflow_stage_t;
    if(reset_kind == CSCO_INIT_TEST) {
      start_phase = CSCO_INIT_TEST_PROP;
    }
    else if(reset_kind == CSCO_RESET) {
      start_phase = CSCO_RESET_PROP;
    }
    else {
      start_phase = CSCO_PRE_RESET_PROP;
    };

    for each (phase) in all_values(csco_testflow_stage_t) {
      if(start_phase.as_a(int) <= phase.as_a(int)) {
        do_testflow_stage(phase);
        if(!is_top_flow and join_phase == phase) {
          message(CSCO_TESTFLOW, NONE,
                  "ending local test flow in ", phase,
                  " to merge with parent test flow, domain=%s\n", domain);
          break;
        };
      };
    };

    if is_top_flow {
      sys.drop_objection(TEST_DONE);
      messagef(NORMAL, NONE, "%s test flow domain has finished\n", domain);
    };
  }; // do_testflow()

  // Executes the specified testflow stage by hierarchically launching the test phase wrappers
  // of the unit tree and waiting until all objections to test phase in unit tree have been 
  // dropped.
  package do_testflow_stage(stage : csco_testflow_stage_t) @sys.any is {
    var obj := append(stage, "_DONE").as_a(csco_testflow_objection_kind_t);
    var timer := get_phase_timer(stage);

      outf("\t\n");
      messagef(NORMAL, LOW, "***********************************************************************************\n");
      messagef(NORMAL, LOW, "%s Domain advancing to %s phase\n", domain, stage);
      messagef(NORMAL, LOW, "***********************************************************************************\n");

      first of {
        {
          run_phase(stage, obj);
          wait_on_cur_test_phase();
        };

        {
          messagef(CSCO_TESTFLOW, MEDIUM, "starting %s countdown for %s phase\n", timer, stage);
          if get_parent_unit().agent() == "NULL_SIM" {
            wait true(sys.time == timer);
          }
          else {
            wait delay(timer);
          };

          check_current_objections() ;
          dut_errorf("%s's %s %s watchdog timer expired!!!! timeout value=%s\n", get_parent_unit().e_path(), domain, stage, timer);
        };
      }; // first of

      // If the unit is currently being rerun, then just wait and this thread will eventually
      // be killed. The wait is necessary to prevent conflict between the normal test flow and
      // the reset.
      if is_being_rerun {
        messagef(CSCO_TESTFLOW, LOW, "%s testflow under %s is being reset, will wait to be restarted\n", domain, get_parent_unit().short_name());
        wait true(is_being_rerun == FALSE);
      };
  }; // do_testflow_stage()

  // Wait until all objections to the current test phase for this unit and all its sub-units
  // have been dropped. 
  package wait_on_cur_test_phase() @sys.any is {
    messagef(CSCO_TESTFLOW, HIGH, "wait_on_cur_test_phase() called during %s phase, domain=%s\n", cur_test_phase, domain);
    assert that cur_test_phase != UNDEFINED else
      error("should not call wait_on_cur_test_phase() when cur_test_phase is UNDEFINED");


    repeat {
      sync cycle @phase_objections_dropped;
      messagef(CSCO_TESTFLOW, LOW, "all objections to %s in %s domain have been dropped\n", cur_test_objection, domain);
      messagef(CSCO_TESTFLOW, LOW, "waiting for 2 cycles of sys.new_time as hysteresis\n");

      // The extra hysteresis wait is to account for phase transition conditions,
      // where an empty method hook will drop the phase objection in the same cycle as 
      // the phase transition. This may cause the <phase>_objections_dropped event to be
      // emitted. If a different thread then subsequently raises an objection in the
      // same cycle, it will cause a race condition with the state transition without
      // the hysteresis delay.
      wait [2]@sys.new_time;
      if get_parent_unit().csco_get_testflow_objection_total(cur_test_objection) != 0 {
        messagef(CSCO_TESTFLOW, LOW, "new objections to %s raised during hysteresis\n", cur_test_objection);
      };
    } until get_parent_unit().csco_get_testflow_objection_total(cur_test_objection) == 0 ;

    messagef(CSCO_TESTFLOW, HIGH, "wait for %s has finished\n", cur_test_objection);
  }; // wait_on_cur_test_phase() 

  // Launch the corresponding user test phase hooks of the unit. A check is made to ensure
  // that a rerun() call is not in effect at the same time as the test phase transition. 
  // If it is, then no more logic is executed for this method, effectively terminating it. 
  // This is to prevent the method from launching more do_wrapper() methods that might start 
  // run away sequences, which will not be killed by rerun() because they were started after 
  // the rerun() call, albeit in the same tick.
  //
  run_phase(stage : csco_testflow_stage_t, obj : csco_testflow_objection_kind_t) is {
    messagef(CSCO_TESTFLOW, MEDIUM, "entering %s phase\n", stage);

    if is_enabled() == FALSE {
      messagef(CSCO_TESTFLOW, LOW, "%s testflow under %s is disabled, dropping objection\n", domain, get_parent_unit().short_name());
      if get_parent_unit().get_objection_counter(obj) != 0 {
        get_parent_unit().drop_objection(obj);
      };
      launch_child_run_phases(stage, obj);
      return;
    };
    if !is_being_rerun {
      cur_test_phase = stage;
      cur_test_objection = obj;
      emit test_phase_change;

      // call the run_phase() methods of my testflow child units. Note that the fact launch_chil_run_phases()
      // is not a TCM guarantees that the 'cur_test_phase' field of all units in the test flow get updated
      // as an atomic step. This prevents the case where a new sequence thread might raise the wrong objection
      // if it gets started in the same tick as a phase transition.
      launch_child_run_phases(stage, obj);

      // Launch the appropriate do_wrapper() for this unit.
      // couple special cases for "_or_later" method variants
      var m : out method_port of csco_testflow_hook_t;
      var phase_methods : list of out method_port of csco_testflow_hook_t;
      if((stage == CSCO_PRE_RESET ||
          (stage == CSCO_RESET_PROP && get_reset_kind() == CSCO_RESET) ||
          (stage == CSCO_INIT_TEST_PROP && get_reset_kind() == CSCO_INIT_TEST)) &&
         do_pre_reset_or_later.is_connected())
      {
        phase_methods.add(do_pre_reset_or_later);
      };
      if((stage == CSCO_RESET ||
          (stage == CSCO_INIT_TEST_PROP && get_reset_kind() == CSCO_INIT_TEST)) &&
         do_reset_or_later.is_connected()) {
        phase_methods.add(do_reset_or_later);
      };
      m = do_phases[stage.as_a(int)];
      if(m.is_connected()) { phase_methods.add(m) };

      if(!phase_methods.is_empty()) {
        start do_wrapper(phase_methods);
      }
      else {
        get_parent_unit().csco_drop_phase_objection();
      };
    }
    else {
      messagef(CSCO_TESTFLOW, MEDIUM, "run_phase() method will be skipped because unit is being rerun()\n");
    };
  }; // run_phase()

  // Method to call the appropriate method output port and
  // drop the objection for the specified phase
  package do_wrapper(phase_methods: list of out method_port of csco_testflow_hook_t) @sys.any is {
    // Call the user hook to launch user defined phase logic
    for each (m) in phase_methods {
      m$();
    };
    get_parent_unit().csco_drop_phase_objection();
  };

  // method to write branch_is_disabled; side effect is to propagate down the hierarchy
  package set_branch_is_disabled(d:bool) is {
    branch_is_disabled = d ;
    if enable == TRUE {  
      propagate_branch_is_disabled() ;
    };
  };

  // method to set branch_is_disabled for all children based on whether this level is enabled
  package propagate_branch_is_disabled() is {
    for each (child_unit) in get_parent_unit().csco_get_testflow_children() {
      child_unit.csco_check_testflow_unit().set_branch_is_disabled(not is_enabled());
    };
  };

}; // unit csco_testflow_u

extend TRUE'is_top_flow csco_testflow_u {
  dut_error_dump() is also {
    var obj_cnt : uint; 
    var total_cnt : uint;
    var obj_units : list of any_unit;

    obj_units = get_current_objecting_units() ;

    print_objecting_units(obj_units) ;

    for each (u) in obj_units {
      u.dut_error_dump();
    };

    if obj_units.is_empty() {
      out(cur_test_phase, " timed out without any outstanding objections, ",
          "this means that the phase timed out during 2 sys.new_time cycles of hysteresis.",
          " There are 2 possibilities:");
      out("1 - The watchdog timer for ", cur_test_phase, 
          " needs to be extended, since it falls so closely to the hystersis wait.");
      outf("\n");
      out("2 - There are simply no (or not enough) sys.new_time events firing during the hysteresis",
          "make sure there's a clock running during the test phase.");
    };
  };
};


// Add a regenerate_data phase to the rerun() sequence
// for non-driver routines.
extend any_unit {
  /////////////////////////////////
  // User Writable Fields
  /////////////////////////////////
  // This method MUST be extended by the calling unit to return the
  // instantiated testflow struct. If the CREATE_TESTFLOW macro is
  // used, then this is done automatically for the user.
  csco_get_testflow_unit() : csco_testflow_u is {
    result = NULL;
  };

  // Method hook to regenerate config/testbench fields on a reset. This
  // method is called after reset_unit() is called and before the unit's
  // run() method is called again.
  csco_regenerate_unit_data() is {
    messagef(CSCO_TESTFLOW, MEDIUM, "regenerate_unit_data() is being called\n");
  };


  /////////////////////////////////
  // User Readable Fields
  /////////////////////////////////
  // Indicates whether this unit has instantiated testflow.
  csco_has_testflow() : bool is {
    result = (csco_get_testflow_unit() != NULL);
  };

  // Returns the testflow domain to which the unit belongs.
  // Only valid if the unit has instantiated testflow.
  csco_get_testflow_domain() : csco_testflow_domain_t is {
    assert that me is not a csco_testflow_u else
      error(appendf("csco_get_testflow_domain() should be called from a functional unit, not its testflow unit\n"));

    result = csco_check_testflow_unit().domain;
  };

  // Returns the current test phase of the unit's testflow domain.
  // Only valid if the unit has instantiated testflow.
  csco_get_cur_test_phase() : csco_testflow_stage_t is {
    assert that me is not a csco_testflow_u else
      error(appendf("csco_get_cur_test_phase() should be called from a functional unit, not its testflow unit\n"));

    return csco_check_testflow_unit().cur_test_phase;
  };

  // Returns the current test phase objection kind of the unit's testflow domain.
  // Only valid if the unit has instantiated testflow.
  csco_get_cur_test_objection() : csco_testflow_objection_kind_t is {
    assert that me is not a csco_testflow_u else
      error(appendf("csco_get_cur_test_objection() should be called from a functional unit, not its testflow unit\n"));

    assert that csco_get_cur_test_phase() != UNDEFINED else
      error("should not call csco_get_cur_test_objection() when cur_test_phase is UNDEFINED");

    result = csco_check_testflow_unit().cur_test_objection;
  };

  // Raises an objection to the current test phase. Can be used by user defined logic outside
  // of the testflow method hooks to extend current testflow. MUST BE MATCHED BY A CORRESPONDING
  // CALL TO drop_phase_objection().
  // Only valid if the unit has instantiated testflow.
  csco_raise_phase_objection() is {
    var cur_objection : csco_testflow_objection_kind_t;

    assert that me is not a csco_testflow_u else
      error(appendf("csco_raise_phase_objection() should be called from a functional unit, not its testflow unit\n"));

    assert that  csco_get_cur_test_phase() != UNDEFINED else
      error(appendf("cannot call csco_raise_phase_objection() during %s phase\n", csco_get_cur_test_phase()));

    assert that csco_check_testflow_unit().is_being_rerun == FALSE else
      error(appendf("cannot call csco_raise_phase_objection() during a reset\n"));

    raise_objection(csco_get_cur_test_objection());
  }; 

  // Drops an objection to the current testflow. Used in conjunction with csco_raise_phase_objection.
  // It's possible, due to resets, that there's no outstanding objection when this method is
  // invoked. Checks are in place to ensure no errors are generated in that case.
  // Only valid if the unit has instantiated testflow.
  csco_drop_phase_objection() is {
    assert that me is not a csco_testflow_u else
      error(appendf("csco_drop_phase_objection() should be called from a functional unit, not its testflow unit\n"));

    if csco_get_cur_test_phase() == UNDEFINED {
      assert that csco_check_testflow_unit().is_being_rerun else
        error(appendf("cannot call csco_drop_phase_objection() during UNDEFINED phase\n"));
    }
    else if get_objection_counter(csco_get_cur_test_objection()) != 0 {
      drop_objection(csco_get_cur_test_objection());
      messagef(CSCO_TESTFLOW, HIGH, "%s dropping objection to %s\n", short_name(), csco_get_cur_test_phase());
    }
    else {
      if csco_check_testflow_unit().get_rerun_counter() == 0 {
        assert FALSE else
          error(appendf("csco_drop_phase_objection() called, but no objection to %s is outstanding and %s has not been reset\n", csco_get_cur_test_phase(), short_name()));
      }
      else {
        messagef(CSCO_TESTFLOW, HIGH, "no need to drop %s objection for %s because none is outstanding\n", short_name(), csco_get_cur_test_phase());
      };
    };
  }; 

  // This method is used to do a reset of the enclosing unit and its testflow
  // sub-units. do_testflow() will be launched by run() following rerun(). It will
  // also automatically re-merge the local testflow with the system test flow.
  csco_reset_unit(kind : csco_testflow_reset_kind_t) is {
    var err_str : string;

    assert(me is not a csco_testflow_u) else
      error("ERR_CSCO_TESTFLOW_INVALID_RESET_UNIT\n",
            "reset_unit() should be called from a functional unit, not its testflow unit");

    assert(kind != CSCO_POWER_ON) else 
      error("ERR_CSCO_TESTFLOW_INVALID_RESET_PHASE\n",
            kind, " is an invalid argument to reset_unit()");

    var start_phase := kind.as_a(string).as_a(csco_testflow_stage_t);
    assert(start_phase.as_a(int) < csco_get_cur_test_phase().as_a(int)) else
      error("ERR_CSCO_TESTFLOW_UNEXPECTED_RESET_PHASE\n",
            "cannot do a reset back to ", start_phase, " phase\n",
            "while unit is in ", csco_get_cur_test_phase() , " phase\n",
            "for unit at: ", e_path());

    outf("\t\n");
    outf("\t\n");
    messagef(NONE, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    messagef(NONE, "initiating %s reset for %s and all its child units...\n", kind, short_name()); 
    messagef(NONE, "domain=%s  cur_test_phase=%s\n",  csco_get_testflow_domain(), csco_get_cur_test_phase());
    messagef(NONE, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    outf("\n");

    // Set is_reset_root field so that the run() method will kick off the do_testflow()
    // method following rerun().
    csco_check_testflow_unit().is_reset_root = TRUE;
    csco_check_testflow_unit().reset_kind = kind;

    rerun();
  }; // csco_reset_unit()


  /////////////////////////////////
  // Private Fields
  /////////////////////////////////
  // Call rerun for all the units whose test phases are launched by this unit.
  package csco_rerun_child_units() is {
    for each (child) in csco_get_testflow_children() do {
      if child.csco_has_testflow() {
        child.csco_check_testflow_unit().reset_kind = csco_check_testflow_unit().reset_kind;
      };
      child.rerun();
    };
  }; // csco_rerun_child_units()

  // Method to return testflow unit if it's instantiated, or an error otherwise.
  package csco_check_testflow_unit() : csco_testflow_u is {
    result = csco_get_testflow_unit();
    assert that result != NULL else
      error(appendf("testflow for %s has not been defined", short_name()));
  };

  package csco_get_testflow_children() : list of any_unit is {
    result = csco_check_testflow_unit().testflow_children;
  };

  package csco_get_testflow_parent() : any_unit is {
    result = csco_check_testflow_unit().testflow_parent;
  };

  package csco_get_testflow_root() : any_unit is {
    if csco_has_testflow() and csco_check_testflow_unit().is_enabled() {
      result = sys.csco_testflow_utils.testflow_roots.first(.csco_get_testflow_domain() == csco_get_testflow_domain());
    };

    assert that result != NULL else
      error(appendf("can not find testflow root for unit: %s\n", short_name()));
  }; 

  package csco_get_all_testflow_children() : list of any_unit is {
    if csco_has_testflow() and csco_check_testflow_unit().is_enabled() {
      result.add(csco_get_testflow_children());
      for each (c) in csco_get_testflow_children() {
        result.add(c.csco_get_all_testflow_children());
      };
    };
  };

  package csco_get_testflow_objection_total(obj : csco_testflow_objection_kind_t) : uint is {
    if csco_has_testflow() and csco_check_testflow_unit().is_enabled() {
      for each (u) in {me; csco_get_all_testflow_children()} {
        if u.csco_check_testflow_unit().is_enabled() {
          result += u.get_objection_counter(obj);
        };
      };
    };
  };

  // run() kicks off the do_testflow() method of the unit if it is the top
  // level unit, or if run() is being invoked as a result of a run-time reset
  // on the unit.
  run() is also {
    var parent_unit : any_unit;
    var children_list : list of any_unit;
    var tu : csco_testflow_u;

    if csco_has_testflow() {
      tu = csco_check_testflow_unit();

      if tu.macro_flag == FALSE {
        messagef(CSCO_TESTFLOW, NONE, "WARNING: %s's testflow was NOT instantiated via the CREATE_TESTFLOW macro\n", short_name());
      };

      if not tu.is_being_rerun {
        // Determine which testflow domain the unit is on.
        parent_unit = get_parent_unit();
        while parent_unit != NULL {
          if parent_unit.csco_has_testflow() {
            if tu.domain == CSCO_GLOBAL {
              tu.domain = parent_unit.csco_get_testflow_domain();
              break;
            }
            else if tu.domain == parent_unit.csco_get_testflow_domain() {
              break;
            };
          };

          parent_unit = parent_unit.get_parent_unit();
        };

        tu.is_top_flow = (parent_unit == NULL);
        tu.is_reset_root = tu.is_top_flow;

        if not tu.is_top_flow {
          // Add myself to the parent unit who is going to launch my test phase hooks. The top-level
          // unit of the environment doesn't have a parent test flow unit. Instead, its do_testflow()
          // method is invoked by run() on start-up to launch the test flow for the whole testbench.
          //
          messagef(CSCO_TESTFLOW, LOW, "adding %s to the testflow children list of %s\n", me, parent_unit);
          tu.testflow_parent = parent_unit;
          children_list = parent_unit.csco_get_testflow_children();
          children_list.add(me);

          parent_unit.csco_check_testflow_unit().propagate_branch_is_disabled(); --learn if this testflow branch is disabled
        }
        else {
          sys.csco_testflow_utils.testflow_roots.add(me);
        };
      }; // if not is_being_rerun

      messagef(CSCO_TESTFLOW, LOW, "domain=%s  is_top_flow=%s  is_reset_root=%s  reset_kind=%s\n", tu.domain, tu.is_top_flow, tu.is_reset_root, tu.reset_kind);
      var start_obj : csco_testflow_objection_kind_t;
      if(tu.reset_kind == CSCO_INIT_TEST) {
        start_obj = CSCO_INIT_TEST_PROP_DONE;
      }
      else if(tu.reset_kind == CSCO_RESET) {
        start_obj = CSCO_RESET_PROP_DONE;
      }
      else {
        start_obj = CSCO_PRE_RESET_PROP_DONE;
      };

      message(CSCO_TESTFLOW, LOW, "run() started, raising objection ",
              start_obj, " and all subsequent phase objections");
      for each (obj) in all_values(csco_testflow_objection_kind_t) {
        if(start_obj.as_a(int) <= obj.as_a(int)) {
          raise_objection(obj);
        };
      };

      // run() is executed at the next sys.new_time after rerun() is called. So at this point,
      // the unit is no longer being rerun().
      tu.is_being_rerun = FALSE;

      // do_testflow is only started for a unit if it is the top level test flow unit or if the
      // unit (and its child units) are being rerun.
      if tu.is_top_flow or tu.is_reset_root {
        start tu.do_testflow();
      };
    }; // if csco_has_testflow() and csco_check_testflow_unit().is_enabled()
  }; // run() is only

  rerun() is first {
    if csco_has_testflow() {
      var tu : csco_testflow_u = csco_check_testflow_unit();

      messagef(NORMAL, HIGH, "resetting %s\n", short_name());
      tu.rerun_counter += 1;

      // Update the cur_test_phase of all the units about to be reset to UNDEFINED. User code should
      // not start a thread in a unit whose cur_tes_phase is UNDEFINED. The user should NEVER start
      // a thread in a unit whose 'is_being_rerun' is set to TRUE.
      tu.cur_test_phase = UNDEFINED;
      tu.is_being_rerun = TRUE;

      // Rerun calls are not automatically propagated to child-units. So we need to
      // call rerun for all the test flow child units of this unit.
      tu.rerun();
      csco_rerun_child_units();
    }; // if csco_has_testflow()
  }; // rerun() 

  quit() is also {
    if csco_has_testflow() {
      messagef(CSCO_TESTFLOW, FULL, "quit() is being called: csco_has_testflow=%s, is_being_rerun=%s\n", csco_has_testflow(), 
                                      csco_check_testflow_unit().is_being_rerun);

      // Make sure regenerate_unit_data is only called during a rerun()
      if csco_has_testflow() and csco_check_testflow_unit().is_being_rerun {
        csco_regenerate_unit_data();
      };
    };
  }; // quit() 

  // Extend drop_objection() method to check whether the sub-tree under the unit
  // has dropped all its objections to the given test phase.
  drop_objection(kind : objection_kind) is also {
    if csco_has_testflow() {
      if kind.as_a(csco_testflow_objection_kind_t) in all_values(csco_testflow_objection_kind_t) {
        csco_check_objections(kind.as_a(csco_testflow_objection_kind_t));
      };
      // don't need TRUE is_reset_root value (if present) any more
      csco_check_testflow_unit().is_reset_root = FALSE ;
    };
  }; // drop_objection()


  package csco_check_objections(obj : csco_testflow_objection_kind_t) is {
    if csco_get_testflow_objection_total(obj) == 0 {
      emit csco_check_testflow_unit().phase_objections_dropped;

      if csco_get_testflow_parent() != NULL {
        csco_get_testflow_parent().csco_check_objections(obj);
      };
    };
  };
}; // extend any_unit


// Make started sequences raise objection to the current test phase when
// it is launched, and drop the objection when it finishes.
extend any_sequence {
  // This field should only be constrained at gen time, it should NEVER NEVER NEVER
  // be modified during the execution of a sequence.
  csco_infinite_len : bool;
    keep soft !csco_infinite_len;
    keep parent_sequence != NULL => !csco_infinite_len;

  // This field is only meaningful is csco_infinite_len is TRUE. In that case, the
  // user may constrain this field to indicate the phase upon which the sequence
  // should be aborted. By default, the field is constrained to POST_TEST_CHECK,
  // which means the csco_infinite_len sequence will run until the driver enters the 
  // POST_TEST_CHECK phase.  However, the user may change it to an earlier phase.
  csco_abort_phase : csco_testflow_stage_t;
    keep soft csco_abort_phase == CSCO_POST_TEST_CHECK;

  package !csco_objection_raised : bool;

  pre_body() @sys.any is also {
    var punit := get_driver();

    if punit.csco_has_testflow() {
      if punit.csco_get_testflow_unit().is_being_rerun {
        messagef(CSCO_TESTFLOW, LOW, "%s sequence from %s is waiting to be terminated because %s is being reset\n", 
                                      nice_string(), punit.short_name(), punit.short_name());
        wait @punit.csco_get_testflow_unit().test_phase_change;
      };

      assert that punit.csco_get_cur_test_phase() != UNDEFINED else
        error(appendf("sequence (%s) on %s is being started outside of test flow\n", nice_string(), punit.short_name()));

      if !csco_infinite_len {
        messagef(CSCO_TESTFLOW, HIGH, "%s sequence from %s raising objection to %s\n", nice_string(), punit.short_name(), punit.csco_get_cur_test_phase());
        punit.csco_raise_phase_objection();
        csco_objection_raised = TRUE;
      }
      else {
        assert parent_sequence == NULL else
          error(appendf("CSCO_TESTFLOW_ERR_DO_INF_LEN: %s on %s has csco_infinite_len=TRUE, but it is not a new thread", nice_string(), punit.short_name()));

        messagef(CSCO_TESTFLOW, LOW, "%s sequence on %s will not raising objection to %s because it's infinite\n", nice_string(), punit.short_name(), 
                                        punit.csco_get_cur_test_phase());

        assert that csco_abort_phase.as_a(uint) > punit.csco_get_cur_test_phase().as_a(uint)  or csco_abort_phase == UNDEFINED else
          error(appendf("infinite-len sequence's abort phase (%s) is earlier than current driver phase (%s)", csco_abort_phase, punit.csco_get_cur_test_phase()));

        messagef(CSCO_TESTFLOW, LOW, "%s sequence on %s will be terminated in %s phase\n", nice_string(), punit.short_name(), csco_abort_phase);
        
        start exe_monitor();
      };
    }
    else {
      messagef(CSCO_TESTFLOW, LOW, "%s sequence on %s will not raising testflow objection because its driver has no test flow\n", nice_string(), punit.short_name());
    };
  }; // pre_body() @sys.any is also

  post_body() @sys.any is also {
    var punit := get_driver();

    if csco_objection_raised {
      messagef(CSCO_TESTFLOW, HIGH, "sequence (%s) from %s dropping objection to %s\n", nice_string(), punit.short_name(), punit.csco_get_cur_test_phase());
      punit.csco_drop_phase_objection();
      csco_objection_raised = FALSE;
    };
  }; // post_body() @sys.any is also

  package exe_monitor() @sys.any is {
    var tfu := get_driver().csco_get_testflow_unit();

    while tfu.cur_test_phase != csco_abort_phase {
      wait @tfu.test_phase_change;
      if tfu.cur_test_phase == csco_abort_phase {
        messagef(LOW, "infinite-len sequence (%s) on %s has reached its abort phase and its tree will be terminated\n", nice_string(), get_driver().short_name());
        me.stop();
      };
    };
  };
}; // extend any_sequence

extend any_sequence_driver {
  #ifdef SPECMAN_VERSION_4_2_1_OR_LATER {             // As of Specman 4.2.1 or later, the MAIN sequence can be disabled from
    post_generate() is also {                         // being automatically launched via the gen_and_start_main field native
      if csco_has_testflow() {                             // to any_sequence_driver.
        gen_and_start_main = FALSE;
      };
    };
  }; // #ifdef SPECMAN_VERSION_4_2_1_OR_LATER 
}; // extend any_sequence_driver


extend sys {
  CREATE_TESTFLOW;

  setup() is also { 
    set_config(run,tick_max,MAX_UINT) ; 
  };
};


'>
