// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_trans.e
//              :
// Project      : csco_trans
//              :
// Created On   : 11/08/2005
//              :
//              :
// Description  : This file contains the base csco_trans_s struct, which should
//                be like-inherited by all transaction types in the TB. The base
//                type instantiates the transaction tracking struct, which comes
//                with transaction linking support, as well as watchdog and 
//                tracking functionalities. The base type also provides several
//                print method hooks, to attempt to standardize the output.
//                
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_trans;


// This is the base transaction type, from which all transactions
// in the testbench should like-inherit.
struct csco_trans_s like any_sequence_item {
  // Whether the parent transaction is a generated stimulus item (GEN)
  // or a transaction collected by the monitor (MON). The enumerated
  // type is defined such that 'MON' is the default value. So for
  // transactions which are not generated (i.e., collected by monitor),
  // this field would take on the correct value upon creation. For stimulus
  // items, the constraint below should also ensure the same.
  origin : csco_trans_origin_t;
    keep soft origin == GEN;
    csco_print_attr origin SHORT enable;

  // Instantiates the transaction tracking struct.
  tracking : csco_trans_tracking_s;
    keep tracking.parent_trans == me;
    keep tracking.origin == read_only(origin);
    attribute tracking deep_copy = ignore;
    csco_print_attr tracking SHORT enable;
    csco_print_attr tracking ALL expand=ON;

  pre_xmt_done() is empty;

  // Hook to execute some non time consuming logic just before the transaction's
  // hop_done event is emitted.
  pre_hop_done() is empty;

  // Hook to execute some non time consuming logic just before the transaction's
  // all_done event is emitted.
  pre_all_done() is empty;

  // Short descriptive name for the transaction, based on info fields
  // in the tracking struct.
  csco_short_name() : string is {
    result = append(tracking.origin, "_",
                    tracking.trans_kind, "_TR",
                    tracking.id_string());
  };

  print_trans_tree() is {
    csco_print_sub_tree(LONG, "tracking.downstream_ptrs");
  };

  // Does a deep_compare_physical on the 2 transactions. The input
  // transaction is considered the actual value. The enclosing
  // transaction is considered the expected value.
  compare_trans(actual : csco_trans_s) : list of string is {
    result.add(deep_compare_physical(me, actual, 50));
  };

  check_trans(actual : csco_trans_s) is {
    var errs : list of string = compare_trans(actual);
    if errs.size() > 0 {
      dut_error("CSCO_TRANS_ERR_TR_MISCOMPARE:\n",
                "expected transaction ", me.csco_short_name(), " is not the same as \n",
                "actual   transaction ", actual.csco_short_name()) {
        out(errs);
      };
    };
  };

  init() is also {
    if tracking == NULL {
      // allocate tracking struct if the logic that instantiated the transaction
      // didn't do so already.
      tracking = new csco_trans_tracking_s with {
        // Set the backpointer in the tracking struct, because some transaction are
        // new'ed, not generated.
        .parent_trans = me;
      };
    };
  };

  run() is also {
    if tracking != NULL {
      tracking.origin = origin;
    };

    // Activate watchdog timer if necessary. Tracking could be NULL if
    // we're dealing with a deep_copied trans instance. 
    if tracking != NULL and !tracking.disable_watchdog {
      tracking.reset_watchdog_timer();
    };
  };
}; // struct csco_trans_s like any_sequence_item


//------------------------------------------------------------
// transaction passing method ports

// non-time consuming PUSH interface.
// a transaction source uses this to PUSH to sink(s) and return immediately.
// sinks must use their own threading for time consuming actions.
// used, eg, for passing transactions from monitors to drivers/scoreboards
method_type csco_trans_push_mp(trans : csco_trans_s);

// time consuming PUSH interface.
// a transaction source uses this to PUSH to sink(s) and block for processing.
// used, eg, for passing transactions from drivers to BFMs in PUSH_MODE
method_type csco_trans_push_tcmp(trans : csco_trans_s) @sys.any;

// non-time consuming PULL interface.
// a transaction sink uses this to PULL from a source.
// NULL is returned if no transaction is availble.
// used, eg, for a try_next_item() style BFM/driver interaction
method_type csco_trans_pull_mp() : csco_trans_s;

// time consuming PULL interface.
// a transaction sink uses this to PULL from a source.
// the call will block until a transaction is available.
// used, eg, for a get_next_item() style BFM/driver interaction
method_type csco_trans_pull_tcmp() : csco_trans_s @sys.any;


'>
