// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_trans_env.e
//              :
// Project      : csco_trans
//              :
// Created On   : 11/08/2005
//              :
//              :
// Description  : The transaction env unit is used to track active 
//                transactions and archive completed transactions. It
//                also periodically checks all the active transactions
//                to see if any of them has timed out. This is more 
//                efficient versus launching an individual monitor for
//                each transaction.  The monitoring is done periodically
//                and so timeout conditions are not detected as soon as
//                they occur, instead they will be flagged at the next
//                checking period, if the transaction's hop still hasn't
//                completed by then.
//
//                Each testbench environment should only have ONE trans.
//                env unit.
//                
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_trans;

import csco_testflow/e/csco_testflow_top ;


// Only generated transaction stimuli are directly added to the
// transaction env unit. The GEN transactions span out to 
// trees containing the monitored transactions.
extend GEN'origin csco_trans_tracking_s {
  add_to_active : bool;
    keep soft add_to_active;

  // Returns global transaction env unit.
  get_trans_env() : csco_trans_env_u is {
    result = sys.csco_trans_env;
  };

  package !trans_env : csco_trans_env_u;

  run() is also {
    // Add transaction to transaction env unit, if one
    // can be found.
    trans_env = get_trans_env();
    if (trans_env != NULL and trans_env.disable == FALSE) {
      trans_env.new_trans(me);
    };
  };

  // User hook that is invoked BEFORE a transaction is moved from the active list
  // to the retired list.
  pre_done_hook() is empty;

  // User hook that is invoked AFTER a transaction is moved from the active list
  // to the retired list.
  post_done_hook() is empty;

  on all_done {
    // On completion, notify the trans. env unit to archive the transaction.
    if trans_env != NULL {
      pre_done_hook();
      trans_env.trans_done(me);
      post_done_hook();
    };
  };
}; // extend GEN csco_trans_tracking_s


// The transaction env unit is used to monitor for timeout conditions 
// of currently active GEN transactions. 
extend csco_trans_env_u {
  CREATE_TESTFLOW;

  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  short_name() : string is only {
    result = "TRANS_HIST";
  };

  // Will prevent transactions from being added to the env, effectively 
  // disabling this unit's features. This field may also be dynamically 
  // set to disable checking by this unit.
  disable : bool;
    keep soft disable == FALSE;

  // How many completed transactions to be kept in the transaction history list.
  trans_hist_max_depth: uint;
    keep soft trans_hist_max_depth == 25;

  //check the transaction history for transactions descriptors that have timed out
  //every this many ns:
  timeout_check_period: time;
    keep soft timeout_check_period == 10000 ns;

  // Wait until active transaction list is empty before leaving POST_TEST. The user
  // may set this to false if he has alternate ways of waiting for system quiesce
  // following MAIN_TEST.
  wait_post_test_quiesce : bool;
    keep soft wait_post_test_quiesce;

  // Wait until active transaction list is empty before leaving
  // POST_TEST_CHECK.
  wait_post_test_check_quiesce : bool;
    keep soft wait_post_test_check_quiesce;


  ////////////////////////
  // User Read-Only  Fields 
  ////////////////////////
  // Prints all transactions in active_trans_list
  dut_error_dump() is also {
    out("printing all ", dec(active.all(.upstream_ptrs.is_empty()).size()), 
        " active transaction trees in system:");
    for each (tr) in active.all(.upstream_ptrs.is_empty()) {
      outf("\nactive transaction tree #%d\n", index);
      tr.parent_trans.print_trans_tree();
    };
  };


  ////////////////////////
  // Private  Fields 
  ////////////////////////
  // The list of currently active GEN transactions.
  package !active : list (key:id) of GEN'origin csco_trans_tracking_s;

  // The list of already completed GEN transactions.
  package !retired : list (key:id) of GEN'origin csco_trans_tracking_s;

  // This method adds a descriptor to the active_trans and checks for 
  // duplicates.
  package new_trans(tr: GEN'origin csco_trans_tracking_s) is {
    // If the transaction is not already on the active_trans then add it.
    if !active.key_exists(tr.id) {
      if tr.add_to_active {
        message(CSCO_TRANS, MEDIUM, "adding TR with ID ", tr.get_id(), " to the active list.");
        active.add(tr); 
      };
    }
    else {
      dut_error_dump();
      error("CSCO_TRANS_ERR_ACTV_DUP: TR ID ", tr.get_id(), " already exists on active list.");
    }; 
  }; // new_trans()

  // This method is called when a transaction's done event fires.  It make 
  // sure a transaction is on the active list and moves it from the active 
  // list to the history list. It also calls quit_tree() for any transactions 
  // that no longer exist on either list (if the transaction history depth 
  // should be exceeded) so that garbage collection will take place.
  package trans_done(tr: GEN'origin csco_trans_tracking_s) is {
    var idx : uint;
    var oldest_tr: csco_trans_tracking_s;

    if tr.add_to_active {
      idx = active.key_index(tr.id);
      if (idx == UNDEF) {
        tr.parent_trans.print_trans_tree();
        error(appendf("CSCO_TRANS_ERR_MISN_TR: %s is not on the active transaction list", tr.parent_trans.csco_short_name()));
      } 
      else {
        messagef(CSCO_TRANS, MEDIUM, "terminating %s by deleting it from active list:\n", tr.parent_trans.csco_short_name());
        messagef(CSCO_TRANS, HIGH, "") {
          tr.parent_trans.print_trans_tree();
        };
        active.delete(idx);

        // maintain max depth limit of transaction history list by
        // removing the oldest entries if necessary.
        if (trans_hist_max_depth > 0)  {
          retired.add(tr);
          if (retired.size() > trans_hist_max_depth) {
            oldest_tr = retired.pop0();
            quit_tree(oldest_tr);
          };
        }
        else {
          quit_tree(tr);
        };
      };
    };
  }; // trans_done()

  // This method calls quit() for a descriptor and its downstream transaction records.
  package quit_tree(tr: csco_trans_tracking_s) is {
    message(CSCO_TRANS, HIGH, "calling quit_tree() for ", tr.parent_trans.csco_short_name());
    for each (t) in tr.get_all_children() {
      t.quit();
    };
    tr.parent_trans.quit();
  }; // quit_tree()

  // Periodically check for transactions that have timed out.
  package timeout_monitor() @sys.any is {
    var tr_list : list of csco_trans_s;

    while (TRUE) {
      wait delay(timeout_check_period);
      messagef(CSCO_TRANS, HIGH, "checking active transactions for timeouts\n");

      for each (tr) in active {
        tr_list = {tr.parent_trans; tr.get_all_children()};
        for each (n) in tr_list.tracking {
          messagef(CSCO_TRANS, FULL, "checking %s for timeout\n", n.parent_trans.csco_short_name());

          if n.disable_watchdog == FALSE and n.timeout_time < sys.time and disable == FALSE {
            if (n.origin == GEN and !n.xmt_done_flag) or (n.origin == MON and !n.hop_done_flag) {
              var tmp_str : string;
              if n.origin == GEN {
                tmp_str = appendf("For GEN TRs, xmt_done is set when all downstream TRs have been linked in.\n");
              } else {
                tmp_str = appendf("For MON TRs, hop_done is set when all downstream TRs have been linked in.\n");
              };
              dut_error("CSCO_TRANS_ERR_TIMEOUT: ", n.parent_trans.csco_short_name(), " timed out after at least ",
                        n.watchdog_timer, ".  The elapsed time is between the start_time and ",
                        "when all expected downstream TRs should have been linked in. ", tmp_str,
                        "  start_time         = ", n.start_time, "\n",
                        "  wachdog expiration = ", n.timeout_time, "\n",
                        "  wachdog sampled @  = ", sys.time, "\n") {
                n.parent_trans.print_trans_tree();
              };
            };
          };
        };
      };
    };
  }; // timeout_monitor()

  // clear the active_trans_list and trans_hist_list. have to do
  // this for when we run a reset test.
  csco_reset() @sys.any is also {
    active.clear();
    retired.clear();
  }; 

  csco_init_test() @sys.any is also {
    start timeout_monitor();
  }; 

  csco_post_test() @sys.any is also {
    if wait_post_test_quiesce {
      wait true(active.is_empty());
    };
  };

  csco_post_test_check() @sys.any is also {
    if wait_post_test_check_quiesce {
      wait true(active.is_empty());
    };
  };

  check() is {
    if not active.all(.disable_watchdog == FALSE).is_empty() {
      dut_errorf("CSCO_TRANS_ERR_ORPHAN: active transaction list is not empty at the end of the test (check_test())\n");
    };
  }; 
}; // unit csco_trans_hist_u


'>
