// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_trans_tracking.e
//              :
// Project      : csco_trans
//              :
// Created On   : 11/08/2005
//              :
//              :
// Description  : This file implements the tracking struct for the baseline 
//                csco_trans. The tracking struct provides support for transaction
//                linking, request/response linking, as well as transaction 
//                tracking. Furthermore, a watchdog timer is implemented for each
//                transaction to detect stalled system interfaces.
//
//                The basic idea of transaction linking is that a unit of stimulus
//                fans out to multiple transactions as it traverses various blocks
//                and their interfaces in the system. We'd like to capture and 
//                track these transformations until all the transactions leave the
//                system. We do this by linking all the downstream transactions with
//                their upstream counterparts and providing methods to maintain, query, 
//                and print the transaction tree.
//
//                An example of a transaction tree is:
//                  PCI transaction stimulus
//                   --> PCI transaction input
//                   -->--> internal transaction 1
//                   -->--> internal transaction 2
//                   -->-->--> HT transaction output
//
//                The progress of a transaction as it traverses the system is tracked
//                via timestamps and flags:
//
//                  start_time : when transaction was generated, or first observed on 
//                               the bus.
//                  xmt_done_time : when transaction was completely transmitted across the
//                                  bus.
//                  hop_done_time : when transaction completely traverses the local
//                                  DUT block.
//                  all_done_time : when all of the transaction's expected downstream
//                                  transactions have been observed.
//
//                Transaction timeouts are monitored on a hop-by-hop basis to maximize
//                problem localization.
//
//                The benefits of all this are easier debugging and easier performance
//                measurements.
//
//                A basic premise of this package is that stimulus transaction is 
//                always generated. While a monitored transaction is always new'ed. This
//                is consistent with a sensible testbench implementation and all the 
//                code comments assuem this convention.  But the struct implementation
//                itself supports both generated and new'ed struct instances.
//                
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_trans;


extend csco_trans_env_u {
  // Globally unique ID maintained by eVC and automatically assigned 
  // to each csco_trans_s instance created during simulation.
  !next_id : uint;
};


// This is a transaction descriptor, which will be instantiated inside
// the csco_trans_s base struct, which in turn should be inherited by
// every transaction defined in the system.
struct csco_trans_tracking_s {

  ////////////////////////
  // User Writable Fields 
  ////////////////////////
  // Whether the parent transaction is a generated stimulus item (GEN)
  // or a transaction collected by the monitor (MON). The enumerated
  // type is defined such that 'MON' is the default value. So for
  // transactions which are not generated (i.e., collected by monitor),
  // this field would take on the correct value upon creation. For stimulus
  // items, the constraint below should also ensure the same.
  origin : csco_trans_origin_t;
    keep soft origin == GEN;
    csco_print_attr origin ALL disable;

  // The interface over which the parent transaction was injected/observed.
  // This is a when determinant that should be set from the parent transaction's
  // init() method by each transaction type.  It can be used to specify 
  // new/alternate behavior for the tracking struct. This field should be set 
  // in the init() hook of like-inherited transaction sub-types.
  trans_kind : csco_trans_kind_t;
    csco_print_attr trans_kind SHORT enable;
    csco_print_attr trans_kind SHORT header=tr_kind;

  // Master enable for watchdog timer of transactions.  Enabled by default
  // since disable_watchdog==FALSE on creation.  The watchdog monitors for 
  // transaction timeout conditions using the watchdog_timer field below.
  disable_watchdog : bool;
    keep soft disable_watchdog == FALSE;

  // For GEN transactions, this is the maximum amount of time that can elapse
  // before its 'xmt_done' is reached.
  // <br>
  // For MON transactions, this is the maximum amount of time that can elapse
  // before its 'hop_done' is reached.
  // <br>
  // For monitored transactions, the default value may be modified by extending 
  // the init() hook of the parent struct.  Each scoreboard may also want to chose 
  // specific values for each transaction instance. That can be done via a call 
  // to the set_watchdog_timer() method.
  watchdog_timer : time;
    keep soft watchdog_timer == 25 us;

  // Hook to execute some non time consuming logic just before the transaction's
  // xmt_done event is emitted.
  pre_xmt_done() is {
    parent_trans.pre_xmt_done();
  };

  // Hook to execute some non time consuming logic just before the transaction's
  // hop_done event is emitted.
  pre_hop_done() is {
    parent_trans.pre_hop_done();
  };

  // Hook to execute some non time consuming logic just before the transaction's
  // all_done event is emitted.
  pre_all_done() is {
    parent_trans.pre_all_done();
  };


  ////////////////////////
  // User Read-Only fields 
  ////////////////////////
  // Returns globally unique transaction ID.
  get_id() : uint is {
    result = id;
  };

  id_string() : string is {
    // delimiter before/after make text search easier
    result = appendf("(id=%x)",id);
  };

  // Returns enclosing transaction, in which the tracking
  // struct is instantiated.
  get_parent_trans() : csco_trans_s is {
    result = parent_trans;
  };

  // Returns transaction depth.
  get_max_depth() : uint is {
    result = max_depth;
  };

  // Time stamp of when the trans. was created. 
  // For GEN transactions, this is the time when the tranaction
  // was generated. For MON transactions, this is the time at which 
  // transmission started and was captured by the monitor.
  get_start_time() : time is {
    result = start_time;
  };

  // Triggered when entire transaction has been sent on bus and
  // observed by monitor. 
  event xmt_done;

  // Indicates that transaction has been entirely observed.
  get_xmt_done_flag() : bool is {
    result = xmt_done_flag;
  };

  // Time stamp of when the transaction was completely sent across the
  // bus. 
  get_xmt_done_time() : time is {
    result = xmt_done_time;
  };

  // Triggered/set when all expected direct child nodes have been observed.
  event hop_done;

  // Indicates that transaction has successfully passed through its local 
  // DUT block (hop). Only valid for MON transactions, declared here to 
  // avoid casting.
  get_hop_done_flag() : bool is {
    result = hop_done_flag;
  };

  // Returns time at which transaction successfully passed through 
  // its local DUT block (hop). Only valid for MON transactions, 
  // declared here to avoid casting.
  get_hop_done_time() : time is {
    result = hop_done_time;
  };

  // Triggered/set when all child nodes have completed. The all_done condition
  // of each node on the tree is re-evaluated whenever a new leaf transaction
  // is linked to the tree.
  event all_done;

  // Indicates that all expected downstream transactions, including
  // non-local hops, have been observed.
  get_all_done_flag() : bool is {
    result = all_done_flag;
  };

  // Returns time when all expected downstream transactions, including
  // non-local hops, have been observed.
  get_all_done_time() : time is {
    result = all_done_time;
  };

  // Changes the maximum allowed life time of the transaction.
  set_watchdog_timer(val : time) is {
    watchdog_timer = val;
    disable_watchdog = FALSE;
    reset_watchdog_timer();
  };

  // Restarts the watchdog timer, effectively extending the life of
  // the packet.
  reset_watchdog_timer() is {
    if not disable_watchdog {
      // calculate the exact time the trans should timeout
      timeout_time = sys.time + watchdog_timer;
      message(CSCO_TRANS, MEDIUM,"timeout time is now set to ", timeout_time, " for TR with ID=", get_id());
    }
    else {
      messagef(CSCO_TRANS, FULL, "%s: timeout time reset called, even though the timer is disabled, reset will not have any effect\n", 
                parent_trans.csco_short_name());
    };
  }; // reset_watchdog_timer() 

  // Adds a new downstream transaction to the current one.
  link_trans(new_tr : csco_trans_s) is {
    message(CSCO_TRANS, MEDIUM, "linking ", new_tr.csco_short_name(), " as a child transaction of ", parent_trans.csco_short_name());

    assert that !all_done_flag else
      error("CSCO_TRANS_ERR_TRANS_LINK: link_trans() being called for ", parent_trans.csco_short_name(), 
              " whose all_done_flag is already set");

    if new_tr.origin == GEN {
      assert that origin == GEN else
        error("CSCO_TRANS_ERR_GEN_LINK: link_trans() being called to link a GEN transaction with a ", origin, " transaction.",
                " GEN transaction can only be linked to another GEN transaction parent.");
    };

    // Detect loops. A transaction cannot appear multiple times on the same branch.
    if new_tr in get_all_parents() {
      for each (r) in get_root_trans() {
        r.print_trans_tree();
      };
      error(appendf("CSCO_TRANS_ERR_LINK_DUP: %s is already a member of the current transaction tree", new_tr.csco_short_name()));
    };

    downstream_ptrs.add(new_tr);
    new_tr.tracking.upstream_ptrs.add(parent_trans);
    for each (tr) in {new_tr; new_tr.tracking.get_all_children()} {
      tr.tracking.max_depth = tr.tracking.get_direct_parents().apply(.tracking.max_depth + 1).max_value(it);
    };

    assert that exp_child_cnt >= downstream_ptrs.size() else
      error(appendf("CSCO_TRANS_ERR_BAD_CHD_CNT: %s's expected child count (%d) is < its actual child count (%d)", 
                      parent_trans.csco_short_name(), exp_child_cnt, 
                                                  downstream_ptrs.size()));

    // Update xmt_done flags/timestamps.
    if new_tr.origin == MON {
      new_tr.tracking.emit_xmt_done();
    };
    if me is a GEN csco_trans_tracking_s (gen_tr) {
      gen_tr.check_xmt_done();
    };

    // Update hop_done/all_done flags/timestamps.
    message(CSCO_TRANS, HIGH, "scheduling status update for ", parent_trans.csco_short_name(), 
                              " at end of current tick due to ", new_tr.csco_short_name(), " being linked in.");
    start update_status();
  }; // link_trans()

  // This methods MUST be called by the appropriate scoreboard to indicate 
  // the number of downstream transactions that are expected to be spawned 
  // by the current transaction for the interface handled by the scoreboard.
  incr_exp_child_cnt(incr : uint) is {
    exp_child_cnt += incr;
    if origin == MON {
      emit_xmt_done();
    };
    message(CSCO_TRANS, HIGH, "calling incr_exp_child_cnt() for ", parent_trans.csco_short_name(), ", exp_child_cnt=", dec(exp_child_cnt));
  };

  // Decrement the expected downstream child count. This adjustment is necessary 
  // when there are downstream transaction drops due to error injection, OIR,
  // etc... This method will take care of adjusting the all_done flags in the 
  // tree if the decrement turns the transaction into a leaf node.
  decr_exp_child_cnt(decr : uint) is {
    assert that exp_child_cnt >= decr else
      error(appendf("CSCO_TRANS_ERR_CHD_CNT_DEC: decrement value (%d) is greater than current value (%d) of exp_child_cnt", decr, exp_child_cnt));

    exp_child_cnt -= decr;
    message(CSCO_TRANS, HIGH, "calling decr_exp_child_cnt() for ", parent_trans.csco_short_name(), "new value is ", dec(exp_child_cnt));
    message(CSCO_TRANS, HIGH, "scheduling status update for ", parent_trans.csco_short_name(), 
                              " at end of current tick due to a call to decr_exp_child_cnt().");
    start update_status();
  };

  // This method is useful during OIR/abort scenarios to remove current
  // downstream expectations on a transaction.
  clear_exp_child_cnt() is {
    exp_child_cnt = downstream_ptrs.size();
    message(CSCO_TRANS, HIGH, "calling clear_exp_child_cnt() for ", parent_trans.csco_short_name());
    message(CSCO_TRANS, HIGH, "scheduling status update for ", parent_trans.csco_short_name(), 
                              " at end of current tick due to a call to clear_exp_child_cnt().");
    start update_status();
  };

  get_exp_child_cnt() : uint is {
    result = exp_child_cnt;
  };

  // Indicates that this transaction is a leaf transaction. This method
  // should be called by a scoreboard if an upstream transaction is
  // terminated inside its DUT block.
  set_leaf() is {
    message(CSCO_TRANS, MEDIUM, "calling set_leaf() for ", parent_trans.csco_short_name());
    if !all_done_flag {
      message(CSCO_TRANS, HIGH, "scheduling status update for ", parent_trans.csco_short_name(), 
                                " at end of current tick due to a call to set_leaf().");
      start update_status();
    };
  };

  // This method should be called for GEN transactions that are executed in
  // bkdoor fashion, in place of link_trans(), incr/decr_exp_child_cnt().
  bkdoor_done() is undefined;

  // Returns the top-level transaction(s) that eventually spawned the current
  // one.
  get_root_trans() : list of csco_trans_s is {
    result = get_all_parents().all(.tracking.max_depth == 0);
  };

  // Get direct downstream nodes.
  get_direct_children() : list of csco_trans_s is {
    result = downstream_ptrs;
  };

  // Returns a list of all the unique child transactions below the current
  // transaction, including non-direct children.
  get_all_children() : list of csco_trans_s is {
    var breadth_first_fifo : list of csco_trans_s;
    var cur_node : csco_trans_s;

    breadth_first_fifo = downstream_ptrs.copy();
    while breadth_first_fifo.size() > 0 {
      cur_node = breadth_first_fifo.pop0();
      if cur_node not in result {
        result.add(cur_node);
      };
      breadth_first_fifo.add(cur_node.tracking.downstream_ptrs);
    };
  };

  // Get direct upstream nodes.
  get_direct_parents() : list of csco_trans_s is {
    result = upstream_ptrs;
  };

  // Returns a list of all the unique parent transactions above the current
  // transaction, including non-direct parents.
  get_all_parents() : list of csco_trans_s is {
    var breadth_first_fifo : list of csco_trans_s;
    var cur_node : csco_trans_s;

    breadth_first_fifo = upstream_ptrs.copy();
    while breadth_first_fifo.size() > 0 {
      cur_node = breadth_first_fifo.pop0();
      if cur_node not in result {
        result.add(cur_node);
      };
      breadth_first_fifo.add(cur_node.tracking.upstream_ptrs);
    };
  };


  ////////////////////////
  // private fields 
  ////////////////////////
  // Globally unique ID across all transaction instances. Automatically 
  // assigned on creation.
  package !id : uint;
    csco_print_attr id SHORT enable;

  package !start_time : time;

  // For GEN transactions, this field is set automatically when all
  // its downstream transactions are linked. For MON transactions, this
  // field is set when the transaction is linked with an upstream trans.
  // The assumption here is that for a MON transaction, it's forwarded
  // to a scoreboard and linked with an upstream transaction as soon as
  // its transmission is observed to be complete.
  package !xmt_done_flag : bool;
  package !xmt_done_time : time;
    csco_print_attr xmt_done_flag ALL disable;
    csco_print_attr xmt_done_time ALL header=xmt_done;

  // Time stamp of when the hop_done event was emitted. Only applicable to
  // MON transactions, but declared in the base type to avoid casting.
  package !hop_done_flag : bool;
  package !hop_done_time : time;
    csco_print_attr hop_done_flag ALL disable;
    csco_print_attr hop_done_time ALL header=hop_done;

  // Time stamp of when the all_done event was emitted.
  package !all_done_flag : bool;
  package !all_done_time : time;
    csco_print_attr all_done_flag ALL disable;
    csco_print_attr all_done_time ALL header=all_done;

  csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is also {
    if pf.name == "xmt_done_time" and !xmt_done_flag {
      result = "n/a";
    };
    if pf.name == "hop_done_time" and !hop_done_flag {
      result = "n/a";
    };
    if pf.name == "all_done_time" and !all_done_flag {
      result = "n/a";
    };
  };

  // Depth of TR in transaction flow (root being depth 0). Automatically 
  // assigned when the transaction is linked into the tree via a call to 
  // link_trans()
  package !max_depth : uint;
    csco_print_attr max_depth ALL disable;

  // Expected number of downstream pointers. 
  package !exp_child_cnt : uint;

  // time (in ns) that this descriptor is due to timeout
  package !timeout_time: time;

  // Pointer is set during run() of parent transaction.
  package parent_trans : csco_trans_s;
    attribute parent_trans deep_copy = ignore;
    csco_print_attr parent_trans ALL disable;

  // List of upstream transactions that caused this one.
  package !upstream_ptrs : list of csco_trans_s;
    attribute upstream_ptrs deep_copy = ignore;
    csco_print_attr upstream_ptrs ALL header=upstream;

  // List of transaction records that this one caused.
  package !downstream_ptrs : list of csco_trans_s;
    attribute downstream_ptrs deep_copy = ignore;
    csco_print_attr downstream_ptrs ALL header=downstream;

  // Since a given TR may be broadcast to multiple downstream
  // scoreboards, we need to give each scoreboard a chance to
  // increment the expected child count before any one of them
  // is allowed to set the 'all_done' flag on the TR.
  package !status_update_scheduled : bool;
    csco_print_attr status_update_scheduled ALL disable;

  package update_status() @sys.any is {
    if !status_update_scheduled {
      status_update_scheduled = TRUE;
      sync cycle@sys.tick_end;
      check_hop_done();
      check_all_done();
      status_update_scheduled = FALSE;
    };
  };

  package check_hop_done() is {
    if origin == MON {
      if downstream_ptrs.size() == exp_child_cnt {
        emit_hop_done();
      };
    }
    else {
      if downstream_ptrs.all(.tracking.hop_done_flag).size() == exp_child_cnt {
        emit_hop_done();
      };
    };

    if hop_done_flag {
      for each (parent) in upstream_ptrs {
        parent.tracking.check_hop_done();
      };
    };
  };

  // Updates the all_done status of all the transactions on the tree due to the
  // observation of a new leaf transaction.
  package check_all_done() is {
    if downstream_ptrs.all(.tracking.all_done_flag == TRUE).size() == exp_child_cnt {
      emit_all_done();

      for each (parent) in upstream_ptrs {
        parent.tracking.check_all_done();
      };
    };
  }; // check_all_done()
  
  package emit_xmt_done() is {
    if !xmt_done_flag {
      pre_xmt_done();
      messagef(CSCO_TRANS, MEDIUM, "%s has finished transmission (xmt_done)\n", parent_trans.csco_short_name());

      xmt_done_flag = TRUE;
      xmt_done_time = sys.time;
      emit xmt_done;
    };
  };

  package emit_hop_done() is {
    if !hop_done_flag {
      pre_hop_done();
      messagef(CSCO_TRANS, MEDIUM, "%s has passed through its local DUT block (hop_done)\n", parent_trans.csco_short_name());

      hop_done_flag = TRUE;
      hop_done_time = sys.time;
      emit hop_done;
    };
  };

  package emit_all_done() is {
    if !all_done_flag {
      pre_all_done();
      messagef(CSCO_TRANS, LOW, "%s has passed through the entire system (all_done)\n", parent_trans.csco_short_name());

      all_done_flag = TRUE;
      all_done_time = sys.time;
      emit all_done;
    };
  };

  init() is also {
    // Assign unique ID to transaction.
    if (id == 0) {
      id = sys.csco_trans_env.next_id;
      sys.csco_trans_env.next_id += 1;
    };

    start_time = sys.time;
  }; 
}; // struct csco_trans_tracking_s


extend GEN'origin csco_trans_tracking_s {
  // Update the xmt_done status of all the transactions on the tree due to the observation
  // of a new MON transaction.
  package check_xmt_done() is {
    if downstream_ptrs.all(.tracking.xmt_done_flag).size() == exp_child_cnt {
      emit_xmt_done();
      for each (parent) in upstream_ptrs {
        if parent.tracking is a GEN csco_trans_tracking_s (gen_tr) {
          gen_tr.check_xmt_done();
        }
        else {
          error("detected a GEN TR with a parent whose origin is ", parent.origin);
        };
      };
    };
  };

  // This will set all the approriate flags for a backdoor transaction,
  // which doesn't require any linking.
  bkdoor_done() is also {
    emit_xmt_done();
    emit_hop_done();
    emit_all_done();
  };
}; // extend GEN'origin csco_trans_tracking_s


extend MON'origin csco_trans_tracking_s {
  // Initialize user settable fields, for cases where the transaction
  // was not generated.
  init() is also {
    watchdog_timer = 25 us;
  }; // init() 
}; // extend MON'origin csco_trans_tracking_s


'>
