--------------------------------------------------------------------------------
--
--   Filename : csco_verbosity_cmds.e
--   Language : 'e'
--   Author   : Gary Singer
--
--   General Description: Control Message Verbosity Through File command 
--                        interface.
--
--   History:
--   8/29/05  GSS  Initial Coding
--   1/9/06   GSS  Replaced dut_error with assert statements
--
--------------------------------------------------------------------------------

<'


// This struct is used for a keyed list of all loggers in the env
struct csco_verbosity_logger_info_s
{
  path   : string;          // String given by logger.e_path()
  logger : message_logger;  // Pointer to logger
};

// This struct is used for loggers that are grouped in the verbose.txt file
struct csco_verbosity_logger_group_s
{
  name    : string;                      // String given by logger.e_path()
  !loggers : list (key : path) of csco_verbosity_logger_info_s;  // Pointers to logger info
};

// Base unit to hold the info
unit csco_verbosity_u
{
  // Keyed list of loggers
  !loggers : list (key : path) of csco_verbosity_logger_info_s;
  
  // Keyed list of logger groups
  !logger_groups : list (key : name) of csco_verbosity_logger_group_s;

  // File name for available loggers
  available_loggers_fname : string;
  keep soft available_loggers_fname == "available_loggers.txt";
  
  // File name verbose file
  verbosity_file : string;
  keep soft verbosity_file == "verbose.txt";

  short_name() : string is
  {
    result = "csco_verbosity";
  };

  get_all_loggers() is
  {
    var logger_info : csco_verbosity_logger_info_s;
    var lfile        : file;

    var template := files.add_file_type("csco_verbosity/e/csco_verbosity_available_loggers_template.txt",".txt",TRUE);  
    lfile = files.open (available_loggers_fname, "w", "Text file");
        files.write (lfile, "### This file contains a list of all of the available loggers in the test env.\n");

    files.write (lfile, "### *** Available tags ***");
    messagef (CSCO_VERBOSITY, MEDIUM, "*** Available tags ***\n");
    files.write (lfile, appendf ("#  %s", all_values (message_tag)));
    messagef (CSCO_VERBOSITY, MEDIUM, " %s\n", all_values (message_tag));
    files.write (lfile, "\n### *** Available message loggers (with attached tags at startup) ***");
    messagef (CSCO_VERBOSITY, MEDIUM, "*** Available message loggers (with attached tags at startup) ***\n");
    for each in (rf_manager.get_all_message_loggers())
    {
      logger_info = new;
      logger_info.logger = it;
      logger_info.path = it.e_path();
      loggers.add (logger_info);
      messagef (CSCO_VERBOSITY, MEDIUM, " %s %s\n", logger_info.path, it.get_tags());
      files.write (lfile, append("#   ", logger_info.path, " ", it.get_tags()));
    };
    files.write (lfile, "\n");
    files.close (lfile);
    files.file_append (template, available_loggers_fname);
  };

  // Read the verbose file (default verbose.txt) at run() and interpret the commands
  read_verbose_file () is 
  {
    var grouping     : bool;
    var logger_group : csco_verbosity_logger_group_s;
    
    if (files.file_exists (verbosity_file))
    {
      messagef (CSCO_VERBOSITY, LOW, "---------------------- Reading verbosity file %s ----------------------\n", verbosity_file);
       
      for each (vline) in file verbosity_file {
        // skip blank lines and lines starting with #. Also skip init statements at this time
        if (vline !~ "/^\s*$/" and vline !~ "/^\s*\#/" and vline !~ "/^\s*init\s+/")
        {
          if (!grouping)
          {
            case TRUE 
            {
              // Set default message level
              vline ~ "/^\s*set_default\s+(\w+)\s*$/" : 
              {
                for each (l) in loggers
                {
                  l.logger.set_actions($1.as_a(message_verbosity), l.logger.tags, "*", "...", replace);
                  l.logger.verbosity = $1.as_a(message_verbosity);
                };
                messagef (CSCO_VERBOSITY, LOW, "SET ALL LOGGERS MESSAGE_VERBOSITY TO %s. (tags are unchanged)", $1);
              };

              // add a logger group
              vline ~ "/^\s*add_group\s+([a-zA-Z0-9_\.]+)\s*$/" : 
              {
                assert !loggers.key_exists ($1) else
                  error ("ERROR: ", verbosity_file," verbosity file had ", $1, " as a logger group name. That is the name of an instanced logger\n", verbosity_file, $1);
                assert !logger_groups.key_exists ($1) else
                  error ("ERROR: ", verbosity_file," verbosity file had ", $1, " as a logger group name. That logger group already exists\n", verbosity_file, $1);
                logger_group = new;
                logger_group.name = $1;
                grouping = TRUE;
                messagef (CSCO_VERBOSITY, HIGH, "CREATING logger group %s", $1);
              };

              // Call one specman function after a delay (multiple commands may be seperated by ";" on one line)
              vline ~ "/^\s*change\s+(\d+)\s+([a-zA-Z]+)\s+(.+)\s*$/" :
              {
                messagef (CSCO_VERBOSITY, HIGH, "AT %s %s execute '%s'", $1, $2, $3);
                start delayed_specman_cmd ($1.as_a(uint), $2.as_a(e_time_units), $3);
              };
              
              // Same as "change" command
              vline ~ "/^\s*at\s+(\d+)\s+([a-zA-Z]+)\s+(.+)\s*$/" :
              {
                messagef (CSCO_VERBOSITY, HIGH, "AT %s %s execute '%s'", $1, $2, $3);
                start delayed_specman_cmd ($1.as_a(uint), $2.as_a(e_time_units), $3);
              };

              // Execute one specman command now!
              vline ~ "/^\s*now\s+(.+)\s*$/" :
              {
                messagef (CSCO_VERBOSITY, HIGH, "Execute '%s' now", $1);
                specman ($1);
              };

              // Call set message for each logger in a group
              vline ~ "/^\s*change_group\s+(\d+)\s+([a-zA-Z]+)\s+([a-zA-Z0-9_\.]+)\s+set\s+message\s+(.+)\s*$/" :
              {
                messagef (CSCO_VERBOSITY, HIGH, "AT %s %s execute 'set message %s' on logger group %s:", $1, $2, $4, $3);
                assert logger_groups.key_exists ($3) else
                  error ("ERROR : logger group ", $3, " is not defined in verbosity file verbosity_file\n");
                
                assert (vline !~ "/-logger/") else
                  error ("ERROR : change_group command may not use '-logger' option to choose a specific logger in verbosity file verbosity_file\n");
                
                for each (l) in logger_groups.key ($3).loggers
                {
                  messagef (CSCO_VERBOSITY, HIGH, "AT %s %s execute 'set message -logger=%s %s'", $1, $2, l.path, $4);
                  start delayed_specman_cmd ($1.as_a(uint), $2.as_a(e_time_units), appendf("set message -logger=%s %s", l.path, $4));
                };
                messagef (CSCO_VERBOSITY, HIGH, "AT %sns UPDATE ALL LOGGERS MESSAGE_VERBOSITY TO %s", $1, $2);
              };
              
              // Add file to logger (This could also be done with the change command)
              vline ~ "/^\s*add_file\s+([a-zA-Z0-9_\.]+)\s+([a-zA-Z0-9_\.]+)\s*$/" :
              {
                if ( !loggers.key_exists ($1))
                {
                  assert FALSE else
                    error ("ERROR: ", verbosity_file, " verbosity file had ", $1 ,"as a path to a logger. The path does not exist\n", verbosity_file, $1);
                }
                else
                {
                  messagef (CSCO_VERBOSITY, HIGH, "ADD logger %s TO LOGGER FILE %s", $1, $2);
                  loggers.key ($1).logger.set_file ($2, on);
                };
              };
              
              default :
              {
                assert FALSE else error ("ERROR : line syntax not understood for <", vline, "> in verbosity file ", verbosity_file, "\n");
              };
            };
          }
          else // busy grouping
          {
            case TRUE
            {
              vline ~ "/^\s*end_group\s*$/" :
              {
                assert logger_group.loggers.size() > 0 else
                  error ("ERROR : No loggers added to logger group ", logger_group.name, " in verbosity file ", verbosity_file, "\n");
                grouping = FALSE;
                logger_groups.add (logger_group);
                messagef (CSCO_VERBOSITY, HIGH, "CREATED logger group %s", logger_group.name);
              };

              vline ~ "/^\s*([a-zA-Z0-9_\.]+)\s*$/" :
              {
                assert loggers.key_exists ($1) else
                 error ("ERROR : logger ", $1, " not found while creating logger group ", logger_group.name, " in verbosity file ", verbosity_file, "\n");
                assert !logger_group.loggers.key_exists ($1) else
                  error ("ERROR : logger ", $1, " is already part of logger group ", logger_group.name, " in verbosity file ", verbosity_file, " \n");
                messagef (CSCO_VERBOSITY, HIGH, "ADDING logger %s to logger group %s", $1, logger_group.name);
                logger_group.loggers.add (loggers.key ($1));
              };
                            
              default :
              {
                assert FALSE else
                  error("ERROR : logger ", vline, " not found while adding a logger group ", logger_group.name, " in verbosity file ", verbosity_file, "\n");
              };
            };
          };

        };
      };
      assert !grouping else
        error ("ERROR : No end_group statement found for logger group ", logger_group.name, " in verbosity file ", verbosity_file, "\n");
        
      messagef (CSCO_VERBOSITY, LOW, "---------------------- Verbosity file %s has been read -----------------\n", verbosity_file);
    }
    else
    {
      messagef (CSCO_VERBOSITY, LOW, "---------------------- No verbosity file %s found ----------------------\n", verbosity_file);
    };
  };

  // Start this method for a delayed time specman command
  delayed_specman_cmd (update_time : uint, 
                       time_unit   : e_time_units,
                       cmd         : string) @sys.any is 
  {
    if agent() == "NULL_SIM"
    {
      if (time_unit != fs)
      {
        messagef (LOW, "WARNING: Running without a simulator. Ensure that the time scale used in verbose.txt is in fs.");
      };
      sync true(sys.time == to_specman_scale (update_time, time_unit));
    }
    else
    {
      wait delay(to_specman_scale (update_time, time_unit));
    };
    messagef (LOW, "EXECUTING specman command '%s'\n", cmd);
    specman (cmd);
  };

  // Read the verbose.txt file (name can't be changed by user) at init() and interpret the "init" commands
  read_verbose_file_at_init () is 
  {
    
    if (files.file_exists ("verbose.txt"))
    {
      for each (vline) in file "verbose.txt" {
        // execute any init commands
        if (vline ~ "/^\s*init\s+(.+)\s*$/")
        {
          specman ($1);
        };
      };
    };
  };

//  set_verbosity (logger     : message_logger,
//                 level      : message_verbosity,
//                 check_tags : bool ) is
//  {
//    if (check_tags && logger.tags == {})
//    {
//      logger.tags = { NORMAL };
//    };
//    logger.set_actions(level, logger.tags, "*", "...", replace);
//    logger.verbosity = level;
//  };

  init() is first
  {
    read_verbose_file_at_init();
  };
  

};

extend sys
{
  csco_verbosity : csco_verbosity_u is instance;
  
  post_generate() is also
  {
    csco_verbosity.get_all_loggers();
  };
  

  run() is first
  {
    csco_verbosity.read_verbose_file();
    messagef(NONE,"eRM Logger Configuration Report:\n") {
      specman("show messages\n");
    };
  }
};


'>
