//--------------------------------------------------------------------------------------------------------------
//
//   Filename : csco_verbosity_format.e
//   Language : 'e'
//   Author   : Joseph Hanli Zhang
//
//   Description: Standard message formatting for RTP ASIC.
//
//   History:
//   10/24/02  JHZ  Added Another format with no time elapsed messages.....
//
//--------------------------------------------------------------------------------------------------------------

// This module creates a new format for message actions called RTP_STD. It is meant to be used as a standard
// means of displaying logging information from a Specman testcase. Features include:
//
// display verbosity of each message action.
// adds a 'time elapsed' indicator for each block of message statements that have a different time stamp as the
// previous block.
// alternating blakc/gray colors for time stamp with different values.
//
// Each message output is preceded by a header of the format:
//
//    <time> ns [<verbosity>] (<short_name_path>):
//
// To use the formatting provided by this module, simply constrain the 'format' field of the logger to 'RTP_STD'
//
//
// A few pecuilar behaviors for the message logger that people should be aware of as of eRM 2.0:
//
// - Any output string supplied as an argument to message() is passed to format_message() as a single string, even
//   if the string contains "\n" characters.
// - Any output string passed to message() via the action block will be split into a list of strings using "\n"
//   as the separator. The "\n" is stripped during the split process. An output string from the message action
//   block will also be split if the string length exceeds the "line_size" Specman config. The resulting list of
//   string is passed to the format_message() method.
// - format_message() returns a list of string to be displayed to the screen (or file). However, the mechanism
//   following format_massge() seems to concatenate result[0] and result[1] while adding a "\n" to all strings from
//   result[1] and onwards. This bug seems to be fixed as of Specman 4.2b2.
// - The Specman string routines do not automatically strip out the coloring meta text when they perform their
//   operations. The user must manually use vt.filter_style() prior to calling the string routines.
//
// format_line() takes a single string (without "\n" characters) and splits it into the appropriate number of lines
// based on the line size. The first line in the result has the message header, the rest of the lines get indentation
// equivalent to the length of the header.
//
// format_inmsg() takes a list of string and for each string, splits into multiple lines using the "\n" as the
// separator. format_line() is then executed on each string in the resulting list of strings.
//
// format_message() can be used to format the message text from a message action into a nice list of lines according
// to the line size and the message header.


<'
import csco_verbosity_types;

extend sys {
  // Configure logger
  keep soft logger.format    == RTP_STD;
};


extend csco_verbosity_utils_u {
  csco_msg_cur_time_style : vt_style;
    keep soft csco_msg_cur_time_style == BLACK;

  csco_msg_cur_time_str : string;
    keep soft csco_msg_cur_time_str == "";

  csco_msg_prev_msg_time : time;
    keep soft csco_msg_prev_msg_time == 0;
};


extend message_logger {
  format_line(str : string, line_len : uint, hdr : string) : list of string is {
    var tmp_str_list : list of string = str_split_all(str, "/\s/");
    var cur_len : uint = 0;
    var tmp_str : string = "";
    var hdr_len := str_len(vt.filter_style(hdr));
    var total_line_len : int = line_len - hdr_len - 10;

    if total_line_len < 0 {
      var err_str := appendf("configured line_size (%d) < message header length (%d) + 10\n", line_len, hdr_len);
      error(err_str);
    };

    -- Also note that if a word (plus the message header) is longer than the
    -- configured line_size, then Specman itself will also add a new line to
    -- the string. In this case, the auto-indent will not work anymore because
    -- Specman doesn't perform auto-indent.

    -- As of Specman 4.2b1, the specman string routines don't seem to support vt_util so
    -- they cannot discern between meta characters used by vt_util. So if a
    -- string contains both white space and style meta characters, the message
    -- may not display correctly because we're using white spaces as the
    -- delimiters when we do a str_split_all.

    for each (str) in tmp_str_list do {
      //outf("DEBUG: format_line_strlist[%d] = %s\n", index, str);
      var tmp_len := str_len(vt.filter_style(str));

      if cur_len + tmp_len > total_line_len {
        result.add(tmp_str);
        tmp_str = str;
        cur_len = tmp_len;
      }
      else {
        tmp_str = str_join({tmp_str; str}, "");
        cur_len += tmp_len;
      };
    };
    result.add(tmp_str);

    // The first string in the result gets the message header prepended to it.
    // Subsequent lines get indentation equivalent to the length of the header.
    for each (str) in result do {
      //outf("DEBUG: format_line_result[%d] = %s\n", index, str);
      if index == 0 {
        tmp_str = appendf("%s:", hdr);
      }
      else {
        tmp_str = str_pad("", hdr_len+1);
        result[index] = str_replace(str, "/^\s/", "");
      };
      tmp_str = (index == 0) ? appendf("%s:", hdr) : str_pad("", hdr_len+1);
      result[index] = appendf("%s %s", tmp_str, result[index]);
    };
  }; -- format_line(str : string, line_len : uint, hdr : string) : list of string is

  format_inmsg(str_list : list of string, line_len : uint, hdr : string) : list of string is {
    var tmp_str_list : list of string;

    // If a string contains "\n", then break it into separate lines using the "\n" as the separator.
    // the "\n" is then discarded.
    tmp_str_list.clear();
    for each (str) in str_list do {
      // Note that the trailing '\n' is stripped on each line since 
      // we're already splitting into multiple lines. Only the input
      // string to the message() statement should have a trailing '\n'.
      tmp_str_list.add(str_split(str_replace(str, "/\n$/", ""), "/\n/"));
    };

    // If a particular string exceeds line_len, then break it into multiple lines.
    for each (str) in tmp_str_list do {
      result.add(format_line(str, line_len, hdr));
    }; -- for each (str) in tmp_str_list do
  }; -- format_inmsg(str_list : list of string, line_len : uint, hdr : string) : list of string is
}; -- extend message_logger


-- IMPORTANT: To use this message logger format, the user must configure line_size in Specman such
-- that it is greater than or equal to the maximum length of any message header plus 10. Otherwise,
-- an error() will be generated during run-time.
extend message_format : [RTP_LONG];
extend RTP_LONG message_logger {
  !elapsed_time : time;

  accept_message() : bool is also {
    if sys.time != sys.csco_verbosity_utils.csco_msg_prev_msg_time && result == TRUE {
        elapsed_time = sys.time - sys.csco_verbosity_utils.csco_msg_prev_msg_time;
        outf("\n");
        out("<", elapsed_time, " elapsed...>\n");
        outf("\n");
        sys.csco_verbosity_utils.csco_msg_prev_msg_time = sys.time;
    };
  }; -- accept_message() : bool is also

  format_message() : list of string is only {
    var vb := appendf("[%s]", get_verbosity());
    var src_struct : any_struct = source_struct();
    var hdr : string;
    var tmp_str : string;
    var tmp_str_list : list of string;
    var msg_list : list of string;
    var msg_style : vt_style;
    var short_name_path := source_struct().get_unit().short_name_path();
    var line_size := get_config(print, line_size);

    // set color of time string so that statements from two successive time blocks would have
    // contrasting black and gray colors. Also front pad time string to 14 characters.
    var time_str := append(sys.time);
    if  time_str != sys.csco_verbosity_utils.csco_msg_cur_time_str {
        sys.csco_verbosity_utils.csco_msg_cur_time_style = (sys.csco_verbosity_utils.csco_msg_cur_time_style == BLACK) ? GRAY : BLACK;
        sys.csco_verbosity_utils.csco_msg_cur_time_str = time_str;
    };
    time_str = vt.filter_style(time_str);
    time_str = vt.text_style(sys.csco_verbosity_utils.csco_msg_cur_time_style, time_str);
    time_str = str_join({str_pad("", 11-str_len(vt.filter_style(time_str))); time_str}, "");

    // replace short name path 'space' separators with '.'
    short_name_path = str_replace(short_name_path, "/\s/", ".");

    // front pad verbosity string to 8 characters
    vb = str_join({str_pad("", 8-str_len(vt.filter_style(vb))); vb}, "");

    // compose message header
    hdr = appendf("%s %s (%s)", time_str, vb, short_name_path);

    // format message text into a list of lines consistent with the line size.
    msg_list = format_inmsg(get_message(), line_size, hdr);

    result.clear();
    msg_style = get_action_style();
    for each (msg) in msg_list do {
      tmp_str = msg;
      if msg != "" {
        // Set color of message body to whatever the logger is configured for.
        // This overrides any coloring already in the message text.
        if msg_style != BLACK {
          tmp_str = vt.filter_style(msg);
          tmp_str = vt.text_style(msg_style, tmp_str);
        };
        result.add(tmp_str);
      };
    };

    #ifndef SPECMAN_VERSION_4_2_OR_LATER then {
      // add "\n" after 1st string to account for Specman bug.
      if result.size() > 1 {
        tmp_str = appendf("%s\n", result[0]);
        result[0] = tmp_str;
      };
    };
  }; -- format_message() : list of string is also
}; -- extend RTP_LONG message_logger


-- IMPORTANT: To use this message logger format, the user must configure line_size in Specman such
-- that it is greater than or equal to the maximum length of any message header plus 10. Otherwise,
-- an error() will be generated during run-time.
extend message_format : [RTP_STD];
extend RTP_STD message_logger {
  format_message() : list of string is only {
    var vb := appendf("[%s]", get_verbosity());
    var src_struct : any_struct = source_struct();
    var hdr : string;
    var tmp_str : string;
    var tmp_str_list : list of string;
    var msg_list : list of string;
    var msg_style : vt_style;
    var short_name_path := source_struct().get_unit().short_name_path();
    var line_size := get_config(print, line_size);

    // set color of time string so that statements from two successive time blocks would have
    // contrasting black and gray colors. Also front pad time string to 14 characters.
    var time_str := append(sys.time);
    if  time_str != sys.csco_verbosity_utils.csco_msg_cur_time_str {
        sys.csco_verbosity_utils.csco_msg_cur_time_style = (sys.csco_verbosity_utils.csco_msg_cur_time_style == BLACK) ? GRAY : BLACK;
        sys.csco_verbosity_utils.csco_msg_cur_time_str = time_str;
    };
    time_str = vt.filter_style(time_str);
    time_str = vt.text_style(sys.csco_verbosity_utils.csco_msg_cur_time_style, time_str);
    time_str = str_join({str_pad("", 11-str_len(vt.filter_style(time_str))); time_str}, "");

    // replace short name path 'space' separators with '.'
    short_name_path = str_replace(short_name_path, "/\s/", ".");

    // front pad verbosity string to 8 characters
    vb = str_join({str_pad("", 8-str_len(vt.filter_style(vb))); vb}, "");

    // compose message header
    hdr = appendf("%s %s (%s)", time_str, vb, short_name_path);

    // format message text into a list of lines consistent with the line size.
    msg_list = format_inmsg(get_message(), line_size, hdr);

    result.clear();
    msg_style = get_action_style();
    for each (msg) in msg_list do {
      tmp_str = msg;
      if msg != "" {
        // Set color of message body to whatever the logger is configured for.
        // This overrides any coloring already in the message text.
        if msg_style != BLACK {
          tmp_str = vt.filter_style(msg);
          tmp_str = vt.text_style(msg_style, tmp_str);
        };
        result.add(tmp_str);
      };
    };

    #ifndef SPECMAN_VERSION_4_2_OR_LATER then {
      // add "\n" after 1st string to account for Specman bug.
      if result.size() > 1 {
        tmp_str = appendf("%s\n", result[0]);
        result[0] = tmp_str;
      };
    };
  }; -- format_message() : list of string is also
}; -- extend RTP_STD message_logger


'>
