// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//
// File name    : csco_verbosity_print_util.e
//              :
// Project      : csco_verbosity
//              :
// Created On   : 01/10/2007
//              :
//              :
// Description  : Printing facility for automatically displaying struct 
//              : contents using standard short and long formats. Custom
//              : formats may also be added in the future.
//              :
// Assumptions  : none
//              :
// Limitations  : none
//              :
// Known Errors : none
//              :
// Notes        :
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Copyright 2003 (c) Cisco Systems Inc.
//
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


<'
package csco_verbosity;


// May be extended, along with necessary methods, to implement additional
// formatting.
type csco_print_format_t : [
  NONE,     // no printing.
  SHORT,    // 2 lines to be printed for each struct. 1 line of field headers and 1 line of values.
  LONG,     // multiple lines will be printed for each struct, 1 or more lines per field.
  ALL       // Used to indicate that a field should be included in all print formats.
];

type csco_print_enable_disable_t : [enable, disable];


// This macro is the user interface for enabling/disabling printing of the given
// field in the given format.
define <CSCO_PRINT_EN'struct_member> 
"csco_print_attr <field'name> (<FMT>SHORT|LONG|ALL)[ ](<EN>enable|disable)" as {
  csco_print_get_fields_hook(fmt : csco_print_format_t, flist : list of csco_print_field_s) is also {
    csco_print_enable_field(flist, "<field'name>", fmt, <FMT>, <EN>);
  };
};

// This macro is the user interface for specifying printing behavior which
// is different from the implemented default. The meaning of each parameter
// is as follows:
//
//   <field'name> : required, specifies the field that will be affected by the attribute.
//   SHORT|LONG|ALL : required, specifies the format which will be affected by the attribute.
//   precede : specifies that the field be printed ahead of <field'name>.
//   header : alternative header to be displayed with field value.
//   radix : override default of HEX.
//   expand : applicable only to struct and list of struct fields. "ON" will cause
//            the field to be expanded during printing.
//
// Note that it is legal for the macro to be called on the same field multiple
// times within a struct. Each subsequent call, for parameters which were specified,
// will override the values set during previous calls. The relative ordering of
// the macro calls is based on Specman code loading order. 
//
define <CSCO_PRINT_ATTR'struct_member> 
"csco_print_attr <field'name> (<FMT>SHORT|LONG|ALL)[ precede[ ]=[ ]<precede'name>][ header[ ]=[ ]<hdr'name>][ radix[ ]=[ ](<RDX>DEC|BIN|HEX|<any>)][ expand[ ]=[ ](<EXP>ON|OFF)]" as {

  csco_print_get_fields_hook(fmt : csco_print_format_t, flist : list of csco_print_field_s) is also {
    csco_print_update_field(flist, "<field'name>", fmt, <FMT>, 
                            "<precede'name>", "<hdr'name>", "<RDX>", "<EXP>");
  };
};

extend sys {
  csco_print_error_tcm(s:any_struct,kind:string,fmt:csco_print_format_t,other:string) @sys.any is {
    out("\n\n\nHere is the error previously mentioned (search backwards for CSCO_VERBOSITY_PRINT_UTIL_INTERNAL)\n");
    if kind ~ "print_str" {
      compute s.csco_print_str(fmt) ;
    } else if kind ~ "print_sub_tree_str" {
      compute s.csco_print_sub_tree_str(fmt,other) ;
    } else if kind ~ "print_field_str" {
      compute s.csco_print_field_str(fmt,other) ;
    };
  };
};

// The three methods defined here represents the main method API
// for printing structs. 
extend any_struct {
  // Standard name to mimic functionality of any_unit's short_name() method.
  // It's called by csco_print routines to display values for unexpanded 
  // struct fields and list elements.
  csco_short_name() : string is {
    result = append(me);
  };

  // Users should call this to do SHORT and LONG printing.
  csco_print(fmt : csco_print_format_t) is {
    try {
      out(csco_print_str(fmt));
    } else {
      out("CSCO_VERBOSITY_PRINT_UTIL_INTERNAL Error doing csco_print for following struct") ;
      print me ;
      out("error details will follow (search forward for CSCO_VERBOSITY_PRINT_UTIL_INTERNAL) ");
      start sys.csco_print_error_tcm(me,"print_str",fmt,"") ;
    };
  };

  csco_print_str(fmt : csco_print_format_t) : list of string is {
    var flist : list of csco_print_field_s;
    var printables : list of csco_printable_s;

    // Grab and expand fields in the current struct.
    var pstruct := csco_print_get_fields(fmt);
    flist = pstruct.expand_fields();

    // Format the fields into lines and tables.
    printables = csco_print_format(fmt, flist, 0);
    csco_print_align_printables(printables);

    // Align and print.
    for each (p) in printables {
      result.add(p.align());
    };
  };

  // Users should call this method to print a tree of structs. The name of the
  // field which holds child structs must be supplied to the method as a string.
  // Note that the child list field path supplied doesn't have to be an immediate
  // field of the enclosing struct. It can be a nested field represented by a 
  // '.' delimited path.
  csco_print_sub_tree(fmt : csco_print_format_t, child_list_path : string) is {
    try {
      out(csco_print_sub_tree_str(fmt, child_list_path));
    } else {
      out("CSCO_VERBOSITY_PRINT_UTIL_INTERNAL Error doing csco_print_sub_tree for following struct") ;
      print me ;
      out("Now illustrating the error");
      start sys.csco_print_error_tcm(me,"print_sub_tree_str",fmt,child_list_path) ;
    };
  };

  csco_print_sub_tree_str(fmt : csco_print_format_t, child_list_path : string) : list of string is {
    var struct_list : list of csco_print_struct_and_depth_s;

    // Get all the structs which are part of the child tree.
    struct_list = csco_print_sub_tree_get_structs(fmt, 0, child_list_path);

    var printables : list of csco_printable_s;
    printables = csco_print_sub_tree_format(fmt, struct_list, child_list_path);
    for each (p) in printables {
      result.add(p.align());
    };
  };

  // Prints a single field of the struct instance from which the method is called.
  // 'fname' must be the name of a direct field inside the struct.
  csco_print_field(fmt : csco_print_format_t, fname : string) is {
    try {
      out(csco_print_field_str(fmt, fname));
    } else {
      out("CSCO_VERBOSITY_PRINT_UTIL_INTERNAL Error doing csco_print_field for following struct") ;
      print me ;
      out("Now illustrating the error");
      start sys.csco_print_error_tcm(me,"print_field_str",fmt,fname) ;
    };
  };

  csco_print_field_str(fmt : csco_print_format_t, fname : string) : list of string is {
    var flist : list of csco_print_field_s;
    var printables : list of csco_printable_s;

    var rf_s := rf_manager.get_exact_subtype_of_instance(me);
    var rf_f := rf_s.get_field(fname);
    assert that rf_f != NULL else
      error("CSCO_PRINT_INVALID_FIELD_NAME:\n", "cannot print field \"", fname, 
            "\" of ", me, " - invalid field name.");

    // Grab and expand fields in the current struct.
    var pstruct := csco_print_get_fields(fmt);
    var tmp_ind := pstruct.fields.first_index(.name == fname);
    if tmp_ind == UNDEF {
      pstruct.fields.clear();
    }
    else {
      pstruct.fields = {pstruct.fields[tmp_ind]};
    };

    flist = pstruct.expand_fields();

    // Format the fields into lines and tables.
    printables = csco_print_format(fmt, flist, 0);
    csco_print_align_printables(printables);

    // Align and print.
    for each (p) in printables {
      result.add(p.align());
    };
  };
};

// Base struct to hold printable items. This struct contains methods the
// user can extend to customize formatting that are applied across all
// struct types.
struct csco_printable_s {
  !indent : uint;

  // Specifies the indent pattern to use when printing trees of structs.
  indent_pattern() : string is {
    result = "-->";
  };

  package get_indent() : string is {
    result = get_default_indent();
    if indent > 0 {
      for i from 0 to indent - 1 {
        result = append(result, indent_pattern());
      };
    };
  };

  // Global indent that's applied to every line displayed by csco_print. Note
  // that this indent is in addition to the indent applied during tree printing.
  get_default_indent() : string is {
    result = " ";
  };

  // Spacing to use when printing multiple fields in a line.
  column_spacing() : string is {
    result = "  ";
  };

  align() : list of string is undefined;
};

// This extension contains methods which can readily be extended by the user
// to customize various aspects of csco_print, on a per struct type basis.
extend any_struct {
  // This is a user hook which is called right before fields are collected
  // for printing.
  csco_print_get_fields_user_hook(format : csco_print_format_t) is empty;

  // Users can override this hook to allow/disallow port fields to be
  // automatically included in printing. Default is DISALLOW for both
  // SHORT and LONG formats.
  csco_print_auto_init_filter_port(fmt : csco_print_format_t, rf_f : rf_field) : bool is {
    result = TRUE;
  };

  // Users can override this hook to allow/disallow struct fields to be
  // automatically included in printing. Default is ALLOW for LONG
  // and DISALLOW for SHORT.
  csco_print_auto_init_filter_struct(fmt : csco_print_format_t, rf_f : rf_field) : bool is {
    result = (fmt == SHORT);
  };

  // Users can override this hook to allow/disallow list fields to be
  // automatically included in printing. Default is DISALLOW for 
  // list of ports. For other element types, DEFAULT is ALLOW for LONG
  // and DISALLOW for SHORT.
  csco_print_auto_init_filter_list(fmt : csco_print_format_t, rf_f : rf_field) : bool is {
    var vholder := rf_f.get_value(me);
    var elem_type := vholder.get_type().as_a(rf_list).get_element_type();
    if elem_type is a rf_port {
      result = TRUE;
    }
    else {
      result = (fmt == SHORT);
    };
  };

  // Users can override this hook to allow/disallow simple fields to be
  // automatically included in printing. Default is ALLOW for LONG
  // and DISALLOW for SHORT.
  csco_print_auto_init_filter_simple(fmt : csco_print_format_t, rf_f : rf_field) : bool is {
    result = (fmt == SHORT);
  };

  // Users can override this hook to allow/disallow Specman private fields to be
  // automatically included in printing. Default is DISALLOW for both
  // SHORT and LONG formats.
  csco_print_auto_init_filter_nonuser(fmt : csco_print_format_t, rf_f : rf_field) : bool is {
    result = TRUE;
  };

  csco_print_get_numeric_field_value(pf : csco_print_field_s) : string is {
    var is_time : bool;
    //var tmp_int : int;

    if pf.vholder.get_type().get_name() == "time" {
      var time_val : int(bits:64) = pf.vholder.get_value().unsafe();
      time_val = from_specman_scale(time_val, ns);
      pf.vholder = rf_manager.get_type_by_name("int(bits:64)").create_holder(time_val.unsafe());
      is_time = TRUE;
    };

    if pf.radix == "DEC" {
      result = dec(pf.vholder.get_type().value_to_string(pf.vholder.get_value()));
    }
    else if pf.radix == "BIN" {
      result = bin(pf.vholder.get_type().value_to_string(pf.vholder.get_value()));
    }
    else if pf.radix == "HEX" {
      result = hex(pf.vholder.get_type().value_to_string(pf.vholder.get_value()));
    }
    else {
      var tmp_int : int(bits:*);
      tmp_int = pf.vholder.get_type().value_to_string(pf.vholder.get_value()).as_a(int);
      result = appendf(pf.radix, tmp_int);
    };

    if is_time {
      result = append(result, " ns");
    };
  };

  csco_print_get_string_field_value(pf : csco_print_field_s) : string is {
    result = pf.vholder.get_type().value_to_string(pf.vholder.get_value());
  };

  csco_print_get_scalar_field_value(pf : csco_print_field_s) : string is {
    result = csco_print_get_string_field_value(pf);
  };

  csco_print_get_struct_field_value(pf : csco_print_field_s) : string is {
    var sptr : any_struct = pf.vholder.get_value().unsafe();
    if sptr != NULL {
      result = sptr.csco_short_name();
    }
    else {
      result = "NULL";
    };
  };

  csco_print_get_list_field_value(pf : csco_print_field_s) : string is {
    result = append(dec(rf_manager.get_list_size(pf.vholder)), " elems");
  };

  csco_print_get_port_field_value(pf : csco_print_field_s) : string is {
    var pt : any_port = pf.vholder.get_value().unsafe();
    if pt.get_bound_set().size() <= 1 {
      if pt.is_bound() {
        result = "external";
      }
      else {
        result = "unconn";
      };
    }
    else {
      result = append(dec(pt.get_bound_set().size() - 1), " conn");
    };
  };

  // User extensible method to define the number of elements displayed
  // per line during simple list table printing. The parameters are as
  // follows:
  //
  //   hdr_len : length of the header string for the list.
  //   max_elem_len : max length of all the elements.
  //   indent_len : initial indentation length.
  //   column_spacing : amount of space between each table column.
  // 
  csco_print_calc_list_width(hdr_len : uint, max_elem_len : uint, indent_len : uint, column_spacing : uint) : uint is {
    var line_width := get_config(print, line_size);
    result = (line_width - hdr_len - indent_len - 30) / (max_elem_len + column_spacing);
    var tmp_int := ilog2(result);
    result = ipow(2, tmp_int);
  };
};


//
// Logic below this point are really of a package private nature and should only
// be customized if the user wishes to introduced major formatting changes on a 
// project wide basis, and should only be undertaken by those who really 
// understands the code.
//


// Holds a printable string and its max width.
struct csco_print_token_s {
  !value : string;

  // 0 means on addition padding.
  !width : uint;

  get_width() : uint is {
    result = max(width, str_len(value));
  };
};

// Represents a printable line, containing a list of tokens as
// fields in the line.
struct csco_printable_line_s like csco_printable_s {
  !tokens : list of csco_print_token_s;

  align() : list of string is {
    for each (t) in tokens {
      t.value = str_pad(t.value, t.get_width());
    };

    var token_str : string;
    if !tokens.is_empty() {
      if tokens.size() > 1 {
        token_str = str_join(tokens.value, column_spacing());
      }
      else {
        token_str = tokens[0].value;
      };
      result.add(str_join({get_indent(); token_str}, ""));
    };
  };
};

// Represents a printable table, containing a list of lines.
struct csco_printable_table_s like csco_printable_s {
  !lines : list of csco_printable_line_s;

  align() : list of string is {
    //?? check empty list of lines.
    //?? check for non-empty tokens.

    assert that !lines.has(.tokens.size() != lines[0].tokens.size()) else
      error("CSCO_PRINT_ERR_MISALIGNED_TABLE:\n", 
            "not all lines have the same # of tokens in table.");

    var max_width : uint;
    for i from 0 to lines[0].tokens.size() - 1 {
      max_width = lines.max_value(.tokens[i].get_width());
      for each (l) in lines {
        l.tokens[i].width = max_width;

        if i == lines[0].tokens.size() - 1 {
          l.indent = indent;
          result.add(l.align());
        };
      };
    };
  };
};

//?? make sure you do caching based on unique when sub-types.

// Base struct for any struct field that should be printed.
struct csco_print_field_s {
  // name of the field inside its struct.
  !name : string;

  // alternative name to be used for prinintg, as specified by
  // the user.
  !header : string;

  // position of field with respect to other fields in the struct.
  !order : uint;

  !format : csco_print_format_t;

  !radix : string;
  init() is also {
    radix = "HEX";
  };

  // Only applicable for struct and list of struct fields. 
  // If TRUE, field will be expanded during printing.
  !expand : bool;

  // Pointer to struct holding the field.
  !holding_struct : any_struct;

  // holds current value of the field which is to be printed.
  !vholder : rf_value_holder;

  // Called by printing routines to obtain a complete list of
  // fields which need to be printed.
  expand_fields() : list of csco_print_field_s is undefined;
};

// Models scalar and string fields.
struct csco_print_simple_field_s like csco_print_field_s {
  expand_fields() : list of csco_print_field_s is {
    result = {me};
  };
};

// Models struct fields.
struct csco_print_struct_field_s like csco_print_field_s {
  !fields : list of csco_print_field_s;

  expand_fields() : list of csco_print_field_s is {
    var tmp_flist : list of csco_print_field_s;
    for each (f) in fields {
      if f is a csco_print_struct_field_s {
        var sptr : any_struct = f.vholder.get_value().unsafe();
        if sptr == NULL or !f.expand {
          result.add(f);
        }
        else {
          var tmp_ps := sptr.csco_print_get_fields(format);
          tmp_flist = tmp_ps.expand_fields();
          for each (tf) in tmp_flist {
            tf.header = append(f.header, ".", tf.header);
          };
          result.add(tmp_flist);
        };
      }
      else {
        tmp_flist = f.expand_fields();
        result.add(tmp_flist);
      };
    }; // for each (f) in fields
  }; // expand_fields() 
};

// Models list fields.
struct csco_print_list_field_s like csco_print_field_s {
  expand_fields() : list of csco_print_field_s is {
    if format == SHORT {
      result.add(me);
    };
  };
};

// Models list of scalar fields.
struct csco_print_simple_list_field_s like csco_print_list_field_s {
  when LONG'format csco_print_simple_list_field_s {
    expand_fields() : list of csco_print_field_s is also {
      result.add(me);
    };
  };
};

// Models list of struct fields.
struct csco_print_compound_list_field_s like csco_print_list_field_s {
  when LONG'format csco_print_compound_list_field_s {
    expand_fields() : list of csco_print_field_s is also {
      if !expand or rf_manager.get_list_size(vholder) == 0 {
        result.add(me);
      }
      else {
        var lsize := rf_manager.get_list_size(vholder);
        var elem_val : rf_value_holder;
        var sptr : any_struct;
        var tmp_sf : csco_print_struct_field_s;
        var tmp_flist : list of csco_print_field_s;
        if lsize > 0 {
          for i from 0 to lsize - 1 {
            elem_val = rf_manager.get_list_element(vholder, i);
            sptr = elem_val.get_value().unsafe();
            if sptr == NULL {
              var tmp_f := new csco_print_simple_field_s with {
                .name = append(header, "[", dec(i), "]");
                .header = .name;
                .vholder = rf_manager.get_type_by_name("string").create_holder("NULL".unsafe());
              };
              result.add(tmp_f);
            }
            else {
              tmp_sf = sptr.csco_print_get_fields(LONG);
              tmp_flist = tmp_sf.expand_fields();
              for each (f) in tmp_flist {
                f.header = append(header, "[", dec(i), "].", f.header);
              };
              result.add(tmp_flist);
            };
          };
        };
      };
    }; // expand_fields() 
  }; // when LONG'format csco_print_compound_list_field_s
};

//?? show physical/non-generate qualifiers when printing fields.
//?? check for inifinite loops involving expand=ON and print sub_tree.
//?? add new dummy macro definition to detect ordering issues.

// Models port (simple, event, method) fields.
struct csco_print_port_field_s like csco_print_field_s {
  when SHORT'format csco_print_port_field_s {
    expand_fields() : list of csco_print_field_s is {
      result.add(me);
    };
  };

  when LONG'format csco_print_port_field_s {
    expand_fields() : list of csco_print_field_s is {
      var pt : any_port = vholder.get_value().unsafe();
      var tmp_field : csco_print_field_s;

      if pt.get_bound_set().size() <= 1 {
        result.add(me);
      }
      else {
        for each (b) in pt.get_bound_set() {
          if b != pt {
            tmp_field = new csco_print_field_s with {
              .name = name;
              .header = append(name, ".bound[", dec(index), "]");
              .holding_struct = b.get_enclosing_unit(any_unit);
              .vholder = rf_manager.get_type_by_name("string").create_holder(b.e_path().unsafe());
            };
            result.add(tmp_field);
          };
        };
      };
    };
  };
};

struct csco_print_struct_and_depth_s {
  !sptr : any_struct;
  !depth : uint;
};


extend any_struct {
  // Grabs all the fields that should be printed for the given format. Note that
  // no field expansion is performed in this method.
  package csco_print_get_fields(format : csco_print_format_t) : csco_print_struct_field_s is {
    csco_print_get_fields_user_hook(format);

    result = new csco_print_struct_field_s with {
      .format = format;
    };

    // For SHORT format, default initialization is to print no fields.
    // For LONG format, default initialization is to print almost all fields. Only
    // port and Specman private fields are omitted.
    result.fields.add(csco_print_auto_init(format));

    // Hook for macro to modify default attributes.
    csco_print_get_fields_hook(format, result.fields);

    // Remove fields that won't be displayed and remember ordering of fields.
    var tmp_ind : uint = 0;
    while tmp_ind < result.fields.size() {
      if result.fields[tmp_ind].format == NONE {
        result.fields.delete(tmp_ind);
      }
      else {
        result.fields[tmp_ind].order = tmp_ind;
        tmp_ind += 1;
      };
    };
  };

  // This is a macro hook. Users shouldn't touch it. 
  csco_print_get_fields_hook(fmt : csco_print_format_t, flist : list of csco_print_field_s) is empty;

  // Autoamatically collect fields to be printed. 
  // For SHORT format, default initialization is to print no fields.
  // For LONG format, default initialization is to print almost all fields. Only
  // port and Specman private fields are omitted.
  package csco_print_auto_init(fmt : csco_print_format_t) : list of csco_print_field_s is {
    var rf_s := rf_manager.get_exact_subtype_of_instance(me);
    var rf_fields := rf_s.get_fields();
    var tmp_ind : int;
    var rf_t : rf_type;
    var remove_f : bool;
    var tmp_field : csco_print_field_s;

    for each (rf_f) in rf_fields {
      rf_t = rf_f.get_type();

      if rf_t is a rf_port {
        tmp_field = new csco_print_port_field_s;
        remove_f = csco_print_auto_init_filter_port(fmt, rf_f);
      }
      else if rf_t is a rf_struct {
        tmp_field = new csco_print_struct_field_s;
        remove_f = csco_print_auto_init_filter_struct(fmt, rf_f);
      }
      else if rf_t is a rf_list (rf_l) {
        if rf_l.get_element_type() is a rf_struct {
          tmp_field = new csco_print_compound_list_field_s;
        }
        else {
          tmp_field = new csco_print_simple_list_field_s;
        };

        remove_f = csco_print_auto_init_filter_list(fmt, rf_f);
      }
      else {
        tmp_field = new csco_print_simple_field_s;
        remove_f = csco_print_auto_init_filter_simple(fmt, rf_f);
      };

      if !rf_f.get_declaration().get_module().is_user_module() {
        if csco_print_auto_init_filter_nonuser(fmt, rf_f) {
          remove_f = TRUE;
        };
      };

      tmp_field.name = rf_f.get_name();
      tmp_field.header = tmp_field.name;
      tmp_field.format = fmt;
      tmp_field.holding_struct = me;
      tmp_field.vholder = rf_f.get_value(me);
      if tmp_field.vholder.get_type().get_name() == "time" {
        tmp_field.radix = "DEC";
      };
      result.add(tmp_field);

      if remove_f {
        tmp_field.format = NONE;
      };
    };  // for each (rf_f) in rf_fields
  };

  // This method is called by the macro to enable/disable fields for a given
  // print format.
  csco_print_enable_field(flist : list of csco_print_field_s, name:string, 
                          print_fmt:csco_print_format_t, update_fmt:csco_print_format_t,
                          action : csco_print_enable_disable_t) is {
    if update_fmt == ALL or update_fmt == print_fmt {
      var ind : int = flist.first_index(.name == name);

      assert that ind != UNDEF else
        error("CSCO_PRINT_ERR: csco_print_udpate_field() cannot define csco_print attribute on '", 
               name, "',  it's not a valid field of '", me, "'.");

      if action == enable {
        flist[ind].format = print_fmt;
      }
      else {
        flist[ind].format = NONE;
      };
    };
  };

  // This method is called by the macro to modify field printing attributes.
  // Users shouldn't touch this method.
  csco_print_update_field(flist : list of csco_print_field_s, name:string, 
                          print_fmt:csco_print_format_t, update_fmt:csco_print_format_t,
                          precede:string, hdr:string, rdx:string, exp:string) is {

    if update_fmt == ALL or update_fmt == print_fmt {
      var ind : int = flist.first_index(.name == name);
      var tmp_field : csco_print_field_s;
      var insert_ind : int = UNDEF;

      assert that precede != "" or hdr != "" or rdx != "" or exp != "" else
        error("CSCO_PRINT_ERR: csco_print_attr statement must specify at least 1 attribute.");

      assert that ind != UNDEF else
        error("CSCO_PRINT_ERR: csco_print_update_field() cannot define csco_print attribute on '", 
               name, "',  it's not a valid field of '", me, "'.");

      // Update existing attributes with new ones.
      insert_ind = ind;
      tmp_field = flist[ind];
      flist.delete(ind);

      if hdr != "" {
        tmp_field.header = hdr;
        assert that hdr not in flist.all(index != ind).header else
          error("CSCO_PRINT_ERR: csco_print_update_field() cannot use duplicate header name ", hdr);
      };

      if rdx != "" { 
        tmp_field.radix = rdx;
      };

      if exp != "" { 
        assert that exp in {"ON"; "OFF"} else
          error("CSCO_PRINT_ERR: legal values for 'expand' option is [ON|OFF], detected ", exp);
        tmp_field.expand = (exp == "ON") ? TRUE : FALSE;
      };

      if precede != "" {
        // Reposition field as required.
        insert_ind = flist.first_index(.name == precede);
        assert that insert_ind != UNDEF else
          error("CSCO_PRINT_ERR: ", precede, " is not a legal field name in ", me);
      };

      if insert_ind == UNDEF {
        flist.add(tmp_field);
      }
      else {
        flist.insert(insert_ind, tmp_field);
      };
    }; 
  };

  // Takes multiple lines and tables, and ensures that the 1st field
  // in each printable is padded to the same width.
  csco_print_align_printables(printables : list of csco_printable_s) is {
    var line_list : list of csco_printable_line_s;
    var max_len : uint;

    for each (p) in printables {
      if p is a csco_printable_line_s (line) {
        line_list.add(line);
        max_len = max(max_len, line.tokens[0].get_width());
      }
      else if p is a csco_printable_table_s (table) {
        line_list.add(table.lines[0]);
        max_len = max(max_len, table.lines.max_value(.tokens[0].get_width()));
      };
    };

    for each (l) in line_list {
      l.tokens[0].width = max_len;
    };
  };

  package csco_print_format(fmt : csco_print_format_t, flist : list of csco_print_field_s, indent : uint) : list of csco_printable_s is {
    if fmt == SHORT {
      result = csco_print_format_short(flist, indent, "");
    }
    else if fmt == LONG {
      result = csco_print_format_long(flist, indent);
    };
  };

  // Format SHORT fields into 2 lines, 1 for header and 1 for values. Each
  // line consists of the same # of tokens as printed fields. Note that 
  // list fields are never expanded in SHORT format.  List fields will print 
  // their size.
  csco_print_format_short(fields : list of csco_print_field_s, indent : uint, prefix : string) : list of csco_printable_s is {
    var token : csco_print_token_s;
    var header : csco_printable_line_s;
    var values : csco_printable_line_s;
    var mline_fields :  list of csco_print_field_s;
    var mline_indices : list of uint;
    var num_extra_lines : uint;

    if fields.size() > 0 {
      header = new;
      values = new;
      if prefix != "" {
        token = new;
        header.tokens.add(token);

        token = new csco_print_token_s with {
          .value = prefix;
        };
        values.tokens.add(token);
      };

      for each (f) in fields {
        // Add token to header.
        token = new csco_print_token_s with {
          .value = f.header;
        };
        header.tokens.add(token);

        // A null vholder is possible due to printing a compound list
        // with different when subtypes amongst elements.
        if f.vholder == NULL {
          token = new csco_print_token_s with {
            .value = "n/a";
          };
        }
        // Compound lists which are expanded will result in each struct element's 
        // csco_short_name() being printed on a separate line.
        else if f is a csco_print_compound_list_field_s and f.expand and rf_manager.get_list_size(f.vholder) > 0 {
          var sptr : any_struct = rf_manager.get_list_element(f.vholder, 0).get_value().unsafe();
          token = new csco_print_token_s with {
            .value = (sptr == NULL ? "NULL" : sptr.csco_short_name());
          };

          if rf_manager.get_list_size(f.vholder) > 1 {
            mline_fields.add(f);
            mline_indices.add(index);
            num_extra_lines = max(num_extra_lines, rf_manager.get_list_size(f.vholder) - 1);
          };
        }
        // All other field types are always printed with a single value.
        else {
          // Add token to values.
          token = new csco_print_token_s with {
            .value = csco_print_get_field_value(f);
          };
        };

        values.tokens.add(token);
      };

      result.resize(1);
      result[0] = new csco_printable_table_s with {
        .indent = indent;
        .lines.add(header);
        .lines.add(values);
      };

      // Add value lines for each expanded compound list, if one exists.
      if num_extra_lines > 0 {
        for i from 1 to num_extra_lines {
          values = new;
          for j from 1 to fields.size() {
            token = new;
            values.tokens.add(token);
          };

          for each (f) in mline_fields {
            if rf_manager.get_list_size(f.vholder) > i {
              var sptr : any_struct = rf_manager.get_list_element(f.vholder, i).get_value().unsafe();
              token = new csco_print_token_s with {
                .value = (sptr == NULL ? "NULL" : sptr.csco_short_name());
              };
              values.tokens[mline_indices[index]] = token;
            };
          };

          if prefix != "" {
            token = new;
            values.tokens.add0(token);
          };

          result[0].as_a(csco_printable_table_s).lines.add(values);
        };
      };
    }
    // If a struct doesn't have any SHORT fiels defined, then simply
    // print its csco_short_name().
    else {
      values = new;
      if prefix != "" {
        token = new csco_print_token_s with {
          .value = prefix;
        };
        values.tokens.add(token);
      };

      token = new csco_print_token_s with {
        .value = csco_short_name();
      };
      values.tokens.add(token);

      values.indent = indent;
      result.add(values);
    };
  };

  // Format LONG fields into multiple lines. 1 line each for simple files.
  // Struct fileds are expanded into multiple lines, 1 for each of the struc's fields.
  // Simple list fileds are expanded into multiple lines, in a tabular form. 
  // Compound list fileds are expanded into multiple lines, where each struct element is again expanded.
  csco_print_format_long(flist : list of csco_print_field_s, indent : uint) : list of csco_printable_s is {
    var line : csco_printable_line_s;
    var token : csco_print_token_s;
    var sptr : any_struct;

    for each (f) in flist {
      if f is a csco_print_struct_field_s {
        sptr = f.vholder.get_value().unsafe();
      };

      // format simple lists into tables.
      if f is a csco_print_simple_list_field_s and rf_manager.get_list_size(f.vholder) > 0 {
        result.add(csco_print_format_simple_list_long(f, indent));
      }
      // format unexpanded compound lists into tables.
      else if f is a csco_print_compound_list_field_s and !f.expand and rf_manager.get_list_size(f.vholder) > 0 {
        result.add(csco_print_format_compound_list_long(f, indent));
      }
      // format unexpanded struct fields using SHORT format.
      else if f is a csco_print_struct_field_s and sptr != NULL {
        var pstruct := sptr.csco_print_get_fields(SHORT);
        result.add(sptr.csco_print_format_short(pstruct.expand_fields(), indent, f.header));
      }
      // simple fields and struct fields that are NULL or have no SHORT format fields.
      else {
        line = new csco_printable_line_s with {
          .indent = indent;
        };

        // 1 token for the field header.
        token = new csco_print_token_s with {
          .value = f.header;
        };
        line.tokens.add(token);

        // 1 token for the field value.
        token = new csco_print_token_s with {
          .value = csco_print_get_field_value(f);
        };
        line.tokens.add(token);

        result.add(line);
      };
    };
  };

  package csco_print_sub_tree_format(fmt : csco_print_format_t, slist : list of csco_print_struct_and_depth_s, clist_path : string) : list of csco_printable_s is {
    if fmt == SHORT {
      result = csco_print_sub_tree_format_short(slist, clist_path);
    }
    else if fmt == LONG {
      result = csco_print_sub_tree_format_long(slist, clist_path);
    };
  };

  // SHORT formatting of a tree print results in short printing for each 
  // of the structs involved plus 1 line of field headers.
  csco_print_sub_tree_format_short(slist : list of csco_print_struct_and_depth_s, clist_path : string) : list of csco_printable_s is {
    // Construct a dummy printable field whose value is a list of
    // structs and then format it like a compound list.
    var pfield : csco_print_field_s = new;
    pfield.vholder = rf_manager.get_type_by_name("list of any_struct").create_holder(slist.sptr.unsafe());
    result = csco_print_format_compound_list_long(pfield, 0);
  };

  // LONG formatting of a tree print results in long printing for each 
  // of the structs involved, with each struct block indented to indicate
  // their depth relative to each other.
  csco_print_sub_tree_format_long(slist : list of csco_print_struct_and_depth_s, clist_path : string) : list of csco_printable_s is {
    var pstruct : csco_print_struct_field_s;
    var flist : list of csco_print_field_s;
    var line : csco_printable_line_s;
    var token : csco_print_token_s;
    var printables : list of csco_printable_s;

    for each (s) in slist {
      // Add a blank line to improve delineation between structs.
      if index > 0 {
        token = new;
        line = new csco_printable_line_s with {
          .tokens.add(token);
        };
        result.add(line);
      };

      // Add a header line containing the struct's csco_short_name()
      // for each struct block of printing.
      token = new csco_print_token_s with {
        .value = s.sptr.csco_short_name();
      };
      line = new csco_printable_line_s with {
        .tokens.add(token);
        .indent = s.depth;
      };
      result.add(line);

      // Grab and expand all the fields and call csco_print_format_long() 
      // for each struct.
      pstruct = s.sptr.csco_print_get_fields(LONG);
      flist = pstruct.expand_fields();
      printables = s.sptr.csco_print_format_long(flist, s.depth);
      csco_print_align_printables(printables);
      result.add(printables);
    };
  };

  // Recursively get all the structs involved in a tree print.
  csco_print_sub_tree_get_structs(fmt : csco_print_format_t, depth : uint, child_list_path : string) : list of csco_print_struct_and_depth_s is {
    var slist : list of any_struct;
    var tmp_result : csco_print_struct_and_depth_s;

    tmp_result = new csco_print_struct_and_depth_s with {
      .sptr = me;
      .depth = depth;
    };
    result.add(tmp_result);

    slist = csco_print_get_nested_structs(child_list_path);
    for each (s) in slist {
      var common_p := csco_print_get_common_like_parent(s);
      assert that common_p.get_name() not in {"any_struct"; "any_unit"} else
        error("CSCO_PRINT_ERR: ", me, " and ", s, " do not share the same base type ",
              "and cannot be printed together using csco_print_sub_tree()");

      result.add(s.csco_print_sub_tree_get_structs(fmt, depth+1, child_list_path));
    };
  };

  // Locate and return a nested field of the enclosing struct.
  package csco_print_get_nested_structs(path : string) : list of any_struct is {
    var str_list := str_split(path, ".");
    var rf_s : rf_struct;
    var rf_f : rf_field;
    var tmp_result : list of any_struct;

    result.add(me);
    for each (path) using index (pindex) in str_list {
      tmp_result.clear();
      for each (r) using index (rindex) in result {
        rf_s = rf_manager.get_exact_subtype_of_instance(r);
        rf_f = rf_s.get_field(path);

        if rf_f != NULL {
          if rf_f.get_type() is a rf_struct {
            var sptr : any_struct = rf_f.get_value(r).get_value().unsafe();
            tmp_result.add(sptr);
          }
          else if rf_f.get_type() is a rf_list (rf_l) { 
            var rf_elemt := rf_l.get_element_type();
            assert that rf_elemt is a rf_struct else
              error("CSCO_PRINT_ERR: ", rf_f.get_name(), " is not a list of structs in ", r,
                    ", as required by csco_print_sub_tree().");

            var slist : list of any_struct = rf_f.get_value(r).get_value().unsafe();
            tmp_result.add(slist);
          };
        };
      }; // for each (r) in result
      result = tmp_result.copy();
    };
  };

  package csco_print_get_common_like_parent(s2 : any_struct) : rf_like_struct is {
    var my_path := csco_print_get_like_parents();
    var s2_path := s2.csco_print_get_like_parents();
    for each (p) in s2_path {
      if p in my_path {
        return p;
      };
    };
  };

  package csco_print_get_like_parents() : list of rf_like_struct is {
    var rf_s := rf_manager.get_struct_of_instance(me).get_when_base();
    while rf_s.get_name() != "base_struct" {
      result.add(rf_s);
      rf_s = rf_s.get_supertype();
    };
  };

  // Returns the unexpanded value of any field of the enclosing struct.
  // The user may override any of the following methods to custom value
  // formatting on a per field basis.
  package csco_print_get_field_value(pfield : csco_print_field_s) : string is {
    var rf_t := pfield.vholder.get_type();

    if rf_t is a rf_numeric {
      result = pfield.holding_struct.csco_print_get_numeric_field_value(pfield);
    }
    else if rf_t is a rf_string {
      result = pfield.holding_struct.csco_print_get_string_field_value(pfield);
    }
    else if rf_t is a rf_scalar {
      result = pfield.holding_struct.csco_print_get_scalar_field_value(pfield);
    }
    else if rf_t is a rf_struct {
      result = pfield.holding_struct.csco_print_get_struct_field_value(pfield);
    }
    else if rf_t is a rf_list {
      result = pfield.holding_struct.csco_print_get_list_field_value(pfield);
    }
    else if rf_t is a rf_port {
      result = pfield.holding_struct.csco_print_get_port_field_value(pfield);
    };
  };

  // This method formats a simple list of scalars/strings into a table. The
  // # of elements displayed on each line is a user extensible function of the
  // screen width as well as the list elements sizes.
  csco_print_format_simple_list_long(pfield : csco_print_field_s, indent : uint) : list of csco_printable_s is {
    var lsize : uint;
    var token : csco_print_token_s;
    var token_list : list of csco_print_token_s;
    var max_elem_len : uint;
    var table : csco_printable_table_s;
    var tmp_field : csco_print_field_s = pfield.copy();

    // Construct a raw list of tokens, 1 for each list element.
    lsize = rf_manager.get_list_size(pfield.vholder);
    if lsize > 0 {
      for i from 0 to lsize - 1 {
        tmp_field.header = dec(i);
        tmp_field.vholder = rf_manager.get_list_element(pfield.vholder, i);
        token = new csco_print_token_s with {
          .value = csco_print_get_field_value(tmp_field);
        };
        token_list.add(token);
        max_elem_len = max(max_elem_len, str_len(token.value));
      };
    };

    // Initialize table.
    table = new csco_printable_table_s with {
      .indent = indent;
    };

    // Determine elements displayed per line. The user can override
    // the called method to define his own scheme.
    var elems_per_line := csco_print_calc_list_width(str_len(pfield.header), 
                                                     max_elem_len, 
                                                     str_len(table.get_indent()), 
                                                     str_len(table.column_spacing()));
    elems_per_line = min(elems_per_line, lsize);

    // Construct each line in the table.
    var tmp_line : csco_printable_line_s;
    var num_rows := div_round_up(lsize, elems_per_line);
    for i from 0 to num_rows - 1 {
      // Each line is preceded by a label indicating the indices
      // displayed on the line.
      token = new csco_print_token_s with {
        .value = append(pfield.header, "[", dec(i*elems_per_line), ":", dec(min((i+1)*elems_per_line-1, lsize-1)), "]:");
      };
      tmp_line = new csco_printable_line_s with {
        .tokens.add(token);
      };

      if i != num_rows - 1 {
        tmp_line.tokens.add(token_list[i*elems_per_line .. ((i + 1)*elems_per_line) - 1]);
      }
      else {
        // For the last line, we fill with NULL strings to pad to
        // the end of the table.
        tmp_line.tokens.add(token_list[i*elems_per_line .. ]);
        token = new;
        tmp_line.tokens.resize(elems_per_line+1, TRUE, token, TRUE);
      };

      table.lines.add(tmp_line);
    };

    result.add(table);
  };

  // This method formats a compound list of scalars/strings into a table, without
  // expanding the struct elements. The # of lines in the resulting table is equal
  // to the list size, plus 1 line of headers. Each non-header line consists of 
  // the SHORT print of the respective struct element.
  csco_print_format_compound_list_long(pfield : csco_print_field_s, indent : uint) : list of csco_printable_s is {
    var lsize : uint = rf_manager.get_list_size(pfield.vholder);
    var sptr : any_struct;
    var elem_val : rf_value_holder;
    var s_ptrs : list of any_struct;
    var pstructs : list of csco_print_struct_field_s;
    var tmp_pstruct : csco_print_struct_field_s;
    var combined_fields : list of csco_print_field_s;
    var tmp_flist : list of csco_print_field_s;

    // For each struct element in the list, grab and expand all the SHORT fields 
    // in the struct and save them for the logic below. Also construct a merged
    // list of fields so that we can uniquify.
    if lsize > 0 {
      for i from 0 to lsize - 1 {
        elem_val = rf_manager.get_list_element(pfield.vholder, i);
        sptr = elem_val.get_value().unsafe();
        if sptr != NULL {
          tmp_pstruct = sptr.csco_print_get_fields(SHORT);
          tmp_flist = tmp_pstruct.expand_fields();
          tmp_pstruct.fields = tmp_flist.copy();
          combined_fields.add(tmp_flist);
        }
        else {
          tmp_pstruct = NULL;
        };
        pstructs.add(tmp_pstruct);
        s_ptrs.add(sptr);
      };
    };

    // Identify all the unique fields across all struct elements and 
    // uniquify.
    combined_fields = combined_fields.sort_by_field(header);
    combined_fields = combined_fields.unique(.header);
    combined_fields = combined_fields.sort_by_field(order);

    var table : csco_printable_table_s = new;
    var header : csco_printable_line_s = new;
    var line : csco_printable_line_s;
    var token : csco_print_token_s = new;
    var tmp_result : list of csco_printable_s;

    // Construct header line of table.
    if !combined_fields.is_empty() {
      header.tokens.add(token);
      for each (f) in combined_fields {
        token = new csco_print_token_s with {
          .value = f.header;
        };
        header.tokens.add(token);
      };
      table.lines.add(header);
    };

    for each (s) using index (sindex) in s_ptrs {
      if s == NULL {
        line = new;
        token = new csco_print_token_s with {
          .value =  append(pfield.header, "[", dec(sindex), "]");
        };
        line.tokens.add(token);

        for each (f) in combined_fields {
          token = new csco_print_token_s with {
            .value = "NULL";
          };
          line.tokens.add(token);
        };
        table.lines.add(line);
      }
      else {
        var tmp_pfield : csco_print_field_s;
        tmp_flist.clear();
        for each (f) using index (findex) in combined_fields {
          tmp_pfield = pstructs[sindex].fields.first(.header == f.header);
          if tmp_pfield != NULL {
            tmp_flist.add(tmp_pfield);
          }
          else {
            f.vholder = NULL;
            tmp_flist.add(f);
          };
        };
        tmp_result = s.csco_print_format_short(tmp_flist, indent, append(pfield.header, "[", dec(sindex), "]"));

        // Remove header line of result returned from format_short().
        if combined_fields.size() > 0 {
          tmp_result[0].as_a(csco_printable_table_s).lines.delete(0);
          table.lines.add(tmp_result[0].as_a(csco_printable_table_s).lines);
        }
        else {
          table.lines.add(tmp_result[0].as_a(csco_printable_line_s));
        };
      };
    };
    table.indent = indent;
    result.add(table);
  };
}; // extend any_struct

extend any_unit {
  // For units, default csco_short_name() to simply be the unit's
  // short_name().
  csco_short_name() : string is {
    result = short_name();
  };
};


extend global {
  setup_test() is also {
    set_config(print, line_size, 180);
  };
};


'>


//
// For legacy reasons only. Should eventually be removed.
//
extend any_struct {
  print_short() is undefined;
  print_long() is undefined;

  format_list_of_bytes ( prefix: string, num_entries : uint, data : list of byte ) : string is {
    var i : uint;
    var j : uint;
    var l_string: list of string;

    if (data.size() !=0) {
      for {i=0; i < data.size()/num_entries; i+=1;} {
        l_string.add(appendf("%s> [%02d-%02d] ",prefix,i*num_entries,i*num_entries+(num_entries-1)));
        for {j=i*num_entries; j < i*num_entries+(num_entries-1); j+=1;} {
          l_string.add(appendf("%02x_",data[j]));
        };
        l_string.add(appendf("%02x\n",data[j]));
      };

      if (data.size() % num_entries != 0) {
        l_string.add(appendf("%s> [%02d-%02d] ",prefix,i*num_entries,data.size()-1));
      for {j=i*num_entries; j < data.size()-1; j+=1; } {
      l_string.add(appendf("%02x_",data[j]));
      };
        l_string.add(appendf("%02x\n",data[j]));
      };
      result = str_join(l_string,"");
    } else {
      result = appendf("%s> EMPTY!\n",prefix);
    };
  };                                                       // end method  format_list_of_bytes()
};
